/*
 * Copyright (C) 2019 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-amiga.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>


#if __BYTE_ORDER == __BIG_ENDIAN
# define ID(a,b,c,d) ((a<<24)|(b<<16)|(c<<8)|(d<<0))
#else
# define ID(a,b,c,d) ((a<<0)|(b<<8)|(c<<16)|(d<<24))
#endif

#define ID_SCREEN  ID(0x12,0x03,0x19,0x90)
#define ID_PICTURE ID(0x06,0x07,0x19,0x63)

#define ID_MEMORYBANK  ID('A','m','B','k')
#define ID_SPRITE_BANK ID('A','m','S','p')
#define ID_ICON_BANK   ID('A','m','I','c')

typedef struct{
    int stride;
    int width;
    int height;
    int depth;
    //int hx;
    //int hy;
    const uint8_t *data;
}sprite_t;

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

typedef struct{
    frame_t *frame;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_ABK,
        T_PAC
    }type;
    page_t *page;
};


static const char **
_amos_get_mime_types()
{
    static const char *mime_types[] = {
        "application/x-amos-memorybank",
        "image/x-amos-spritebank",
        "image/x-amos-iconbank",
        "image/x-amos-picturepacker",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_amos_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strcasecmp(mime_type,"image/x-amos-picturepacker")){
        h->type = T_PAC;
    }else{
        h->type = T_ABK;
    }
    h->info = info;
    info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_amos_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int j;
        for(j=0;j<h->info->page_count;++j){
            int i;
            for(i=0;i<h->info->frame_count;++i)
                if(h->page[j].frame[i].surface)
                    cairo_surface_destroy(h->page[j].frame[i].surface);
            free(h->page[j].frame);
        }
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get24(a) (get16(a)<<8|get8(a+2))
#define get32(a) be32toh(*(uint32_t*)(a))


static int
_decompress(uint8_t *dst,
            int len,
            const uint8_t *table1,
            const uint8_t *table2,
            const uint8_t *pointers)
{
    int i;
    for(i=0;i<len;++i){
        if(!(i&7))
            if(pointers[i>>6] & (128>>((i>>3)&7)))
                ++table2;
        if(*table2 & (128>>(i&7)))
            ++table1;
        *dst++ = *table1;
    }
    return 0;
}


static cairo_surface_t *
_surface_from_buffer(const uint8_t *src,
                     int tx,
                     int ty,
                     int tcar,
                     int bitplanes,
                     uint8_t *palette,
                     int bpc,
                     nile_format_t format)
{
    int width = tx*8;
    int height = ty*tcar;
    int planesize = tx*ty*tcar;
    const uint8_t *src_plane[8];
    off_t src_offs = 0;
    cairo_surface_t *surface;
    uint8_t* dst;
    int stride;
    int i;
    for(i=0;i<bitplanes;++i)
        src_plane[i] = src + i*planesize;
    if(NILE_FORMAT_ne_r5g6b5 == format){
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width,height);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    }
    dst = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);
    while(ty-->0){
        int x;
        for(x=0;x<tx;++x){
            int y;
            for(y=0;y<tcar;++y){
                int i;
                for(i=0;i<8;++i){
                    int bit;
                    int pixel = 0;
                    for(bit=0;bit<bitplanes;++bit){
                        if(src_plane[bit][src_offs] & (128>>i))
                            pixel |= 1<<bit;
                    }
                    memcpy(dst + (x*8+i)*bpc + y*stride, palette + pixel*bpc, bpc);
                }
                ++src_offs;
            }
        }
        dst += tcar*stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_sprite(sprite_t *sprite, const uint8_t *palette, int bpc, nile_format_t format)
{
    int planesize = sprite->stride*sprite->height;
    const uint8_t *src_plane[8];
    off_t src_offs = 0;
    cairo_surface_t *surface;
    uint8_t* dst;
    int stride;
    int y;
    int i;
    for(i=0;i<sprite->depth;++i)
        src_plane[i] = sprite->data + i*planesize;
    if(NILE_FORMAT_ne_r5g6b5 == format){
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,sprite->width,sprite->height);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,sprite->width,sprite->height);
    }
    dst = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);
    for(y=0;y<sprite->height;++y){
        int byte;
        for(byte=0;byte<sprite->stride;++byte){
            int i;
            for(i=0;i<8;++i){
                int bit;
                int pixel = 0;
                for(bit=0;bit<sprite->depth;++bit){
                    if(src_plane[bit][src_offs] & (128>>i))
                        pixel |= 1<<bit;
                }
                memcpy(dst + (byte*8+i)*bpc, palette + pixel*bpc, bpc);
            }
            ++src_offs;
        }
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_handle_pacpic(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    uint8_t palette[32*4];
    uint8_t *buffer;
    int datasize;
    int bplcon0;
    int colors;
    int bitplanes;
    int tx;
    int ty;
    int tcar;

    if(ID_SCREEN==*(uint32_t*)data){
        /* got optional screen header */
        bplcon0 = get16(data+20);
        colors = get16(data+22);
#if 1
        nil_amiga_copy_colors(palette, 4, NIL_CAIRO_FORMAT_RGB24, (uint16_t*)(data+26), colors);
#else
        amiga_decode_palette((uint16_t*)(data+26),palette,colors);
#endif
        data+=90;
        len-=90;
    }else{
        static const uint16_t default_palette[32] = {
            0x0000,0xfe0f,0xec0e,0xea0e,0xda0f,0xa40e,0x820d,0x700d,
            0x4404,0x5505,0x7707,0x8808,0xaa0a,0xbb0b,0xcc0c,0xee0e,
            0xff0d,0xee0b,0xdd0a,0xcd09,0xbc07,0xab06,0x8a05,0x7904,
            0x6903,0x5803,0x5802,0x4702,0x4602,0x3601,0x2601,0x0500
        };
        bplcon0 = 0;
        colors = 32;
#if 1
        nil_amiga_copy_colors(palette, 4, NIL_CAIRO_FORMAT_RGB24, default_palette, 32);
#else
        amiga_decode_palette(default_palette,palette,32);
#endif
    }
    if(*(uint32_t*)data != ID_PICTURE)
        return -1;

    if(bplcon0 & BPLCON0_HIRES){
        h->info->pixel_ratio = 8.0/15;
    }else{
        h->info->pixel_ratio = 16.0/15;
    }

    tx = get16(data+8);
    ty = get16(data+10);
    tcar = get16(data+12);
    h->info->width = tx*8;
    h->info->height = ty * tcar;
    bitplanes = get16(data+14);
    datasize = tx * ty * tcar * bitplanes;


    buffer = malloc(datasize);
    _decompress(buffer,datasize,(uint8_t*)data+24,(uint8_t*)data+get32(data+16),(uint8_t*)data+get32(data+20));
    h->page = (page_t*)malloc(sizeof(page_t));
    h->page[0].frame = (frame_t*)malloc(sizeof(frame_t));
    h->page[0].frame[0].surface = _surface_from_buffer(buffer,tx,ty,tcar,bitplanes,palette,4,NIL_CAIRO_FORMAT_RGB24);
    free(buffer);

    return 0;
}


static int
_handle_spritebank(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    uint8_t palette[64*4];
    sprite_t *sprite;
    int i;
    h->info->frame_count = get16(data+0);
    data += 2;
    len -= 2;
    sprite = (sprite_t*)malloc(h->info->frame_count*sizeof(sprite_t));
    for(i=0;i<h->info->frame_count;++i){
        int size;
        sprite[i].stride = get16(data)*2;
        sprite[i].width = sprite[i].stride*8;
        sprite[i].height = get16(data+2);
        sprite[i].depth = get16(data+4);
        //sprite[i].hx = get16(data+6);
        //sprite[i].hy = get16(data+8);
        sprite[i].data = (const uint8_t*)(data+10);
        size = 10+sprite[i].stride*sprite[i].height*sprite[i].depth;
        data += size;
        len -= size;
    }
    DEBUG_p("bytes left: %d",(int)len);
#if 1
    nil_amiga_copy_colors(palette, 4, NIL_CAIRO_FORMAT_RGB24, (uint16_t*)data, 16);
#else
    amiga_decode_palette((uint16_t*)data,palette,16);
#endif
    /* decode sprites */
    h->page = (page_t*)malloc(sizeof(frame_t));
    h->page[0].frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        h->page[0].frame[i].surface = _surface_from_sprite(&sprite[i],palette,4,NIL_CAIRO_FORMAT_RGB24);
        h->page[0].frame[i].duration = 5.0/60;
        if(sprite[i].width > h->info->width)
            h->info->width = sprite[i].width;
        if(sprite[i].height > h->info->height)
            h->info->height = sprite[i].height;
    }
    free(sprite);
    h->info->pixel_ratio = 16.0/15;
    return 0;
}


static int
_handle_iconbank(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    uint8_t palette[64*4];
    sprite_t *sprite;
    int i;
    h->info->page_count = get16(data+0);
    data += 2;
    len -= 2;
    sprite = (sprite_t*)malloc(h->info->page_count*sizeof(sprite_t));
    for(i=0;i<h->info->page_count;++i){
        int size;
        sprite[i].stride = get16(data)*2;
        sprite[i].width = sprite[i].stride*8;
        sprite[i].height = get16(data+2);
        sprite[i].depth = get16(data+4);
        //sprite[i].hx = get16(data+6);
        //sprite[i].hy = get16(data+8);
        sprite[i].data = (const uint8_t*)(data+10);
        size = 10+sprite[i].stride*sprite[i].height*sprite[i].depth;
        data += size;
        len -= size;
    }
    DEBUG_p("bytes left: %d",(int)len);
#if 1
    nil_amiga_copy_colors(palette, 4, NIL_CAIRO_FORMAT_RGB24, (uint16_t*)data, 16);
#else
    amiga_decode_palette((uint16_t*)data,palette,16);
#endif
    /* decode sprites */
    h->page = (page_t*)malloc(h->info->page_count*sizeof(page_t));
    for(i=0;i<h->info->page_count;++i){
        h->page[i].frame = (frame_t*)malloc(sizeof(frame_t));
        h->page[i].frame[0].surface = _surface_from_sprite(&sprite[i],palette,4,NIL_CAIRO_FORMAT_RGB24);
        h->page[i].frame[0].duration = 0;
        if(sprite[i].width > h->info->width)
            h->info->width = sprite[i].width;
        if(sprite[i].height > h->info->height)
            h->info->height = sprite[i].height;
    }
    free(sprite);
    h->info->pixel_ratio = 16.0/15;
    return 0;
}


static int
_handle_memorybank(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    //int banknum = get16(data+0);
    //int memtype = get16(data+2);
    int size = get32(data+4) & 0x0fffffff;
    DEBUG_p("len: %d, size: %d",(int)len,(int)size);
    /* FIXME!! validate size */
    if(!strncmp(data+8,"Pac.Pic.",8))
        return _handle_pacpic(h,data+16,len-16);
    h->info->error = "Unknown data type";
    return -1;
}


static int
_amos_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    switch(h->type){
    case T_ABK:
        switch(*(uint32_t*)data){
        case ID_SPRITE_BANK:
            return _handle_spritebank(h,data+4,len-4);
        case ID_ICON_BANK:
            return _handle_iconbank(h,data+4,len-4);
        case ID_MEMORYBANK:
            return _handle_memorybank(h,data+4,len-4);
        default:
            h->info->error = "Invalid header";
            return -1;
        }
    case T_PAC:
        return _handle_pacpic(h,data,len);
    }
    return -1;
}


static cairo_surface_t *
_amos_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->page[page].frame[frame].surface);
}


static void
_amos_get_duration(abydos_plugin_handle_t *h,int page,double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->page[0].frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _amos_get_mime_types,
        .new = _amos_new,
        .free = _amos_free,
        .create_from_data = _amos_create_from_data,
        .get_image_surface = _amos_get_image_surface,
        .get_duration = _amos_get_duration
    };
    return &p;
}
