/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <nil-amiga.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t page[2];
};


static const char **
_wb_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-amiga-icon",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_wb_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->info->page_count = 2;
    h->page[0].surface = NULL;
    h->page[1].surface = NULL;
    return h;
}


static void
_wb_free(abydos_plugin_handle_t *h)
{
    int i;
    for(i=0;i<h->info->page_count;++i)
        if(h->page[i].surface)
            cairo_surface_destroy(h->page[i].surface);
    free(h);
}


#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static int
_page_from_data(page_t *p, const char *data, const char *end, const uint8_t *palette)
{
    uint8_t *bitplanes[8];
    int depth;
    int stride;
    int i;
    p->width = get16(data+4);
    p->height = get16(data+6);
    depth = get16(data+8);
    stride = nil_amiga_plane_stride(p->width);
    for(i=0;i<depth;++i)
        bitplanes[i] = (uint8_t*)data + 20 + stride*p->height*i;
    p->surface = nil_cairo_surface_create_from_bitplanes(bitplanes, p->width, p->height, stride, depth, palette, 1<<depth, 3, NILE_FORMAT_r8g8b8, 0);
    return 20 + stride*p->height*depth;
}


static int
_wb_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    static const uint8_t palette1[3*4] = {0x55,0xAA,0xFF, 0xFF,0xFF,0xFF, 0x00,0x00,0x00, 0xFF,0x88,0x00};
    static const uint8_t palette2[3*8] = {0x95,0x95,0x95, 0x00,0x00,0x00, 0xFF,0xFF,0xFF, 0x3B,0x67,0xA2,
                                          0x7B,0x7B,0x7B, 0xAF,0xAF,0xAF, 0xAA,0x90,0x7C, 0xFF,0xA9,0x97};
    const uint8_t *palette;
    const char *end;
    int bytes;

    if(len<98)
        return -1;
    if(get16(data) != 0xe310 || data[2] != 0 || data[3] != 1)
        return -1;

    switch(data[45]){
    case 0: /* Old Style icon */
        h->info->pixel_ratio = 0.5;
        palette = palette1;
        break;
    case 1: /* NewIcons */
        palette = palette2;
        break;
    default:
        return -1;
    }

    end = data + len;
    data += 78 + (get32(data+66) ? 56 : 0);
    bytes = _page_from_data(&h->page[0], data, end, palette);
    if(bytes < 0)
        return -1;
    data += bytes;
    bytes = _page_from_data(&h->page[1], data, end, palette);
    if(bytes < 0)
        return -1;

    return 0;
}


static cairo_surface_t *
_wb_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


static void
_wb_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
    info->pixel_ratio = 0.5;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _wb_get_mime_types,
        .new = _wb_new,
        .free = _wb_free,
        .create_from_data = _wb_create_from_data,
        .get_image_surface = _wb_get_image_surface,
        .get_page_info = _wb_get_page_info
    };
    return &p;
}
