/*
 * ANSI graphics decoder
 *
 * Copyright (C) 2006-2020 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/** @file
  ANSI graphics decoder.

  Optional defines:
    ANSI_CPR                            Support Cursor Position Report
    ANSI_UNIX                           Assume UNiX (original) style ANSi
*/
#include <stdlib.h>                     /* malloc() */

#include "ansi_dec.h"
#include "term_if.h"

#define ANSI_CHECK_ERROR
#define ANSI_CPR
#define ANSI_MAX_CODE_LENGTH 255
#define ANSI_MAX_PARAMS 16


struct _AnsiDecoder{
    const TerminalInterface* term_if;
    Terminal* term;

    int incode;
    char code[ANSI_MAX_CODE_LENGTH+1];
    int codeptr;

    int save_cursorx;
    int save_cursory;

    AnsiDecoderCallback cpr_func;
    void* cpr_data;
};


#ifdef ANSI_CPR
static char* strcati(char* s,int i){
    if(i){
        /* FIXME!! reverse string */
        while(i){
            *(s++) = '0'+(i%10);
            i /= 10;
        }
    }else{
        *(s++)='0';
    }
    return s;
}
#endif /* ANSI_CPR */


/** Translate ANSI colors to EGA colors */
static int ansi2ega[8] = {0,4,2,6,1,5,3,7};


AnsiDecoder* ansi_decoder_new(const TerminalInterface* term_if,Terminal* term,AnsiDecoderCallback func,void* data){
    AnsiDecoder* dec;
    dec = (AnsiDecoder*)malloc(sizeof(AnsiDecoder));
    dec->term_if = term_if;
    dec->term = term;
    dec->incode = 0;
    dec->cpr_func = func;
    dec->cpr_data = data;
    return dec;
}


static void ansi_decoder_interpret_string(AnsiDecoder* dec,char* s){
    int param[ANSI_MAX_PARAMS];
    int paramc;
    int isparam;
    int ti1;
    int ti2;

    paramc = 0;
    isparam = 0;

    for(;*s;++s){
        if(*s>='0' && *s<='9'){
            isparam = 1;
            param[paramc] *= 10;
            param[paramc] += *s-'0';
        }else{
            switch(*s){
            case 27: /* ESC */
                paramc = 0;
                isparam = 0;
                break;
            case '[':
            case ';':
                if(isparam){
                    ++paramc;
                    isparam = 0;
                }
                param[paramc] = 0;
#ifdef ANSI_UNIX
            case '?':
            case '(':
#endif
                break;
            default: /* Perform command */
                if(isparam)
                    paramc++;
                switch(*s){
                case 'A': /* go up */
                    dec->term_if->up(dec->term,paramc?param[0]:1);
                    break;
                case 'B': /* go down */
                    dec->term_if->down(dec->term,paramc?param[0]:1);
                    break;
                case 'C': /* go right */
                    dec->term_if->right(dec->term,paramc?param[0]:1);
                    break;
                case 'D': /* go left */
                    dec->term_if->left(dec->term,paramc?param[0]:1);
                    break;
                case 'f':
                case 'H': /* gotoxy */
                    dec->term_if->gotoxy(dec->term,paramc>=2?param[1]-1:0,paramc?param[0]-1:0);
                    break;
                case 'h': /* select mode */
                case 'J':
#ifndef ANSI_UNIX
                    dec->term_if->gotoxy(dec->term,0,0);
#endif
                    if(2==param[0])
                        dec->term_if->clrscr(dec->term);
                    break;
                case 'K':
                /* case 'k': */
                    dec->term_if->clreol(dec->term);
                    break;
                case 'l': /* reset mode */
                    dec->term_if->set_attribute(dec->term,0x07);
                    break;
                case 'm': /* Set text attributes */
                    for(ti1=0;ti1<paramc;++ti1){
                        ti2 = param[ti1];
                        if(0==ti2){ /* all atributes off */
                            dec->term_if->set_attribute(dec->term,0x07);
                        }else if(1==ti2){ /* bald = highlight foreground */
                            dec->term_if->foreground(dec->term,dec->term_if->get_foreground(dec->term)|0x08);
                        /* }else if(2==ti2){ */ /* dim, not supported */
                        /* }else if(3==ti2){ */ /* italic, not supported */
                        /* }else if(4==ti2){ */ /* underline, not supported */
                        }else if(5==ti2){ /* blink ( = hilight background if iCE-color) */
                            dec->term_if->background(dec->term,dec->term_if->get_background(dec->term)|0x08);
                        /* }else if(6==ti2){ */ /* ???, not supported */
                        }else if(7==ti2){ /* reverse video */
                            dec->term_if->set_attribute(dec->term,(dec->term_if->get_foreground(dec->term)<<4) + dec->term_if->get_background(dec->term));
                        }else if(8==ti2){ /* invisible */
                            dec->term_if->set_attribute(dec->term,0x00);
                        }else if(ti2>=30 && ti2<=37){ /* set foreground */
                            dec->term_if->foreground(dec->term,(dec->term_if->get_foreground(dec->term) & 0x08)  + ansi2ega[ti2-30]);
                        }else if(ti2>=38 && ti2<=39){ /* default foreground ??? */
                            dec->term_if->foreground(dec->term,7);
                        }else if(ti2>=40 && ti2<=47){ /* set background */
                            dec->term_if->background(dec->term,(dec->term_if->get_background(dec->term) & 0x08) + ansi2ega[ti2-40]);
                        }else if(ti2>=48 && ti2<=49){ /* default background ??? */
                            dec->term_if->background(dec->term,0);
                        }
                    }
                    break;
#ifdef ANSI_CPR
                case 'n': /* Cursor Position Report */
                    if(6==param[0] && dec->cpr_func){
                        char ts1[12];
                        char* ts2;
                        ts2 = ts1;
                        *(ts2++) = 27;
                        *(ts2++) = '[';
                        ts2 = strcati(ts2,dec->term_if->get_cursory(dec->term)+1);
                        *(ts2++) = ';';
                        ts2 = strcati(ts2,dec->term_if->get_cursorx(dec->term)+1);
                        *(ts2++) = 'R';
                        dec->cpr_func(dec->cpr_data,ts1,ts2-ts1);
                    }
                    break;
#endif /* ANSI_CPR */
                case 's': /* save cursor position */
                    dec->save_cursorx = dec->term_if->get_cursorx(dec->term);
                    dec->save_cursory = dec->term_if->get_cursory(dec->term);
                    break;
                case 'u': /* restore curosr position */
                    dec->term_if->gotoxy(dec->term,dec->save_cursorx,dec->save_cursory);
                    break;
                }
            }
        }
    }
}



#ifdef ANSI_CHECK_ERROR
static void codeerror(AnsiDecoder* dec){
    dec->incode = 0;
}
#endif


int ansi_decoder_write(AnsiDecoder* dec,char ch){
    switch(dec->incode){
    case 0: /* expect ESC */
        if(27==ch){
            dec->codeptr = 0;
            dec->code[dec->codeptr++] = ch;
            dec->incode = 1;
        }else{
            /* This character is not part of an ansi code */
            return 0;
        }
        break;
    case 1: /* expect [ */
        dec->code[dec->codeptr++] = ch;
        dec->incode = 3;
#ifdef ANSI_CHECK_ERROR
        if(ch!='[')
            codeerror(dec);
#endif
        break;
    case 2: /* expect digit */
#ifdef ANSI_CHECK_ERROR
        dec->code[dec->codeptr++] = ch;
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else{
            codeerror(dec);
        }
        if(dec->codeptr>ANSI_MAX_CODE_LENGTH)
            --dec->codeptr;
        break;
#endif
    case 3: /* excpect digit or terminator */
#ifdef ANSI_CHECK_ERROR
        dec->code[dec->codeptr++] = ch;
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else if((ch>='A' && ch<='Z') || (ch>='a' && ch<='z')){
            dec->code[dec->codeptr] = 0; /* terminate string */
            ansi_decoder_interpret_string(dec,dec->code);
            dec->incode = 0;
        }else{
            codeerror(dec);
        }
        if(dec->codeptr>ANSI_MAX_CODE_LENGTH)
            --dec->codeptr;
        break;
#endif
    case 4: /* excpect digit or separator or terminator */
        dec->code[dec->codeptr++] = ch;
        if(ch>='0' && ch<='9'){
            dec->incode = 4;
        }else if(';'==ch){
            dec->incode = 2;
        }else if((ch>='A' && ch<='Z') || (ch>='a' && ch<='z')){
            dec->code[dec->codeptr] = 0; /* terminate string */
            ansi_decoder_interpret_string(dec,dec->code);
            dec->incode = 0;
#ifdef ANSI_CHECK_ERROR
        }else{
            codeerror(dec);
#endif
        if(dec->codeptr>ANSI_MAX_CODE_LENGTH)
            --dec->codeptr;
        }
    }
    return 1;
}
