/*
 * Dynamically expanding terminal emulator
 *
 * Copyright (C) 2006, 2010, 2011 Magnus Bergman
 *
 * Author: Magnus Bergman <fabel@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>                     /* malloc(), free() */

#include "debug.h"
#include "cons_if.h"
#include "term_dyn.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

struct _Terminal{
    const ConsoleInterface* cons_if;
    Console* cons;

    int x;
    int y;
    int attr;
    int fill_attr;
};


static void term_dyn_resize(Terminal* term){
#if 1
    int extend = term->y+1 - term->cons_if->height(term->cons);
    if(extend>0)
        term->cons_if->extend(term->cons,extend,term->fill_attr);
#else
    //int fill_start;
    //int fill_end;
    //short* dst;
    //int cell;
    //int i;
    if(term->y >= term->cons_if.height(term->cons)){
        fill_start = term->cons_if.height(term->cons) * term->cons_if.width(term->cons);
        term->cons_if_resize(term->cons,80,term->y+1);
        fill_end = term->cons_if.height(term->cons) * term->cons_if.width(term->cons);
        DEBUG_p("resizing to %i lines",term->scrn->height);
        if(term->scrn->screen){
            term->scrn->screen = realloc(term->scrn->screen,term->scrn->width * term->scrn->height * sizeof(unsigned short));
        }else{
            term->scrn->screen = malloc(term->scrn->width * term->scrn->height * sizeof(unsigned short));
        }

        dst =
        cell = ' ' | term->fill_attr << 8;
        for(i=fill_start;i<fill_end;++i)
            term->scrn->screen[i] = cell;
    }
#endif
}


Terminal* term_dyn_new(const ConsoleInterface* cons_if,Console* cons){
    Terminal* term;
    term = (Terminal*)malloc(sizeof(Terminal));
    term->cons_if = cons_if;
    term->cons = cons;
    term->x = 0;
    term->y = 0;
    term->attr = 0x07;
    term->fill_attr = 0x07;
    return term;
}



void term_dyn_gotoxy(Terminal* term,int x,int y){
    term->x = MIN(x,term->cons_if->width(term->cons)-1);
    term->y = y;
}


void term_dyn_up(Terminal* term,int steps){
    term->y = MAX(term->y-steps,0);
}


void term_dyn_down(Terminal* term,int steps){
    term->y += steps;
}


void term_dyn_left(Terminal* term,int steps){
    term->x = MAX(term->x-steps,0);
}


void term_dyn_right(Terminal* term,int steps){
    term->x = MIN(term->x+steps,term->cons_if->width(term->cons)-1);
}



void term_dyn_foreground(Terminal* term,int foreground){
    DEBUG_p("begin");
    term->attr = (term->attr & 0xf0) | foreground;
}


void term_dyn_background(Terminal* term,int background){
    DEBUG_p("begin");
    term->attr = (term->attr & 0x0f) | (background << 4);
}


void term_dyn_set_attribute(Terminal* term,int attr){
    DEBUG_p("begin");
    term->attr = attr & 0xff;
}



void term_dyn_write(Terminal* term,int ch){
    switch(ch){
    case 8: /* backspace */
        term_dyn_left(term,1);
        break;
    case 9: /* tab */
        term_dyn_right(term,8-term->x%8);
        break;
    case 10: /* line feed */
        ++term->y;
#if 0
        term_dyn_clreol(term);
#endif
        break;
    case 13: /* carrage return */
        term->x = 0;
        break;
    default:
        term_dyn_resize(term);
        term->cons_if->charxya(term->cons,term->x,term->y,term->attr,ch);
        if(++term->x >= term->cons_if->width(term->cons)){
            ++term->y;
            term->x = 0;
        }
    }
}


void term_dyn_clrscr(Terminal* term){
    //int cell;
    //int i;
    int x;
    int y;

    DEBUG_p("begin");
    term->fill_attr = term->attr;
#if 1
    for(y=0;y<term->cons_if->height(term->cons);++y)
        for(x=0;x<term->cons_if->width(term->cons);++x)
            term->cons_if->charxya(term->cons,x,y,term->attr,' ');
#else
    if(!term->scrn->height)
        return;
    cell = ' ' | (term->attr << 8);
    for(i=0;i<term->x + term->y * term->scrn->width;++i)
        term->scrn->screen[i] = cell;
#endif
}


void term_dyn_clreol(Terminal* term){
    //int cell;
    int x;
    //cell = ' ' | (term->attr << 8);
    term_dyn_resize(term);
    for(x=term->x;x<term->cons_if->width(term->cons);++x)
        term->cons_if->charxya(term->cons,x,term->y,term->attr,' ');
}



int term_dyn_get_foreground(Terminal* term){
    return term->attr & 0x0f;
}


int term_dyn_get_background(Terminal* term){
    return term->attr >> 4;
}


int term_dyn_get_cursorx(Terminal* term){
    return term->x;
}


int term_dyn_get_cursory(Terminal* term){
    return term->y;
}



const TerminalInterface term_dyn_if = {
    .gotoxy = term_dyn_gotoxy,
    .up = term_dyn_up,
    .down = term_dyn_down,
    .left = term_dyn_left,
    .right = term_dyn_right,

    .foreground = term_dyn_foreground,
    .background = term_dyn_background,
    .set_attribute = term_dyn_set_attribute,

    .write = term_dyn_write,
    .clrscr = term_dyn_clrscr,
    .clreol = term_dyn_clreol,

    .get_foreground = term_dyn_get_foreground,
    .get_background = term_dyn_get_background,
 /* term_dyn_get_attribute, */
    .get_cursorx = term_dyn_get_cursorx,
    .get_cursory = term_dyn_get_cursory
};
