/*
 * Tiny Stuff image loader
 *
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */

#include <abydos-plugin.h>
#include <nil-atarist.h>
#include <nil-cycle.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "atarist.h"


typedef cairo_surface_t*(*decode_image_func_t)(uint16_t*, atarist_cairo_color_t*);

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    int resolution;
    decode_image_func_t decode_image;
    atarist_cairo_color_t cairo_palette[16];
    uint16_t atari_image[16000];
    nil_cycle_t cycle;
    cairo_surface_t *cached_surface;
};


static const char **
_tiny_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-tiny-stuff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tiny_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->cached_surface = NULL;
    return h;
}


static void
_tiny_free(abydos_plugin_handle_t *h)
{
    if(h->cached_surface)
        cairo_surface_destroy(h->cached_surface);
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get16be(a) (*(uint16_t*)(a))


static int
_unpack(const uint8_t *src, const uint8_t *val, const uint8_t *end, uint16_t *dst)
{
    int bitplane;
    int count = 0;
    for(bitplane=0; bitplane < 4; ++bitplane){
        int x;
        for(x=bitplane; x < 80; x += 4){
            int offs;
            for(offs = x; offs < 16000; offs += 80){
                int b;
                if(count<=0){
                    if(src >= end)
                        return -1;
                    b = *src++;
                    if (b < 128) {
                        if(0==b || 1==b){
                            if(src+2 > end)
                                return -1;
                            count = *src++;
                            count <<= 8;
                            count |= *src++;
                        }else{
                            count = b;
                        }
                        if(1==b){
                            b = -1;
                        }else{
                            if(val+2 > end)
                                return -1;
                            b = get16be(val);
                            val += 2;
                        }
                    }else{
                        count = 256-b;
                        b = -1;
                    }
                }
                if(b>=0){
                    dst[offs] = b;
                }else{
                    if(val+2 > end)
                        return -1;
                    dst[offs] = get16be(val);
                    val += 2;
                }
                --count;
            }
        }
    }
    return 0;
}


static int
_setup_resolution(abydos_plugin_handle_t *h)
{
    switch(h->resolution){
    case 0:
        h->info->width = 320;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = atarist_decode_image_0;
        break;
    case 1:
        h->info->width = 640;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/12;
        h->decode_image = atarist_decode_image_1;
        break;
    case 2:
        h->info->width = 640;
        h->info->height = 400;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = atarist_decode_image_2;
        break;
    default:
        return -1;
    }
    return 0;
}


static int
_tiny_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const uint8_t *p = (const uint8_t*)data;
    const uint8_t *end;
    uint16_t *palette;
    int control_len;
    int value_len;

    if(len < 42)
        return -1;

    end = p+len;
    h->resolution = get8(p);
    if(h->resolution > 2){
        int ticks;
        if(h->resolution > 5)
            return -1;
        h->resolution -= 3;
        h->cycle.first = p[1] >> 4;
        h->cycle.last = p[1] & 15;
        ticks = p[2];
        if(ticks){
            if(ticks >= 128){
                h->cycle.direction = NIL_CYCLE_LEFT;
                ticks = 256-ticks;
            }else{
                h->cycle.direction = NIL_CYCLE_RIGHT;
            }
            h->cycle.duration = (double)ticks / 60;
            h->info->frame_count = nil_cycle_total_frames(&h->cycle,1);
        }else{
            h->cycle.direction = NIL_CYCLE_OFF;
        }
        p += 4;
    }
    palette = (uint16_t*)&p[1];

    control_len = get16(p+33);
    value_len = get16(p+35)*2;

    p += 37;

    if(p + control_len + value_len > end)
        return -1;

    if(_unpack(p, p+control_len, end, h->atari_image)<0)
        return -1;

    if(_setup_resolution(h) < 0)
        return -1;
    atarist_decode_palette(palette, h->cairo_palette);
    if(h->info->frame_count <= 1){
        h->cached_surface = h->decode_image(h->atari_image, h->cairo_palette);
    }
    return 0;
}


static cairo_surface_t *
_tiny_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    if(h->info->frame_count > 1){
        atarist_cairo_color_t cairo_palette[16];
        memcpy(cairo_palette,h->cairo_palette,16*sizeof(atarist_cairo_color_t));
        nil_cycle_apply_time(&h->cycle, time, cairo_palette, sizeof(atarist_cairo_color_t));
        return h->decode_image(h->atari_image, cairo_palette);
    }else{
        return cairo_surface_reference(h->cached_surface);
    }
}


static void
_tiny_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->cycle.duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tiny_get_mime_types,
        .new = _tiny_new,
        .free = _tiny_free,
        .create_from_data = _tiny_create_from_data,
        .get_image_surface = _tiny_get_image_surface,
        .get_duration = _tiny_get_duration
    };
    return &p;
}
