/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdint.h>                     /* uint8_t */
#include <stdlib.h>
#include <string.h>

#include "nil-c64.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_geop_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-geopaint",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_geop_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_geop_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get8(a) (*(uint8_t*)(a))

static int
_unpack(uint8_t *dst, const uint8_t *dstend, const char *src, const char *srcend)
{
    const char *p = src;
    for(;;){
        int b;
        if(p >= srcend)
            return -1;
        b = get8(p++);
        if(b <= 0){
            DEBUG_p("%d bytes missing", (int)(dstend - dst));
            return p - src;
        }else if(b <= 0x3f){
            if(dst + b > dstend || p + b > srcend)
                return -1;
            memcpy(dst, p, b);
            dst += b;
            p += b;
        }else if(b <= 0x40){
            return -1;
        }else if(b <= 0x7f){
            b -= 0x40;
            if(dst + 8*b > dstend || p + 8 > srcend)
                return -1;
            while(b-->0){
                memcpy(dst, p, 8);
                dst += 8;
            }
            p += 8;
        }else if(b <= 0x80){
            return -1;
        }else{
            int d;
            b -= 0x80;
            if(dst + b > dstend || p >= srcend)
                return -1;
            d = get8(p++);
            memset(dst, d, b);
            dst += b;
        }
    }
}


static int
_draw_line(uint32_t *dst, int stride, const uint8_t *src, const uint8_t *colors, const uint32_t *palette){
    int i;
    for(i=0; i<80; ++i){
        int bg = palette[*colors & 15];
        int fg = palette[*colors >> 4];
        int y;
        for(y=0; y<8; ++y){
            int b = *src++;
            int x;
            for(x=0; x<8; ++x){
                dst[x + y*stride] = b & 128 ? fg : bg;
                b <<= 1;
            }
        }
        dst += 8;
        ++colors;
    }
    return 0;
}


static int
_geop_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data+len;
    uint8_t buffer[1448];
    uint32_t *dst;
    int stride;
    uint32_t palette[16];

    if(len < 0x2fa)
        return -1;
    data += 0x2fa;

    h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, 640, 720);
    dst = (uint32_t*)cairo_image_surface_get_data(h->surface);
    stride = cairo_image_surface_get_stride(h->surface) / sizeof(uint32_t);

    nil_c64_copy_palette_colors((uint8_t*)palette, 4, NIL_CAIRO_FORMAT_RGB24, NIL_C64_COLODORE, 16);
    for(;;){
        int bytes = _unpack(buffer, buffer+1448, data, end);
        DEBUG_p("%d bytes unpacked", bytes);
        if(bytes <= 0)
            break;
        bytes += 253 - (bytes + 253) % 254;
        data += bytes;
        _draw_line(dst, stride, buffer, buffer + 1288, palette);
        dst += 8 * stride;
        _draw_line(dst, stride, buffer + 640, buffer + 1368, palette);
        dst += 8 * stride;
    }
    cairo_surface_mark_dirty(h->surface);
    h->info->width = 640;
    h->info->height = 720;
    return 0;
}


static cairo_surface_t *
_geop_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _geop_get_mime_types,
        .new = _geop_new,
        .free = _geop_free,
        .create_from_data = _geop_create_from_data,
        .get_image_surface = _geop_get_image_surface
    };
    return &p;
}
