/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <string.h>                     /* memcpy() */

#include "nile-blit.h"
#include "nil-c64.h"


/* Colodore */
static const uint8_t _colodore_palette[16*3] = {
    0x00,0x00,0x00, /* black */
    0xff,0xff,0xff, /* white */
    0x81,0x33,0x38, /* dark red */
    0x75,0xce,0xc8, /* cyan */
    0x8e,0x3c,0x97, /* purple */
    0x56,0xac,0x4d, /* dark green */
    0x2e,0x2c,0x9b, /* dark blue */
    0xed,0xf1,0x71, /* yellow */
    0x8e,0x50,0x29, /* orange */
    0x55,0x38,0x00, /* brown */
    0xc4,0x6c,0x71, /* light red */
    0x4a,0x4a,0x4a, /* dark gray */
    0x7b,0x7b,0x7b, /* mid gray */
    0xa9,0xff,0x9f, /* light green */
    0x70,0x6d,0xeb, /* light blue */
    0xb2,0xb2,0xb2  /* light gray */
};

/* GoDot */
static const uint8_t _godot_palette[16*3] = {
    0x00,0x00,0x00, /* black */
    0xff,0xff,0xff, /* white */
    0x88,0x00,0x00, /* dark red */
    0xaa,0xff,0xee, /* cyan */
    0xcc,0x44,0xcc, /* purple */
    0x00,0xcc,0x55, /* dark green */
    0x00,0x00,0xaa, /* dark blue */
    0xee,0xee,0x77, /* yellow */
    0xdd,0x88,0x55, /* orange */
    0x66,0x44,0x00, /* brown */
    0xff,0x77,0x77, /* light red */
    0x33,0x33,0x33, /* dark gray */
    0x77,0x77,0x77, /* mid gray */
    0xaa,0xff,0x66, /* light green */
    0x00,0x88,0xff, /* light blue */
    0xbb,0xbb,0xbb  /* light gray */
};

static const uint8_t _pepto_palette[16*3] = {
    0x00,0x00,0x00, /* black */
    0xff,0xff,0xff, /* white */
    0x68,0x37,0x2b, /* dark red */
    0x70,0xa4,0xb2, /* cyan */
    0x6f,0x3d,0x86, /* purple */
    0x58,0x8d,0x43, /* dark green */
    0x35,0x28,0x79, /* dark blue */
    0xb8,0xc7,0x6f, /* yellow */
    0x6f,0x4f,0x25, /* orange */
    0x43,0x39,0x00, /* brown */
    0x9a,0x67,0x59, /* light red */
    0x44,0x44,0x44, /* dark gray */
    0x6c,0x6c,0x6c, /* mid gray */
    0x9a,0xd2,0x84, /* light green */
    0x6c,0x5e,0xb5, /* light blue */
    0x95,0x95,0x95  /* light gray */
};


void
nil_c64_copy_palette_colors(uint8_t *dst, int dst_bpc, nile_format_t dst_format, nil_c64_palette_t pal, int count)
{
    const uint8_t *src;
    if(count > 16)
        count = 16;
    switch(pal){
    case NIL_C64_COLODORE:
        src = _colodore_palette;
        break;
    case NIL_C64_GODOT:
        src = _godot_palette;
        break;
    case NIL_C64_PEPTO:
        src = _pepto_palette;
        break;
    default:
        return;
    }
    nile_blit_colors_rgb(dst, dst_bpc, dst_format, src, 3, NILE_FORMAT_r8g8b8, count);
}


int
nil_c64_palette(nile_t *ni,nil_c64_palette_t pal)
{
    if(!ni->bpc)
        ni->bpc = 3;
    if(!ni->format)
        ni->format = NILE_FORMAT_r8g8b8;
    nile_ensure_palette(ni, 16, ni->bpc);
    nil_c64_copy_palette_colors(ni->palette, ni->bpc, ni->format, pal, 16);
    return 0;
}


static uint8_t
_get_color(const uint8_t *screen, const uint8_t *color, int background,int x,int y,int c)
{
    switch(c){
    default:
    case 0:return background;
    case 1:return screen[x+40*y]>>4;
    case 2:return screen[x+40*y]&15;
    case 3:return color[x+40*y]&15;
    }
}


int
nil_c64_decode(nile_t *ni,const uint8_t *bitmap, const uint8_t *screen, const uint8_t *color, int background){
    uint8_t *dst = ni->pixels;
    int stride = nile_get_stride(ni);
    int y;
    const uint8_t *src = bitmap;
    for(y=0;y<25;++y){
        int x;
        for(x=0;x<40;++x){
            int i;
            for(i=0;i<8;++i){
                int pix = *src++;
                dst[x*4+0+(y*8+i)*stride] = _get_color(screen,color,background,x,y,(pix>>6) & 3);
                dst[x*4+1+(y*8+i)*stride] = _get_color(screen,color,background,x,y,(pix>>4) & 3);
                dst[x*4+2+(y*8+i)*stride] = _get_color(screen,color,background,x,y,(pix>>2) & 3);
                dst[x*4+3+(y*8+i)*stride] = _get_color(screen,color,background,x,y,(pix>>0) & 3);
            }
        }
    }
    return 0;
}


int
nil_c64_vram_decode(nile_t *ni,const nil_c64_vram_t *vram)
{
#if 1
    return nil_c64_decode(ni, vram->bitmap, vram->screen, vram->color, vram->background);
#else
    uint8_t *dst = ni->pixels;
    int stride = nile_get_stride(ni);
    int y;
    const uint8_t *src = vram->bitmap;
    for(y=0;y<25;++y){
        int x;
        for(x=0;x<40;++x){
            int i;
            for(i=0;i<8;++i){
                int pix = *src++;
                dst[x*4+0+(y*8+i)*stride] = _get_color(vram,x,y,(pix>>6) & 3);
                dst[x*4+1+(y*8+i)*stride] = _get_color(vram,x,y,(pix>>4) & 3);
                dst[x*4+2+(y*8+i)*stride] = _get_color(vram,x,y,(pix>>2) & 3);
                dst[x*4+3+(y*8+i)*stride] = _get_color(vram,x,y,(pix>>0) & 3);
            }
        }
    }
    return 0;
#endif
}
