/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <fitsio.h>
#include <nile-blit-comp.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>
//#include <stdio.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_cfits_get_mime_types()
{
    static const char *mime_types[] = {
        "application/fits", "image/fits",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_cfits_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_cfits_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#if 0
static int
_cfits_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
}
#endif


static cairo_surface_t *
_surface_from_image(uint8_t *img, int width, int height, int comps, int prec)
{
#if 1
    nile_format_t format;
    cairo_format_t cairo_format;
    cairo_surface_t *surface;
    uint8_t *dst;
    int dst_stride;
    uint8_t *src[4];
    int src_bpc;
    int i;
    nile_blit_t blit;

    switch(prec){
    case 8:
        switch(comps){
        case 1: /* gray scale */
            format = NILE_FORMAT_g8;
            break;
        case 2: /* gray scale + alpha */
            format = NILE_FORMAT_g8a8;
            break;
        case 3: /* RGB */
            format = NILE_FORMAT_r8g8b8;
            break;
        case 4: /* RGB + alpha */
            format = NILE_FORMAT_r8g8b8a8;
            break;
        default:
            return NULL;
        }
        src_bpc = 1;
        break;
    case 16:
        switch(comps){
        case 1: /* gray scale */
            format = NILE_FORMAT_ne_g16;
            break;
        case 2: /* gray scale + alpha */
            format = NILE_FORMAT_ne_g16a16;
            break;
        case 3: /* RGB */
            format = NILE_FORMAT_ne_r16g16b16;
            break;
        case 4: /* RGB + alpha */
            format = NILE_FORMAT_ne_r16g16b16a16;
            break;
        default:
            return NULL;
        }
        src_bpc = 2;
        break;
    case 32:
        switch(comps){
        case 1: /* gray scale */
            format = NILE_FORMAT_ne_g32;
            break;
        case 2: /* gray scale + alpha */
            format = NILE_FORMAT_ne_g32a32;
            break;
        case 3: /* RGB */
            format = NILE_FORMAT_ne_r32g32b32;
            break;
        case 4: /* RGB + alpha */
            format = NILE_FORMAT_ne_r32g32b32a32;
            break;
        default:
            return NULL;
        }
        src_bpc = 4;
        break;
    default:
        return NULL;
    }

    for(i=0; i<comps; ++i){
        src[i] = img + i * width * height * src_bpc;
    }

    cairo_format = nil_cairo_format_from_nile(format);
    surface = cairo_image_surface_create(cairo_format, width, height);
    dst = cairo_image_surface_get_data(surface);
    dst_stride = cairo_image_surface_get_stride(surface);

    nile_blit_init_from_components_rgb(&blit, nil_cairo_format_to_bpc(cairo_format), nil_cairo_format_to_nile(cairo_format),
                                              src_bpc, format);
    nile_blit_from_components(&blit, dst, dst_stride, src, width * src_bpc, width, height, NILE_FLIPPED);
    cairo_surface_mark_dirty(surface);
    return surface;
#else
    int stride;
    uint8_t *comp[4];
    nile_format_t format;
    int bpc;
    cairo_surface_t *surface;
    //nile_blit_t blit;
    nile_t dst;
    nile_t src;
    int i;
    switch(channels){
    case 3:
        switch(prec){
        case 8:
            stride = width;
            for(i=0; i<channels; ++i)
                comp[i] = img + i * stride * height;
            return abydos_surface_from_comps(comp, width, height, stride, channels, prec, NILE_FLIPPED);
            format = NILE_FORMAT_r8g8b8;
            bpc = 1;
            break;
        default:
            return NULL;
        }
        break;
    default:
        return NULL;
    }
    surface = cairo_image_surface_create(nil_cairo_format_from_nile(format), width, height);
    nil_cairo_from_surface(&dst, surface);
    nile_copy_identity(&src, &dst);
    src.format = bpc;
#endif
}


static int
_cfits_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    fitsfile *ff;
    int ret = -1;
    int status = 0;
    int bitpix;
    //int naxis;
    int anynul;
    int channeltype;
    int channelsize;
    int dim_count;
    long dim[3];
    long fpixel[3] = {1,1,1};
    void *img = NULL;
    int bpc;

    // Open the FITS file
    if(fits_open_file(&ff, filename, READONLY, &status)){
        return -1;
    }

    // Read the image dimensions and data type
    if(fits_get_img_param(ff, 3, &bitpix, &dim_count, dim, &status)){
        goto out;
    }

    switch(bitpix){
    case 1:
        channeltype = TBIT;
        channelsize = 1;
        break;
    case 8:
        channeltype = TBYTE;
        channelsize = sizeof(uint8_t);
        break;
    case 16:
        channeltype = TSHORT;
        channelsize = sizeof(uint16_t);
        break;
    case 32:
        channeltype = TINT;
        channelsize = sizeof(uint32_t);
        break;
    case 64:
        channeltype = TFLOAT;
        channelsize = sizeof(float);
        break;
    default:
        goto out;
    }

    // Allocate memory for the image data
    //npixels = dim[0] * dim[1];
    bpc = dim[2] * channelsize;
    img = malloc(dim[0] * dim[1] * bpc);

    // Read the image data
    if(fits_read_pix(ff, channeltype, fpixel, dim[0] * dim[1] * dim[2], NULL, img, &anynul, &status)){
        goto out;
        return -1;
    }

#if 1
    h->surface = _surface_from_image(img, dim[0], dim[1], dim[2], bitpix);
#else
    switch(dim[2]){
    case 1:
        goto out;
        break;
    case 2:
        goto out;
        break;
    case 3:
        switch(channeltype){
        case TBYTE:
            //h->surface = nil_cairo_surface_create_from_direct(img, dim[0], dim[1], 0, bpc, NILE_FORMAT_r8g8b8, 0);
            //nile_blit_init_rgb(&blit,
            break;
        default:
            goto out;
        }
        break;
    case 4:
        goto out;
        break;
    }
#endif

    // Print the image dimensions and the first pixel value
    DEBUG_p("image dimensions: %ld x %ld\n", dim[0], dim[1]);
    //DEBUG_p("first pixel value: %f\n", image_data[0]);
    h->info->width = dim[0];
    h->info->height = dim[1];
    ret = 0;

out:
    fits_close_file(ff, &status);
    if(img)
        free(img);

    return ret;
}


static cairo_surface_t *
_cfits_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _cfits_get_mime_types,
        .new = _cfits_new,
        .free = _cfits_free,
        //.create_from_data = _cfits_create_from_data,
        .create_from_file = _cfits_create_from_file,
        .get_image_surface = _cfits_get_image_surface
    };
    return &p;
}
