/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>                     /* memcpy() */

#include "cgmread.h"
#include "cgmread-binary.h"


#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif

#define get8s(d) *(int8_t*)(d)
#define get8u(d) *(uint8_t*)(d)
#define get16s(d) (int16_t)be16toh(*(int16_t*)(d))
#define get16u(d) be16toh(*(uint16_t*)(d))
#define get32s(d) (int32_t)be16toh(*(int32_t*)(d))
#define get32u(d) be16toh(*(uint32_t*)(d))

int _get_int_x(cgm_read_t *r, int size)
{
    int v;
    if(r->data + size > r->end)
        return 0;
    switch(size){
    case 1:
        v = get8s(r->data);
        break;
    case 2:
        v = get16s(r->data);
        break;
    case 3:
        v = (*(int8_t*)&r->data[0]) << 16 | r->data[1] << 8 | r->data[2];
        break;
    case 4:
        v = get32s(r->data);
        break;
    default:
        return 0;
    }
    r->data += size;
    return v;
}


unsigned int _get_uint_x(cgm_read_t *r, int size)
{
    unsigned int v;
    if(r->data + size > r->end)
        return 0;
    switch(size){
    case 1:
        v = get8u(r->data);
        break;
    case 2:
        v = get16u(r->data);
        break;
    case 3:
        v = r->data[0] << 16 | r->data[1] << 8 | r->data[2];
        break;
    case 4:
        v = get32u(r->data);
        break;
    default:
        return 0;
    }
    r->data += size;
    return v;
}


static void
_swap4(uint8_t *dst, const uint8_t *src)
{
    int i;
    for(i=0;i<4;++i)
        dst[i] = src[i^3];
}


static void
_swap8(uint8_t *dst, const uint8_t *src)
{
    int i;
    for(i=0;i<8;++i)
        dst[i] = src[i^7];
}


double _get_real_x(cgm_read_t *r, real_type_t type, int size)
{
    const void*   pPtr;
    uint8_t   aBuf[8];
    double  nRetValue;
    double  fDoubleBuf;
    float   fFloatBuf;
#if __BYTE_ORDER == __BIG_ENDIAN
    const int bCompatible = 1;
#else
    const int bCompatible = 0;
#endif

    if(r->data + size > r->end)
        return 0;

    if ( bCompatible )
    {
        pPtr = r->data;
    }
    else
    {
        if ( size == 4 )
            _swap4(aBuf, r->data);
        else
            _swap8(aBuf, r->data);
        pPtr = &aBuf;
    }
    if ( type == RT_FLOAT )
    {
        if ( size == 4 )
        {
            memcpy( &fFloatBuf, pPtr, 4 );
            nRetValue = fFloatBuf;
        }
        else
        {
            memcpy( &fDoubleBuf, pPtr, 8 );
            nRetValue = fDoubleBuf;
        }
    }
    else // ->RT_FIXED
    {
        long    nVal;
        const int nSwitch = bCompatible ? 0 : 1 ;
        if ( size == 4 )
        {
            const uint16_t* pShort = pPtr;
            nVal = pShort[ nSwitch ];
            nVal <<= 16;
            nVal |= pShort[ nSwitch ^ 1 ];
            nRetValue = nVal;
            nRetValue /= 65536;
        }
        else
        {
            const int32_t* pLong = pPtr;
            nRetValue = ABS( pLong[ nSwitch ] );
            nRetValue *= 65536;
            nVal = pLong[ nSwitch ^ 1 ];
            nVal >>= 16;
            nRetValue += nVal;
            if ( pLong[ nSwitch ] < 0 )
            {
                nRetValue = -nRetValue;
            }
            nRetValue /= 65536;
        }
    }
    r->data += size;
    return nRetValue;
}


static int
_binary_eoe(cgm_read_t *r)
{
    return r->data >= r->end;
}

/* --- Basic types --- */

static int
_binary_get_int(cgm_read_t *r)
{
    return _get_int_x(r, r->int_precision);
}


static unsigned int
_binary_get_index(cgm_read_t *r)
{
    return _get_int_x(r, r->index_precision);
}


static unsigned int
_binary_get_color_channel(cgm_read_t *r)
{
    return _get_uint_x(r, r->color_precision);
}


static unsigned int
_binary_get_color_index(cgm_read_t *r)
{
    return _get_int_x(r, r->color_index_precision);
}


static double
_binary_get_real(cgm_read_t *r)
{
    return _get_real_x(r, r->real_type, r->real_size);
}


static const char *
_binary_get_string(cgm_read_t *r, int *len)
{
    const uint8_t *s;
    *len = _get_uint_x(r, 1);
    s = r->data;
    if(s + *len > r->end)
        *len = r->end - s;
    r->data += *len;
    return (char*)s;
}


static double
_binary_get_vdc(cgm_read_t *r)
{
    switch(r->vdc_type){
    case VDC_INT:
        return _get_int_x(r, r->vdc_int_precision);
    case VDC_REAL:
        return _get_real_x(r, r->vdc_real_type, r->vdc_real_size);
    }
    /* not reached */
    return 0;
}


static int
_binary_get_int_precision_bytes(cgm_read_t *r)
{
    int v;
    if(r->data + 2 > r->end)
        return 0;
    v = be16toh(*(uint16_t*)r->data);
    r->data += 2;
    return v / 8;
}

/* --- Enumeration types --- */

int _get_enum(cgm_read_t *r)
{
    int v;
    if(r->data + 2 > r->end)
        return 0;
    v = be16toh(*(uint16_t*)r->data);
    r->data += 2;
    return v;
};


static vdc_type_t
_binary_get_vdc_type(cgm_read_t *r)
{
    return (vdc_type_t)_get_enum(r);
}


static ScalingMode
_binary_get_scaling_mode(cgm_read_t *r)
{
    return (ScalingMode)_get_enum(r);
}


static DeviceViewPortMode
_binary_get_device_view_port_mode(cgm_read_t *r)
{
    return (DeviceViewPortMode)_get_enum(r);
}


static DeviceViewPortMap
_binary_get_device_view_port_map(cgm_read_t *r)
{
    return (DeviceViewPortMap)_get_enum(r);
}


DeviceViewPortMapH
_binary_get_device_view_port_map_h(cgm_read_t *r)
{
    return (DeviceViewPortMapH)_get_enum(r);
}


static DeviceViewPortMapV
_binary_get_device_view_port_map_v(cgm_read_t *r)
{
    return (DeviceViewPortMapV)_get_enum(r);
}


static ClipIndicator
_binary_get_clip_indicator(cgm_read_t *r)
{
    return (ClipIndicator)_get_enum(r);
}


static ColorSelectionMode
_binary_get_color_selection_mode(cgm_read_t *r)
{
    return (ColorSelectionMode)_get_enum(r);
}


static close_type_t
_binary_get_close_type(cgm_read_t *r)
{
    return (close_type_t)_get_enum(r);
}


static void
_binary_get_char_set_type(cgm_read_t *r)
{
    return (void)_get_enum(r);
}


static CharacterCodingA
_binary_get_character_coding_a(cgm_read_t *r)
{
    return (CharacterCodingA)_get_enum(r);
}


static TextPrecision
_binary_get_text_precision(cgm_read_t *r)
{
    return (TextPrecision)_get_enum(r);
}


static TextPath
_binary_get_text_path(cgm_read_t *r)
{
    return (TextPath)_get_enum(r);
}


static TextAlignmentH
_binary_get_text_alignment_h(cgm_read_t *r)
{
    return (TextAlignmentH)_get_enum(r);
}


static TextAlignmentV
_binary_get_text_alignment_v(cgm_read_t *r)
{
    return (TextAlignmentV)_get_enum(r);
}


static UnderlineMode
_binary_get_underline_mode(cgm_read_t *r)
{
    return (UnderlineMode)_get_enum(r);
}


static FinalFlag
_binary_get_final_flag(cgm_read_t *r)
{
    return (FinalFlag)_get_enum(r);
}


static SpecMode
_binary_get_spec_mode(cgm_read_t *r)
{
    return (SpecMode)_get_enum(r);
}


static EdgeVisibility
_binary_get_edge_visibility(cgm_read_t *r)
{
    return (EdgeVisibility)_get_enum(r);
}


static edge_flag_t
_binary_get_edge_flag(cgm_read_t *r)
{
    return (edge_flag_t)_get_enum(r);
}


static MarkerType
_binary_get_marker_type(cgm_read_t *r)
{
    return (MarkerType)_get_enum(r);
}


static Transparency
_binary_get_transparency(cgm_read_t *r)
{
    return (Transparency)_get_enum(r);
}


static interior_style_t
_binary_get_interior_style(cgm_read_t *r)
{
    return (interior_style_t)_get_enum(r);
}


static int
_binary_get_asf_mask(cgm_read_t *r)
{
    return 1<<_get_enum(r);
}


static asf_value_t
_binary_get_asf_value(cgm_read_t *r)
{
    return (asf_value_t)_get_enum(r);
}



static ssize_t
_handle_entry(cgm_read_t *r, const uint8_t *start, size_t len, cgm_handler_t userfunc, void *userdata)
{
    const uint8_t *data = start;
    const uint8_t *end = data + len;
    cgm_element_t e;

    if(data + 2 > end)
        return -1;
    int mnEscape = be16toh(*(uint32_t*)data);
    data += 2;
    e.klass = mnEscape >> 12;
    e.id = ( mnEscape & 0x0fe0 ) >> 5;
    int bytes = mnEscape & 0x1f;
    DEBUG_p("class: %d, id: 0x%02x, size: %d", e.klass, e.id, bytes);

    if ( bytes == 31 ){
        if(data + 2 > end)
            return -1;
        bytes = be16toh(*(uint16_t*)data);
        DEBUG_p("new size: %d", bytes);
        data += 2;
    }
    if(data + bytes > end){
        DEBUG_p("entry past end of data (%d > %d)", bytes, (int)(end-data));
#if 0
        return -1;
#else
        bytes = end - data;
#endif
    }

    cgm_read_set(r, data, bytes);
    userfunc(userdata, &e, r);

    bytes += data - start;
    if(bytes & 1)
        ++bytes;
    return bytes;
}


int
cgm_read_binary(const uint8_t *data, size_t len, cgm_handler_t userfunc, void *userdata)
{
    static const cgm_read_interface_t binary_text_vt = {
        .eoe = _binary_eoe,
        .get_I = _binary_get_int,
        .get_IX = _binary_get_index,
        .get_R = _binary_get_real,
        .get_S = _binary_get_string,

        .get_color_channel = _binary_get_color_channel,
        .get_color_index = _binary_get_color_index,
        .get_vdc = _binary_get_vdc,
        .get_int_precision_bytes = _binary_get_int_precision_bytes,

        .get_vdc_type = _binary_get_vdc_type,
        .get_scaling_mode = _binary_get_scaling_mode,
        .get_device_view_port_mode = _binary_get_device_view_port_mode,
        .get_device_view_port_map = _binary_get_device_view_port_map,
        .get_device_view_port_map_h = _binary_get_device_view_port_map_h,
        .get_device_view_port_map_v = _binary_get_device_view_port_map_v,
        .get_clip_indicator = _binary_get_clip_indicator,
        .get_color_selection_mode = _binary_get_color_selection_mode,
        .get_close_type = _binary_get_close_type,
        .get_char_set_type = _binary_get_char_set_type,
        .get_character_coding_a = _binary_get_character_coding_a,
        .get_text_precision = _binary_get_text_precision,
        .get_text_path = _binary_get_text_path,
        .get_text_alignment_h = _binary_get_text_alignment_h,
        .get_text_alignment_v = _binary_get_text_alignment_v,
        .get_interior_style = _binary_get_interior_style,
        .get_underline_mode = _binary_get_underline_mode,
        .get_final_flag = _binary_get_final_flag,
        .get_spec_mode = _binary_get_spec_mode,
        .get_edge_visibility = _binary_get_edge_visibility,
        .get_edge_flag = _binary_get_edge_flag,
        .get_marker_type = _binary_get_marker_type,
        .get_transparency = _binary_get_transparency,
        .get_interior_style = _binary_get_interior_style,

        .get_asf_mask = _binary_get_asf_mask,
        .get_asf_value = _binary_get_asf_value
    };
    const uint8_t *end = data + len;
    cgm_read_t r;
    cgm_read_init(&r);
    r.vt = &binary_text_vt;
    while(data < end){
        ssize_t bytes = _handle_entry(&r, data, end-data, userfunc, userdata);
        if(bytes < 0)
            return -1;
        data += bytes;
    }
    DEBUG_p("successfully done");
    return 0;
}
