/*
 * Copyright (C) 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <nile.h>
#include <nil-cairo.h>
#include <endian.h>                       /* le16toh() */
#include <stdint.h>
#include <stdlib.h>                       /* malloc() */
#include <string.h>                       /* memset(), memcpy() */


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


#define get8(a) (*(uint8_t*)(a))
#define get16(a) le16toh(*(uint16_t*)(a))

static const char **
_drhalo_get_mime_types()
{
    static const char *mime_types[] = {
        "application/dr-halo", "image/x-cut",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_drhalo_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_drhalo_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static void
_guess_palette(nile_t *ni)
{
    int max = 0;
    int i;

    for(i=0; i< ni->width * ni->height; ++i)
        if(ni->pixels[i] > max)
            max = ni->pixels[i];

    for(i=0; i<=max; ++i)
        ni->palette[i] = 255 * i /max;
}


static int
_drhalo_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    nile_t ni;
    uint8_t *dst;
    const char *end;
    int y;

    if(len < 6)
        return -1;
    nile_init(&ni);
    ni.width = h->info->width = get16(data);
    ni.height = h->info->height = get16(data+2);
    nile_ensure_indexed(&ni, ni.width, ni.height);
    dst = ni.pixels;

    end = data + len;
    data += 6;

    for(y=0; y<h->info->height; ++y){
        const char *lineend;
        int c;
        if(data + 2 > end)
            goto error;
        c = get16(data);
        data += 2;
        lineend = data + c;
        if(lineend > end)
            goto error;
        for(;;){
            if(data >= lineend)
                goto error;
            c = get8(data++);
            if(c & 0x80){
                c &= 0x7f;
                if(data >= lineend)
                    goto error;
                memset(dst, get8(data++), c);
            }else if(c){
                if(data + c > lineend)
                    goto error;
                memcpy(dst, data, c);
                data += c;
            }else{
                break;
            }
            dst += c;
        }
    }

    nile_ensure_palette(&ni, 256, 1);
    ni.format = NILE_FORMAT_g8;
    _guess_palette(&ni);

    h->surface = nil_cairo_surface_create(&ni, 0);
    nile_done(&ni);
    return 0;
error:
    nile_done(&ni);
    return -1;
}


static cairo_surface_t *
_drhalo_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _drhalo_get_mime_types,
        .new = _drhalo_new,
        .free = _drhalo_free,
        .create_from_data = _drhalo_create_from_data,
        .get_image_surface = _drhalo_get_image_surface
    };
    return &p;
}
