/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <nile.h>
#include <nil-cairo.h>
#include <nil-iff.h>
#include <stdlib.h>
#include <string.h>

#include "nil-fax.h"


#define IFF_FAXX IFF_ID('F','A','X','X')
#define FAXX_FXHD IFF_ID('F','X','H','D')
#define FAXX_GPHD IFF_ID('G','P','H','D')
#define FAXX_PAGE IFF_ID('P','A','G','E')


typedef struct{
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_faxx_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-faxx",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_faxx_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->page = NULL;
    return h;
}


static void
_faxx_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i].surface)
                cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))


typedef struct{
    abydos_plugin_handle_t *h;
    nile_t ni;
    int compression;
}parse_context_t;


static int
_on_chunk(void *userdata, uint32_t id, const char *data, size_t len)
{
    static const uint8_t palette[6] = {0,0,0,255,255,255};
    parse_context_t *pc = userdata;
    int res;
    DEBUG_p("id: %.4s, length: %d", (char*)&id, (int)len);
    switch(id){
    case FAXX_FXHD:
        if(len < 4)
            return -1;
        pc->ni.width = get16(data);
        pc->ni.height = get16(data+2);
        //pc->ni.stride = NILE_STRIDE_ALIGN16;
        pc->compression = get8(data+8);
        break;
    case FAXX_GPHD:

        break;
    case FAXX_PAGE:
        DEBUG_p("compression: %d", pc->compression);
        switch(pc->compression){
        case 0: /* No compression -- available under Group IV */
            break;
        case 1: /* One-dimensional (MH) coding */
            //nile_ensure_bitplanes(&pc->ni, pc->ni.width, pc->ni.height, 1, 1);
            nile_ensure_indexed(&pc->ni, pc->ni.width, pc->ni.height);
            pc->ni.palette = (uint8_t*)palette;
            pc->ni.colors = 2;
            pc->ni.bpc = 3;
            pc->ni.format = NILE_FORMAT_r8g8b8;
            return nil_fax_decode(&pc->ni, NIL_FAX_MH, (uint8_t*)data, len);
        case 2: /* Two-dimensional (MR) coding */
            break;
        case 4: /* Modified Two-dimensional (MMR) coding */
            break;
        }
        break;
    case IFF_FORM:
        DEBUG_p("embedded form: %.4s", data);
        switch(*(uint32_t*)data){
        case IFF_ILBM:
            {
                abydos_t *ar = abydos_create("image/x-ilbm");
                if(!ar)
                    return -1;
                res = abydos_from_data(ar, data - 8, len + 8);
                if(!res){
                    abydos_get_size(ar, &pc->h->info->width, &pc->h->info->height);
                    pc->h->page = malloc(sizeof(page_t));
                    pc->h->page->surface = abydos_get_image_surface(ar);
                }
                abydos_destroy(ar);
                if(res)
                    return res;
            }
            break;
        }
        break;
    }
    return 0;
}


static int
_faxx_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    parse_context_t pc;
    pc.h = h;
    nile_init(&pc.ni);
    if(0 == nil_iff_parse(IFF_FAXX, 0, data, len, _on_chunk, &pc)){
        if(!h->page && pc.ni.width > 0 && pc.ni.height > 0){
            h->info->width = pc.ni.width;
            h->info->height = pc.ni.height;
            h->page = malloc(sizeof(page_t));
            h->page->surface = nil_cairo_surface_create(&pc.ni, 0);
        }
    }
    nile_done(&pc.ni);
    return h->page ? 0 : -1;
}


static cairo_surface_t *
_faxx_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _faxx_get_mime_types,
        .new = _faxx_new,
        .free = _faxx_free,
        .create_from_data = _faxx_create_from_data,
        .get_image_surface = _faxx_get_image_surface
    };
    return &p;
}
