/*
 * Copyright (C) 2020 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <gdal.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


/* #define ENABLE_BMP */
/* #define ENABLE_BPG */
/* #define ENABLE_DDS */
/* #define ENABLE_DXF */
/* #define ENABLE_GIF */
/* #define ENABLE_PDF */
/* #define ENABLE_PNM */ /* pbm not supported */
/* #define ENABLE_SVG */
/* #define ENABLE_XPM */

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    const char *driver;
    int vector;
    cairo_surface_t *surface;
};


static const char **
_gdal_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_BMP
        "image/bmp",
#endif
#ifdef ENABLE_BPG
        "image/bpg", "image/x-bpg",
#endif
#ifdef ENABLE_DDS
        "image/x-dds", "image/x-direct-draw-surface",
#endif
#ifdef ENABLE_DXF
        "application/dxf", "application/x-dxf", "image/vnd.dxf",
#endif
        "image/x-fit",
#ifdef ENABLE_GIF
        "image/gif",
#endif
        "image/jpeg",
        "application/vnd.nitf",
#ifdef ENABLE_PDF
        "application/pdf",
#endif
        "image/png",
#ifdef ENABLE_PNM
        "image/x-portable-anymap",
#endif
        "image/x-portable-graymap", "image/x-portable-greymap",
        "image/x-portable-pixmap",
        "image/x-rgb", "image/x-sgi",
#ifdef ENABLE_SVG
        "image/svg+xml",
#endif
        "image/x-vicar",
#ifdef ENABLE_XPM
        "image/x-xpixmap", "image/x-xpmi",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gdal_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_BMP
    if(!strcasecmp(mime_type, "image/bmp")){
        h->driver = "BMP";
    }else
#endif
#ifdef ENABLE_BGP
    if(!strcasecmp(mime_type, "image/bpg") ||
       !strcasecmp(mime_type, "image/x-bpg"))
    {
        h->driver = "BPG";
    }else
#endif
#ifdef ENABLE_DDS
    if(!strcasecmp(mime_type, "image/x-dds") ||
       !strcasecmp(mime_type, "image/x-direct-draw-surface"))
    {
        h->driver = "DDS";
    }else
#endif
#ifdef ENABLE_DXF
    if(!strcasecmp(mime_type, "application/dxf") ||
       !strcasecmp(mime_type, "application/x-dxf") ||
       !strcasecmp(mime_type, "image/vnd.dxf"))
    {
        h->driver = "DXF";
    }else
#endif
    if(!strcasecmp(mime_type, "image/x-fit")){
        h->driver = "FIT";
    }else
#ifdef ENABLE_GIF
    if(!strcasecmp(mime_type, "image/gif")){
        h->driver = "GIF";
    }else
#endif
    if(!strcasecmp(mime_type, "image/jpeg")){
        h->driver = "JPEG";
    }else if(!strcasecmp(mime_type, "application/vnd.nitf")){
        h->driver = "NITF";
    }else
#ifdef ENABLE_PDF
    if(!strcasecmp(mime_type, "application/pdf")){
        h->driver = "PDF";
    }else
#endif
    if(!strcasecmp(mime_type, "image/png")){
        h->driver = "PNG";
    }else
#ifdef ENABLE_PNM
    if(!strcasecmp(mime_type, "image/x-portable-anymap")){
        h->driver = "PNM";
    }else
#endif
    if(!strcasecmp(mime_type, "image/x-portable-graymap") ||
       !strcasecmp(mime_type, "image/x-portable-greymap") ||
       !strcasecmp(mime_type, "image/x-portable-pixmap"))
    {
        h->driver = "PNM";
    }else if(!strcasecmp(mime_type, "image/x-rgb") ||
             !strcasecmp(mime_type, "image/x-sgi"))
    {
        h->driver = "SGI";
    }else
#ifdef ENABLE_SVG
    if(!strcasecmp(mime_type, "image/svg+xml")){
        h->driver = "SVG";
    }else
#endif
    if(!strcasecmp(mime_type, "image/x-vicar")){
        h->driver = "VICAR";
    }else
#ifdef ENABLE_XPM
    if(!strcasecmp(mime_type, "image/x-xpixmap") ||
             !strcasecmp(mime_type, "image/x-xpmi"))
    {
        h->driver = "XPM";
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;

    return h;
}


static void
_gdal_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


typedef void(*blit_func_t)(uint32_t*, uint8_t**, int);


static void
_blit_g(uint32_t *dst, uint8_t **line, int count)
{
    int i;
    for(i=0; i<count; ++i){
        int g = line[0][i];
        *dst++ = g<<16|g<<8|g;
    }
}


static void
_blit_rgb(uint32_t *dst, uint8_t **line, int count)
{
    int i;
    for(i=0; i<count; ++i){
        int r = line[0][i];
        int g = line[1][i];
        int b = line[2][i];
        *dst++ = r<<16|g<<8|b;
    }
}


static void
_blit_rgba(uint32_t *dst, uint8_t **line, int count)
{
    int i;
    for(i=0; i<count; ++i){
        int a = line[3][i];
        int r = a * line[0][i] / 255;
        int g = a * line[1][i] / 255;
        int b = a * line[2][i] / 255;
        *dst++ = a<<24|r<<16|g<<8|b;
    }
}


static cairo_surface_t *
_surface_from_bands(cairo_format_t format, GDALRasterBandH *band, int bands, blit_func_t func, int width, int height)
{
    uint8_t *line[4];
    int i;
    cairo_surface_t *surface = cairo_image_surface_create(format, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    line[0] = malloc(bands * width * sizeof(uint8_t));
    for(i=0; i<bands-1; ++i)
        line[i+1] = line[i] + width;
    for(y=0;y<height;++y){
        for(i=0; i<bands; ++i)
            if(CE_None != GDALRasterIO(band[i], GF_Read, 0, y, width, 1, line[i], width, 1, GDT_Byte, 0, 0 ))
                goto error;
        func(dst, line, width);
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    free(line[0]);
    return surface;
error:
    free(line[0]);
    cairo_surface_destroy(surface);
    return NULL;
}


#if 0
static int
_test_vector(GDALDatasetH dataset)
{
    OGRLayerH layer;
    OGRFeatureDefnH defn;
    OGRFeatureH feature;

    layer = GDALDatasetGetLayerByName(dataset, "point");
    if(!layer)
        return -1;
    defn = OGR_L_GetLayerDefn(layer);
    OGR_L_ResetReading(layer);
    while((feature = OGR_L_GetNextFeature(layer))){
        //int iField;
        OGRGeometryH hGeometry;
        int i;

        for(i=0;i<OGR_FD_GetFieldCount(defn);++i){
            OGRFieldDefnH field = OGR_FD_GetFieldDefn(defn,i);

            switch(OGR_Fld_GetType(field)){
            case OFTInteger:
                printf( "%d,", OGR_F_GetFieldAsInteger(feature,i));
                break;
            case OFTInteger64:
                printf( CPL_FRMT_GIB ",", OGR_F_GetFieldAsInteger64(feature,i));
                break;
            case OFTReal:
                printf( "%.3f,", OGR_F_GetFieldAsDouble(feature,i));
                break;
            case OFTString:
                printf( "%s,", OGR_F_GetFieldAsString(feature,i));
                break;
            default:
                printf( "%s,", OGR_F_GetFieldAsString(feature,i));
                break;
            }
        }

        hGeometry = OGR_F_GetGeometryRef(feature);
        if( hGeometry && wkbFlatten(OGR_G_GetGeometryType(hGeometry)) == wkbPoint ){
            DEBUG_p("%.3f,%3.f", OGR_G_GetX(hGeometry, 0), OGR_G_GetY(hGeometry, 0) );
        }else{
            DEBUG_p("no point geometry");
        }
        OGR_F_Destroy(feature);
    }
    return 0;
}
#endif


#if 0
static void
_render_vector(GDALDatasetH dataset, int layers)
{
    int i;
    DEBUG_p("got %d layers", layers);
    for(i=0; i<layers; ++i){
        OGRLayerH layer = GDALDatasetGetLayer(dataset, i);
        OGRFeatureH feature;
        //OGR_L_ResetReading(layer);
        while((feature = OGR_L_GetNextFeature(layer))){
            OGRGeometryH g = OGR_F_GetGeometryRef(feature);
            OGRwkbGeometryType t;
            if(!g)
                continue;
            //DEBUG_p("geometry: %s", OGR_G_GetGeometryName(hGeometry));
            t = OGR_G_GetGeometryType(g);
            DEBUG_p("geotype: %s", OGRGeometryTypeToName(t));
            switch(t){
            case wkbUnknown:
                break;
            case wkbPoint:
                break;
            case wkbLineString:
                {
                    int num = OGR_G_GetPointCount(g);
                    int i;
                    for(i=0; i<num; ++i){
                        double x = OGR_G_GetX(g, i);
                        double y = OGR_G_GetY(g, i);
                        DEBUG_p("point: (%.2f, %.2f)\n", x, y);
                    }
                }
                break;
            case wkbPolygon:
                break;

            case wkbMultiPoint:
            case wkbMultiLineString:
            case wkbMultiPolygon:
                break;
            case wkbGeometryCollection:
            case wkbCircularString:
            case wkbCompoundCurve:
            case wkbCurvePolygon:
            case wkbMultiCurve:
            case wkbMultiSurface:
            case wkbCurve:
            case wkbSurface:
            case wkbPolyhedralSurface:
            case wkbTIN:
            case wkbTriangle:
            case wkbNone:
            case wkbLinearRing:
            case wkbCircularStringZ:
            case wkbCompoundCurveZ:
            case wkbCurvePolygonZ:
            case wkbMultiCurveZ:
            case wkbMultiSurfaceZ:
            case wkbCurveZ:
            case wkbSurfaceZ:
            case wkbPolyhedralSurfaceZ:
            case wkbTINZ:
            case wkbTriangleZ:
            case wkbPointM:
            case wkbLineStringM:
            case wkbPolygonM:
            case wkbMultiPointM:
            case wkbMultiLineStringM:
            case wkbMultiPolygonM:
            case wkbGeometryCollectionM:
            case wkbCircularStringM:
            case wkbCompoundCurveM:
            case wkbCurvePolygonM:
            case wkbMultiCurveM:
            case wkbMultiSurfaceM:
            case wkbCurveM:
            case wkbSurfaceM:
            case wkbPolyhedralSurfaceM:
            case wkbTINM:
            case wkbTriangleM:
            case wkbPointZM:
            case wkbLineStringZM:
            case wkbPolygonZM:
            case wkbMultiPointZM:
            case wkbMultiLineStringZM:
            case wkbMultiPolygonZM:
            case wkbGeometryCollectionZM:
            case wkbCircularStringZM:
            case wkbCompoundCurveZM:
            case wkbCurvePolygonZM:
            case wkbMultiCurveZM:
            case wkbMultiSurfaceZM:
            case wkbCurveZM:
            case wkbSurfaceZM:
            case wkbPolyhedralSurfaceZM:
            case wkbTINZM:
            case wkbTriangleZM:
            case wkbPoint25D:
            case wkbLineString25D:
            case wkbPolygon25D:
            case wkbMultiPoint25D:
            case wkbMultiLineString25D:
            case wkbMultiPolygon25D:
            case wkbGeometryCollection25D:
                break;
            }


            OGR_F_Destroy(feature);
        }
    }
}
#endif


static int
_gdal_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    const char *drivers[2];
    GDALDatasetH dataset;
    int bands;
    GDALRasterBandH band[4];
    //int layers;

    drivers[0] = h->driver;
    drivers[1] = NULL;
    dataset = GDALOpenEx(filename, GA_ReadOnly, drivers, NULL, NULL);
    if(!dataset){
        DEBUG_p("failed to open %s",filename);
        return -1;
    }

#if 0
    layers = GDALDatasetGetLayerCount(dataset);
    if(layers > 0)
        _render_vector(dataset, layers);
#endif

    h->info->width = GDALGetRasterXSize(dataset);
    h->info->height = GDALGetRasterYSize(dataset);
    bands = GDALGetRasterCount(dataset);
#if 0
    if(bands > 4)
        bands = 3;
#endif
    if(bands <= 4){
        int i;
        for(i=0; i<bands; ++i){
            band[i] = GDALGetRasterBand(dataset, i+1);
            DEBUG_p("channel type %d", GDALGetRasterDataType(band[i]));
        }
        switch(bands){
        case 1:
            h->surface = _surface_from_bands(CAIRO_FORMAT_RGB24, band, bands, _blit_g, h->info->width, h->info->height);
            break;
        case 3:
            h->surface = _surface_from_bands(CAIRO_FORMAT_RGB24, band, bands, _blit_rgb, h->info->width, h->info->height);
            break;
        case 4:
            h->surface = _surface_from_bands(CAIRO_FORMAT_ARGB32, band, bands, _blit_rgba, h->info->width, h->info->height);
            break;
        }
    }

    GDALClose(dataset);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_gdal_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gdal_get_mime_types,
        .new = _gdal_new,
        .free = _gdal_free,
        .create_from_file = _gdal_create_from_file,
        .get_image_surface = _gdal_get_image_surface
    };
    GDALAllRegister();
    return &p;
}
