/*
 * Copyright (C) 2019 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <gegl.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */


#define ENABLE_EXR
#define ENABLE_GIF
#define ENABLE_JPEG
#define ENABLE_JPEG2000
#define ENABLE_PNG
#define ENABLE_PPM
#define ENABLE_RADIANCE
#define ENABLE_TIFF
/* #define ENABLE_WEBP */ /* Not working? */

/*

  gegl-0.2/exr-load.so
  gegl-0.2/ff-load.so
  gegl-0.2/jp2-load.so
  gegl-0.2/jpg-load.so
 (gegl-0.2/magick-load.so)
  gegl-0.2/png-load.so
  gegl-0.2/ppm-load.so
 (gegl-0.2/raw-load.so)
  gegl-0.2/rgbe-load.so
  gegl-0.2/svg-load.so

  gegl-0.3/exr-load.so
  gegl-0.3/ff-load.so
  gegl-0.3/jp2-load.so
  gegl-0.3/jpg-load.so
  gegl-0.3/png-load.so
  gegl-0.3/ppm-load.so
 (gegl-0.3/raw-load.so)
  gegl-0.3/rgbe-load.so
  gegl-0.3/svg-load.so
+ gegl-0.3/tiff-load.so
+ gegl-0.3/webp-load.so

  gegl-0.4/exr-load.so                  OpenEXR
  gegl-0.4/ff-load.so                   FFmpeg
+ gegl-0.4/gif-load.so
  gegl-0.4/jp2-load.so
  gegl-0.4/jpg-load.so
  gegl-0.4/pdf-load.so
  gegl-0.4/png-load.so
  gegl-0.4/ppm-load.so
 (gegl-0.4/raw-load.so)
  gegl-0.4/rgbe-load.so                 Radiance
  gegl-0.4/svg-load.so
  gegl-0.4/tiff-load.so
  gegl-0.4/webp-load.so

*/

#if 0 == GEGL_MAJOR_VERSION
# if GEGL_MINOR_VERSION < 3 && defined(ENABLE_TIFF)
#  undef ENABLE_TIFF
# endif
# if GEGL_MINOR_VERSION < 3 && defined(ENABLE_WEBP)
#  undef ENABLE_WEBP
# endif
# if GEGL_MINOR_VERSION < 4 && defined(ENABLE_GIF)
#  undef ENABLE_GIF
# endif
#endif

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    const char *plugin;
    int can_have_alpha;
    GeglNode *node;
};


static const char **
_gegl_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_EXR
        "image/x-exr",
#endif
#ifdef ENABLE_GIF
        "image/gif",
#endif
#ifdef ENABLE_JPEG
        "image/jpeg",
#endif
#ifdef ENABLE_JPEG2000
        "image/jp2",
        "image/jpeg2000",
        "image/jpx",
#endif
#ifdef ENABLE_PNG
        "image/png",
#endif
#ifdef ENABLE_PPM
        "image/x-portable-graymap",
        "image/x-portable-greymap",
        "image/x-portable-pixmap",
#endif
#ifdef ENABLE_RADIANCE
        "image/vnd.radiance",
#endif
#ifdef ENABLE_TIFF
        "image/tiff",
#endif
#ifdef ENABLE_WEBP
        "image/webp",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gegl_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->can_have_alpha = 1;
#ifdef ENABLE_EXR
    if(!strcasecmp(mime_type, "image/x-exr")){
        h->plugin = "gegl:exr-load";
    }else
#endif
    if(!strcasecmp(mime_type, "image/gif")){
        h->plugin = "gegl:gif-load";
    }else
#ifdef ENABLE_JPEG
    if(!strcasecmp(mime_type, "image/jpeg")){
        h->plugin = "gegl:jpg-load";
        h->can_have_alpha = 0;
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type, "image/jp2") ||
       !strcasecmp(mime_type, "image/jpeg2000") ||
       !strcasecmp(mime_type, "image/jpx"))
    {
        //h->plugin = "gegl:jp2-load";
        h->plugin = "gegl:load";
    }else
#endif
#ifdef ENABLE_PNG
    if(!strcasecmp(mime_type, "image/png")){
        h->plugin = "gegl:png-load";
    }else
#endif
#ifdef ENABLE_PPM
    if(!strcasecmp(mime_type, "image/x-portable-graymap") ||
       !strcasecmp(mime_type, "image/x-portable-greymap") ||
       !strcasecmp(mime_type, "image/x-portable-pixmap"))
    {
        h->plugin = "gegl:ppm-load";
    }else
#endif
#ifdef ENABLE_RADIANCE
    if(!strcasecmp(mime_type, "image/vnd.radiance")){
        h->plugin = "gegl:rgbe-load";
        h->can_have_alpha = 0;
    }else
#endif
#ifdef ENABLE_TIFF
    if(!strcasecmp(mime_type, "image/tiff")){
        h->plugin = "gegl:tiff-load";
    }else
#endif
#ifdef ENABLE_WEBP
    if(!strcasecmp(mime_type, "image/webp")){
        h->plugin = "gegl:webp-load";
    }else
#endif
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->node = NULL;
    return h;
}


static void
_gegl_free(abydos_plugin_handle_t *h)
{
    if(h->node)
        g_object_unref(h->node);
    free(h);
}


static int
_gegl_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    GeglRectangle rect;
    h->node = gegl_node_new_child(NULL,
                                  "operation", h->plugin,
                                  "path", filename,
                                  NULL);
    rect = gegl_node_get_bounding_box(h->node);
    h->info->width = rect.width;
    h->info->height = rect.height;

    /* Using a premultiply node doesn't work (gets replaced by a passthrough node). */
    //h->premul = gegl_node_new_child(NULL, "operation", "premultiply", NULL);
    //gegl_node_link(h->node, h->premul);
    return 0;
}


static void
_premultiply(uint8_t *p, int stride, int width , int height)
{
    int skip = stride - width * 4;
    while(height-->0){
        int x;
        for(x=0; x<width; ++x){
#if __BYTE_ORDER == __BIG_ENDIAN
            int a = p[0];
            p[1] = a * p[1] / 255;
            p[2] = a * p[2] / 255;
            p[3] = a * p[3] / 255;
#else
            int a = p[3];
            p[0] = a * p[0] / 255;
            p[1] = a * p[1] / 255;
            p[2] = a * p[2] / 255;
#endif
            p += 4;
        }
        p += skip;
    }
}


static void
_gegl_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *layer_visibility,
             int variant,
             double time,
             int frame)
{
    double x1;
    double y1;
    double x2;
    double y2;
    GeglRectangle roi;
    cairo_surface_t *surface;
    uint8_t *data;
    int stride;
    const Babl *format;

    cairo_clip_extents(cr,&x1,&y1,&x2,&y2);
    roi.x = x1;
    roi.y = y1;
    roi.width  = ceil(x2-x1);
    roi.height = ceil(y2-y1);

    if(h->can_have_alpha){
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,roi.width,roi.height);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,roi.width,roi.height);
    }
    data = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);

    /* R'G'B'A doesn't seem to work. */
    format = babl_format_new(babl_model ("RGBA"), babl_type ("u8"),
#if __BYTE_ORDER == __BIG_ENDIAN
                             babl_component ("A"),
                             babl_component ("R"),
                             babl_component ("G"),
                             babl_component ("B"),
#else
                             babl_component ("B"),
                             babl_component ("G"),
                             babl_component ("R"),
                             babl_component ("A"),
#endif
                             NULL);

    gegl_node_blit(h->node,
                   1,
                   &roi,
                   format,
                   (void*)data,
                   stride,
                   GEGL_BLIT_CACHE);
    if(h->can_have_alpha)
        _premultiply(data, stride, roi.width, roi.height);
    cairo_surface_mark_dirty(surface);

    cairo_set_source_surface(cr, surface, x1, y1);
    cairo_paint(cr);

    cairo_surface_destroy(surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = gegl_exit,
        .get_mime_types = _gegl_get_mime_types,
        .new = _gegl_new,
        .free = _gegl_free,
        .create_from_file = _gegl_create_from_file,
        .render = _gegl_render
    };
    gegl_init(0, NULL);
    return &p;
}
