/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <gif_lib.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


#define GIF_TRANSPARENT_MASK 0x01
#define GIF_DISPOSE_MASK     0x07
#define GIF_DISPOSE_SHIFT    2


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};

typedef struct{
    abydos_plugin_handle_t *handle;
    GifColorType *shared_palette;
    enum{
        EXT_NETSCAPE2_0 = 1,
        EXT_XMP_dataXMP,
        EXT_ICCRGBG1012,
    }ext;
    int transparent_index;
    int loop_count;
}decode_context_t;



static const char **
_gif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/gif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h;
    h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_gif_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static void
_setup_palette(const GifColorType *src,uint32_t *dst,int transparent_index)
{
    int i;
    for(i=0;i<256;++i){
        if(i == transparent_index){
            dst[i] = 0;
        }else{
            dst[i] = src[i].Blue | src[i].Green << 8 | src[i].Red << 16 | 255 << 24;
        }
    }
}


static cairo_bool_t
_handle_extension(decode_context_t *c,frame_t *frame,ExtensionBlock *b)
{
    switch (b->Function) {
    case GRAPHICS_EXT_FUNC_CODE:
        {
            int flags;
            int dispose;
            double delay;
            if (b->ByteCount != 4)
                return 0;
            flags = b->Bytes[0];
            dispose = (flags >> GIF_DISPOSE_SHIFT) & GIF_DISPOSE_MASK;
            delay = b->Bytes[1] | (b->Bytes[2] << 8);  // In 10 ms units.

            frame->duration = delay/100;
            if(frame->duration < 0.02)
                frame->duration = 0.02;
            switch(dispose){
            case 2:
                //frame.dispose_method = WEBP_MUX_DISPOSE_BACKGROUND;
                DEBUG_p("dispose background");
                break;
            case 3:
                DEBUG_p("dispose restore");
                DEBUG_p("WARNING: GIF_DISPOSE_RESTORE not supported");
                /* failsafe. TODO(urvang): emulate the correct behaviour by
                 * recoding the whole frame.
                 */
                //frame.dispose_method = WEBP_MUX_DISPOSE_BACKGROUND;
                break;
            default:
                DEBUG_p("dispose none");
                //frame.dispose_method = WEBP_MUX_DISPOSE_NONE;
                break;
            }
            c->transparent_index = (flags & GIF_TRANSPARENT_MASK) ? b->Bytes[3] : -1;
        }
        break;
    case APPLICATION_EXT_FUNC_CODE:
        if (b->ByteCount != 11)
            break; /* Chunk is too short */
        if (!memcmp(b->Bytes, "NETSCAPE2.0", 11)) {
            /* Recognize and parse Netscape2.0 NAB extension for loop count. */
            c->ext = EXT_NETSCAPE2_0;
        } else if (!memcmp(b->Bytes, "XMP dataXMP", 11)) {
            /* Read XMP metadata. */
            c->ext = EXT_XMP_dataXMP;
        } else if (!memcmp(b->Bytes, "ICCRGBG1012", 11)) {
            /* Read ICC profile. */
            c->ext = EXT_ICCRGBG1012;
        } else {
            DEBUG_p("unhandled extension %11s",b->Bytes);
            c->ext = 0;
        }
        break;
    case CONTINUE_EXT_FUNC_CODE:
        switch(c->ext){
        case EXT_NETSCAPE2_0:
            if (b->ByteCount != 3 && b->Bytes[0] != 1)
                break; /* wrong size/marker */
            c->loop_count = b->Bytes[1] | (b->Bytes[2] << 8);
            DEBUG_p("loop count: %d\n", c->loop_count);
            break;
        case EXT_XMP_dataXMP:
#if 0
            WebPData xmp;
            if (DGifGetExtensionNext(gif, &data) == GIF_ERROR)
                return 0;
            if (!data)
                return 0;
            xmp.bytes = (uint8_t*)data;
            xmp.size = data[0] + 1;
            WebPMuxSetChunk(mux, "XMP ", &xmp, 1);
            DEBUG_p("XMP size: %d", (int)xmp.size);
#endif
            break;
        case EXT_ICCRGBG1012:
#if 0
            WebPData icc;
            if (DGifGetExtensionNext(gif, &data) == GIF_ERROR)
                goto error;
            if (!data)
                goto error;
            icc.bytes = (uint8_t*)data;
            icc.size = data[0] + 1;
            WebPMuxSetChunk(mux, "ICCP", &icc, 1);
            DEBUG_p("ICC size: %d", (int)icc.size);
#endif
            break;
        }
        break;
    default:
        DEBUG_p("unhandled function %d",b->Function);
        break;
    }
    return 1;
}


static cairo_surface_t *
_surface_from_pixels(abydos_plugin_handle_t *h, uint8_t *src, uint32_t *palette){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - h->info->width;
    int x;
    int y;
#if 1
    for(y=0;y<h->info->height;++y){
        for(x=0;x<h->info->width;++x)
            *dst++ = palette[*src++];
        dst += dst_skip;
    }
#endif
    cairo_surface_mark_dirty(surface);
    return surface;
}


static void
_setup_frame(abydos_plugin_handle_t *h,decode_context_t *c,SavedImage *src,frame_t *dst)
{
    uint32_t palette[256];
    int i;
    for(i=0;i<src->ExtensionBlockCount;++i)
        _handle_extension(c,dst,&src->ExtensionBlocks[i]);

    if(src->ImageDesc.ColorMap){
        _setup_palette(src->ImageDesc.ColorMap->Colors,palette,c->transparent_index);
    }else if(c->shared_palette){
        _setup_palette(c->shared_palette,palette,c->transparent_index);
    }else{
        /* FIXME!! grayscale? */
    }
    dst->surface = _surface_from_pixels(h,src->RasterBits,palette);
}


static int
_gif_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    decode_context_t c;
    int i;
    GifFileType *gif = DGifOpenFileName((char*)filename,NULL);

    if(!gif){
        h->info->error = "Failed to open file";
        return -1;
    }

    DGifSlurp(gif);
    h->frame = (frame_t*)malloc(gif->ImageCount*sizeof(frame_t));
    h->info->width = gif->SWidth;
    h->info->height = gif->SHeight;
    h->info->frame_count = gif->ImageCount;

    c.handle = h;
    c.shared_palette = gif->SColorMap ? gif->SColorMap->Colors : NULL;

    for(i=0;i<gif->ImageCount;++i)
        _setup_frame(h,&c,&gif->SavedImages[i],&h->frame[i]);

    DGifCloseFile(gif,NULL);
    return 0;
}


static cairo_surface_t *
_gif_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_gif_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gif_get_mime_types,
        .new = _gif_new,
        .free = _gif_free,
        .create_from_file = _gif_create_from_file,
        .get_image_surface = _gif_get_image_surface,
        .get_duration = _gif_get_duration
    };
    return &p;
}
