/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_hsi_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-hsi-raw",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_hsi_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_hsi_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) be16toh(*(uint16_t*)(a))


/*

    Header
    """"""

    Variant with palette:

  _=mhwanh_____________________
  0x6d 0x68 0x77 0x61 0x6e 0x68 0x00 0x04
  _width___ _height__
  0x01 0x40 0x00 0xc8 0x01 0x00 0xff 0xfa

  0xff 0xfb 0x00 0x00 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

    Other variant:

  _=mhwanh_____________________
  0x6d 0x68 0x77 0x61 0x6e 0x68 0x00 0x04
  _width___ _height__
  0x00 0x82 0x00 0x26 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

  @0x143c:

  _=mhwanh_____________________
  0x6d 0x68 0x77 0x61 0x6e 0x68 0x00 0x04
  _width___ _height__
  0x00 0x79 0x00 0x26 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

  @0x2722:

  _=mhwanh_____________________
  0x6d 0x68 0x77 0x61 0x6e 0x68 0x00 0x04
  _width___ _height__
  0x00 0x77 0x00 0x25 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00

  0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00


*/

typedef struct{
    const uint8_t *src;
    const uint8_t *end;
    unsigned int buf;
    int len;
}bitbuf_t;


static bitbuf_t *
_bitbuf_init(bitbuf_t *bb, const uint8_t *src, const uint8_t *end)
{
    bb->src = src;
    bb->end = end;
    bb->buf = 0;
    bb->len = 0;
    return bb;
}


static int
_bitbuf_get(bitbuf_t *bb, int bits){
    int ret;
    while(bb->len < bits){
        if(bb->src >= bb->end)
            return -1;
        bb->buf <<= 8;
        bb->buf |= *bb->src++;
        bb->len += 8;
    }
    ret = bb->buf >> (bb->len - bits);
    bb->len -= bits;
    bb->buf &= 0xffff >> (16-bb->len);
    return ret;
}


static cairo_surface_t *
_surface_from_data(const char *data, const char *end, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - width;
    bitbuf_t bb;

    _bitbuf_init(&bb, (uint8_t*)data, (uint8_t*)end);

    while(height-->0){
        int x;
        for(x=0; x<width; ++x){
#if 1
            int v = _bitbuf_get(&bb, 8);
#else
            int v = _bitbuf_get(&bb, 4);
            //v = v << 2 | v;
            v = v << 4 | v;
#endif
            *dst++ = v << 16 | v << 8 | v;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_hsi_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data + len;

    if(len < 32)
        return -1;
    if(strncmp(data, "mhwanh", 6))
        return -1;

    h->info->width = get16(data+8);
    h->info->height = get16(data+10);
    DEBUG_p("size: %dx%d", h->info->width, h->info->height);

    if(data[12]){
        if(len < 32 + 768 + h->info->width * h->info->height)
            return -1;


        h->surface = nil_cairo_surface_create_from_indexed((uint8_t*)data + 800,
                                                           h->info->width,
                                                           h->info->height,
                                                           -1,
                                                           (uint8_t*)data + 32,
                                                           256,
                                                           3,
                                                           NILE_FORMAT_r8g8b8,
                                                           0);
    }else{
        h->surface = _surface_from_data(data + 12, end, h->info->width, h->info->height);
    }
    return 0;
}


static cairo_surface_t *
_hsi_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _hsi_get_mime_types,
        .new = _hsi_new,
        .free = _hsi_free,
        .create_from_data = _hsi_create_from_data,
        .get_image_surface = _hsi_get_image_surface
    };
    return &p;
}
