/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <byteswap.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* memset() */

#include "nil-amiga.h"
#include "nil-cycle.h"
#include "ni-ehb.h"
#include "ni-ilbm.h"

#define CRNG_MAX 16

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

typedef struct{
    int compression;
    int mask;
    int width;
    int height;
    int x;
    int y;
    int duration;
    uint32_t bits;
}delta_info_t;

typedef struct{
    uint32_t type;
    double pixel_ratio;
    int screen_width;
    int screen_height;
    int compression;
    int use_palette_stepping;
    int is_hame;

    uint32_t camg;

    nil_cycle_t cycle[CRNG_MAX];
    int cycle_count;

    int frame_count;
    double duration_multiplier;
    int default_duration;
    delta_info_t di;
    nile_t src;
    nile_t *dst;
}frame_info_t;

typedef struct{
    int frame_count;
    int bpc;
    nile_format_t format;
    ni_ilbm_request_func_t request_func;
    ni_ilbm_frame_func_t frame_func;
    void *userdata;
}context_t;


static frame_info_t *
_frame_info_init(frame_info_t *fi)
{
    fi->pixel_ratio = 1;
    fi->use_palette_stepping = 0;
    fi->is_hame = 0;
    fi->camg = 0;
    fi->cycle_count = 0;
    fi->frame_count = -1;
    /* Use jiffies (1/60 sec) as the default. */
    fi->duration_multiplier = 1.0/60;
    fi->default_duration = 0;
    nile_init(&fi->src);
    fi->dst = NULL;
    return fi;
}


static void
_frame_info_done(frame_info_t *fi)
{
    nile_done(&fi->src);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static double
_pixel_ratio_from_screen_size(int width, int height)
{
    /* try to guess aspect ratio from screen size */
    if(height <= 480){
        /* assume this is an 4:3 screen */
        return ((double)height/width) * (4.0/3);
    }else{
        /* higher resolutions are assumed to have square pixels independant
           of their aspect ratio. */
        return 1;
    }
}


static int
_handle_bmhd(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    /*
       0 u16 width
       2 u16 height
       4 u16 xoff
       6 u16 yoff
       8 u8  planes
       9 u8  masking
      10 u8  compression
      11 u8  pad
      12 u16 transparent
      14 u8  x aspect
      15 u8  y aspect
      16 u16 screen width
      18 u16 screen height
    */
    int x_aspect;
    int y_aspect;

    if(len<20)
        return -1;

    fi->src.width = get16(data+0);
    fi->src.height = get16(data+2);
    fi->src.depth = get8(data+8);
    fi->compression = get8(data+10);

    x_aspect = get8(data+14);
    y_aspect = get8(data+15);
    fi->screen_width = get16(data+16);
    fi->screen_height = get16(data+18);
    DEBUG_p("X aspect = %d, Y aspect = %d",x_aspect,y_aspect);
    DEBUG_p("screen width : %d",fi->screen_width);
    DEBUG_p("screen height: %d",fi->screen_height);
    if(x_aspect != y_aspect){
        //fi->pixel_ratio = (double)x_aspect/y_aspect;
        fi->pixel_ratio = (double)x_aspect/y_aspect;
    }else if(fi->screen_width && fi->screen_height){
        fi->pixel_ratio = _pixel_ratio_from_screen_size(fi->screen_width, fi->screen_height);
    }

    return 0;
}


static int
_handle_anhd(delta_info_t *di, const uint8_t *data, ssize_t len)
{
    /*
       0 u8  compression
       1 u8  mask
       2 u16 width
       4 u16 height
       6 u16 x
       8 u16 y
      10 u32 abstime
      14 u32 reltime
      18 u8  interleave
      19 u8  pad
      20 u32 bits

      additional 16 bytes padding
    */

    di->compression = get8(data+0);
    di->mask = get8(data+1);
    di->width = get16(data+2);
    di->height = get16(data+4);
    di->x = get16(data+6);
    di->y = get16(data+8);
    di->duration = get32(data+14);
    di->bits = get32(data+20);

    return 0;
}


static int
_handle_cmap(nile_t *ni, const uint8_t *data, ssize_t len)
{
    int colors;
    int r_o;
    int g_o;
    int b_o;
    int i;
    colors = len/3;
    DEBUG_p("%d colors", colors);
    /* Allocate one palette per line just in case. */
    if(!ni->bpc)
        ni->bpc = 3;
    if(!ni->format)
        ni->format = NILE_FORMAT_r8g8b8;
    nile_ensure_palette(ni, colors * ni->height, ni->bpc);
    ni->colors = colors;
    r_o = NILE_OFFSET_RED(ni->format);
    g_o = NILE_OFFSET_GREEN(ni->format);
    b_o = NILE_OFFSET_BLUE(ni->format);
    for(i=0;i<ni->colors;++i){
        ni->palette[i * ni->bpc + r_o] = get8(data+i*3+0);
        ni->palette[i * ni->bpc + g_o] = get8(data+i*3+1);
        ni->palette[i * ni->bpc + b_o] = get8(data+i*3+2);
    }
    return 0;
}


static int
_handle_camg(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    if(len>=4)
        fi->camg = get32(data);
    return 0;
}


static int
_handle_crng(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    int rate;
    int flags;
    int first;
    int last;
    nil_cycle_t *cycle;
    if(fi->cycle_count >= CRNG_MAX){
        return 0;
    }
    if(len < 8)
        return -1;
    flags = get16(data+4);
    if(!flags)
        return 0;
    rate = get16(data+2);
    /* this special case also means off */
    if(0==rate || (1==flags && 36==rate))
        return 0;
    DEBUG_p("rate: %d",rate);
    first = (uint8_t)data[6];
    last = (uint8_t)data[7];
    if(first == last)
        return 0;
    cycle = &fi->cycle[fi->cycle_count++];
    cycle->direction = 2==flags ? NIL_CYCLE_RIGHT : NIL_CYCLE_LEFT;
    cycle->duration = 4096.0 / (15*rate);
    cycle->first = first;
    cycle->last = last;
    return 0;
}


static int
_handle_ctbl(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    int entries = len/2;
    /*
    int r_o;
    int g_o;
    int b_o;
    int i;
    */

    if(fi->src.palette){
        entries = MIN(entries, fi->src.colors * fi->src.height);
    }else{
        if(!fi->src.bpc)
            fi->src.bpc = 3;
        nile_ensure_palette(&fi->src, entries*2, fi->src.bpc);
        fi->src.colors = entries/fi->src.height;
        if(!fi->src.format)
            fi->src.format = NILE_FORMAT_r8g8b8;
    }
    fi->use_palette_stepping = 1;

#if 1
    nil_amiga_copy_colors(fi->src.palette, fi->src.bpc, fi->src.format, (uint16_t*)data, entries);
#else
    r_o = NILE_OFFSET_RED(fi->src.format);
    g_o = NILE_OFFSET_RED(fi->src.format);
    b_o = NILE_OFFSET_RED(fi->src.format);
    for(i=0;i<entries;++i){
        int col = get16(data+i*sizeof(uint16_t));
        int r = (col>>8)&0x0f;
        int g = (col>>4)&0x0f;
        int b = (col>>0)&0x0f;
        fi->src.palette[i * fi->src.bpc + r_o] = r<<4|r;
        fi->src.palette[i * fi->src.bpc + g_o] = g<<4|g;
        fi->src.palette[i * fi->src.bpc + b_o] = b<<4|b;
    }
#endif
    return 0;
}


typedef struct{
    const uint8_t *start;
    const uint8_t *tree;
    const uint8_t *data;
    const uint8_t *end;
    int bits;
}huffman_state_t;


static void
_huffman_init(huffman_state_t *state, const uint8_t *data, size_t len, size_t tree_len)
{
    state->start = data;
    state->tree = data+tree_len-2;
    state->data = data+tree_len;
    state->end = data+len;
    state->bits = 0;
}


static int
_huffman_get_bytes(huffman_state_t *state, uint8_t *dst, int bytes)
{
    while(bytes-->0){
        const uint8_t *tree = state->tree;
        for(;;){
            int bit;
            if(!(state->bits&0xff)){
                if(state->data>=state->end)
                    return -1;
                state->bits = (*state->data++)<<1 | 1;
            }
            bit = state->bits&256;
            state->bits<<=1;
            if(bit){
                int hi = tree[0];
                int lo = tree[1];
                if(hi < 0x80){
                    *dst++ = lo;
                    break;
                }
                tree += (hi - 0x100) << 8 | lo;
                if(tree < state->start)
                    return -1;
            }else{
                tree -= 2;
                if(tree < state->start)
                    return -1;
                if((tree[0] & 0x81) == 1){
                    *dst++ = tree[1];
                    break;
                }
            }
        }
    }
    DEBUG_p("%p < %p > %p",state->start,state->data,state->end);
    return 0;
}


static int
_handle_pchg(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    int compression;
    int flags;
    //int startline;
    int line_count;
    //int min_reg;
    //int max_reg;
    int tree_size;
    int original_size;
    int line_length;
    const uint8_t *linemask;
    int linemask_len;
    size_t len2;
    uint8_t *allocated_data = NULL;
    const uint8_t *pchg_data;
    int pchg_len;
    huffman_state_t huff;
    int i;

    DEBUG_p("chunk size: %d",(int)len);

    if(len<20)
        return -1;
    compression = get16(data+0);
    flags = get16(data+2);
    /* FIXME!! use startline */
    //startline = get16(data+4);
    line_count = get16(data+6);
    //min_reg = get16(data+8);
    //max_reg = get16(data+10);
    tree_size = get16(data+12);
    original_size = get32(data+16);

    /* FIXME!! make use of startline */

    DEBUG_p("original size: %d",original_size);

    data += 20;
    len -= 20;

    linemask_len = ((line_count+31)>>5)*4;
    DEBUG_p("line mask size: %d",linemask_len);

    switch(compression){
    case 0:
        DEBUG_p("uncompressed");
        pchg_data = (uint8_t*)data;
        pchg_len = len;
        break;
    case 1:
        DEBUG_p("huffman compressed");
        pchg_len = get32(data+4);
        pchg_data = allocated_data = malloc(pchg_len);
        len2 = get32(data);
        DEBUG_p("tree size %d / %d / %d",tree_size,(int)len2,(int)len);
        _huffman_init(&huff, (const uint8_t*)data+8, len-8, len2);
        if(_huffman_get_bytes(&huff, allocated_data, pchg_len)<0)
            goto error;
        DEBUG_p("successfully decompressed %d bytes",(int)pchg_len);
        break;
    default:
        DEBUG_p("unknown compression");
        return -1;
    }

    if(linemask_len > pchg_len)
        goto error;
    linemask = (const uint8_t*)pchg_data;
    pchg_data += linemask_len;
    pchg_len -= linemask_len;

    /* FIXME!! what if there was no CMAP? */

    line_length = fi->src.colors * fi->src.bpc;
    fi->use_palette_stepping = 1;

    DEBUG_p("flags: 0x%x",flags);
    if(!(flags&1)){
        for(i=0;i<line_count;++i){
            int change_count;
            int r_o;
            int g_o;
            int b_o;
            int j;
            if(i > 0)
                memcpy(fi->src.palette + i * line_length ,fi->src.palette + (i-1) * line_length, line_length);
            DEBUG_p("mask: 0x%x",linemask[i>>3]);
            if(!(linemask[i>>3] & (128>>(i&31))))
                continue;
            DEBUG_p("there are changes for line %04d",i);
            if(pchg_len < 2)
                goto error;
            change_count = get16(pchg_data);
            len2 = 2 + change_count*6;
            DEBUG_p("changes: %d",change_count);
            if(len2 > pchg_len){
                DEBUG_p("out of data");
                goto error;
            }
            r_o = NILE_OFFSET_RED(fi->src.format);
            g_o = NILE_OFFSET_GREEN(fi->src.format);
            b_o = NILE_OFFSET_BLUE(fi->src.format);
            for(j=0;j<change_count;++j){
                int index = get16(pchg_data+2+j*6);
                int r = get8(pchg_data+2+j*6+3);
                int g = get8(pchg_data+2+j*6+4);
                int b = get8(pchg_data+2+j*6+5);
                fi->src.palette[(i * fi->src.colors + index) * fi->src.bpc + r_o] = r<<4|r;
                fi->src.palette[(i * fi->src.colors + index) * fi->src.bpc + g_o] = g<<4|g;
                fi->src.palette[(i * fi->src.colors + index) * fi->src.bpc + b_o] = b<<4|b;
            }

            pchg_data += len2;
            pchg_len -= len2;
        }
    }else{
        for(i=0;i<line_count;++i){
            int change_count16;
            int change_count32;
            int r_o;
            int g_o;
            int b_o;
            int j;
            if(i > 0)
                memcpy(fi->src.palette + i * line_length, fi->src.palette + (i-1) * line_length, line_length);
            //DEBUG_p("mask: 0x%x",linemask[i>>3]);
            if(!(linemask[i>>3] & (128>>(i&7))))
                continue;
            DEBUG_p("there are changes for line %04d",i);
            if(pchg_len < 2)
                goto error;
            change_count16 = get8(pchg_data);
            change_count32 = get8(pchg_data+1);
            len2 = 2 + change_count16*2 + change_count32*2;
            DEBUG_p("changes16: %d",change_count16);
            DEBUG_p("changes32: %d",change_count32);
            if(len2 > pchg_len){
                DEBUG_p("out of data");
                goto error;
            }
            r_o = NILE_OFFSET_RED(fi->src.format);
            g_o = NILE_OFFSET_GREEN(fi->src.format);
            b_o = NILE_OFFSET_BLUE(fi->src.format);
            for(j=0;j<change_count16;++j){
                int change = get16(pchg_data+2+j*2);
                int r = (change>>8) & 0xf;
                int g = (change>>4) & 0xf;
                int b = (change>>0) & 0xf;
                fi->src.palette[(i * fi->src.colors + (change>>12)) * fi->src.bpc + r_o] = r<<4|r;
                fi->src.palette[(i * fi->src.colors + (change>>12)) * fi->src.bpc + g_o] = g<<4|g;
                fi->src.palette[(i * fi->src.colors + (change>>12)) * fi->src.bpc + b_o] = b<<4|b;
            }
            for(j=0;j<change_count32;++j){
                int change = get16(pchg_data+2+change_count16*2+j*2);
                int r = (change>>8) & 0xf;
                int g = (change>>4) & 0xf;
                int b = (change>>0) & 0xf;
                fi->src.palette[(i * fi->src.colors + 16 + (change>>12)) * fi->src.bpc + r_o] = r<<4|r;
                fi->src.palette[(i * fi->src.colors + 16 + (change>>12)) * fi->src.bpc + g_o] = g<<4|g;
                fi->src.palette[(i * fi->src.colors + 16 + (change>>12)) * fi->src.bpc + b_o] = b<<4|b;
            }
            pchg_data += len2;
            pchg_len -= len2;
        }
    }
    if(allocated_data)
        free(allocated_data);
    return 0;
error:
    if(allocated_data)
        free(allocated_data);
    return -1;
}


static int
_handle_sham(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    int entries;
    int stripes;
    data += 2;
    len -= 2;
    if(fi->src.height*32 == len){
        fi->src.colors = 16;
        return _handle_ctbl(fi, data, len);
    }
    entries = len/2;
    stripes = (fi->src.height-1)/2+1;
    if(entries == 16*stripes){
        int r_o;
        int g_o;
        int b_o;
        int y;
        DEBUG_p("colors: %d",fi->src.colors);

        if(fi->src.palette){
            entries = MIN(entries, fi->src.colors * stripes);
        }else{
            if(!fi->src.bpc)
                fi->src.bpc = 3;
            nile_ensure_palette(&fi->src, entries*2, fi->src.bpc);
            fi->src.colors = entries / stripes;
            if(!fi->src.format)
                fi->src.format = NILE_FORMAT_r8g8b8;
        }
        fi->use_palette_stepping = 1;

        r_o = NILE_OFFSET_RED(fi->src.format);
        g_o = NILE_OFFSET_GREEN(fi->src.format);
        b_o = NILE_OFFSET_BLUE(fi->src.format);
        for(y=0;y<fi->src.height;++y){
            if(y&1){
                memcpy(&fi->src.palette[y * fi->src.colors * fi->src.bpc],
                       &fi->src.palette[(y-1) * fi->src.colors * fi->src.bpc],
                       fi->src.colors * fi->src.bpc);
            }else{
                int i;
                for(i=0;i<fi->src.colors;++i){
                    int col = get16(data + (i + (y/2)*fi->src.colors) * sizeof(uint16_t));
                    int r = (col>>8)&0x0f;
                    int g = (col>>4)&0x0f;
                    int b = (col>>0)&0x0f;
                    fi->src.palette[(i + y * fi->src.colors) * fi->src.bpc + r_o] = r<<4|r;
                    fi->src.palette[(i + y * fi->src.colors) * fi->src.bpc + g_o] = g<<4|g;
                    fi->src.palette[(i + y * fi->src.colors) * fi->src.bpc + b_o] = b<<4|b;
                }
            }
        }
        return 0;
    }
    return -1;
}


static int
_get_bitplanes(nile_t *dst, const uint8_t *src, ssize_t len, frame_info_t *fi)
{
    int stride; // = nil_amiga_plane_stride(fi->width);
    //uint8_t **bitplane;
    int y;

    nile_drop_pixels(dst);
    dst->stride = NILE_STRIDE_ALIGN16;
    nile_ensure_bitplanes(dst, fi->src.width, fi->src.height, fi->src.depth, 1);
    stride = nile_get_stride(dst);

    if(2 == fi->compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        int plane;
        for(plane=0;plane<fi->src.depth;++plane){
            const uint8_t *next_plane;
            const uint16_t *vals;
            int x;
            if(ILBM_VDAT != *(uint32_t*)src)
                return -1;
            DEBUG_p("valid VDAT plane");
            next_plane = src + 8 + get32(src+4);
            vals = (uint16_t*)(src + 8 + get16(src+8));
            src += 10;
            for(x=0;x<stride;x+=2){
                for(y=0;y<fi->src.height;++y){
                    int bits;
                    if(bytes <= 0){
                        if(src>=end)
                            return -1;
                        bytes = *src++;
                        if(bytes>=128){
                            bytes = 256-bytes;
                            runlength = -1;
                        }else if(0==bytes){
                            if(vals+1 > (uint16_t*)end){
                                DEBUG_p("out of data 1");
                                return -1;
                            }
                            bytes = be16toh(*vals++);
                            runlength = -1;
                        }else if(1==bytes){
                            /* it's a long runlength */
                            if(vals+2 > (uint16_t*)end){
                                DEBUG_p("out of data 2");
                                return -1;
                            }
                            bytes = be16toh(*vals++);
                            runlength = *vals++;
                        }else{
                            /* it's a short runlength */
                            if(vals+1 > (uint16_t*)end){
                                DEBUG_p("out of data 3");
                                return -1;
                            }
                            runlength = *vals++;
                        }
                    }
                    --bytes;
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src+1 > end)
                            return -1;
                        bits = *vals++;
                    }
                    *(uint16_t*)(&dst->bitplanes[plane][x+y*stride]) = bits;
                }
            }
            src = next_plane;
        }
    }else if(fi->compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        for(y=0;y<fi->src.height;++y){
            int plane;
            int x;
            for(plane=0;plane<fi->src.depth;++plane){
                for(x=0;x<stride;++x){
                    int bits;
                    if(bytes-- <= 0){
                        if(src>=end)
                            return -1;
                        bytes = *src++;
                        if(bytes>=128){
                            /* it's a runlength */
                            bytes = 256-bytes;
                            if(src>=end)
                                return -1;
                            runlength = *src++;
                        }else{
                            runlength = -1;
                        }
                    }
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src>=end)
                            return -1;
                        bits = *src++;
                    }
                    dst->bitplanes[plane][x+y*stride] = bits;
                }
            }
        }
    }else{
        /* Just uninterlace the bitplanes */
        /* FIXME!! store pointers instead of copying bitplanes */
        for(y=0;y<fi->src.height;++y){
            int plane;
            for(plane=0;plane<fi->src.depth;++plane)
                memcpy(dst->bitplanes[plane] + y*stride, src + (y*fi->src.depth + plane)*stride, stride);
        }
    }
    return 0;
}


#if 0
static uint8_t **
_clone_bitplanes(uint8_t **src, int width, int height, int depth)
{
    int stride = nil_amiga_plane_stride(width);
    uint8_t **dst;
    int y;
    dst = (uint8_t**)malloc(height*sizeof(uint8_t*));
    dst[0] = (uint8_t*)malloc(stride*height*depth);
    for(y=1;y<depth;++y)
        dst[y] = dst[0] + stride*y*height;
    memcpy(dst[0],src[0],stride*height*depth);
    return dst;
}
#endif


static int
_apply_delta(uint8_t **bitplane, int planes, int width, int height, const uint8_t *data, ssize_t len, int type)
{
    //const char *end = data+len;
    int stride;
    int plane;
    switch(type){
    case 5:
        if(len<planes*4){
            DEBUG_p("delta chunk too small");
            return -1;
        }
        stride = nil_amiga_plane_stride(width);
        for(plane=0;plane<planes;++plane){
            off_t offs = get32(data+plane*4);
            int x;
            for(x=0;x<stride;++x){
                int y;
                if(offs>=len){
                    DEBUG_p("trying to get operation count past en of chunk");
                    return -1;
                }
                int count = get8(data+offs++);
                y = 0;
                while(count-->0){
                    int op;
                    if(offs>=len){
                        DEBUG_p("trying to get operation past end of chunk");
                        return -1;
                    }
                    op = get8(data+offs++);

                    if(op & 128){
                        int bytes = op & ~128;
                        while(bytes-->0){
                            if(y >= height){
                                DEBUG_p("trying to apply operation past image height");
                                return -1;
                            }
                            if(offs >= len){
                                DEBUG_p("trying to get pixels past end of chunk (op=%d)",op);
                                return -1;
                            }
                            bitplane[plane][x+y*stride] = get8(data+offs++);
                            ++y;
                        }
                    }else if(0==op){
                        int bytes;
                        int val;
                        if(offs+1 >= len){
                            DEBUG_p("trying to get pixels past end of chunk (op=%d)",op);
                            return -1;
                        }
                        bytes = get8(data+offs++);
                        val = get8(data+offs++);
                        while(bytes-->0){
                            if(y >= height){
                                DEBUG_p("trying to apply operation past image height");
                                return -1;
                            }
                            bitplane[plane][x+y*stride] = val;
                            ++y;
                        }
                    }else{
                        y += op;
                    }
                }
            }
        }
    }
    return 0;
}



static int
_decode_indexed(nile_t *ni, const uint8_t *src, ssize_t len, int compression)
{
    uint8_t *dst;
    int dst_skip;
    int src_stride = ni->width + (ni->width&1);
    nile_ensure_indexed(ni, ni->width, ni->height);
    dst = ni->pixels;
    dst_skip = nile_get_skip(ni);

    if(compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        int y = ni->height;
        while(y-->0){
            int x;
            for(x=0;x<src_stride;++x){
                uint8_t p;
                if(bytes-- <= 0){
                    if(src>=end)
                        goto error;
                    bytes = *src++;
                    if(bytes>=128){
                        /* it's a runlength */
                        bytes = 256-bytes;
                        if(src>=end)
                            goto error;
                        runlength = *src++ % ni->colors;
                    }else{
                        runlength = -1;
                    }
                }
                if(runlength>=0){
                    p = runlength;
                }else{
                    if(src>=end)
                        goto error;
                    p = *src++ % ni->colors;
                }
                if(x < ni->width)
                    *dst++ = p;
            }
            dst += dst_skip;
        }
    }else{
        int dst_stride = nile_get_stride(ni);
        int y = ni->height;
        if(src_stride * ni->height > len)
            goto error;
        while(y-->0){
            memcpy(dst, src, ni->width);
            src += src_stride;
            dst += dst_stride;
        }
    }
    return 0;
error:
    return -1;
}


static int
_handle_interlaced_rgb(nile_t *ni, const uint8_t *src, ssize_t len, int compression)
{
    uint8_t *dst;
    int dst_stride;
    int src_stride = nil_amiga_plane_stride(ni->width-1);
    uint32_t mask[24];
    int plane;
    nile_ensure_direct(ni, ni->width, ni->height, 3);
    if(!ni->format)
        ni->format = NILE_FORMAT_r8g8b8;
    dst = ni->pixels;
    dst_stride = nile_get_stride(ni);
    {
        int r_o = NILE_OFFSET_RED(ni->format);
        int g_o = NILE_OFFSET_GREEN(ni->format);
        int b_o = NILE_OFFSET_BLUE(ni->format);
        for(plane=0;plane<ni->depth;++plane){
            switch(plane/8){
            case 0: /* Red */
                //mask[plane] = be32toh(1<<(r_o*8+plane%8));
                mask[plane] = 1<<(r_o*8+plane%8);
                break;
            case 1: /* Blue */
                //mask[plane] = be32toh(1<<(b_o*8+plane%8));
                mask[plane] = 1<<(g_o*8+plane%8);
                break;
            case 2: /* Green */
                //mask[plane] = be32toh(1<<(g_o*8+plane%8));
                mask[plane] = 1<<(b_o*8+plane%8);
                break;
            }
        }
    }
    if(compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        int y;
        for(y=0;y<ni->height;++y){
            memset(dst, 0, ni->width * ni->bpc);
            for(plane=0;plane<ni->depth;++plane){
                int x;
                for(x=0;x<src_stride;++x){
                    int bits;
                    int b;
                    if(bytes-- <= 0){
                        if(src>=end)
                            return -1;
                        bytes = *src++;
                        if(bytes>=128){
                            /* it's a runlength */
                            bytes = 256-bytes;
                            if(src>=end)
                                return -1;
                            runlength = *src++;
                        }else{
                            runlength = -1;
                        }
                    }
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src>=end)
                            return -1;
                        bits = *src++;
                    }
                    for(b=0;b<8;++b){
                        if(x*8+b >= ni->width)
                            break;
                        if(bits & 1<<(7-b))
                            *((uint32_t*)dst) |= mask[plane];
                        dst += ni->bpc;
                    }
                }
                dst -= ni->width * ni->bpc;
            }
            dst += dst_stride;
        }
    }else{
        int y;
        if(src_stride * ni->height * ni->depth > len)
            return -1;
        for(y=0;y<ni->height;++y){
            memset(dst, 0, ni->width * ni->bpc);
            for(plane=0;plane<ni->depth;++plane){
                int x;
                for(x=0;x<src_stride;++x){
                    int bits = *src++;
                    int b;
                    for(b=0;b<8;++b){
                        if(x*8+b >= ni->width)
                            break;
                        if(bits & 1<<(7-b))
                            *((uint32_t*)dst) |= mask[plane];
                        dst += ni->bpc;
                    }
                }
                dst -= ni->width * ni->bpc;
            }
            dst += dst_stride;
        }
    }
    return 0;
}


static int
_move(nile_t *dst, nile_t *src, int camg, int palette_stepping)
{
    switch(src->type){
    case NILE_TYPE_NONE:
        return -1;
    case NILE_TYPE_BITPLANES:
    case NILE_TYPE_BITPLANES_DIRECT:
        if(src->width >= 256 && src->height >= 3 && nil_amiga_is_dctv(src, camg)){
            return nil_amiga_decode_dctv(dst, src, camg);
        }else if((camg & BPLCON0_HAM) || (palette_stepping>0)){
            nil_amiga_direct_from_bitplanes(dst, src, camg, palette_stepping, 0);
            return 0;
        }else if(camg & BPLCON0_UHRES){
            nile_t tmp;
            nile_copy(&tmp, src, 0);
            --tmp.depth;
            ni_ehb_expand_palette(&tmp);
            nile_move(dst, &tmp);
            return 0;
        }else{
            /* No amiga specific feature enabled. */
            nile_move(dst, src);
            return 0;
        }
    case NILE_TYPE_INDEXED:
        /* palette stepping not supported (?) */
        nile_move(dst, src);
        return 0;
    case NILE_TYPE_DIRECT:
        nile_move(dst, src);
        return 0;
    }
    /* not reached */
    return -1;
}


static int
_move_from_frame_info(nile_t *dst, frame_info_t *fi)
{
    if(fi->is_hame){
        nil_amiga_direct_from_hame(fi->dst, &fi->src, fi->camg);
        return 0;
    }else{
        int palette_stepping = fi->use_palette_stepping ? fi->src.colors : 0;
        return _move(dst, &fi->src, fi->camg, palette_stepping);
    }
}


static void
_sanity_check(frame_info_t *fi){
    DEBUG_p("width      : %d",fi->src.width);
    DEBUG_p("height     : %d",fi->src.height);
    DEBUG_p("depth      : %d/%d",fi->src.depth,(fi->camg>>12) & 7);
    DEBUG_p("camg       : %x",fi->camg);
    DEBUG_p("LACE       : %s",fi->camg & BPLCON0_LACE ? "yes" : "no");
    DEBUG_p("UHRES      : %s",fi->camg & BPLCON0_UHRES ? "yes" : "no");
    DEBUG_p("HAM        : %s",fi->camg & BPLCON0_HAM ? "yes" : "no");
    DEBUG_p("HIRES      : %s",fi->camg & BPLCON0_HIRES ? "yes" : "no");
    DEBUG_p("pixel ratio: %f",fi->pixel_ratio);
    DEBUG_p("colors     : %d",fi->src.colors);

    /*
        When EHB should not be forced:

            width      : 640
            height     : 480
            depth      : 4/0
            camg       : 8004
            LACE       : yes
            UHRES      : no
            HAM        : no
            HIRES      : yes
            pixel ratio: 0.909091
            colors     : 16

        When EHB should be forced:

            width      : 320
            height     : 512
            depth      : 6/4
            camg       : 4084
            LACE       : yes
            UHRES      : yes
            HAM        : no
            HIRES      : no
            pixel ratio: 1.818182
            colors     : 64

    */
    if(6 == fi->src.depth && fi->src.colors<64 && !(fi->camg & BPLCON0_DPF) && !(fi->camg & BPLCON0_HAM)){
        fi->camg |= BPLCON0_UHRES;
    }

    if(fi->camg & BPLCON0_UHRES && fi->src.colors > 32 && fi->src.colors >= 1<<fi->src.depth){
        fi->src.colors /= 2;
    }

    if(fi->camg & BPLCON0_HAM && fi->src.colors > 32 && fi->src.colors >= 1<<fi->src.depth){
        fi->src.colors /= 2;
    }

    if(fi->screen_width && fi->screen_height){
        double physical_ratio = fi->pixel_ratio*fi->screen_width/fi->screen_height;
        DEBUG_p("claimed physical screen ratio: %f",physical_ratio);
        if(physical_ratio < 1 || physical_ratio > 2.35){
            /* unreasonable screen ratio */
            fi->pixel_ratio = _pixel_ratio_from_screen_size(fi->screen_width, fi->screen_height);
        }
    }
}


static void
_request_destination(context_t *c, frame_info_t *fi)
{
    if(!fi->dst)
        c->request_func(c->userdata, fi->src.width, fi->src.height, fi->pixel_ratio, c->frame_count, &fi->dst);
}


static int
_handle_ilbm(context_t *c, frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    while(len >= 8){
        ssize_t chunk_len = get32(data+4);
        if(chunk_len > len-8)
            return -1;
        switch(*(uint32_t*)data){
        case ILBM_BMHD:
            if(_handle_bmhd(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_ANHD:
            if(_handle_anhd(&fi->di,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_ANNO:
            DEBUG_p("ANNO");
            break;
        case ILBM_CMAP:
            if(_handle_cmap(&fi->src, data+8, chunk_len) < 0)
                return -1;
            break;
        case ILBM_CAMG:
            if(_handle_camg(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_DPAN: /* Deluxe Paint Animation */
            DEBUG_p("DPAN size: %d",(int)chunk_len);
            if(chunk_len>=8){
                DEBUG_p("DPAN version: %d", get16(data+8+0));
                DEBUG_p("DPAN frames : %d", get16(data+8+2));
                DEBUG_p("DPAN FPS    : %d", get8(data+8+4));
                /* Set duration multiplier based on FPS. */
                fi->duration_multiplier = 1.0/get8(data+8+4);
            }
            break;
        case ILBM_DPPS: /* Deluxe Paint Preview Storage or Deluxe Paint III Perspective? */
            DEBUG_p("DPPS");
            /* FIXME!! some bits might be related to animation speed */
            if(chunk_len>=108){
                DEBUG_p("DPPS offset  74: %d", get16(data+8+74));
                DEBUG_p("DPPS offset  90: %d", get16(data+8+90));
                DEBUG_p("DPPS offset 106: %d", get16(data+8+106));
            }
            break;
        case ILBM_CRNG:
            if(_handle_crng(fi,data+8,chunk_len) < 0)
                return -1;
            break;
        case ILBM_CTBL:
        case ILBM_BEAM:
            if(_handle_ctbl(fi,data+8,chunk_len) < 0)
                return -1;
            break;
        case ILBM_DLTA:
            if(_apply_delta(fi->src.bitplanes,fi->src.depth,fi->src.width,fi->src.height,data+8,chunk_len,fi->di.compression) < 0)
                return -1;
            //frame->surface = abydos_image_surface_create_from_amiga_bitplanes((const uint8_t**)fi->bitplane,fi->width,fi->height,fi->depth,fi->camg,fi->palette,fi->colors,0);
            /* Nothing else should appear after the DLTA. */
            return 0;
        case ILBM_PCHG:
            if(_handle_pchg(fi,data+8,chunk_len) < 0)
                return -1;
            break;
        case ILBM_SHAM:
            if(_handle_sham(fi,data+8,chunk_len) < 0)
                return -1;
            break;
        case ILBM_BODY:
            if(fi->src.width <= 0 || fi->src.height <= 0)
                return -1;
            switch(fi->type){
            case IFF_ILBM:
                if(fi->src.colors <= 0){
                    _request_destination(c, fi);
                    if(fi->dst->type)
                        nile_copy_pixels(&fi->src, fi->dst, 0);
                    if(_handle_interlaced_rgb(&fi->src, (uint8_t*)data+8,chunk_len, fi->compression) < 0)
                        return -1;
                }else{
                    _sanity_check(fi);
                    if(_get_bitplanes(&fi->src, (uint8_t*)data+8,chunk_len,fi) < 0)
                        return -1;
                    if(4==fi->src.depth && nil_amiga_is_hame(&fi->src, 0)){
                        fi->src.width /= 2;
                        fi->pixel_ratio *= 2;
                        fi->is_hame = 1;
                        //nile_copy(&fi->src, fi->dst, 0);
                        //_image_done(&fi->image);
                        //_image_init(&fi->image);
                    }
                }
                break;
            case IFF_PBM:
                if(fi->src.colors <= 0)
                    return -1;
                if(_decode_indexed(&fi->src, data+8,chunk_len,fi->compression) < 0)
                    return -1;
                break;
            }
            /* Nothing else should appear after the BODY. */
            return 0;
        default:
            DEBUG_p("unhandled chunk (%4s)",data);
            break;
        }
        chunk_len += chunk_len&1;
        data += chunk_len+8;
        len -= chunk_len+8;
    }
    /* Body not found. */
    return -1;
}


static int
_handle_adat(uint8_t *bitplane, frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    const int8_t *p = (int8_t*)data+2;
    uint16_t *dst;
    const uint16_t *dst_end;
    const uint16_t *src;
    const uint16_t *src_end;
    int stride;

    if(len<2)
        return 0;

    src = (uint16_t*)(data+get16(data));
    src_end = (uint16_t*)(data+len);
    dst = (uint16_t*)bitplane;
    stride = fi->src.width / 16;
    dst_end = dst + fi->src.height * stride;

    while(src < src_end && p < (int8_t*)src_end){
        int b = *p++;
        if(0==b){
            b = (int16_t)be16toh(*src++);
            DEBUG_p("move %d bytes", b);
            dst += b/2;
            continue;
        }

        if(dst < (uint16_t*)bitplane){
            DEBUG_p("bitplane underrun by %d words", (int)((uint16_t*)bitplane - dst));
            return -1;
        }else if(dst >= dst_end){
            DEBUG_p("bitplane overrun by %d words", (int)(dst - dst_end));
            return -1;
            //return 0;
        }

        if(b<0){
            if(-1 == b){
                b = be16toh(*src++);
            }else{
                b = -1 -b;
            }
            DEBUG_p("copy %d words",b);
            if(src + b > src_end)
                return -1;
            while(b-->0){
                *dst ^= *src++;
                dst += stride;
            }
        }else{
            int bits = *src++;
            DEBUG_p("repeat %d words",b);
            while(b-->0){
                *dst ^= bits;
                dst += stride;
            }
        }
    }
    return 0;
}


static int
_handle_vdlt(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    int plane = 0;
    while(len >= 8 && plane < fi->src.depth){
        ssize_t chunk_len = get32(data+4);
        if(chunk_len > len-8)
            return -1;
        switch(*(uint32_t*)data){
        case VDLT_ADAT:
            if(_handle_adat(fi->src.bitplanes[plane],fi,data+8,chunk_len)<0)
                return -1;
            ++plane;
            break;
        }
        chunk_len += chunk_len&1;
        data += chunk_len+8;
        len -= chunk_len+8;
    }
    return 0;
}


static int
_count_frames(const uint8_t *data, ssize_t len)
{
    int count = 0;
    while(len >= 8){
        uint32_t chunk_type = *(uint32_t*)data;
        ssize_t chunk_len = get32(data+4);
        data += 8;
        if(IFF_FORM == chunk_type){
            switch(*(uint32_t*)data){
            case IFF_ILBM:
            case IFF_PBM:
            case IFF_VDLT:
                ++count;
                break;
            }
        }
        data += chunk_len;
        len -= chunk_len+8;
    }
    return count;
}


static int
_handle_dpah(frame_info_t *fi, const uint8_t *data, ssize_t len)
{
    if(len>=2){
        fi->default_duration = get16(data+0);
        DEBUG_p("default duration: %d",fi->default_duration);
    }
    if(len>=4)
        fi->frame_count = get16(data+2);
    return 0;
}


static int
_handle_anim(context_t *c, const uint8_t *data, ssize_t len, int buffers)
{
    frame_info_t fi;
    nile_t buffer[4];
    int ret = -1;
    int i;

    c->frame_count = _count_frames(data,len);
    if(c->frame_count<=0)
        return -1;
    DEBUG_p("found %d frames",c->frame_count);
    _frame_info_init(&fi);
    fi.src.bpc = c->bpc;
    fi.src.format = c->format;
    for(i=0;i<c->frame_count;){
        uint32_t chunk_type = *(uint32_t*)data;
        ssize_t chunk_len = get32(data+4);
        data += 8;
        switch(chunk_type){
        case DPST_DPAH:
            if(_handle_dpah(&fi,data,len)<0)
                return -1;
            break;
        case IFF_FORM:
            fi.type = *(uint32_t*)data;
            switch(fi.type){
            case IFF_ILBM:
            case IFF_VDLT:
                DEBUG_p("frame %d", i);
                if(i>0)
                    nile_copy_pixels(&fi.src, &buffer[i%buffers], 0);
                fi.di.duration = fi.default_duration;
                if(IFF_ILBM == fi.type){
                    ret = _handle_ilbm(c, &fi,data+4,chunk_len-4);
                }else{
                    ret = _handle_vdlt(&fi,data+4,chunk_len-4);
                }
                if(ret)
                    goto out;
                if(0==i){
                    int j;
                    nile_init(&buffer[0]);
                    nile_move(&buffer[0], &fi.src);
                    for(j=1;j<buffers;++j){
                        nile_init(&buffer[j]);
                        nile_copy(&buffer[j], &buffer[0], 0);
                        nile_copy_pixels(&buffer[j], &buffer[0], 1);
                    }
                }
                _request_destination(c, &fi);
                _move_from_frame_info(&fi.dst[i], &fi);
                ret = c->frame_func(c->userdata, i, fi.di.duration * fi.duration_multiplier);
                if(ret)
                    goto out;
                ++i;
                break;
            }
            /*
            if(fi.width > h->info->width)
                h->info->width = fi.width;
            if(fi.height > h->info->height)
                h->info->height = fi.height;
            */
            break;
        }
        data += chunk_len;
        len -= chunk_len+8;
    }
    ret = 0;
out:
    _frame_info_done(&fi);
    return ret;
}


int
ni_ilbm_decode(const uint8_t *data, size_t len, int bpc, nile_format_t format, ni_ilbm_request_func_t request_func, ni_ilbm_frame_func_t frame_func, void *userdata)
{
    /* for single frame images */
    context_t c;
    frame_info_t fi;
    int ret;
    size_t form_len;
    if(len < 12)
        return -1;
    if(*(uint32_t*)data != IFF_FORM)
        return -1;
    form_len = get32(data+4)-4;
    if(form_len > len-12){
        DEBUG_p("truncated data");
        return -1;
    }
    c.frame_count = 1;
    c.bpc = bpc;
    c.format = format;
    c.request_func = request_func;
    c.frame_func = frame_func;
    c.userdata = userdata;
    fi.type = *(uint32_t*)(data+8);
    data += 12;
    switch(fi.type){
    case IFF_ILBM:
    case IFF_PBM:
        _frame_info_init(&fi);
        fi.src.bpc = c.bpc;
        fi.src.format = c.format;
        ret = _handle_ilbm(&c, &fi, data, form_len);
        if(0==ret){
            if(fi.cycle_count>0){
                int i;
                DEBUG_p("got %d color cycles",fi.cycle_count);
#if 1
                c.frame_count = nil_cycle_total_frames(fi.cycle, fi.cycle_count);
                DEBUG_p("total frames: %d", c.frame_count);
#else
                for(i=0;i<fi.cycle_count;++i){
                    nil_cycle_t *c = &fi.cycle[i];
                    DEBUG_p("%d: duration: %f",i,c->duration);
                    DEBUG_p("%d: range %d - %d",i,c->first,c->last);
                }
#endif
                _request_destination(&c, &fi);
                for(i=0;i<c.frame_count;++i){
                    _move_from_frame_info(&fi.dst[i], &fi);
                    ret = frame_func(userdata, i, fi.cycle[0].duration);
                    if(ret)
                        break;
                    /* cycle the original palette since we create a surface
                       for every frame then ditch the palette (for now). */
                    nil_cycle_apply_step(&fi.cycle[0], fi.src.palette, fi.src.bpc);
                }
            }else{
                _request_destination(&c, &fi);
                _move_from_frame_info(&fi.dst[0], &fi);
                frame_func(userdata, 0, 0);
            }
        }
        _frame_info_done(&fi);
        return ret;
    case IFF_ANIM:
        return _handle_anim(&c, data, form_len, 2);
    case IFF_DPST:
        return _handle_anim(&c, data, form_len, 1);
    default:
        DEBUG_p("unhandled type (%4s)",data);
        return -1;
    }
}
