/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <byteswap.h>
#include <debug.h>
#include <nile-blit.h>
#include <nil-cairo.h>
#include <stdint.h>                     /* uint16_t */
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_ingr_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-intergraph-raster",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ingr_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_ingr_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16n(a) (*(uint16_t*)(a))
#define get32n(a) (*(uint32_t*)(a))
#define get16b(a) be16toh(get16n(a))
#define get32b(a) be32toh(get32n(a))
#define get16l(a) le16toh(get16n(a))
#define get32l(a) le32toh(get32n(a))


static int
_verify_len(abydos_plugin_handle_t *h, int bpc, const char *data, const char *end)
{
    return data + h->info->width * h->info->height * bpc > end ? -1 : 0;
}


/* FIXME!! this might not be needed for older versions of cairo */
static uint32_t
_scan_max(const char *src, int width, int height, int stride, int bpc, nile_format_t format)
{
    int skip = stride > 0 ? stride - width : 0;
    int r_o = NILE_OFFSET_CHANNEL(format,0);
    int g_o = NILE_OFFSET_CHANNEL(format,1);
    int b_o = NILE_OFFSET_CHANNEL(format,2);
    uint32_t max;
    switch(NILE_GET_BYTES_PER_CHANNEL(format)){
    case 1:
        return UINT8_MAX;
    case 2:
        max = UINT8_MAX << 8;
        while(height-->0){
            int x;
            for(x=0; x<width; ++x){
                uint32_t val = get16n(src+r_o) | get16n(src+g_o) | get16n(src+b_o);
                if((format & NILE_ENDIANESS) != NILE_ENDIAN_NATIVE)
                    val = bswap_16(val);
                if(val & 0xff)
                    return UINT16_MAX;
                src += bpc;
            }
            src += skip;
        }
        break;
    case 4:
        max = UINT8_MAX << 24;
        while(height-->0){
            int x;
            for(x=0; x<width; ++x){
                uint32_t val = get32n(src+r_o) | get32n(src+g_o) | get32n(src+b_o);
                if((format & NILE_ENDIANESS) != NILE_ENDIAN_NATIVE)
                    val = bswap_32(val);
                if(val & 0xffff){
                    return UINT32_MAX;
                }else if(val & 0xffffff){
                    max = UINT16_MAX << 16;
                }
                src += bpc;
            }
            src += skip;
        }
        break;
    default:
        /* should not be reached */
        break;
    }
    return max;
}


#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
static void
_blit16be(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    uint32_t max = *(uint32_t*)blit->userdata;
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = (float)be16toh(*(uint16_t*)&src[blit->src_o[0]]) / max;
        *(float*)&dst[blit->dst_o[1]] = (float)be16toh(*(uint16_t*)&src[blit->src_o[1]]) / max;
        *(float*)&dst[blit->dst_o[2]] = (float)be16toh(*(uint16_t*)&src[blit->src_o[2]]) / max;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}
#endif


#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
static void
_blit32be(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    uint32_t max = *(uint32_t*)blit->userdata;
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = (float)be32toh(*(uint32_t*)&src[blit->src_o[0]]) / max;
        *(float*)&dst[blit->dst_o[1]] = (float)be32toh(*(uint32_t*)&src[blit->src_o[1]]) / max;
        *(float*)&dst[blit->dst_o[2]] = (float)be32toh(*(uint32_t*)&src[blit->src_o[2]]) / max;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}
#endif


static int
_handle_environv_palette(float *dst, const char *src, const char *end, int count)
{
    /* The 16-bit values might acually be 8-bit numbers so we have to make
     * sure they are acually 16-bit.
     */
    uint32_t max = _scan_max(src, count, 1, -1, 8, NILE_FORMAT_le_x16r16g16b16);

    while(count-->0){
        int i = get16l(src);
        if(i >= 256)
            continue;
        dst[i*3+0] = (float)get16l(src+2) / max;
        dst[i*3+1] = (float)get16l(src+4) / max;
        dst[i*3+2] = (float)get16l(src+6) / max;
        src += 8;
    }
    return 0;
}


#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
static cairo_surface_t *
_surface_from_data(const char *data, int width, int height, int bpc, nile_format_t format)
{
    uint32_t max = _scan_max(data, width, height, -1, 4, format);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB96F, width, height);
    uint8_t *dst = cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface);
    nile_blit_t blit;

    nile_blit_init_rgb(&blit, 12, NIL_CAIRO_FORMAT_RGB96F, bpc, format);
    /* override with a custom blit func */
    switch(NILE_GET_BYTES_PER_CHANNEL(format)){
    case 16:
        blit.blit_func = _blit16be;
        break;
    case 32:
        blit.blit_func = _blit32be;
        break;
    }
    blit.userdata = &max;
    nile_blit(&blit, dst, dst_stride, (uint8_t*)data, width * bpc, width, height, 0);
    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


static int
_ingr_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    const char *end = data + len;
    int WordsToFollow;
    int DataTypeCode;
    unsigned int NextImageOffset;
    int ColorTableType;
    int ColorTableEntries;
    int StartOfAppPacket;
    const char *color_table;
    if(len < 1024)
        return -1;
    if(data[0] != 8 || data[1] != 9)
        return -1;
    WordsToFollow = get16l(data+2);
    DEBUG_p("words to follow: %d", WordsToFollow);
    DataTypeCode = get16l(data+4);
    DEBUG_p("data type code: %d", DataTypeCode);
    h->info->width = get32l(data+184);
    h->info->height = get32l(data+188);
    DEBUG_p("size: %dx%d", h->info->width, h->info->height);

    if(WordsToFollow >= 510){
        DEBUG_p("there is a second header block");
        NextImageOffset = get32l(data + 512 + 16);
        ColorTableType = get16l(data + 512 + 20);
        ColorTableEntries = get32l(data + 512 + 24);
        StartOfAppPacket = get32l(data + 512 + 28);
        color_table = data + 512 * 2;
    }else{
        NextImageOffset = 0;
        ColorTableType = 0;
    }
    DEBUG_p("next image offset: %d", NextImageOffset);
    DEBUG_p("color table type: %d (%d)", ColorTableType, ColorTableEntries);
    DEBUG_p("start of app packet: %d", StartOfAppPacket);

    data += 4 + WordsToFollow * 2;
    DEBUG_p("bytes left: %d", (int)(end-data));
    switch(DataTypeCode){
    case 2: /* BYTE array 8 bits per pixel */
        if(_verify_len(h, 1, data, end) < 0)
            return -1;
        switch(ColorTableType){
        case 0: /* Grayscale */
            h->surface = nil_cairo_surface_create_from_direct((uint8_t*)data, h->info->width, h->info->height, -1, 1, NILE_FORMAT_g8, 0);
            break;
        case 1: /* IGDS color table */
            /* FIXME!! */
            return -1;
        case 2: /* Environ V color table */
            {
                float *palette = calloc(256, sizeof(float) * 3);
                _handle_environv_palette(palette, color_table, end, ColorTableEntries);
                h->surface = nil_cairo_surface_create_from_indexed((uint8_t*)data, h->info->width, h->info->height, -1,
                                                                   (uint8_t*)palette, 256, sizeof(float) * 3, NILE_FORMAT_f_r32g32b32, 0);
                free(palette);
            }
            break;
        default:
            return -1;
        }
        break;
    case 3: /* WORD array 16 bits per pixel */
        if(_verify_len(h, 4, data, end) < 0)
            return -1;
        switch(ColorTableType){
        case 0: /* Grayscale */
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            h->surface = _surface_from_data(data, h->info->width, h->info->height, 2, NILE_FORMAT_be_g16);
#else
            h->surface = nil_cairo_surface_create_from_direct((uint8_t*)data, h->info->width, h->info->height, -1, 4, NILE_FORMAT_g32|NILE_ENDIAN_BIG, 0);
#endif
            break;
        case 1: /* IGDS color table */
            /* not supported */
        case 2: /* Environ V color table */
            /* not supported */
        default:
            return -1;
        }
        break;
    case 4: /* DWORD array 32 bits per pixel */
        if(_verify_len(h, 4, data, end) < 0)
            return -1;
        switch(ColorTableType){
        case 0: /* Grayscale */
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            h->surface = _surface_from_data(data, h->info->width, h->info->height, 4, NILE_FORMAT_be_g32);
#else
            h->surface = nil_cairo_surface_create_from_direct((uint8_t*)data, h->info->width, h->info->height, -1, 4, NILE_FORMAT_g32|NILE_ENDIAN_BIG, 0);
#endif
            break;
        case 1: /* IGDS color table */
            /* not supported */
        case 2: /* Environ V color table */
            /* not supported */
        default:
            return -1;
        }
        break;
    case 9: /* Run-Length Encoded 1-bit */
        /* FIXME!! how is this encoded? (frmt09.cot) */
        return -1;
    case 10: /* Run-Length Encoded Grayscale or color */
        /* FIXME!! how is this encoded? (frmt10.cot) */
        return -1;
    case 24: /* CCITT Group 4 1-bit */
        /* FIXME!! how is this encoded? (frmt24.cot) */
        return -1;
    case 27: /* Adaptive RLE RGB */
        /* FIXME!! how is this encoded? (frmt27.cot) */
        return -1;
    case 28:
        if(_verify_len(h, 3, data, end) < 0)
            return -1;
        h->surface = nil_cairo_surface_create_from_direct((uint8_t*)data, h->info->width, h->info->height, -1, 3, NILE_FORMAT_r8g8b8, 0);
        break;
    case 29: /* Adaptive RLE Grayscale */
        /* FIXME!! how is this encoded? (frmt29.cot) */
        return -1;
    case 65: /* Tiled */
        /* FIXME!! how is this encoded? (frmt09t.cot) */
        return -1;
    default:
        DEBUG_p("unhandled data type: %d", DataTypeCode);
        return -1;
    }
    return 0;
}


static cairo_surface_t *
_ingr_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ingr_get_mime_types,
        .new = _ingr_new,
        .free = _ingr_free,
        .create_from_data = _ingr_create_from_data,
        .get_image_surface = _ingr_get_image_surface
    };
    return &p;
}
