/*
 * Copyright (C) 2022 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <nil-color.h>
#include <stdlib.h>

#include <jxl/decode.h>

#define TEST_DIRECT

#if __BYTE_ORDER == __BIG_ENDIAN
# define NIL_JXL_FORMAT_NATIVE_rgb8 (NILE_FORMAT_a8b8g8r8)
#else
# define NIL_JXL_FORMAT_NATIVE_rgb8 (NILE_FORMAT_r8g8b8a8)
#endif


typedef struct{
    cairo_surface_t *surface;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_jpegxl_get_mime_types()
{
    static const char *mime_types[] = {
        "image/jxl",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_jpegxl_new(const char *mime_type, abydos_plugin_info_t *info){
    abydos_plugin_handle_t *h = malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->frame = NULL;

    return h;
}


static void
_jpegxl_free(abydos_plugin_handle_t *h){
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static nile_format_t
_convert_format(JxlPixelFormat *src, int *bpc)
{
    switch(src->data_type){
    case JXL_TYPE_FLOAT:
        switch(src->num_channels){
        case 3:
            *bpc = 12;
            return NILE_FORMAT_f_r32g32b32;
        case 4:
            *bpc = 16;
            return NILE_FORMAT_f_r32g32b32a32;
        }
        break;
    case JXL_TYPE_FLOAT16:
        break;
    case JXL_TYPE_UINT16:
        switch(src->num_channels){
        case 3:
            *bpc = 6;
            return NILE_FORMAT_ne_r16g16b16;
        case 4:
            *bpc = 8;
            return NILE_FORMAT_ne_r16g16b16a16;
        }
        break;
    case JXL_TYPE_UINT8:
        *bpc = 4;
        switch(src->endianness){
        case JXL_NATIVE_ENDIAN:
            return NIL_JXL_FORMAT_NATIVE_rgb8;
        case JXL_LITTLE_ENDIAN:
            return NILE_FORMAT_a8b8g8r8;
        case JXL_BIG_ENDIAN:
            return NILE_FORMAT_r8g8b8a8;
        }
        break;
    }
    return NILE_FORMAT_UNKNOWN;
}


#ifdef TEST_DIRECT
static JxlPixelFormat
_format_from_surface(cairo_surface_t *surface)
{
    JxlPixelFormat format = {4, JXL_TYPE_UINT8, JXL_LITTLE_ENDIAN, 0};
    switch(cairo_image_surface_get_format(surface)){
    case CAIRO_FORMAT_ARGB32:
        format.data_type = JXL_TYPE_UINT8;
        break;
    case CAIRO_FORMAT_RGBA128F:
        format.data_type = JXL_TYPE_FLOAT;
        break;
    }
    return format;
}


static void
_premultiply_surface(cairo_surface_t *surface)
{
    nile_t ni;
    nile_init(&ni);
    nil_cairo_from_surface(&ni, surface);
    nil_color_premultiply(&ni);
    nile_done(&ni);
}
#endif


static int
_jpegxl_create_from_data(abydos_plugin_handle_t *h, const char *sdata, size_t len)
{
    JxlDecoder *dec = JxlDecoderCreate(NULL);
#ifdef TEST_DIRECT
    cairo_format_t format = CAIRO_FORMAT_ARGB32;
    cairo_surface_t *surface = NULL;
#else
    JxlPixelFormat format = {4, JXL_TYPE_UINT8, JXL_NATIVE_ENDIAN, 0};
    uint8_t *buffer = NULL;
#endif
    JxlDecoderSubscribeEvents(dec, JXL_DEC_BASIC_INFO | JXL_DEC_FULL_IMAGE);
    JxlDecoderSetInput(dec, (uint8_t*)sdata, len);
    h->frame = NULL;
    h->info->frame_count = 0;
    for(;;){
        JxlBasicInfo info;
        switch(JxlDecoderProcessInput(dec)){
        case JXL_DEC_ERROR:
            DEBUG_p("error");
            goto done;
        case JXL_DEC_NEED_MORE_INPUT:
            DEBUG_p("need more data");
            goto done;
        case JXL_DEC_BASIC_INFO:
            if(JxlDecoderGetBasicInfo(dec, &info) != JXL_DEC_SUCCESS)
                goto done;
            DEBUG_p("size: %dx%d", info.xsize, info.ysize);
            h->info->width = info.xsize;
            h->info->height = info.ysize;
            switch(info.num_color_channels){
            case 1:
            case 3:
#ifndef TEST_DIRECT
                format.num_channels = 3;
#endif
                break;
            case 2:
            case 4:
#ifndef TEST_DIRECT
                format.num_channels = 4;
#endif
                break;
            default:
                goto done;
            }
            if(info.bits_per_sample > 8){
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
# ifdef TEST_DIRECT
                format = CAIRO_FORMAT_RGBA128F;
# else
                format.data_type = JXL_TYPE_FLOAT;
# endif
#elif CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,12,0)
                /* Prefer 16 bit channels which can be converted to 10 bit. */
                format.data_type = JXL_TYPE_UINT16;
#endif
            }
            break;
        case JXL_DEC_NEED_IMAGE_OUT_BUFFER:
            {
#ifdef TEST_DIRECT
                JxlPixelFormat jxl_format;
                surface = cairo_image_surface_create(format, info.xsize, info.ysize);
                jxl_format = _format_from_surface(surface);
                if(JxlDecoderSetImageOutBuffer(dec,
                                               &jxl_format,
                                               cairo_image_surface_get_data(surface),
                                               cairo_image_surface_get_stride(surface) * info.ysize) != JXL_DEC_SUCCESS)
                    goto done;
#else
                size_t buffer_size;
                if(JxlDecoderImageOutBufferSize(dec, &format, &buffer_size) != JXL_DEC_SUCCESS)
                    goto done;
                if(buffer_size < info.xsize * info.ysize * 4)
                    goto done;
                buffer = malloc(buffer_size);
                if(JxlDecoderSetImageOutBuffer(dec, &format, buffer, buffer_size) != JXL_DEC_SUCCESS)
                    goto done;
#endif
            }
            break;
        case JXL_DEC_FULL_IMAGE:
            DEBUG_p("full image");
#ifdef TEST_DIRECT
            if(!surface)
                goto done;
#else
            if(!buffer)
                goto done;
#endif
            if(h->frame){
                h->frame = realloc(h->frame, sizeof(frame_t) * (h->info->frame_count + 1));
            }else{
                h->frame = malloc(sizeof(frame_t));
            }
#ifdef TEST_DIRECT
            _premultiply_surface(surface);
            h->frame[h->info->frame_count++].surface = surface;
            surface = NULL;
#else
            {
                nile_format_t nile_format;
                int bpc; /* bytes per color */
                nile_format = _convert_format(&format, &bpc);
                if(NILE_FORMAT_UNKNOWN == nile_format)
                    goto done;
                h->frame[h->info->frame_count++].surface = nil_cairo_surface_create_from_direct(buffer,
                                                                                                info.xsize,
                                                                                                info.ysize,
                                                                                                info.xsize * bpc,
                                                                                                bpc,
                                                                                                nile_format,
                                                                                                0);
            }
#endif
            break;
        case JXL_DEC_SUCCESS:
            DEBUG_p("success");
            goto done;
        default:
            DEBUG_p("unhandled event");
            break;
        }
    }
done:
    JxlDecoderReleaseInput(dec);
    JxlDecoderDestroy(dec);
#ifdef TEST_DIRECT
    if(surface)
        cairo_surface_destroy(surface);
#else
    if(buffer)
        free(buffer);
#endif
    return h->frame ? 0 : -1;
}


static cairo_surface_t *
_jpegxl_get_image_surface(abydos_plugin_handle_t *h,
                          int page,
                          cairo_bool_t *layer_visibility,
                          int variant,
                          double time,
                          int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _jpegxl_get_mime_types,
        .new = _jpegxl_new,
        .free = _jpegxl_free,
        .create_from_data = _jpegxl_create_from_data,
        .get_image_surface = _jpegxl_get_image_surface
    };
    return &p;
}
