/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |   20 | ID = "KYGformat"
    20 |   80 | message
   100 |    1 | oem
   101 |    1 | padding
   102 |    2 | width
   104 |    2 | height
   106 |    2 | x position
   108 |    2 | y position
   110 |    2 | padding
   112 |    4 | size
   116 |    2 | type
   118 |    2 | color
   120 |   12 | reserved

*/

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_kyg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-kyss-graphics",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_kyg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_kyg_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))


static void
_decompress(uint16_t *dst, size_t dst_len, int stride, const uint8_t *src, size_t src_len)
{
    uint8_t nrn = 0;
    int i = 0;
    int dp = 0;
    while(i < dst_len){
        uint16_t base;
        uint8_t count;
        uint8_t rd;
        if((rd = src[dp] & 0x0F) == 0 || nrn > 0){
            if(nrn){
                --nrn;
            }else{
                nrn = (src[dp++] >> 4) - 1;
            }
            base = get16(src + dp);
            dp += 2;
            count = base & 0x8000 ? 1 : src[dp++];
            base &= 0x7FFF;
        }else{
            nrn = src[dp++] >> 4;
            count = rd & 0x08 ? 1 : src[dp++];
            base = dst[i - stride - 4 + (rd & 0x07)];
        }
        for(;count>0;--count,++i)
            dst[i] = base;
    }
}


static cairo_surface_t *
_surface_from_pixels(uint16_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    while(height-->0){
        int x;
        for(x=width;x>0;--x){
            int pix = *src++;
            int r = (pix >> 5) & 31;
            int g = pix >> 10;
            int b = pix & 31;
            r = r << 3 | r >> 2;
            g = g << 3 | g >> 2;
            b = b << 3 | b >> 2;
            *dst++ = r << 16 | g << 8 | b << 0;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_kyg_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    size_t src_len;
    uint16_t *pixels;

    if(len < 128){
        DEBUG_p("file too short");
        return -1;
    }
    if(strncmp(data, "KYGformat", 9 )){
        DEBUG_p("not KYG format");
        return -1;
    }
    h->info->width = get16(data+102);
    h->info->height = get16(data+104);
    src_len = get32(data+112);

    DEBUG_p("type: %d, color: %d", get16(data+116), get16(data+118));

    data += 132;
    len -= 132;

    if(len < src_len){
        DEBUG_p("file too short");
        return -1;
    }

    pixels = (uint16_t*)malloc(h->info->width * h->info->height * sizeof(uint16_t));

    _decompress(pixels, h->info->width * h->info->height, h->info->width, (uint8_t*)data, src_len);
    h->surface = _surface_from_pixels(pixels, h->info->width, h->info->height);

    free(pixels);
    return 0;
}


static cairo_surface_t *
_kyg_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _kyg_get_mime_types,
        .new = _kyg_new,
        .free = _kyg_free,
        .create_from_data = _kyg_create_from_data,
        .get_image_surface = _kyg_get_image_surface
    };
    return &p;
}
