/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <loa-buffer.h>
#include <loa-cairo.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    loa_point_t *point;
    loa_segment_t *segment;
    loa_object_t *object;
    int object_count;
};

#define POLYGON_FILL    (1<<0)
#define POLYGON_OUTLINE (1<<1)

struct _loa_object_t{
    enum{
        OT_POLYGON,
        OT_TEXT
    }type;
    int color;
    union{
        struct{
            int mode;
            loa_path_t path;
        }polygon;
        struct{
            int point;
            int font;
            int width;
            int height;
            int mode;
            char *string;
        }text;
    };
};

static const char **
_lotus_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-lotus-1-2-3-chart",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_lotus_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->point = NULL;
    h->segment = NULL;
    h->object = NULL;
    return h;
}


static void
_lotus_free(abydos_plugin_handle_t *h)
{
    if(h->point)
        free(h->point);
    if(h->segment)
        free(h->segment);
    if(h->object){
        int i;
        for(i=0;i<h->object_count;++i){
            loa_object_t *o = &h->object[i];
            if(OT_TEXT == o->type)
                free(o->text.string);
        }
        free(h->object);
    }
    free(h);
}

/*
01 00 00 00 01 00 08 00 44 00 00 00 0C 7F 09 06

                                       _w___ _h___
01 00 00 00 01 00 08 00 44 00 00 00 00 0c 7f 09 06
*/

#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))

typedef struct{
    abydos_plugin_info_t *info;
    loa_buffer_t buf;
    int color;
    int x;
    int y;
    int font;
    int width;
    int height;
    loa_object_t *shape;
}read_context_t;


static void
_get_path(const char *data, int points, loa_path_t *path, loa_buffer_t *buf)
{
    loa_point_t *p;
    int i;
    path->segment = loa_buffer_segment_count(buf);
    path->segment_count = 0;
    for(i=0;i<points;++i){
        p = loa_buffer_push_segment(buf, path, i ? LOA_LINE : LOA_GOTO);
        p[0].x = get16(data);
        p[0].y = get16(data+2);
        data += 4;
    }
    loa_buffer_push_segment(buf, path, LOA_CLOSE);
}


static int
_lotus_parse(const char *data, size_t len, read_context_t *rc)
{
    const char *end = data + len;
    if(len < 17)
        return -1;

    rc->info->width = get16(data+13) + 1;
    rc->info->height = get16(data+15) + 1;

    data += 17;
    for(;;){
        int cmd;
        if(data+1 > end){
            DEBUG_p("no more data");
            return 0;
        }
        cmd = get8(data);
        ++data;
        switch(cmd>>4){
        case 0x3:
            switch(cmd&0xf){
            case 0x0:
                DEBUG_p("filled polygon");
                {
                    int num;
                    if(data+1 > end)
                        return 1;
                    num = get8(data++) + 1;
                    DEBUG_p("polygon with %d points", num);
                    if(data+num*4 > end)
                        return 1;
                    rc->shape = loa_buffer_add_object(&rc->buf);
                    rc->shape->type = OT_POLYGON;
                    rc->shape->color = rc->color;
                    rc->shape->polygon.mode = POLYGON_FILL;
                    _get_path(data, num, &rc->shape->polygon.path, &rc->buf);
                    data += num*4;
                }
                continue;
            }
            break;
        case 0x6:
            DEBUG_p("end of file");
            return 0;
        case 0xb:
            rc->color = cmd&0xf;
            continue;
        case 0xa:
            switch(cmd&0xf){
            case 0x0: /* move */
                if(data+4 > end)
                    return 1;
                rc->shape = NULL;
                rc->x = get16(data);
                rc->y = get16(data+2);
                data += 4;
                if(rc->shape){
                    loa_point_t *p = loa_buffer_push_segment(&rc->buf, &rc->shape->polygon.path, LOA_GOTO);
                    p[0].x = rc->x;
                    p[0].y = rc->y;
                }
                continue;
            case 0x2: /* line */
                {
                    loa_point_t *p;
                    if(data+4 > end){
                        DEBUG_p("only %d bytes left", (int)(end-data));
                        return 1;
                    }
                    if(!rc->shape){
                        rc->shape = loa_buffer_add_object(&rc->buf);
                        rc->shape->type = OT_POLYGON;
                        rc->shape->color = rc->color;
                        rc->shape->polygon.mode = POLYGON_OUTLINE;
                        rc->shape->polygon.path.segment = loa_buffer_segment_count(&rc->buf);
                        rc->shape->polygon.path.segment_count = 0;
                        p = loa_buffer_push_segment(&rc->buf, &rc->shape->polygon.path, LOA_GOTO);
                        p[0].x = rc->x;
                        p[0].y = rc->y;
                    }
                    p = loa_buffer_push_segment(&rc->buf, &rc->shape->polygon.path, LOA_LINE);
                    p[0].x = rc->x = get16(data);
                    p[0].y = rc->y = get16(data+2);
                    data += 4;
                }
                continue;
            case 0x7: /* font */
                if(data+1 > end)
                    return 1;
                rc->font = get8(data++);
                DEBUG_p("set font %d", rc->font);
                continue;
            case 0x8: /* text */
                {
                    int mode;
                    const char *strend;
                    loa_point_t *p;
                    if(data+2 > end)
                        return 1;
                    mode = get8(data++);
                    strend = memchr(data, '\0', end-data);
                    if(!strend)
                        return 1;
                    //DEBUG_p("text mode %d", mode);
                    rc->shape = loa_buffer_add_object(&rc->buf);
                    rc->shape->type = OT_TEXT;
                    rc->shape->color = rc->color;
                    rc->shape->text.point = loa_buffer_point_count(&rc->buf);
                    p = loa_buffer_add_points(&rc->buf, 1);
                    p[0].x = rc->x;
                    p[0].y = rc->y;
                    rc->shape->text.font = rc->font;
                    rc->shape->text.width = rc->width;
                    rc->shape->text.height = rc->height;
                    rc->shape->text.mode = mode;
                    rc->shape->text.string = strndup(data, strend-data);
                    data = strend + 1;
                }
                continue;
            case 0xc: /* text size */
                if(data+4 > end)
                    return 1;
                rc->width = get16(data);
                rc->height = get16(data+2);
                data += 4;
                continue;
            }
            break;
        case 0xd:
            switch(cmd&0xf){
            case 0x0:
                DEBUG_p("filled polygon with outline");
                {
                    int num;
                    if(data+1 > end)
                        return 1;
                    num = get8(data++) + 1;
                    DEBUG_p("polygon with %d points", num);
                    if(data+num*4 > end)
                        return 1;
                    rc->shape = loa_buffer_add_object(&rc->buf);
                    rc->shape->type = OT_POLYGON;
                    rc->shape->color = rc->color;
                    rc->shape->polygon.mode = POLYGON_FILL | POLYGON_OUTLINE;
                    _get_path(data, num, &rc->shape->polygon.path, &rc->buf);
                    data += num*4;
                }
                continue;
            }
            break;
        }
        DEBUG_p("unknown command: %02x", cmd);
    }
    return 1;
}


static int
_lotus_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    read_context_t rc;
    int res;
    loa_buffer_init(&rc.buf, sizeof(loa_object_t), 0, 0, 0);
    rc.info = h->info;
    rc.color = 0;
    rc.x = 0;
    rc.y = 0;
    rc.shape = NULL;
    res = _lotus_parse(data, len, &rc);

    h->point = loa_buffer_copy_points(&rc.buf);
    h->segment = loa_buffer_copy_segments(&rc.buf);
    h->object = loa_buffer_copy_objects(&rc.buf);
    h->object_count = loa_buffer_object_count(&rc.buf);

    loa_buffer_done(&rc.buf);
    return res ? -1 : 0;
}


static void
_lotus_render(abydos_plugin_handle_t *h,
              cairo_t *cr,
              int page,
              cairo_bool_t *layer_visibility,
              int variant,
              double time,
              int frame)
{
    static const struct{
        double r;
        double g;
        double b;
    }palette[] = {
        {    0,     0,     0},
        {    0,     0, 2.0/3},
        {    0, 2.0/3,     0},
        {    0, 2.0/3, 2.0/3},
        {2.0/3,     0,     0},
        {2.0/3,     0, 2.0/3},
        {2.0/3, 1.0/3,     0},
        {2.0/3, 2.0/3, 2.0/3},
        {1.0/3, 1.0/3, 1.0/3},
        {1.0/3, 1.0/3,     1},
        {1.0/3,     1, 1.0/3},
        {1.0/3,     1,     1},
        {    1, 1.0/3, 1.0/3},
        {    1, 1.0/3,     1},
        {    1,     1, 1.0/3},
        {    1,     1,     1}
    };
    int i;
    cairo_save(cr);
    cairo_translate(cr, 0, h->info->height);
    cairo_scale(cr, 1, -1);
    for(i=0;i<h->object_count;++i){
        loa_object_t *o = &h->object[i];
        cairo_set_source_rgb(cr, palette[o->color].r,
                                 palette[o->color].g,
                                 palette[o->color].b);
        switch(o->type){
        case OT_POLYGON:
            loa_cairo_draw_path(cr, &o->polygon.path, h->segment, h->point);
            switch(o->polygon.mode){
            case POLYGON_FILL:
                cairo_fill(cr);
                break;
            case POLYGON_FILL | POLYGON_OUTLINE:
                cairo_fill_preserve(cr);
                /* Assume the outline is black. */
                cairo_set_source_rgb(cr, 0, 0, 0);
                /* fall through */
            case POLYGON_OUTLINE:
                cairo_stroke(cr);
                break;
            }
            break;
        case OT_TEXT:
            {
                loa_point_t *p = &h->point[o->text.point];
                cairo_text_extents_t ext;
#ifdef _USE_DEBUG
                cairo_save(cr);
                cairo_arc(cr, p->x, p->y, 10, 0, 2*M_PI);
                cairo_set_source_rgb(cr, 0, 1, 0);
                cairo_fill(cr);
                cairo_restore(cr);
#endif
                cairo_save(cr);
                /* Assume text is always black. */
                cairo_set_source_rgb(cr, 0, 0, 0);
                cairo_translate(cr, p->x, p->y);
                cairo_set_font_size(cr, 0.6 * o->text.height);
                cairo_scale(cr, 1, -1);
                cairo_text_extents(cr, o->text.string, &ext);
                //DEBUG_p("mode: %x", o->text.mode);

                /* Rotate text */
                switch(o->text.mode >> 4){
                case 1:
                    DEBUG_p("up (%x)", o->text.mode);
                    cairo_rotate(cr, 3*M_PI/2);
                    break;
                case 2:
                    cairo_rotate(cr, 2*M_PI/2);
                    break;
                case 3:
                    DEBUG_p("down");
                    cairo_rotate(cr, 1*M_PI/2);
                    break;
                }

                /* Align text */
                cairo_translate(cr, -ext.x_bearing, -ext.y_bearing);
                switch(o->text.mode & 0xf){
                case 0x0: /* center aligned */
                    cairo_translate(cr, -ext.width/2, -ext.height/2);
                    break;
                case 0x1: /* left center aligned */
                    cairo_translate(cr, 0, -ext.height/2);
                    break;
                case 0x2: /* top center aligned */
                    cairo_translate(cr, -ext.width/2, 0);
                    break;
                case 0x3: /* right center aligned */
                    cairo_translate(cr, -ext.width, -ext.height/2);
                    break;
                case 0x4: /* bottom center aligned */
                    cairo_translate(cr, -ext.width/2, 0);
                    break;
                case 0x5: /* top left aligned */
                    break;
                case 0x6: /* top right aligned */
                    cairo_translate(cr, -ext.width, 0);
                    break;
                case 0x7: /* bottom left aligned */
                    cairo_translate(cr, 0, -ext.height);
                    break;
                case 0x8: /* bottom right aligned */
                    cairo_translate(cr, -ext.width, -ext.height);
                    break;
                }
                cairo_move_to(cr, 0, 0);
                cairo_show_text(cr, o->text.string);
                cairo_restore(cr);
            }
            break;
        }
    }
    cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _lotus_get_mime_types,
        .new = _lotus_new,
        .free = _lotus_free,
        .create_from_data = _lotus_create_from_data,
        .render = _lotus_render
    };
    return &p;
}
