/*
 * Copyright (C) 2018 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#if IMAGEMAGICK_MAJOR_VERSION >= 7
# include <MagickCore/MagickCore.h>
# include <MagickCore/colorspace.h>
#else
# include <magick/MagickCore.h>
#endif
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


/* #define ENABLE_ALIAS_PIX */ /* not working (null image) */
#define ENABLE_AVSX
/* #define ENABLE_BPG */ /* not working (invalid file) */
#define ENABLE_CINEON
/* #define ENABLE_CUT */ /* not working (invalid file) depricated? */
#define ENABLE_DICOM
/* #define ENABLE_DJVU */ /* not working? */
/* #define ENABLE_EPS */ /* not working */
#define ENABLE_EXR
/* #define ENABLE_FITS */ /* not working (just black) */
/* #define ENABLE_FPX */ /* not working ("invalid image") */
#define ENABLE_GIF
#define ENABLE_HEIF
/* #define ENABLE_JBIG */ /* not working ("invalid image") */
#define ENABLE_JPEG
#define ENABLE_JPEG2000
#define ENABLE_MTV
#define ENABLE_MNG
#define ENABLE_PHOTOCD
#define ENABLE_PCX
/* Define ENABLE_PICT */ /* not working (just black) */
#define ENABLE_PHOTOSHOP
#define ENABLE_SGI
/* #define ENABLE_SUNRASTER */ /* not working ("insufficient image data in file") */
#define ENABLE_TARGA
#define ENABLE_TIFF
#define ENABLE_XBITMAP
#define ENABLE_XPIXMAP

#define QUANTUM_MAX (0xffffffff>>(32-MAGICKCORE_QUANTUM_DEPTH))

/*

Formats supported by imagemagick:

Type code | Image type                                                      | Notes
----------+-----------------------------------------------------------------+
* AAI     | AAI Dune image                                                  |
  ART     | PFS: 1st Publisher                                              | Format originally used on the Macintosh (MacPaint?) and later used for PFS: 1st Publisher clip art.
  ARW     | Sony Digital Camera Alpha Raw Image Format                      |
  AVI     | Microsoft Audio/Visual Interleaved                              |
* AVS     | AVS X image                                                     |
* BPG     | Better Portable Graphics                                        |
  BMP     | Microsoft Windows bitmap                                        |
  CALS    | Continuous Acquisition and Life-cycle Support Type 1 image      | Specified in MIL-R-28002 and MIL-PRF-28002. Standard blueprint archive format as used by the US military to replace microfiche.
  CGM     | Computer Graphics Metafile                                      | Requires ralcgm to render CGM files.
* CIN     | Kodak Cineon Image Format
  CR2     | Canon Digital Camera Raw Image Format
  CRW     | Canon Digital Camera Raw Image Format
  CUBE    | Cube Color lookup table converted to a HALD image
  CUR     | Microsoft Cursor Icon                                           |
* CUT     | DR Halo                                                         |
* DCM     | Digital Imaging and Communications in Medicine (DICOM) image
  DCR     | Kodak Digital Camera Raw Image File
* DCX     | ZSoft IBM PC multi-page Paintbrush image
  DDS     | Microsoft Direct Draw Surface                                   |
  DIB     | Microsoft Windows Device Independent Bitmap                     | DIB is a BMP file without the BMP header. Used to support embedded images in compound formats like WMF.
* DJVU    |                                                                 |
  DNG     | Digital Negative                                                |
  DOT     | Graph Visualization                                             |
  DPX     | SMPTE Digital Moving Picture Exchange 2.0 (SMPTE 268M-2003)     |
  EPDF    | Encapsulated Portable Document Format                           |
  EPI     | Adobe Encapsulated PostScript Interchange format                | Requires Ghostscript to read.
* EPS     | Adobe Encapsulated PostScript                                   | Requires Ghostscript to read.
  EPSF    | Adobe Encapsulated PostScript                                   | Requires Ghostscript to read.
  EPSI    | Adobe Encapsulated PostScript Interchange format                | Requires Ghostscript to read.
  EPT     | Adobe Encapsulated PostScript Interchange format w/TIFF preview | Requires Ghostscript to read.
* EXR     | High dynamic-range (HDR) developed by Industrial Light & Magic  | See High Dynamic-Range Images for details on this image format.
  FAX     | Group 3 TIFF                                                    | This format is a fixed width of 1728 as required by the standard. See TIFF format. Note that FAX machines use non-square pixels which are 1.5 times wider than they are tall but computer displays use square pixels so FAX images may appear to be narrow unless they are explicitly resized using a resize specification of 100x150%.
  FIG     | FIG graphics format                                             | Requires TransFig.
* FITS    | Flexible Image Transport System                                 |
  FPX     | FlashPix Format                                                 | FlashPix has the option to store mega- and giga-pixel images at various resolutions in a single file which permits conservative bandwidth and fast reveal times when displayed within a Web browser. Requires the FlashPix SDK.
* GIF     | CompuServe Graphics Interchange Format                          |
  GPLT    | Gnuplot plot files                                              | Requires gnuplot4.0.tar.Z or later.
* HDR     | Radiance RGBE image format                                      |
* HEIC    | Apple High efficiency Image Format                              |
  HPGL    | HP-GL plotter language                                          | Requires hp2xx-3.4.4.tar.gz
  HRZ     | Slow Scane TeleVision                                           |
* ICO     | Microsoft Windows icon                                          |
* JBIG    | Joint Bi-level Image experts Group file interchange format      | Also known as BIE and JBG. Requires jbigkit-1.6.tar.gz.
* JNG     | Multiple-image Network Graphics                                 | JPEG in a PNG-style wrapper with transparency. Requires libjpeg and libpng-1.0.11 or later, libpng-1.2.5 or later recommended.
* JP2     | JPEG-2000 JP2 File Format Syntax
  JPT     | JPEG-2000 Code Stream Syntax
  J2C     | JPEG-2000 Code Stream Syntax
  J2K     | JPEG-2000 Code Stream Syntax
* JPEG    | Joint Photographic Experts Group JFIF format
  JXR     | JPEG extended range                                             | Requires the jxrlib delegate library. Put the JxrDecApp and JxrEncApp applications in your execution path.
  MAN     | Unix reference manual pages                                     | Requires that GNU groff and Ghostcript are installed.
  MAT     | MATLAB image format                                             |
* MIFF    | Magick image file format                                        |
  MONO    | Bi-level bitmap in least-significant-byte first order           |
* MNG     | Multiple-image Network Graphics                                 |
  M2V     | Motion Picture Experts Group file interchange format (v 2)      | Requires ffmpeg.
  MPEG    | Motion Picture Experts Group file interchange format (v 1)      | Requires ffmpeg.
  MPC     | Magick Persistent Cache image file format                       | The most efficient data processing pattern is a write-once, read-many-times pattern. The image is generated or copied from source, then various analyses are performed on the image pixels over time. MPC supports this pattern. MPC is the native in-memory ImageMagick uncompressed file format. This file format is identical to that used by ImageMagick to represent images in memory and is read by mapping the file directly into memory. The MPC format is not portable and is not suitable as an archive format. It is suitable as an intermediate format for high-performance image processing. The MPC format requires two files to support one image. Image attributes are written to a file with the extension .mpc, whereas, image pixels are written to a file with the extension .cache.
  MRW     | Sony (Minolta) Raw Image File                                   |
  MSL     | Magick Scripting Language                                       | MSL is the XML-based scripting language supported by the conjure utility. MSL requires the libxml2 delegate library.
* MTV     | MTV Raytracing image format
  MVG     | Magick Vector Graphics.                                         | The native ImageMagick vector metafile format. A text file containing vector drawing commands accepted by convert's -draw option.
  NEF     | Nikon Digital SLR Camera Raw Image File
  ORF     | Olympus Digital Camera Raw Image File
  OTB     | On-the-air Bitmap
  P7      | Xv's Visual Schnauzer thumbnail format
  PALM    | Palm pixmap
  PBM     | Portable bitmap format (black and white)
  PCD     | Photo CD                                                        | The maximum resolution written is 768x512 pixels since larger images require huffman compression (which is not supported). Use -bordercolor to specify the border color (e.g. -bordercolor black).
  PCDS    | Photo CD                                                        | Decode with the sRGB color tables.
* PCX     | ZSoft IBM PC Paintbrush file
  PDB     | Palm Database ImageViewer Format
  PDF     | Portable Document Format
  PEF     | Pentax Electronic File
  PES     | Embrid Embroidery Format
  PFA     | Postscript Type 1 font (ASCII)
  PFB     | Postscript Type 1 font (binary)
* PFM     | Portable float map format
  PGM     | Portable graymap format (gray scale)
  PICON   | Personal Icon
* PICT    | Apple Macintosh QuickDraw/PICT file
* PIX     | Alias/Wavefront RLE image format
* PNG     | Portable Network Graphics                                       | Requires libpng-1.0.11 or later, libpng-1.2.5 or later recommended.
  PNM     | Portable anymap                                                 | PNM is a family of formats
  PPM     | Portable pixmap format (color)                                  |
  PS      | Adobe PostScript file                                           | Requires Ghostscript to read.
  PS2     | Adobe Level II PostScript file                                  | Requires Ghostscript to read.
  PS3     | Adobe Level III PostScript file                                 | Requires Ghostscript to read.
  PSB     | Adobe Large Document Format                                     |
* PSD     | Adobe Photoshop bitmap file                                     |
  PTIF    | Pyramid encoded TIFF                                            | Multi-resolution TIFF containing successively smaller versions of the image down to the size of an icon.
* PWP     | Seattle File Works multi-image file
  RAD     | Radiance image file                                             | Requires that ra_ppm from the Radiance software package be installed.
  RAF     | Fuji CCD-RAW Graphic File
  RGF     | LEGO Mindstorms EV3 Robot Graphics File
  RLA     | Alias/Wavefront image file
  RLE     | Utah Run length encoded image file
  SCT     | Scitex Continuous Tone Picture
* SFW     | Seattle File Works image
* SGI     | Irix RGB image
  SID     | MrSID Multiresolution seamless image                            | Requires the mrsidgeodecode command line utility that decompresses MG2 or MG3 SID image files.
* SUN     | SUN Rasterfile                                                  |
  SVG     | Scalable Vector Graphics                                        | ImageMagick utilizes inkscape if its in your execution path otherwise RSVG. If neither are available, ImageMagick reverts to its internal SVG renderer.
  TEXT    | text file                                                       | Requires an explicit format specifier to read, e.g. text:README.txt.
* TGA     | Truevision Targa image                                          | Also known as formats ICB, VDA, and VST.
* TIFF    | Tagged Image File Format                                        | Requires tiff-v3.6.1.tar.gz or later.
  TIM     | PSX TIM file
  TTF     | TrueType font file                                              | Requires freetype 2.
  UYVY    | Interleaved YUV raw image                                       |
* VICAR   | VICAR rasterfile format                                         |
* VIFF    | Khoros Visualization Image File Format                          |
* WBMP    | Wireless bitmap                                                 | Support for uncompressed monochrome only.
  WDP     | JPEG extended range                                             | Requires the jxrlib delegate library.
  WEBP    | Weppy image format                                              | Requires the WEBP delegate library.
  WMF     | Windows Metafile                                                | Requires libwmf. By default, renders WMF files using the dimensions specified by the metafile header.
  WPG     | Word Perfect Graphics File
* XBM     | X Windows system bitmap, black and white only                   | Used by the X Windows System to store monochrome icons.
* XCF     | GIMP image
* XPM     | X Windows system pixmap, also known as PM.                      | Used by the X Windows System to store color icons.
* XWD     | X Windows system window dump
  X3F     | Sigma Camera RAW Picture File

*/

#define QUIRK_DISCARD_FIRST_LAYER (1<<0)

typedef struct{
    cairo_surface_t *surface;
    ssize_t xoffset;
    ssize_t yoffset;
    union{
        struct{
            int width;
            int height;
        }size;
        double duration;
    };
}multi_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    const char *magick_type;
    enum{
        MT_NONE,
        MT_PAGE,
        MT_LAYER,
        MT_VARIANT,
        MT_FRAME
    }multi_type;
    int quirk;
    multi_t *multi;
    int multi_count;
};


static void
_magick_done()
{
    MagickCoreTerminus();
}


static const char **
_magick_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_ALIAS_PIX
        "image/x-alias-pix",
#endif
#ifdef ENABLE_AVSX
        "image/x-avsx",
#endif
#ifdef ENABLE_BPG
        "image/bpg", "image/x-bpg",
#endif
#ifdef ENABLE_CINEON
        "image/x-cineon", "image/x-cin",
#endif
        "image/x-dcx",
#ifdef ENABLE_DICOM
        "application/dicom", "image/dicom-rle", "image/x-dicom",
#endif
#ifdef ENABLE_DJVU
        "image/vnd.djvu",
#endif
#ifdef ENABLE_CUT
        "application/dr-halo", "image/x-cut",
#endif
        "image/x-dune",
#ifdef ENABLE_EPS
        "image/x-eps",
#endif
#ifdef ENABLE_EXR
        "image/x-exr",
#endif
#ifdef ENABLE_FITS
        "application/fits", "image/fits",
#endif
#ifdef ENABLE_FPX
        "image/vnd.fpx",
#endif
#ifdef ENABLE_GIF
        "image/gif",
#endif
#ifdef ENABLE_HEIF
        "image/heic", "image/heic-sequence", "image/heif", "image/heif-sequence",
#endif
        "image/vnd.microsoft.icon",
#ifdef ENABLE_JBIG
        "image/x-jbig",
#endif
        "image/x-jng", "video/x-jng",
#ifdef ENABLE_JPEG
        "image/jpeg",
#endif
#ifdef ENABLE_JPEG2000
        "image/jp2", "image/jpeg2000", "image/jpx",
#endif
        "image/x-miff", "image/miff",
#ifdef ENABLE_MNG
        "video/x-mng", "image/x-mng",
#endif
#ifdef ENABLE_MTV
        "image/x-mtv",
#endif
#ifdef ENABLE_PHOTOCD
        "image/x-photo-cd",
#endif
#ifdef ENABLE_PCX
        "image/vnd.zbrush.pcx",
#endif
#ifdef ENABLE_PHOTOSHOP
        "image/vnd.adobe.photoshop",
#endif
#ifdef ENABLE_PICT
        "image/pict", "image/x-pict",
#endif
        "image/png",
        "image/x-portable-floatmap",
        "image/vnd.radiance",
        "image/x-seattle-filmworks",
#ifdef ENABLE_SGI
        "image/x-rgb", "image/x-sgi",
#endif
#ifdef ENABLE_SUNRASTER
        "image/x-sun-raster",
#endif
#ifdef ENABLE_TARGA
        "image/x-tga",
#endif
#ifdef ENABLE_TIFF
        "image/tiff",
#endif
        "image/x-vicar",
        "image/x-viff",
        "image/vnd.wap.wbmp", "image/x-wireless-bitmap",
#ifdef ENABLE_XBITMAP
        "image/x-xbitmap", "image/x-xbmi",
#endif
        "image/xcf", "image/x-xcf",
#ifdef ENABLE_XPIXMAP
        "image/x-xpixmap", "image/x-xpmi",
#endif
        "image/x-xwindowdump",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_magick_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->multi_type = MT_NONE;
    h->quirk = 0;
#ifdef ENABLE_ALIAS_PIX
    if(!strcasecmp(mime_type,"image/x-alias-pix")){
        h->magick_type = "PIX";
    }else
#endif
#ifdef ENABLE_AVSX
    if(!strcasecmp(mime_type,"image/x-avsx")){
        h->magick_type = "AVS";
    }else
#endif
    if(!strcasecmp(mime_type,"image/bpg") ||
       !strcasecmp(mime_type,"image/x-bpg"))
    {
        h->magick_type = "BPG";
    }else
#ifdef ENABLE_CINEON
    if(!strcasecmp(mime_type,"image/x-cineon") ||
       !strcasecmp(mime_type,"image/x-cin"))
    {
        h->magick_type = "CIN";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-dcx")){
        h->magick_type ="DCX";
        h->multi_type = MT_PAGE;
    }else
#ifdef ENABLE_DICOM
    if(!strcasecmp(mime_type,"application/dicom") ||
       !strcasecmp(mime_type,"image/dicom-rle") ||
       !strcasecmp(mime_type,"image/x-dicom"))
    {
        h->magick_type = "DCM";
    }else
#endif
#ifdef ENABLE_DJVU
    if(!strcasecmp(mime_type,"image/vnd.djvu")){
        h->magick_type = "DJVU";
        h->multi_type = MT_PAGE;
    }else
#endif
#ifdef ENABLE_CUT
    if(!strcasecmp(mime_type, "application/dr-halo") ||
       !strcasecmp(mime_type, "image/x-cut"))
    {
        h->magick_type = "CUT";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-dune")){
        h->magick_type = "AAI";
    }else
#ifdef ENABLE_EPS
    if(!strcasecmp(mime_type,"image/x-eps")){
        h->magick_type = "EPS";
    }else
#endif
#ifdef ENABLE_EXR
    if(!strcasecmp(mime_type,"image/x-exr")){
        h->magick_type = "EXR";
    }else
#endif
#ifdef ENABLE_FITS
    if(!strcasecmp(mime_type,"application/fits")||
       !strcasecmp(mime_type,"image/fits"))
    {
        h->magick_type = "FITS";
    }else
#endif
#ifdef ENABLE_FPX
    if(!strcasecmp(mime_type,"image/vnd.fpx")){
        h->magick_type = "FPX";
    }else
#endif
#ifdef ENABLE_GIF
    if(!strcasecmp(mime_type,"image/gif")){
        h->magick_type = "GIF";
        h->multi_type = MT_FRAME;
    }else
#endif
#ifdef ENABLE_HEIF
    if(!strcasecmp(mime_type,"image/heic")||
       !strcasecmp(mime_type,"image/heic-sequence")||
       !strcasecmp(mime_type,"image/heif")||
       !strcasecmp(mime_type,"image/heif-sequence"))
    {
        h->magick_type = "HEIC";
    }else
#endif
    if(!strcasecmp(mime_type,"image/vnd.microsoft.icon")){
        h->magick_type = "ICO";
        h->multi_type = MT_VARIANT;
    }else
#ifdef ENABLE_JBIG
    if(!strcasecmp(mime_type,"image/x-jbig")){
        h->magick_type = "JBIG";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-jng")||
       !strcasecmp(mime_type,"video/x-jng"))
    {
        h->magick_type = "JNG";
    }else
#ifdef ENABLE_JPEG
    if(!strcasecmp(mime_type,"image/jpeg")){
        h->magick_type = "JPEG";
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type,"image/jp2")||
       !strcasecmp(mime_type,"image/jpeg2000")||
       !strcasecmp(mime_type,"image/jpx"))
    {
        h->magick_type = "JP2";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-miff")||
       !strcasecmp(mime_type,"image/miff"))
    {
        h->magick_type = "MIFF";
    }else
#ifdef ENABLE_MNG
    if(!strcasecmp(mime_type,"video/x-mng") ||
       !strcasecmp(mime_type,"image/x-mng"))
    {
        h->magick_type = "MNG";
        h->multi_type = MT_FRAME;
    }else
#endif
#ifdef ENABLE_MTV
    if(!strcasecmp(mime_type,"image/x-mtv")){
        h->magick_type = "MTV";
    }else
#endif
#ifdef ENABLE_PHOTOCD
    if(!strcasecmp(mime_type,"image/x-photo-cd")){
        h->magick_type = "PCD";
    }else
#endif
#ifdef ENABLE_PCX
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")){
        h->magick_type = "PCX";
    }else
#endif
#ifdef ENABLE_PHOTOSHOP
    if(!strcasecmp(mime_type,"image/vnd.adobe.photoshop")){
        h->magick_type = "PSD";
        h->multi_type = MT_LAYER;
        h->quirk = QUIRK_DISCARD_FIRST_LAYER;
    }else
#endif
#ifdef ENABLE_PICT
    if(!strcasecmp(mime_type,"image/pict")||
       !strcasecmp(mime_type,"image/x-pict"))
    {
        h->magick_type = "PICT";
    }else
#endif
    if(!strcasecmp(mime_type,"image/png")){
        h->magick_type = "PNG";
    }else if(!strcasecmp(mime_type,"image/x-portable-floatmap")){
        h->magick_type = "PFM";
    }else if(!strcasecmp(mime_type,"image/vnd.radiance")){
        h->magick_type = "HDR";
    }else if(!strcasecmp(mime_type,"image/x-seattle-filmworks")){
        h->magick_type = "SFW";
    }else
#ifdef ENABLE_SGI
    if(!strcasecmp(mime_type,"image/x-rgb") ||
       !strcasecmp(mime_type,"image/x-sgi"))
    {
        h->magick_type = "SGI";
    }else
#endif
#ifdef ENABLE_SUNRASTER
    if(!strcasecmp(mime_type,"image/x-sun-raster")){
        h->magick_type = "SUN";
    }else
#endif
#ifdef ENABLE_TARGA
    if(!strcasecmp(mime_type,"image/x-tga")){
        h->magick_type = "TGA";
    }else
#endif
#ifdef ENABLE_TIFF
    if(!strcasecmp(mime_type,"image/tiff")){
        h->magick_type = "TIFF";
        h->multi_type = MT_PAGE;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-vicar")){
        h->magick_type = "VICAR";
    }else if(!strcasecmp(mime_type,"image/x-viff")){
        h->magick_type = "VIFF";
    }else if(!strcasecmp(mime_type,"image/vnd.wap.wbmp") ||
             !strcasecmp(mime_type,"image/x-wireless-bitmap"))
    {
        h->magick_type = "WBMP";
    }else
#ifdef ENABLE_XBITMAP
    if(!strcasecmp(mime_type,"image/x-xbitmap")||
       !strcasecmp(mime_type,"image/x-xbmi"))
    {
        h->magick_type = "XBM";
    }else
#endif
    if(!strcasecmp(mime_type,"image/xcf") ||
       !strcasecmp(mime_type,"image/x-xcf"))
    {
        h->magick_type = "XCF";
        h->multi_type = MT_LAYER;
    }else
#ifdef ENABLE_XPIXMAP
    if(!strcasecmp(mime_type,"image/x-xpixmap")||
       !strcasecmp(mime_type,"image/x-xpmi"))
    {
        h->magick_type = "XPM";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-xwindowdump")){
        h->magick_type = "XWD";
    }else
    {
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;
    h->multi = NULL;
    h->multi_count = 0;
    return h;
}


static void
_magick_free(abydos_plugin_handle_t *h)
{
    if(h->multi){
        int i;
        for(i=0;i<h->multi_count;++i)
            cairo_surface_destroy(h->multi[i].surface);
        free(h->multi);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_image(Image *image,ExceptionInfo *exception)
{
    cairo_surface_t *surface;
#if IMAGEMAGICK_MAJOR_VERSION >= 7
    const Quantum *src;
    int channels;
#else
    PixelPacket *src;
#endif
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
    float *dst;
    int dst_channels;
#else
    uint32_t *dst;
#endif
    int rowskip;
    int x;
    int y;

#if IMAGEMAGICK_MAJOR_VERSION >= 7
    TransformImageColorspace(image, sRGBColorspace, exception);
#else
    TransformImageColorspace(image, sRGBColorspace);
#endif

#if MagickLibVersion >= 0x640
    src = GetVirtualPixels(image,0,0,image->columns,image->rows,exception);
#else
    src = (PixelPacket*)AcquireImagePixels(image,0,0,image->columns,image->rows,exception);
#endif
    if(exception->severity){
        /* Can't get the pixels for the sub image, strange but not fatal. */
        DEBUG_p("failed to get sub image");
        ClearMagickException(exception);
        return NULL;
    }

#if IMAGEMAGICK_MAJOR_VERSION >= 7
    channels = GetPixelChannels(image);
    DEBUG_p("channels: %d",channels);
#endif

#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
    if(GetImageAlphaChannel(image)){
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGBA128F,image->columns,image->rows);
        dst_channels = 4;
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB96F,image->columns,image->rows);
        dst_channels = 3;
    }
    dst = (float*)cairo_image_surface_get_data(surface);
    if(BottomLeftOrientation == image->orientation){
        int stride = cairo_image_surface_get_stride(surface)/sizeof(float_t);
        dst += stride*(image->rows-1);
        rowskip = -image->columns * dst_channels -stride;
    }else{
        rowskip = cairo_image_surface_get_stride(surface)/sizeof(float_t) - image->columns * dst_channels;;
    }
#else
    if(GetImageAlphaChannel(image)){
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,image->columns,image->rows);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,image->columns,image->rows);
    }
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    if(BottomLeftOrientation == image->orientation){
        int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
        dst += stride*(image->rows-1);
        rowskip = -image->columns -stride;
    }else{
        rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - image->columns;;
    }
#endif

#if IMAGEMAGICK_MAJOR_VERSION >= 7
    if(GetImageAlphaChannel(image)){
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
# if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
                float a = (float)GetPixelAlpha(image,src) / QUANTUM_MAX;
                *dst++ = a * GetPixelRed(image, src) / QUANTUM_MAX;
                *dst++ = a * GetPixelGreen(image, src) / QUANTUM_MAX;
                *dst++ = a * GetPixelBlue(image, src) / QUANTUM_MAX;
                *dst++ = a;
# else
                int alpha = GetPixelAlpha(image,src);
                int r = alpha*GetPixelRed(image,src)/QUANTUM_MAX;
                int g = alpha*GetPixelGreen(image,src)/QUANTUM_MAX;
                int b = alpha*GetPixelBlue(image,src)/QUANTUM_MAX;
                *dst++ = (r>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16|
                         (g>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 8|
                         (b>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 0|
                         (alpha>>(MAGICKCORE_QUANTUM_DEPTH-8))<<24;
# endif
                src += channels;
            }
            dst += rowskip;
        }
    }else{
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
# if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
                *dst++ = (float)GetPixelRed(image, src) / QUANTUM_MAX;
                *dst++ = (float)GetPixelGreen(image, src) / QUANTUM_MAX;
                *dst++ = (float)GetPixelBlue(image, src) / QUANTUM_MAX;
# else
                *dst++ = (GetPixelRed(image,src)>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16|
                         (GetPixelGreen(image,src)>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 8|
                         (GetPixelBlue(image,src)>>(MAGICKCORE_QUANTUM_DEPTH-8))<< 0;
# endif
                src += channels;
            }
            dst += rowskip;
        }
    }
#else
    if(GetImageAlphaChannel(image)){
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
# if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
                *dst++ = (float)src->red / MAGICKCORE_QUANTUM_DEPTH;
                *dst++ = (float)src->green / MAGICKCORE_QUANTUM_DEPTH;
                *dst++ = (float)src->blue / MAGICKCORE_QUANTUM_DEPTH;
                *dst++ = 1.0 - (float)src->opacity / MAGICKCORE_QUANTUM_DEPTH;
# else
                *dst++ = ((src->red>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16)
                        |((src->green>>(MAGICKCORE_QUANTUM_DEPTH-8))<<8)
                        |((src->blue>>(MAGICKCORE_QUANTUM_DEPTH-8))<<0)
                        |((255-(src->opacity>>(MAGICKCORE_QUANTUM_DEPTH-8)))<<24);
# endif
                ++src;
            }
            dst += rowskip;
        }
    }else{
        for(y=0;y<image->rows;++y){
            for(x=0;x<image->columns;++x){
# if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2) && MAGICKCORE_QUANTUM_DEPTH > 8
                *dst++ = (float)src->red / MAGICKCORE_QUANTUM_DEPTH;
                *dst++ = (float)src->green / MAGICKCORE_QUANTUM_DEPTH;
                *dst++ = (float)src->blue / MAGICKCORE_QUANTUM_DEPTH;
# else
                *dst++ = ((src->red>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16)
                        |((src->green>>(MAGICKCORE_QUANTUM_DEPTH-8))<<8)
                        |((src->blue>>(MAGICKCORE_QUANTUM_DEPTH-8))<<0);
# endif
                ++src;
            }
            dst += rowskip;
        }
    }
#endif
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_magick_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    ExceptionInfo *exception = AcquireExceptionInfo();
    int quirky_offset = 0;
    ImageInfo *info = CloneImageInfo(NULL);
    Image *image = NULL;
    int ret = -1;
    int i;

    strcpy(info->magick,h->magick_type);
    image = BlobToImage(info,data,len,exception);
    if(exception->severity){
        DEBUG_p("invalid image");
        h->info->error = "Invalid image";
        goto out;
    }
    if(!image){
        h->info->error = "Internal error";
        goto out;
    }
    h->info->width = image->columns;
    h->info->height = image->rows;
    h->multi_count = GetImageListLength(image);
    DestroyImageInfo(info);

    if(h->multi_count>1 && h->quirk & QUIRK_DISCARD_FIRST_LAYER){
        --h->multi_count;
        ++quirky_offset;
    }

    DEBUG_p("multi count: %d",h->multi_count);
    switch(h->multi_type){
    case MT_NONE:
        if(h->multi_count>1)
            DEBUG_p("multiple sub images was not expected for an image of type %s",h->magick_type);
        break;
    case MT_PAGE:
        h->info->page_count = h->multi_count;
        break;
    case MT_LAYER:
        h->info->layer_count = h->multi_count;
        break;
    case MT_VARIANT:
        h->info->variant_count = h->multi_count;
        break;
    case MT_FRAME:
        h->info->frame_count = h->multi_count;
        break;
    }

    h->multi = (multi_t*)malloc(h->multi_count * sizeof(multi_t));
    for(i=0;i<h->multi_count;++i){
        Image *subimage = GetImageFromList(image,quirky_offset+i);
        h->multi[i].surface = _surface_from_image(subimage,exception);
        h->multi[i].xoffset = subimage->page.x;
        h->multi[i].yoffset = subimage->page.y;
        switch(h->multi_type){
        case MT_NONE:
            break;
        case MT_PAGE:
        case MT_VARIANT:
            h->multi[i].size.width = subimage->columns;
            h->multi[i].size.height = subimage->rows;
            break;
        case MT_LAYER:
            //h->multi[i].visible = 1;
            break;
        case MT_FRAME:
            if(subimage->ticks_per_second > 0){
                h->multi[i].duration = (double)subimage->delay / subimage->ticks_per_second;
            }else{
                h->multi[i].duration = -1;
            }
            break;
        }
    }
    ret = 0;
out:
    if(image)
        DestroyImage(image);
    if(exception)
        DestroyExceptionInfo(exception);
    return ret;
}


static cairo_surface_t *
_surface_from_layers(abydos_plugin_handle_t *h,cairo_bool_t *visible_layers)
{
    cairo_surface_t *surface;
    int visible_count = 0;
    int has_offset = 0;
    int i;
    if(1==h->info->layer_count)
        return cairo_surface_reference(h->multi[0].surface);
    for(i=0;i<h->info->layer_count;++i)
        if(visible_layers[i]){
            ++visible_count;
            has_offset |= h->multi[i].xoffset | h->multi[i].yoffset;
        }
    if(1==visible_count && !has_offset)
        for(i=0;i<h->info->layer_count;++i)
            if(visible_layers[i])
                return cairo_surface_reference(h->multi[i].surface);

    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    if(visible_count>0){
        cairo_t *cr = cairo_create(surface);
        for(i=0;i<h->info->layer_count;++i)
            if(visible_layers[i]){
                cairo_set_source_surface(cr,h->multi[i].surface,h->multi[i].xoffset,h->multi[i].yoffset);
                cairo_paint(cr);
            }
        cairo_destroy(cr);
    }
    return surface;
}


static cairo_surface_t *
_magick_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int variant,double time,int frame)
{
    int index;
    switch(h->multi_type){
    case MT_NONE:
        index = 0;
        break;
    case MT_PAGE:
        index = page;
        break;
    case MT_LAYER:
        return _surface_from_layers(h,visible_layers);
    case MT_VARIANT:
        index = variant;
        break;
    case MT_FRAME:
        index = frame;
        break;
    }
    return cairo_surface_reference(h->multi[index].surface);
}


static void
_magick_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    for(i=0;i<h->info->variant_count;++i){
        size[i].width = h->multi[i].size.width;
        size[i].height = h->multi[i].size.height;
    }
}


static void
_magick_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->multi[i].duration;
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = _magick_done,
        .get_mime_types = _magick_get_mime_types,
        .new = _magick_new,
        .free = _magick_free,
        .create_from_data = _magick_create_from_data,
        .get_image_surface = _magick_get_image_surface,
        .get_variant_size = _magick_get_variant_size,
        .get_duration = _magick_get_duration
    };
    //if(!IsMagickInstantiated())
        MagickCoreGenesis(NULL,0);
    return &p;
}
