/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_NIA,
        T_NIE
    }type;
    frame_t *frame;
};


static const char **
_naive_get_mime_types()
{
    static const char *mime_types[] = {
        "image/nia",
        "image/nie",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_naive_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    if(!strcasecmp(mime_type,"image/nia")){
        h->type = T_NIA;
    }else{
        h->type = T_NIE;
    }
    h->frame = NULL;
    return h;
}


static void
_naive_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


#define get32(a) le32toh(*(uint32_t*)(a))
#define get64(a) le64toh(*(uint64_t*)(a))


static int
_handle_nie(abydos_plugin_handle_t *h, frame_t *frame, int bpp, const uint8_t *p, size_t len)
{
    nile_format_t format;

    if(len < 16)
        return -1;
    if(get32(p) != 0x45AFC36E)
        return -1;
    if(p[4] != 0xff)
        return -1;

    switch(p[7]){
    case '4':
        if(bpp>0 && bpp!=4)
            return -1;
        bpp = 4;
        break;
    case '8':
        if(bpp>0 && bpp!=8)
            return -1;
        bpp = 8;
        break;
    default:
        /* unknown BPP */
        return -1;
    }

    if(h->info->width > 0){
        if(h->info->width != get32(p+8))
            return -1;
    }else{
        h->info->width = get32(p+8);
    }
    if(h->info->height > 0){
        if(h->info->height != get32(p+12))
            return -1;
    }else{
        h->info->height = get32(p+12);
    }

    switch(p[5]){
    case 'b':
        if(8==bpp){
            format = NILE_FORMAT_be_b16g16r16a16;
        }else{
            format = NILE_FORMAT_b8g8r8a8;
        }
        break;
    case 'r':
        if(8==bpp){
            format = NILE_FORMAT_be_r16g16b16a16;
        }else{
            format = NILE_FORMAT_r8g8b8a8;
        }
        break;
    default:
        /* unknown pixel format */
        return -1;
    }

    switch(p[6]){
    case 'n':
        break;
    case 'p':
        format |= NILE_PREMULTIPLIED;
        break;
    default:
        /* unknown */
        return -1;
    }

    p += 16;
    len -= 16;

    if(len < h->info->width * h->info->height * bpp)
        return -1;

    frame->surface = nil_cairo_surface_create_from_direct(p, h->info->width, h->info->height, -1, 4, format, 0);

    return 0;
}


static int
_handle_nia(abydos_plugin_handle_t *h, const uint8_t *p, size_t len)
{
    int bpp;
    size_t frame_size;
    long long last_cdd;
    int i;

    if(len < 24)
        return -1;
    if(get32(p) != 0x41AFC36E)
        return -1;
    if(p[4] != 0xff)
        return -1;
    if(get32(p+len-4) != 0x80000000)
        return -1;

    switch(p[7]){
    case '4':
        bpp = 4;
        break;
    case '8':
        bpp = 8;
        break;
    default:
        /* unknown BPP */
        return -1;
    }

    h->info->width = get32(p+8);
    h->info->height = get32(p+12);

    frame_size = 8 + 16 + h->info->width * h->info->height * bpp;
    /* 8 byte alignment */
    frame_size += frame_size & 4;
    DEBUG_p("frame size: %d",(int)frame_size);
    h->info->frame_count = (len-24) / frame_size;
    DEBUG_p("frame count: %d",h->info->frame_count);
    if(len > 24 * h->info->frame_count * frame_size){
        /* The bytes don't add up */
        return -1;
    }

    p += 16;
    h->frame = (frame_t*)calloc(h->info->frame_count, sizeof(frame_t));
    last_cdd = 0;
    for(i=0;i<h->info->frame_count;++i){
        long long cdd = get64(p);
        if(_handle_nie(h, &h->frame[i], bpp, p+8, frame_size-8)<0)
            return -1;
        h->frame[i].duration = (double)(cdd-last_cdd) / 705600000;
        last_cdd = cdd;
        p += frame_size;
    }
    return 0;
}


static int
_naive_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    switch(h->type){
    case T_NIA:
        return _handle_nia(h, (uint8_t*)data, len);
    case T_NIE:
        h->frame = (frame_t*)calloc(1, sizeof(frame_t));
        return _handle_nie(h, &h->frame[0], -1, (uint8_t*)data, len);
    }
    /* not reached */
    return -1;
}


static cairo_surface_t *
_naive_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_naive_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _naive_get_mime_types,
        .new = _naive_new,
        .free = _naive_free,
        .create_from_data = _naive_create_from_data,
        .get_image_surface = _naive_get_image_surface,
        .get_duration = _naive_get_duration
    };
    return &p;
}
