/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <ctype.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>


#define MAX_IMAGE_SIZE 32767


typedef enum{
    BPP_UNKNOWN = -1,
    BPP_8 = 8,
    BPP_16 = 16
}bpp_t;

typedef enum{
    ENC_UNKNOWN = -1,
    ENC_NONE,
    ENC_GZIP
}enc_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_nrrd_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-nrrd",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_nrrd_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_nrrd_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_atoin(const char *s,size_t len)
{
    const char *end = s+len;
    int neg;
    int res;
    while(s<end && isspace(*s))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res = 0;
    while(s<end && *s>='0' && *s<='9'){
        res *= 10;
        res += *s-'0';
        ++s;
    }
    return neg ? -res : res;
}


static const char *
_strnchrnul(const char *s, size_t len, int c)
{
    const char *end = s+len;
    for(;s<end && *s!=c;++s);
    return s;
}


static void
_trim(const char **r_str, size_t *r_len)
{
    const char *str = *r_str;
    size_t len = *r_len;
    while(len>0 && isspace(str[len-1]))
        --len;
    while(len>0 && isspace(str[0])){
        ++str;
        --len;
    }
    *r_str = str;
    *r_len = len;
}


static int
_split(const char *s,size_t len,const char *sep,const char **args,int max)
{
    int count = 0;
    const char *end = len>=0 ? s+len : strchr(s,'\0');
    while(s<end){
        args[2*count++] = s;
        if(count>=max){
            args[2*count-1] = end;
            break;
        }
        while(s<end && !strchr(sep,*s))
            ++s;
        args[2*count-1] = s;
        ++s;
    }
    return count;
}


static int
_splitnum(const char *s,size_t len,const char *sep,int *nums,int max)
{
    int count = 0;
    const char *end = len>=0 ? s+len : strchr(s,'\0');
    while(s<end){
        const char *start = s;
        while(s<end && !strchr(sep,*s))
            ++s;
        nums[count++] = _atoin(start,s-start);
        if(count>=max)
            break;
        ++s;
    }
    return count;
}


static int
_uncompress_gzip(uint8_t *dst, size_t dstlen, const uint8_t *src, size_t srclen)
{
    z_stream strm;
    strm.next_in = (uint8_t*)src;
    strm.avail_in = srclen;
    strm.total_out = 0;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;

    if(inflateInit2(&strm,(16+MAX_WBITS)) != Z_OK)
        return -1;
    memset(dst, 0, dstlen);

    for(;;){
        int err;
        /* If our output buffer is too small */
        if(strm.total_out >= dstlen)
            return -1;

        strm.next_out = dst+strm.total_out;
        strm.avail_out = dstlen-strm.total_out;

        err = inflate(&strm, Z_SYNC_FLUSH);
        if(err != Z_OK)
            break;
    }

    if(inflateEnd (&strm) != Z_OK)
        return -1;

    return 0;
}


static int
_nrrd_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data+len;
    const char *s = data;
    int channels = 1;
    bpp_t bpp = BPP_UNKNOWN;
    enc_t enc = ENC_NONE;
    nile_t ni;
    size_t rawsize;
    nile_init(&ni);
    for(;;){
        const char *args[4];
        const char *p;
        size_t len;
        if(s >= end)
            goto out;
        p = _strnchrnul(s,end-s,'\n');
        len = p-s;
        _trim(&s,&len);
        if(len<=0){
            data = p+1;
            break;
        }
        if(*s!='#'){
            if(_split(s,len,":",args,2)>=2){
                size_t len1 = args[1]-args[0];
                size_t len2 = args[3]-args[2];
                _trim(&args[0],&len1);
                _trim(&args[2],&len2);
                DEBUG_p("'%.*s' : '%.*s'",(int)len1,args[0],(int)len2,args[2]);
                if(!strncasecmp(args[0],"type",len1)){
                    if(!strncasecmp(args[2],"signed char",len2) ||
                       !strncasecmp(args[2],"int8",len2) ||
                       !strncasecmp(args[2],"int8_t",len2))
                    {
                        bpp = BPP_8;
                    }else if(!strncasecmp(args[2],"uchar",len2) ||
                             !strncasecmp(args[2],"unsigned char",len2) ||
                             !strncasecmp(args[2],"uint8",len2) ||
                             !strncasecmp(args[2],"uint8_t",len2))
                    {
                        DEBUG_p("bpp 8");
                        bpp = BPP_8;
                    }else if(!strncasecmp(args[2],"short",len2) ||
                             !strncasecmp(args[2],"short int",len2) ||
                             !strncasecmp(args[2],"signed short",len2) ||
                             !strncasecmp(args[2],"signed short int",len2) ||
                             !strncasecmp(args[2],"int16",len2) ||
                             !strncasecmp(args[2],"int16_t",len2))
                    {
                        bpp = BPP_16;
                    }else if(!strncasecmp(args[2],"ushort",len2) ||
                             !strncasecmp(args[2],"unsigned short",len2) ||
                             !strncasecmp(args[2],"unsigned short int",len2) ||
                             !strncasecmp(args[2],"uint16",len2) ||
                             !strncasecmp(args[2],"uint16_t",len2))
                    {
                        bpp = BPP_16;
                    }else{
                        DEBUG_p("unknown type: %.*s",(int)len2,args[2]);
                        bpp = BPP_UNKNOWN;
                    }
                }else if(!strncasecmp(args[0],"encoding",len1)){
                    if(!strncasecmp(args[2],"gzip",len2)){
                        enc = ENC_GZIP;
                    }else if(!strncasecmp(args[2],"raw",len2)){
                        enc = ENC_NONE;
                    }else{
                        DEBUG_p("unknown encoding: %.*s",(int)len2,args[2]);
                        enc = ENC_UNKNOWN;
                    }
                }else if(!strncasecmp(args[0],"sizes",len1)){
                    int nums[3];
                    int c = _splitnum(args[2],len2," ",nums,3);
                    if(c>=3){
                        channels = nums[0];
                        ni.width = nums[1];
                        ni.height = nums[2];
                    }else if(c>=2){
                        ni.width = nums[0];
                        ni.height = nums[1];
                    }
                }
            }
        }
        s = p+1;
    }
    if(ni.width <= 0 || ni.height <= 0){
        DEBUG_p("invalid size");
        goto out;
    }
    if(ni.width > MAX_IMAGE_SIZE || ni.height > MAX_IMAGE_SIZE){
        DEBUG_p("image too large");
        goto out;
    }
    if(BPP_UNKNOWN == bpp){
        DEBUG_p("unknown BPP");
        goto out;
    }
    if(ENC_UNKNOWN == enc){
        DEBUG_p("unknown encoding");
        goto out;
    }
    switch(channels){
    case 1:
        ni.bpc = 1;
        ni.format = NILE_FORMAT_g8;
        break;
    case 3:
        switch(bpp){
        case BPP_UNKNOWN:
            /* not reached */
            goto out;
        case BPP_8:
            ni.bpc = 3;
            ni.format = NILE_FORMAT_r8g8b8;
            break;
        case BPP_16:
            ni.bpc = 6;
            ni.format = NILE_FORMAT_be_r16g16b16;
            break;
        }
        break;
    case 4:
        switch(bpp){
        case BPP_UNKNOWN:
            /* not reached */
            goto out;
        case BPP_8:
            ni.bpc = 4;
            ni.format = NILE_FORMAT_r8g8b8a8;
            break;
        case BPP_16:
            ni.bpc = 8;
            ni.format = NILE_FORMAT_be_r16g16b16a16;
            break;
        }
        break;
    default:
        DEBUG_p("cannot handle %d channels",channels);
        goto out;
    }
    rawsize = (size_t)ni.width * ni.height * ni.bpc;
    DEBUG_p("rawsize: %d",(int)rawsize);
    DEBUG_p("%d bytes starting with %x",(int)(end-data),*data);
    if(ENC_NONE==enc){
        if(data + rawsize > end)
            goto out;
        ni.pixels = (uint8_t*)data;
    }else{
        switch(enc){
        case ENC_UNKNOWN:
        case ENC_NONE:
            /* note reached */
            goto out;
        case ENC_GZIP:
            nile_ensure_direct(&ni, ni.width, ni.height, ni.bpc);
            /* FIXME!! check if it's successful */
            if(_uncompress_gzip(ni.pixels, rawsize, (uint8_t*)data, end-data) < 0)
                goto out;
            break;
        }
    }
    h->info->width = ni.width;
    h->info->height = ni.height;
    h->surface = nil_cairo_surface_create(&ni, 0);
out:
    nile_done(&ni);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_nrrd_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _nrrd_get_mime_types,
        .new = _nrrd_new,
        .free = _nrrd_free,
        .create_from_data = _nrrd_create_from_data,
        .get_image_surface = _nrrd_get_image_surface
    };
    return &p;
}
