/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "nil-pc88.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_dv_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-davinci",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_dv_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_dv_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_dv_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    nile_t ni;
    const uint8_t *p = (uint8_t*)data;
    const uint8_t *end = p + len;
    int offs;
    int c = 0;

    nile_init(&ni);

    nile_ensure_bitplanes(&ni, 640, 200, 3, 1);
    for(offs=0;offs<16000;++offs){
        int b[3];
        if(c <= 0){
            if(p+4 > end)
                goto error_truncated;
            c = *p++;
            if(c & 128){
                /* It's a runlength. */
                c -= 128;
                b[0] = *p++;
                b[1] = *p++;
                b[2] = *p++;
            }else{
                b[0] = -1;
                if(p + c*3 > end)
                    goto error_truncated;
            }
        }
        if(b[0] >= 0){
            ni.bitplanes[0][offs] = b[0];
            ni.bitplanes[1][offs] = b[1];
            ni.bitplanes[2][offs] = b[2];
        }else{
            ni.bitplanes[0][offs] = *p++;
            ni.bitplanes[1][offs] = *p++;
            ni.bitplanes[2][offs] = *p++;
        }
        --c;
    }

    ni.bpc = 2;
    ni.format = NILE_FORMAT_ne_r5g6b5;
    nil_pc88_default_palette(&ni);

    h->surface = nil_cairo_surface_create(&ni, 0);
    nile_done(&ni);

    h->info->width = 640;
    h->info->height = 200;
    h->info->pixel_ratio = 5.0/12;

    return 0;
error_truncated:
    nile_done(&ni);
    DEBUG_p("truncated data");
    h->info->error = "Truncated data";
    return -1;
}


static cairo_surface_t *
_dv_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _dv_get_mime_types,
        .new = _dv_new,
        .free = _dv_free,
        .create_from_data = _dv_create_from_data,
        .get_image_surface = _dv_get_image_surface
    };
    return &p;
}
