/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#include <debug.h>
#undef new
}
#include <PGFimage.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static void
_premultiply32(UINT8 *p, int width, int height, int stride)
{
    while(height-->0){
        for(int x=0;x<width;++x){
            int alpha = p[x*4+3];
            p[x*4+0] = alpha * p[x*4+0] / 255;
            p[x*4+1] = alpha * p[x*4+1] / 255;
            p[x*4+2] = alpha * p[x*4+2] / 255;
        }
        p += stride;
    }
}


static int
_pgf_create_from_stream(abydos_plugin_handle_t *h, CPGFStream *stream)
{
    CPGFImage img;
    UINT8 *buffer;
    int stride;
    try{
        img.ConfigureDecoder(true);
        img.Open(stream);
        img.Read();
    }catch(const std::exception &e){
        return -1;
    }
    h->info->width = img.Width();
    h->info->height = img.Height();
    switch(img.Mode()){
    case ImageModeBitmap:
    case ImageModeGrayScale:
    case ImageModeIndexedColor:
    case ImageModeRGBColor:
    case ImageModeCMYKColor:
    case ImageModeHSLColor:
    case ImageModeHSBColor:
    case ImageModeMultichannel:
    case ImageModeDuotone:
    case ImageModeLabColor:
    case ImageModeRGBA:
        h->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, h->info->width, h->info->height);
        buffer = cairo_image_surface_get_data(h->surface);
        stride = cairo_image_surface_get_stride(h->surface);
        img.GetBitmap(stride, buffer, 32);
        _premultiply32(buffer, h->info->width, h->info->height, stride);
        cairo_surface_mark_dirty(h->surface);
        return 0;
    case ImageModeGray16:
        DEBUG_p("gray 16 mode not supported");
        return -1;
    case ImageModeRGB48:
        DEBUG_p("RGB 48 mode not supported");
        return -1;
    case ImageModeLab48:
        DEBUG_p("lab 48 mode not supported");
        return -1;
    case ImageModeCMYK64:
        DEBUG_p("CMYK 64 mode not supported");
        return -1;
    case ImageModeDeepMultichannel:
        DEBUG_p("deep multichannel mode not supported");
        return -1;
    case ImageModeDuotone16:
        DEBUG_p("duotone 16 mode not supported");
        return -1;
    case ImageModeGray32:
        DEBUG_p("gray 32 mode not supported");
        return -1;
    case ImageModeRGB12:
        DEBUG_p("RGB 12 mode not supported");
        return -1;
    case ImageModeRGB16:
        DEBUG_p("RGB 16 mode not supported");
        return -1;
    }
    DEBUG_p("unknown mode");
    return -1;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-pgf",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            h->info = info;
            h->info->threadsafe = 1;
            h->surface = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            delete h;
        },
        .create_from_data = [](abydos_plugin_handle_t *h,const char *data, size_t len){
            CPGFMemoryStream stream((UINT8*)data, len);
            return _pgf_create_from_stream(h, &stream);
        },
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int variant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->surface);
        }
    };
    return &p;
}
