/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*

,-----v-----------v------v------v------------------------------------------.
| Pos | Field     | Type | Size | Description                              |
>-----+-----------+------+------+------------------------------------------<
|   0 | Signature | char | 3    | Constant string "PKM"                    |
|   3 | Version   | byte | 1    | For the moment, it can take only the     |
|     |           |      |      | value 0. Other packing methods may       |
|     |           |      |      | change this field but there is only one  |
|     |           |      |      | for now...                               |
|   4 | Pack_byte | byte | 1    | Value of the recognition byte for color  |
|     |           |      |      | repetitions that are coded on 1 byte.    |
|   5 | Pack_word | byte | 1    | Value of the recognition byte for color  |
|     |           |      |      | repetitions that are coded on 2 bytes.   |
|   6 | Width     | word | 2    | Picture width (in pixels)                |
|   8 | Height    | word | 2    | Picture height (in pixels)               |
|  10 | Palette   | byte | 768  | RGB palette (RGB RGB ... 256 times) with |
|     |           |      |      | values from 0 to 63.                     |
| 778 | PH_size   | word | 2    | Post-header size. This is the number of  |
|     |           |      |      | bytes between the header and the picture |
|     |           |      |      | data. This value can be equal to 0.      |
`-----^-----------^------^------^------------------------------------------'

*/

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    int original_screen_width;
    int original_screen_height;
    int transparent_index;
    cairo_surface_t *surface;
};


static const char **
_pkm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-pkm",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_pkm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->original_screen_width = 0;
    h->original_screen_height = 0;
    h->transparent_index = -1;
    h->surface = NULL;
    return h;
}


static void
_pkm_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get8(a) (*(uint8_t*)(a));
#define get16(a) le16toh(*(uint16_t*)(a))


static void
_expand_palette(uint32_t *dst, const uint8_t *src, int colors)
{
    while(colors-->0){
        int r = *src++;
        int g = *src++;
        int b = *src++;
        r = r<<2 | r>>6;
        g = g<<2 | g>>6;
        b = b<<2 | b>>6;
        *dst++ = r<<16 | g<<8 | b | 0xff<<24;
    }
}


static cairo_surface_t *
_surface_from_ckm(const uint8_t *p, int len, int width, int height, int rec1, int rec2, const uint32_t *palette)
{
    const uint8_t *end = p + len;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    int count = 0;
    while(height-->0){
        int b;
        int i;
        for(i=width;i>0;--i){
            if(count<=0){
                if(p >= end)
                    goto error;
                b = *p++;
                if(b==rec1){
                    if(p+2 > end)
                        goto error;
                    b = *p++;
                    count = *p++;
                }else if(b==rec2){
                    if(p+3 > end)
                        goto error;
                    b = *p++;
                    count = get16(p);
                    p += 2;
                }else{
                    count = 1;
                }
            }
            *dst++ = palette[b];
            --count;
        }
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
error:
    cairo_surface_destroy(surface);
    return NULL;
}


static int
_handle_postheader(abydos_plugin_handle_t *h,const uint8_t *p, int len)
{
    const uint8_t *end = p+len;

    while(p < end){
        int b = *p++;
        len = *p++;
        if(p+len > end)
            return -1;

        switch(b){
        case 0: /* Comment */
            DEBUG_p("comment");
            break;
        case 1: /* Original screen dimensions */
            if(len<4)
                return -1;
            h->original_screen_width = get16(p+0);
            h->original_screen_height = get16(p+2);
            DEBUG_p("original screen dimensions: %dx%d",h->original_screen_width,h->original_screen_height);
            break;
        case 2: /* Transparent color */
            if(len<1)
                return -1;
            h->transparent_index = p[0];
            DEBUG_p("transparent index: %d",h->transparent_index);
            break;
        }
        p += len;
    }
    return 0;
}


static int
_pkm_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int rec1;
    int rec2;
    uint32_t palette[256];
    int postheader_len;

    if(len < 780)
        return -1;
    /* Check for 3 byte identifier string and version 0. */
    if(memcmp(data,"PKM\0",4))
        return -1;
    rec1 = get8(data+4);
    rec2 = get8(data+5);
    h->info->width = get16(data+6);
    h->info->height = get16(data+8);
    _expand_palette(palette, (uint8_t*)data+10, 256);
    postheader_len = get16(data+778);
    data += 780;
    len -= 780;

    if(postheader_len > 0){
        if(len < postheader_len)
            return -1;
        if(_handle_postheader(h, (uint8_t*)data, postheader_len)<0)
            return -1;
        data += postheader_len;
        len -= postheader_len;
        /* FIXME!! calculate pixel ratio from original screen size? */
        if(h->transparent_index >= 0)
            palette[h->transparent_index] = 0;
    }

    h->surface = _surface_from_ckm((uint8_t*)data, len, h->info->width, h->info->height, rec1, rec2, palette);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_pkm_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _pkm_get_mime_types,
        .new = _pkm_new,
        .free = _pkm_free,
        .create_from_data = _pkm_create_from_data,
        .get_image_surface = _pkm_get_image_surface
    };
    return &p;
}
