/*
 * Copyright (C) 2019 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <limits.h>                     /* INT_MAX */
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "recoil.h"

#define MAX_LAYERS 4

#define ENABLE_ILBM
/* #define ENABLE_PC1 */
#define ENABLE_SPECTRUM512

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif
#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#define ATARI_ST_LOW_PIXEL_RATIO (5.0/6)
#define ATARI_ST_MED_PIXEL_RATIO (5.0/12)
#define ATARI_ST_HIGH_PIXEL_RATIO (5.0/6)
#if 0
#define C64_LOW_PIXEL_RATIO (5.0/3)
#define C64_HIGH_PIXEL_RATIO (5.0/6)
#endif
#define FALCON_PIXEL_RATIO (5.0/6)

typedef bool(*decode_func_t)(RECOIL *, uint8_t const *, int);

typedef struct{
    cairo_surface_t *surface;
}layer_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    decode_func_t decode_func;
    /* This is used for a hack to get individual "layers" out of recoil */
    int supports_layers;
    /* This is set if the ratio should be set from the recoil resolution */
    enum{
        RATIO_CALCULATE,
        RATIO_CONST,
        RATIO_FROM_RESOLUTION,
        RATIO_TO_43
    }ratio;
    layer_t layer[MAX_LAYERS];
};


static const char **
_recoil_get_mime_types()
{
    static const char *mime_types[] = {
        "application/microsoft-paint",
        "image/x-artist-64",
        "image/x-atari-cad",
        "image/x-atari-duo",
        "image/x-atari-rgb-intermediate",
        "image/x-blazing-paddles",
        "image/x-cheese",
        "image/x-draz-paint",
        "image/x-face-painter",
        "image/x-gem",
        "image/x-gfa-artist",
        "image/x-hi-eddi",
#ifdef ENABLE_ILBM
        "image/x-ilbm",
#endif
        "image/x-imagelab",
        "image/x-macpaint", "application/mac-paint",
        "image/x-maki-chan",
        "image/x-multi-palette-picture",
        "image/x-neochrome","image/x-neo",
        "image/x-pablo-packed-picture",
#ifdef ENABLE_PC1
        "image/x-pc1",
        "image/x-pc2",
        "image/x-pc3",
#endif
        "image/x-picasso-64",
        "image/x-photochrome-screen",
        "image/x-prism-paint",
        "image/x-quantum-paint",
        "image/x-saracen-paint",
#ifdef ENABLE_SPECTRUM512
        "image/x-spectrum512-compressed",
        "image/x-spectrum512-smooshed",
        "image/x-spectrum512-uncompressed",
#endif
        "image/x-stad",
        "image/x-tobias-richter-fullscreen-slideshow",
        "image/x-tricolor-rgb",
        "image/x-vidcom-64",
        "image/x-xga",
        "image/x-zx-spectrum-standard-screen",
        NULL
    };
    return mime_types;
}


#include "recoil.c"


static bool
_decode_a64_wig(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10242)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x2002, 0x2402, 0x2801);
}


static bool
_decode_blazing_paddles(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10242)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x2002, 0x2402, 0x1f82);
}


static bool
_decode_che(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 20482)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x4202, 0x4802, 0x4fff);
}


static bool
_decode_fcp_ftp(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10004)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x1f42, 0x232a, 0x2712);
}


static bool
_decode_hi_eddi(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 9218)
        return false;
    return RECOIL_DecodeC64Hires(r, data, 2, 8194);
}


#ifdef ENABLE_ILBM
static bool
_decode_ilbm(RECOIL *r, uint8_t const *data, int len)
{
    return RECOIL_DecodeIff(r, data, len, RECOILResolution_AMIGA1X1);
}
#endif


static bool
_decode_neo(RECOIL *r, uint8_t const *data, int len)
{
    return RECOIL_DecodeNeo(r, ".neo", data, len);
}


static bool
_decode_p64(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10050)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 0x802, 0x402, 2, 0x801);
}


static bool
_decode_sar(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10219)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 0x402, 2, 0x2402, 0x3f2);
}


static bool
_decode_zx_scr(RECOIL *r, uint8_t const *data, int len)
{
    int attroffs;
    switch(len){
    case 6144:
        attroffs = -1;
        break;
    case 6912:
    case 6913:
        attroffs = 0x1800;
        break;
    default:
        return false;
    }
    RECOIL_SetZx(r, RECOIL_GetResolutionSpectrum1x1(r), 1);
    RECOIL_DecodeZx(r, data, 0, attroffs, RECOIL_ZX_ATTRIBUTES8X8, 0);
    return true;
}


static abydos_plugin_handle_t *
_recoil_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->supports_layers = 0;
    h->ratio = RATIO_CALCULATE;
    if(!strcasecmp(mime_type, "application/microsoft-paint")){
        h->decode_func = RECOIL_DecodeMsp;
    }else if(!strcasecmp(mime_type,"image/x-artist-64")){
        h->decode_func = _decode_a64_wig;
    }else if(!strcasecmp(mime_type,"image/x-atari-cad")){
        //h->info->pixel_ratio = 2.0/3;
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeDrg;
    }else if(!strcasecmp(mime_type,"image/x-atari-duo")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeDuo;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-atari-rgb-intermediate")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeStRgb;
    }else if(!strcasecmp(mime_type,"image/x-blazing-paddles")){
        h->decode_func = _decode_blazing_paddles;
    }else if(!strcasecmp(mime_type,"image/x-cheese")){
        h->decode_func = _decode_che;
    }else if(!strcasecmp(mime_type,"image/x-draz-paint")){
        h->decode_func = RECOIL_DecodeDrz;
    }else if(!strcasecmp(mime_type,"image/x-face-painter")){
        h->decode_func = _decode_fcp_ftp;
    }else if(!strcasecmp(mime_type,"image/x-gem")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeStImg;
    }else if(!strcasecmp(mime_type,"image/x-gfa-artist")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeGfaArtist;
    }else if(!strcasecmp(mime_type,"image/x-hi-eddi")){
        h->decode_func = _decode_hi_eddi;
    }else
#ifdef ENABLE_ILBM
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->decode_func = _decode_ilbm;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-imagelab")){
        h->decode_func = RECOIL_DecodeBw;
    }else if(!strcasecmp(mime_type,"image/x-macpaint") ||
             !strcasecmp(mime_type,"application/mac-paint"))
    {
        h->decode_func = RECOIL_DecodeMac;
    }else if(!strcasecmp(mime_type, "image/x-maki-chan")){
        h->decode_func = RECOIL_DecodeMag;
        h->ratio = RATIO_FROM_RESOLUTION;
    }else if(!strcasecmp(mime_type,"image/x-multi-palette-picture")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeMpp;
        /* not tested */
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-neo") ||
             !strcasecmp(mime_type,"image/x-neochrome"))
    {
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = _decode_neo;
    }else if(!strcasecmp(mime_type,"image/x-pablo-packed-picture")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeStPpp;
    }else
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->recoil_type = ".pc1";
    }else if(!strcasecmp(mime_type,"image/x-pc2")){
        h->info->pixel_ratio = ATARI_ST_MED_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->recoil_type = ".pc2";
    }else if(!strcasecmp(mime_type,"image/x-pc3")){
        h->info->pixel_ratio = ATARI_ST_HIGH_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->recoil_type = ".pc3";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-picasso-64")){
        h->decode_func = _decode_p64;
    }else if(!strcasecmp(mime_type,"image/x-photochrome-screen")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodePcs;
    }else if(!strcasecmp(mime_type,"image/x-prism-paint")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeFalconPnt;
    }else if(!strcasecmp(mime_type,"image/x-quantum-paint")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodePbx;
    }else if(!strcasecmp(mime_type,"image/x-saracen-paint")){
        h->decode_func = _decode_sar;
    }else
#ifdef ENABLE_SPECTRUM512
    if(!strcasecmp(mime_type,"image/x-spectrum512-compressed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeStSpc;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-smooshed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeSps;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-uncompressed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodeSpu;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-stad")){
        h->info->pixel_ratio = ATARI_ST_HIGH_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodePac;
    }else if(!strcasecmp(mime_type,"image/x-tobias-richter-fullscreen-slideshow")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->ratio = RATIO_CONST;
        h->decode_func = RECOIL_DecodePci;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-tricolor-rgb")){
        h->decode_func = RECOIL_Decode3;
        h->ratio = RATIO_CONST;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-vidcom-64")){
        h->decode_func = RECOIL_DecodeDol;
    }else if(!strcasecmp(mime_type,"image/x-xga")){
        h->ratio = RATIO_TO_43;
        h->decode_func = RECOIL_DecodeXga;
    }else if(!strcasecmp(mime_type,"image/x-zx-spectrum-standard-screen")){
        h->decode_func = _decode_zx_scr;
    }else{
        /* not reached */
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    h->layer[0].surface = NULL;
    return h;
}


static void
_recoil_free(abydos_plugin_handle_t *h){
    if(h->layer[0].surface){
        int i;
        for(i=0;i<h->info->layer_count;++i)
            cairo_surface_destroy(h->layer[i].surface);
    }
    free(h);
}


static double
_ratio_from_resolution(RECOILResolution res)
{
    switch(res){
    case RECOILResolution_MSX21X1:
        return 53.0 / 48;
    case RECOILResolution_MSX21X1I:
        return 4.0 / 3;
    case RECOILResolution_MSX22X1I:
        return 11.0 / 6;
    case RECOILResolution_MSX21X2:
    case RECOILResolution_MSX2_PLUS1X1:
        return 53.0 / 96;
    case RECOILResolution_MSX2_PLUS2X1I:
        return 53.0 / 24;
    case RECOILResolution_PC801X2:
    case RECOILResolution_PC881X2:
        return 5.0 / 12;
    case RECOILResolution_PC88_VA1X1:
    case RECOILResolution_PC981X1:
    case RECOILResolution_X68_K1X1:
        return 5.0 / 6;
    default:
        DEBUG_p("unhandled resolution: %d", res);
        return 1;
    }
}


static cairo_surface_t *
_surface_from_image(abydos_plugin_handle_t *h, const int *src, int xscale, int yscale)
{
    int height = h->info->height;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - h->info->width;
    while(height-->0){
        int x;
        for(x=0;x<h->info->width;++x){
            *dst++ = *src;
            src += xscale;
        }
        src += h->info->width * (yscale-1);
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


#if 0
static cairo_surface_t *
_surface_from_image_with_palette(abydos_plugin_handle_t *h, const unsigned char *src, const uint32_t *palette)
{
    int height = h->info->height;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - h->info->width;
    while(height-->0){
        int x;
        for(x=0;x<h->info->width;++x){
            *dst++ = palette[*src];
            src += h->xscale;
        }
        src += h->info->width * (h->yscale-1);
        dst += rowskip;
    }
    return surface;
}
#endif


#if 0
static void
_sort_palette(uint32_t *dst, const int *reference)
{
    int i;
    for(i=0;i<16;++i){
        int best_index = 0;
        int best_val = INT_MAX;
        int j;
        for(j=0;j<16;++j){
            int val = ABS(reference[i] - RECOIL_C64_PALETTE[j]);
            if(val<best_val){
                best_index = j;
                best_val = val;
            }
        }
        DEBUG_p("mapping %02d -> %02d",i,best_index);
        dst[i] = c64_color[best_index];
    }
}
#endif


static int
_recoil_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    int res = -1;
    RECOIL *recoil = RECOIL_New();
    if(h->decode_func(recoil,(const unsigned char*)data,len)){
        int xscale;
        int yscale;
        int i;
        DEBUG_p("success");
        h->info->width = RECOIL_GetOriginalWidth(recoil);
        h->info->height = RECOIL_GetOriginalHeight(recoil);
        xscale = RECOIL_GetWidth(recoil) / h->info->width;
        yscale = RECOIL_GetHeight(recoil) / h->info->height;
        switch(h->ratio){
        case RATIO_CALCULATE:
            {
                int xppm = RECOIL_GetXPixelsPerMeter(recoil);
                int yppm = RECOIL_GetYPixelsPerMeter(recoil);
                if(xppm > 0 && yppm > 0){
                    h->info->pixel_ratio = (double)(yppm * xscale) / (xppm * yscale);
                }else if(1 == h->info->pixel_ratio){
                    DEBUG_p("pixel ratio information not available");
                    h->info->pixel_ratio = (double)xscale / yscale;
                }
            }
            break;
        case RATIO_CONST:
            /* Pixel ratio already set. */
            break;
        case RATIO_FROM_RESOLUTION:
            DEBUG_p("size %dx%d", h->info->width, h->info->height);
            h->info->pixel_ratio = _ratio_from_resolution(recoil->resolution);
            break;
        case RATIO_TO_43:
            h->info->pixel_ratio = (double)(h->info->height * 4) / (h->info->width * 3);
            break;
        }
        DEBUG_p("pixel ratio %f", h->info->pixel_ratio);
        if(h->supports_layers)
            h->info->layer_count = MIN(RECOIL_GetFrames(recoil), MAX_LAYERS);
        DEBUG_p("layers: %d",h->info->layer_count);
        for(i=0;i<h->info->layer_count;++i)
            h->layer[i].surface = _surface_from_image(h,RECOIL_GetPixels(recoil)+i*h->info->width*h->info->height, xscale, yscale);
        res = 0;
    }else{
        DEBUG_p("failure");
    }
    RECOIL_Delete(recoil);
    return res;
}


static cairo_surface_t *
_recoil_get_image_surface(abydos_plugin_handle_t *h,
                          int page,
                          cairo_bool_t *visible_layers,
                          int prefered_variant,
                          double time,
                          int frame)
{
    cairo_surface_t *surface;
    cairo_t *cr;
    int visible_count = 0;
    int i;
    if(1==h->info->layer_count)
        return cairo_surface_reference(h->layer[0].surface);
    for(i=0;i<h->info->layer_count;++i)
        if(visible_layers[i])
            ++visible_count;
    DEBUG_p("number of visible layers: %d",visible_count);
    switch(visible_count){
    case 0:
        return cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    case 1:
        for(i=0;i<h->info->layer_count;++i)
            if(visible_layers[i])
                return cairo_surface_reference(h->layer[i].surface);
        /* not reached */
    default:
        /* blend two or more layers */
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,h->info->height);
        cr = cairo_create(surface);
        visible_count = 0;
        for(i=0;i<h->info->layer_count;++i){
            if(!visible_layers[i])
                continue;
            ++visible_count;
            cairo_set_source_surface(cr,h->layer[i].surface,0,0);
            cairo_paint_with_alpha(cr,1.0/visible_count);
        }
        cairo_destroy(cr);
        return surface;
    }
}

const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _recoil_get_mime_types,
        .new = _recoil_new,
        .free = _recoil_free,
        .create_from_data = _recoil_create_from_data,
        .get_image_surface = _recoil_get_image_surface
    };
    return &p;
}
