/*
 * Copyright (C) 2006 - 2020  Magnus Bergman
 *
 * Author: Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memmove() */

#include "cons_buf.h"
#include "debug.h"

#ifdef USE_DEBUG
# undef USE_DEBUG
#endif

struct _Console{
    int width;
    int height;
    int allocation;
    ConsoleCell* screen;
    /* palette */
    unsigned char* palette;
    /* font */
    unsigned char* font;
};


Console* cons_buf_new(){
    Console* c = (Console*)calloc(1,sizeof(Console));
    return c;
}


void cons_buf_free(Console* c){
    if(c->allocation)
        free(c->screen);
    free(c);
}


void cons_buf_set_size(Console* c,int width,int height,int attr){
    int i;
    if(width == c->width && height > c->height){
        cons_buf_extend(c,height-c->height,0x07);
        return;
    }

    DEBUG_p("resizing to %ix%i (cellsize %d)",width,height,(int)sizeof(ConsoleCell));
    if(width*height > c->width*c->height){
        if(width*height > c->allocation){
            if(c->screen){
                c->screen = realloc(c->screen,width * height * sizeof(ConsoleCell));
            }else{
                c->screen = malloc(width * height * sizeof(ConsoleCell));
            }
            c->allocation = width*height;
        }
        for(i=c->width*c->height;i<width*height;++i){
            c->screen[i].ch = ' ';
            c->screen[i].attr = attr;
        }
    }
    c->width = width;
    c->height = height;
}


/*
void text_screen_put(TextScreen* scrn,int c,int y,unsigned short cell){
    scrn->screen[(x)+(y)*scrn->width]=cell;
}
*/
void cons_buf_charxya(Console* c,int x,int y,int attr,int ch){
    if(x >= c->width)
        return;
    if(y >= c->height)
        cons_buf_set_size(c,c->width,y+1,attr);
    c->screen[x + y * c->width].ch = ch;
    c->screen[x + y * c->width].attr = attr;
    //DEBUG_p("attr: 0x%x",c->screen[x+y*c->width].attr);
}


void cons_buf_scroll(Console* c,int steps,int attr){
    int i;
    memmove(c->screen,&c->screen[c->width * steps],c->width * (c->height-steps) * 2);
    for(i=c->width*(c->height-steps);i<c->width*c->height;++i){
        c->screen[i].ch = ' ';
        c->screen[i].attr = attr;
    }
}


int cons_buf_width(Console* c){
    return c->width;
}


int cons_buf_height(Console* c){
    return c->height;
}


void cons_buf_extend(Console* c,int steps,int attr){
    int i;
    DEBUG_p("extending height to %i",c->height+steps);
    if(c->screen){
        c->screen = realloc(c->screen,c->width * (c->height+steps) * sizeof(ConsoleCell));
    }else{
        c->screen = malloc(c->width * (c->height+steps) * sizeof(ConsoleCell));
    }
    for(i=0;i<c->width*steps;++i){
        c->screen[c->width * c->height + i].ch = ' ';
        c->screen[c->width * c->height + i].attr = attr;
    }
    c->height += steps;
}



void cons_buf_reshape(Console* c,int width){
    int size = c->width*c->height;
    c->width = width;
    c->height = size/width;
}


const ConsoleCell* cons_buf_buffer(Console* c){
    return c->screen;
}


void cons_buf_setmem(Console *c, ConsoleCell *mem, int width, int height){
    if(c->allocation)
        free(c->screen);
    c->width = width;
    c->height = height;
    c->allocation = 0;
    c->screen = mem;
}


static void cons_buf_null(){}


const ConsoleInterface cons_buf_if = {
    .set_size = cons_buf_set_size,
    .cursorxy = cons_buf_null,
    .charxya = cons_buf_charxya,
    .scroll = cons_buf_scroll
#ifndef TERMINAL_FIXED_WIDTH
    ,.width = cons_buf_width
#endif
#ifndef TERMINAL_FIXED_HEIGHT
    ,.height = cons_buf_height,
    .extend = cons_buf_extend
#endif
};
