/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <debug.h>
#include <errno.h>
#include <fcntl.h>                      /* O_NONBLOCK */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <unistd.h>                     /* dup2() */


#include "converter.h"


#define BUFFER_SIZE 4096

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif


static void
_cleanup_child(converter_t *c)
{
    if(c->input[0] >= 0){
        close(c->input[0]);
        c->input[0] = -1;
    }
    if(c->input[1] >= 0){
        close(c->input[1]);
        c->input[1] = -1;
    }
    if(c->output[0] >= 0){
        close(c->output[0]);
        c->output[0] = -1;
    }
    if(c->output[1] >= 0){
        close(c->output[1]);
        c->output[1] = -1;
    }
    if(c->pid > 0){
        int status;
        kill(c->pid,SIGTERM);
        waitpid(c->pid, &status, 0);
        c->result = WIFEXITED(status) && 0==WEXITSTATUS(status) ? 0 : -1;
        c->pid = 0;
    }
}


converter_t *
abydos_converter_init(converter_t *c, const char *prog, const char *mime_type, int flags)
{
    c->pid = 0;
    c->input[0] = -1;
    c->input[1] = -1;
    c->output[0] = -1;
    c->output[1] = -1;
    memset(c->input, 0, sizeof(c->input));
    c->ar = abydos_create(mime_type);
    if(!c->ar)
        return NULL;
    c->flags = flags;

    pipe(c->input);
    pipe(c->output);
    c->pid = fork();
    if(0==c->pid){
        close(c->output[1]);
        close(c->input[0]);
        close(0);
        dup2(c->output[0],0);
        close(1);
        dup2(c->input[1],1);
        execlp(prog, prog, NULL);
        exit(1);
    }else if(c->pid < 0){
        /* fork() failed */
        return NULL;
    }
    DEBUG_p("started child with pid %d",c->pid);
    close(c->output[0]);
    close(c->input[1]);
    c->input[1] = -1;
    c->output[0] = -1;
    c->eof = 0;
    return c;
}


void
abydos_converter_done(converter_t *c)
{
    if(c->ar)
        abydos_destroy(c->ar);
    _cleanup_child(c);
}


static int
_communicate(abydos_plugin_handle_t *h, const char *data, size_t len, int timeout_ms)
{
    for(;;){
        fd_set fdread;
        fd_set fdwrite;
        struct timeval timeout;
        ssize_t bytes;
        int fdmax = h->input[0];
        FD_ZERO(&fdread);
        FD_SET(h->input[0], &fdread);
        if(len > 0){
            fdmax = MAX(fdmax,h->output[1]);
            FD_ZERO(&fdwrite);
            FD_SET(h->output[1], &fdwrite);
        }
        timeout.tv_sec = timeout_ms / 1000;
        timeout.tv_usec = timeout_ms % 1000;
        if(select(fdmax+1, &fdread, len>0 ? &fdwrite : NULL, NULL, len>0 ? NULL : &timeout) < 1){
            DEBUG_p("timeout");
            return 0;
        }
        if(FD_ISSET(h->input[0], &fdread)){
            char buffer[BUFFER_SIZE];
            bytes = read(h->input[0],buffer,BUFFER_SIZE);
            int res;
            if(0==bytes){
                int status;
                DEBUG_p("got eof");
                //kill(h->pid,SIGTERM);
                waitpid(h->pid, &status, 0);
                close(h->input[0]);
                h->eof = 1;
                h->result = WIFEXITED(status) && 0==WEXITSTATUS(status) ? 0 : -1;
                return h->result;
            }else if(bytes<0){
                switch(errno){
                case EAGAIN:
                    //DEBUG_p("more bytes to come");
                    return 0;
                default:
                    DEBUG_p("got error");
                    return -1;
                }
            }
            //DEBUG_p("got %d bytes", (int)bytes);
            res = abydos_load_feed(h->ar, buffer, bytes);
            if(res)
                return res;
        }
        if(len>0 && FD_ISSET(h->output[1], &fdwrite)){
            bytes = MIN(len,BUFFER_SIZE);
            //DEBUG_p("writing %d bytes",(int)bytes);
            bytes = write(h->output[1], data, bytes);
            if(bytes<0)
                return -1;
            data += bytes;
            len -= bytes;
        }
    }
}


static void
_on_info(void *userdata, abydos_t *ar)
{
    converter_t *c = userdata;
    abydos_get_size(c->ar, &c->info->width, &c->info->height);
    if(c->info_func)
        c->info_func(c->userdata);
}


static void
_on_update(void *userdata, abydos_t *ar, cairo_rectangle_int_t *rect)
{
    converter_t *c = userdata;
    if(c->progress_func)
        c->progress_func(c->userdata, rect);
}


void
abydos_converter_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    converter_t *c = (converter_t*)h;
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;


    abydos_load_begin(h->ar, _on_info, _on_update, c);
}


int
abydos_converter_feed(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    if(h->eof)
        return h->result;
    return _communicate(h, data, len, 0);
}


int
abydos_converter_end(abydos_plugin_handle_t *h)
{
    DEBUG_p("closing pipe");
    close(h->output[1]);
    h->output[1] = -1;
    {
        int status;
        if(waitpid(h->pid, &status, WNOHANG) > 0){
            close(h->input[0]);
            h->eof = 1;
            h->result = WIFEXITED(status) && 0==WEXITSTATUS(status) ? 0 : -1;
        }else{
            DEBUG_p("The child (%d) did not exit",h->pid);
        }
    }
    if(!h->eof){
        /* Make input pipe blocking again to let the converter have a
           chance to finish. */
        //int flags = fcntl(h->input[0],F_GETFL,0);
        //fcntl(h->input[0], F_SETFL, flags & ~O_NONBLOCK);
        _communicate(h, NULL, 0, 3000);
    }
    return abydos_load_end(h->ar);
}


cairo_surface_t *
abydos_converter_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    converter_t *c = (converter_t*)h;
    if(h->flags & CONVERTER_FLIP){
        cairo_surface_t *src = abydos_get_image_surface(c->ar);
        cairo_surface_t *dst = cairo_surface_create_similar(src, CAIRO_CONTENT_COLOR_ALPHA, h->info->width, h->info->height);
        cairo_t *cr = cairo_create(dst);
        cairo_scale(cr, 1, -1);
        cairo_set_source_surface(cr, src, 0, -h->info->height);
        cairo_paint(cr);
        cairo_destroy(cr);
        cairo_surface_destroy(src);
        return dst;
    }else{
        return abydos_get_image_surface(c->ar);
    }
}
