/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <nil-cairo.h>
#include <spng.h>
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_spng_get_mime_types()
{
    static const char *mime_types[] = {
        "image/png",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_spng_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_spng_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static int
_spng_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    spng_ctx *ctx = spng_ctx_new(0);
    size_t src_len;
    uint8_t *src;
    struct spng_ihdr ihdr;
    int bpc;
    nile_format_t format;

    spng_set_png_buffer(ctx, data, len);

    /* Calculate output image size */
    spng_decoded_image_size(ctx, SPNG_FMT_RGBA8, &src_len);
    src = malloc(src_len);

    /* Get an 8-bit RGBA image regardless of PNG format */
    spng_decode_image(ctx, src, src_len, SPNG_FMT_RAW, 0);

    spng_get_ihdr(ctx, &ihdr);
    switch(ihdr.bit_depth){
    case 8:
        switch(ihdr.color_type){
        case SPNG_COLOR_TYPE_GRAYSCALE:
            bpc = 1;
            format = NILE_FORMAT_g8;
            break;
        case SPNG_COLOR_TYPE_TRUECOLOR:
            bpc = 3;
            format = NILE_FORMAT_r8g8b8;
            break;
        case SPNG_COLOR_TYPE_INDEXED:
            /* FIXME!! */
            goto out;
        case SPNG_COLOR_TYPE_GRAYSCALE_ALPHA:
            bpc = 2;
            format = NILE_FORMAT_g8a8;
            break;
        case SPNG_COLOR_TYPE_TRUECOLOR_ALPHA:
            bpc = 4;
            format = NILE_FORMAT_r8g8b8a8;
            break;
        }
        break;
    case 16:
        switch(ihdr.color_type){
        case SPNG_COLOR_TYPE_GRAYSCALE:
            bpc = 2;
            format = NILE_FORMAT_be_g16;
            break;
        case SPNG_COLOR_TYPE_TRUECOLOR:
            bpc = 6;
            format = NILE_FORMAT_be_r16g16b16;
            break;
        case SPNG_COLOR_TYPE_INDEXED:
            /* FIXME!! */
            goto out;
        case SPNG_COLOR_TYPE_GRAYSCALE_ALPHA:
            bpc = 4;
            format = NILE_FORMAT_be_g16a16;
            break;
        case SPNG_COLOR_TYPE_TRUECOLOR_ALPHA:
            bpc = 8;
            format = NILE_FORMAT_be_r16g16b16a16;
            break;
        }
        break;
    default:
        goto out;
    }
    h->surface = nil_cairo_surface_create_from_direct(src, ihdr.width, ihdr.height, ihdr.width * bpc, bpc, format, 0);

out:
    free(src);
    spng_ctx_free(ctx);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_spng_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


static void
_spng_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _spng_get_mime_types,
        .new = _spng_new,
        .free = _spng_free,
        .create_from_data = _spng_create_from_data,
        .get_image_surface = _spng_get_image_surface,
        .get_duration = _spng_get_duration
    };
    return &p;
}
