/*
 * Copyright (C) 2023 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <limits.h>                     /* INT_MAX */
#include <loa-buffer.h>
#include <loa-cairo.h>
#include <math.h>
#include <stdint.h>                     /* uint8_t */
#include <stdlib.h>
#include <string.h>

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define MAXSIZE 4096
#define DEFAULT_WIDTH 2

#define NAME 1
#define EXTENTS 2
#define COLOR_TABLE 3
#define BACKGROUND 4
#define LAYER_TABLE 5
/*
Width		= 7
NotifyTable	= 8
*/
#define TRANSPARENT 9
/*
Transform	= 10
Flags		= 11

*/
#define POINT 16
#define MOVE_TO 17
#define LINE_TO 18
#define POLY_LINE 19
#define REL_MOVE_TO 20
#define REL_LINE_TO 21
#define REL_POLY_LINE 22
#define RECTANGLE 23
#define CIRCLE 24
#define ARC 25
/*
Bezier		= 26
*/
#define TEXT 27
#define TEXT_HEIGHT 28
#define SET_COLOR 29
#define SET_LAYER 30
#define SET_PEN_WIDTH 31
#define FILL_MODE 32
#define DATA 34
#define X_1D_LINK 35
#define X_2D_LINK 36
#define INVISIBLE 37

#define SET_OBJECT_ID 39

typedef enum{
    SH_PATH,
    SH_RECTANGLE,
    SH_CIRCLE,
    SH_PIE
}shape_type_t;

struct _loa_object_t{
    shape_type_t type;
    union{
        loa_path_t path;
        struct{
            double x;
            double y;
            double w;
            double h;
        }rectangle;
        struct{
            double x;
            double y;
            double r;
        }circle;
        struct{
            double x;
            double y;
            double r;
            double t0;
            double t1;
        }pie;
    };
    int color;
    int width;
    int fill;
};

typedef struct{
    double r;
    double g;
    double b;
}color_t;

typedef struct{
    int visible;
    loa_array_t object_array;
}layer_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    int background;
    color_t palette[256];
    loa_buffer_t loa;
    layer_t *layer;
    double scale_factor;
};


static void
_create_layers(abydos_plugin_handle_t *h, int count)
{
    int i;
    h->layer = malloc(sizeof(layer_t) * count);
    h->info->layer_count = count;
    for(i=0; i<h->info->layer_count; ++i){
        loa_array_init(&h->layer[i].object_array, sizeof(loa_object_t));
    }
}


static void
_destroy_layers(abydos_plugin_handle_t *h)
{
    int i;
    for(i=0; i<h->info->layer_count; ++i){
        loa_array_done(&h->layer[i].object_array);
    }
}


#if 0
static const color_t default_palette[256] = {
    /* 0 - 15 */
    /* SVFC_Black 0 */
    /* SVFC_Blue 3 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,3}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 16 - 31 */
    /* SVFC_Green 28 */
    /* SVFC_Cyan 31 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,255,0}, {0,0,0}, {0,0,0}, {0,255,255},
    /* 32 - 47 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 48 - 63 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 64 - 79 */
    /* SVFC_DarkGray 73 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0.29,0.29,0.29}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 80 - 95 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 96 - 111 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 112 - 127 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 128 - 143 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 144 - 159 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 160 - 175 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 176 - 191 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 192 - 207 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 208 - 223 */
    /* SVFC_LightGray 219 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0.86,0.86,0.86}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 224 - 239 */
    /* SVFC_Red 224 */
    /* SVFC_Magenta 227 */
    {1,0,0}, {0,0,0}, {0,0,0}, {1,0,1}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0},
    /* 240 - 255 */
    /* SVFC_Gray 146 */
    /* SVFC_Yellow 252 */
    /* SVFC_White 255 */
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {0.57,0.57,0.57}, {0,0,0},
    {0,0,0}, {0,0,0}, {0,0,0}, {0,0,0}, {1,1,0}, {0,0,0}, {0,0,0}, {1,1,1}
};
#endif


static const char **
_svf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/vnd-svf", "image/vnd.svf", "vector/vnd.svf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_svf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    int i;
    h->info = info;
    h->info->threadsafe = 1;
    for(i=0; i<256; ++i){
        h->palette[i].r = (double)(i >> 5) / 7;
        h->palette[i].g = (double)((i >> 2) & 7) / 7;
        h->palette[i].b = (double)(i & 3) / 3;
    }
    loa_buffer_init(&h->loa, 0, 0, 0, 0);
    h->layer = NULL;
    h->scale_factor = 1;
    return h;
}


static void
_svf_free(abydos_plugin_handle_t *h)
{
    loa_buffer_done(&h->loa);
    if(h->layer)
        _destroy_layers(h);
    free(h);
}


typedef struct{
    const char *str;
    size_t len;
}str_t;

typedef struct{
    abydos_plugin_handle_t *h;
    int background;
    int transparent;
    //color_t *palette;
    loa_buffer_t *loa;
    //layer_t *layer;
    loa_object_t *stroke_path;
    double xmin;
    double ymin;
    double xmax;
    double ymax;
    int layer; /* current layer */
    int x;
    int y;
    int color;
    int width;
    int fill;
}read_context_t;


static int
_read_byte(const char **data, const char *end)
{
    if(*data + 1 > end)
        return -1;
    return *(uint8_t*)(*data)++;
}

static int
_read_sint(const char **data, const char *end, int argtype)
{
    int val;
    switch(argtype){
    case 0:
        if(*data + 1 > end)
            return -1;
        val = *(int8_t*)*data;
        *data += 1;
        return val;
    case 1:
        if(*data + 2 > end)
            return -1;
        val = (int16_t)be16toh(*(uint16_t*)*data);
        *data += 2;
        return val;
    case 2:
        if(*data + 4 > end)
            return -1;
        val = (int32_t)be32toh(*(uint32_t*)*data);
        *data += 4;
        return val;
    case 3:
        DEBUG_p("floating points not handled");
        return -1;
    }
    /* not reached */
    return -1;
}


static unsigned  int
_read_uint(const char **data, const char *end, int argtype)
{
    unsigned int val;
    switch(argtype){
    case 0:
        return _read_byte(data, end);
    case 1:
        if(*data + 2 > end)
            return -1;
        val = be16toh(*(uint16_t*)*data);
        *data += 2;
        return val;
    case 2:
        if(*data + 4 > end)
            return -1;
        val = be32toh(*(uint32_t*)*data);
        *data += 4;
        return val;
    case 3:
        DEBUG_p("floating points not handled");
        return -1;
    }
    /* not reached */
    return -1;
}


static float
_read_angle(const char **data, const char *end, int argtype)
{
    double circ;
    switch(argtype){
    case 0:
        circ = 0x80;
        break;
    case 1:
        circ = 0x8000;
        break;
    case 2:
        circ = 0x80000000;
        break;
    case 3:
        DEBUG_p("floating points not handled");
        return 0;
    }
    return M_PI * _read_uint(data, end, argtype) / circ;
}


static str_t
_read_str(const char **data, const char *end)
{
    str_t s;
    s.str = *data;
    s.len = strnlen(*data, end - *data);
    *data += s.len;
    /* Skip the terminating zero, if there is one. */
    if(*data < end)
        *data += 1;
    return s;
}


static loa_object_t *
_add_shape(read_context_t *c, shape_type_t type)
{
#if 1
    if(!c->h->layer)
        _create_layers(c->h, 1);
    loa_object_t *obj = (loa_object_t*)loa_array_append(&c->h->layer[c->layer].object_array, 1);
#else
    loa_object_t *obj = loa_buffer_add_object(c->loa);
#endif
    obj->type = type;
    obj->color = c->color;
    obj->width = c->width;
    obj->fill = c->fill;
    return obj;
}


static void
_end_stroke_path(read_context_t *c){
    if(!c->stroke_path)
        return;
    loa_buffer_end_path(c->loa, &c->stroke_path->path);
    c->stroke_path = NULL;
}


static loa_object_t *
_ensure_stroke_path(read_context_t *c, int close){
    loa_point_t *pt;
    if(c->stroke_path && close)
        _end_stroke_path(c);
    if(!c->stroke_path){
        c->stroke_path = _add_shape(c, SH_PATH);
        c->stroke_path->path.segment_count = 0;
        c->stroke_path->path.point_count = 0;
        loa_buffer_begin_path(c->loa, &c->stroke_path->path);
        pt = loa_buffer_push_segment(c->loa, &c->stroke_path->path, LOA_GOTO);
        pt[0].x = c->x;
        pt[0].y = c->y;
    }
    return c->stroke_path;
}


static ssize_t
_read_token(read_context_t *c, const char *data, const char *end)
{
    const char *p = data;
    int command;
    int argtype;
    int count;
    int val;
    str_t s;
    int i;
    loa_object_t *shape;
    loa_point_t *pt;
    command = *(uint8_t*)p++;
    argtype = command >> 6;
    command &= 0x3f;
    //DEBUG_p("command: %d (%d)", command, argtype);
    switch(command){
    case NAME:
        s = _read_str(&p, end);
        DEBUG_p("name: %.*s", (int)s.len, s.str);
        break;
    case EXTENTS:
        c->xmin = _read_uint(&p, end, argtype);
        c->ymin = _read_uint(&p, end, argtype);
        c->xmax = _read_uint(&p, end, argtype);
        c->ymax = _read_uint(&p, end, argtype);
        break;
    case COLOR_TABLE:
        count = _read_uint(&p, end, argtype);
        for(i=0; i<count; ++i){
            int r = _read_uint(&p, end, argtype);
            int g = _read_uint(&p, end, argtype);
            int b = _read_uint(&p, end, argtype);
            c->h->palette[i].r = (double)r / 255;
            c->h->palette[i].g = (double)g / 255;
            c->h->palette[i].b = (double)b / 255;
        }
        break;
    case BACKGROUND:
        c->background = _read_uint(&p, end, argtype) & 0xff;
        break;
    case LAYER_TABLE:
        count = _read_uint(&p, end, argtype);
        _create_layers(c->h, count);
        for(i=0; i<count; ++i){
            c->h->layer[i].visible = _read_byte(&p, end);
            /* FIXME!! store layer name? */
            _read_str(&p, end);
        }
        break;
    case TRANSPARENT:
        c->transparent = _read_uint(&p, end, argtype) & 0xff;
        break;
    case POINT:
        /* FIXME!! */
        _read_uint(&p, end, argtype);
        _read_uint(&p, end, argtype);
        break;
    case MOVE_TO:
        _end_stroke_path(c);
        c->x = _read_uint(&p, end, argtype);
        c->y = _read_uint(&p, end, argtype);
        break;
    case LINE_TO:
        shape = _ensure_stroke_path(c, 0);
        c->x = _read_uint(&p, end, argtype);
        c->y = _read_uint(&p, end, argtype);
        //DEBUG_p("line to %d, %d", c->x ,c->y);
        pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_LINE);
        pt[0].x = c->x;
        pt[0].y = c->y;
        break;
    case POLY_LINE:
        if(c->fill){
            /* draw a filled polygon */
            _end_stroke_path(c);
            shape = _add_shape(c, SH_PATH);
            loa_buffer_begin_path(c->loa, &shape->path);
            pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_GOTO);
            pt[0].x = c->x;
            pt[0].y = c->y;
        }else{
            /* add points to current path */
            shape = _ensure_stroke_path(c, 0);
        }
        {
            int count = _read_uint(&p, end, argtype);
            while(count-->0){
                c->x = _read_uint(&p, end, argtype);
                c->y = _read_uint(&p, end, argtype);
                //DEBUG_p("(poly) line to %d, %d", c->x ,c->y);
                pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_LINE);
                pt[0].x = c->x;
                pt[0].y = c->y;
            }
        }
        if(c->fill){
            loa_buffer_push_segment(c->loa, &shape->path, LOA_CLOSE);
            loa_buffer_end_path(c->loa, &shape->path);
        }
        break;
    case REL_LINE_TO:
        shape = _ensure_stroke_path(c, 0);
        c->x += _read_sint(&p, end, argtype);
        c->y += _read_sint(&p, end, argtype);
        //DEBUG_p("(rel) line to %d, %d", c->x ,c->y);
        pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_LINE);
        pt[0].x = c->x;
        pt[0].y = c->y;
        break;
    case REL_POLY_LINE:
        if(c->fill){
            /* draw a filled polygon */
            _end_stroke_path(c);
            shape = _add_shape(c, SH_PATH);
            loa_buffer_begin_path(c->loa, &shape->path);
            pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_GOTO);
            pt[0].x = c->x;
            pt[0].y = c->y;
        }else{
            /* add points to current path */
            shape = _ensure_stroke_path(c, 0);
        }
        {
            int count = _read_uint(&p, end, argtype);
            while(count-->0){
                c->x += _read_sint(&p, end, argtype);
                c->y += _read_sint(&p, end, argtype);
                //DEBUG_p("(poly) line to %d, %d", c->x ,c->y);
                pt = loa_buffer_push_segment(c->loa, &shape->path, LOA_LINE);
                pt[0].x = c->x;
                pt[0].y = c->y;
            }
        }
        if(c->fill){
            loa_buffer_push_segment(c->loa, &shape->path, LOA_CLOSE);
            loa_buffer_end_path(c->loa, &shape->path);
        }
        break;
    case RECTANGLE:
        _end_stroke_path(c);
        shape = _add_shape(c, SH_RECTANGLE);
        shape->rectangle.x = c->x;
        shape->rectangle.y = c->y;
        shape->rectangle.w = _read_uint(&p, end, argtype);
        shape->rectangle.h = _read_uint(&p, end, argtype);
        break;
    case CIRCLE:
        _end_stroke_path(c);
        shape = _add_shape(c, SH_CIRCLE);
        shape->circle.x = c->x;
        shape->circle.y = c->y;
        shape->circle.r = _read_uint(&p, end, argtype);
        break;
    case ARC:
        if(c->fill){
            _end_stroke_path(c);
            shape = _add_shape(c, SH_PIE);
            shape->pie.x = c->x;
            shape->pie.y = c->y;
            shape->pie.r = _read_uint(&p, end, argtype);
            shape->pie.t0 = _read_angle(&p, end, argtype);
            shape->pie.t1 = _read_angle(&p, end, argtype);
        }else{
            /* FIXME!! add arch to path */
            DEBUG_p("arc on path");
        }
        break;
    case TEXT:
        /* FIXME!! */
        _read_uint(&p, end, argtype);
        _read_str(&p, end);
        break;
    case TEXT_HEIGHT:
        val = _read_uint(&p, end, argtype);
        DEBUG_p("text height: %d", val);
        break;
    case SET_COLOR:
        _end_stroke_path(c);
        c->color = _read_uint(&p, end, argtype);
        //DEBUG_p("color: %d", c->color);
        break;
    case SET_LAYER:
        val = _read_uint(&p, end, argtype) & c->h->info->layer_count;
        DEBUG_p("layer: %d", val);
        c->layer = val;
        break;
    case SET_PEN_WIDTH:
        _end_stroke_path(c);
        c->width = _read_uint(&p, end, argtype);
        DEBUG_p("pen width: %d", c->width);
        if(!c->width)
            c->width = DEFAULT_WIDTH;
        break;
    case FILL_MODE:
        _end_stroke_path(c);
        c->fill = _read_uint(&p, end, argtype);
        //DEBUG_p("fill mode %s", c->fill ? "on" : "off");
        break;
    case X_1D_LINK:
    case X_2D_LINK:
        _read_str(&p, end);
        _read_str(&p, end);
        break;
    default:
        DEBUG_p("unknown command: %d", command);
        return -1;
    }
    return p - data;
}


static ssize_t
_read_intro(read_context_t *c, const char *data, const char *end)
{
    const char *p = data;
    str_t s;
    if(p + 5 > end)
        return -1;
    if(!strncmp(p, "SVFv", 4)){
        p += 4;
    }else if(!strncmp(p, "SVF v", 5)){
        p += 5;
    }else{
        return -1;
    }
    s = _read_str(&p, end);
    DEBUG_p("version: %.*s", (int)s.len, s.str);
    return p - data;
}


#if 0
static ssize_t
_read_header(read_context_t *c, const char *data, const char *end)
{
    return 0;
}
#endif


static ssize_t
_read_body(read_context_t *c, const char *data, const char *end)
{
    ssize_t bytes;
    while(data < end){
        bytes = _read_token(c, data, end);
        if(bytes < 0)
            return bytes;
        data += bytes;
        while((data < end) && !*data)
            ++data;
    }
    return 0;
}


static int
_svf_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const char *end = data + len;
    read_context_t c;
    ssize_t bytes;

    bytes = _read_intro(&c, data, end);
    if(bytes < 0)
        return -1;
    data += bytes;

#if 0
    bytes = _read_header(&c, data, end);
    if(bytes < 0)
        return -1;
    data += bytes;
#endif

    c.h = h;
    //data += 16;
    c.background = 0;
    c.transparent = -1;
    //c.palette = h->palette;
    c.loa = &h->loa;
    c.stroke_path = NULL;
    c.xmin = INT_MAX;
    c.ymin = INT_MAX;
    c.xmax = 0;
    c.ymax = 0;
    c.layer = 0;
    c.x = 0;
    c.y = 0;
    c.color = 255;
    c.width = DEFAULT_WIDTH;
    c.fill = 0;
    bytes = _read_body(&c, data, end);
    _end_stroke_path(&c);
    if(bytes < 0)
        return -1;

    h->background = c.background == c.transparent ? -1 : c.background;
    if(c.xmin > c.xmax || c.ymin > c.ymax){
        int count = loa_buffer_point_count(&h->loa);
        loa_point_t *p = loa_buffer_get_point(&h->loa, 0);
        int i;
        DEBUG_p("points: %d", count);
        for(i=0; i<count; ++i){
            if(p[i].x < c.xmin)
                c.xmin = p[i].x;
            if(p[i].y < c.ymin)
                c.ymin = p[i].y;
            if(p[i].x > c.xmax)
                c.xmax = p[i].x;
            if(p[i].y > c.ymax)
                c.ymax = p[i].y;
        }
    }
    {
        double size;
        DEBUG_p("dimensions: %fx%f - %fx%f", c.xmin, c.ymin, c.xmax, c.ymax);
        size = MAX(c.xmax, c.ymax);
        if(size > MAXSIZE)
            h->scale_factor = size / MAXSIZE;
        h->info->width = c.xmax / h->scale_factor;
        h->info->height = c.ymax / h->scale_factor;
    }
    return 0;
}


#if 1
static void
_set_color(cairo_t *cr, color_t *c){
    cairo_set_source_rgb(cr, c->r, c->g, c->b);
}
#else
static void
_set_color(cairo_t *cr, int color)
{
    int r = color >> 5;
    int g = (color >> 2) & 7;
    int b = color & 3;
    cairo_set_source_rgb(cr, (double)r / 7, (double)g / 7, (double)b / 3);
}
#endif


static void
_svf_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visible,
            int variant,
            double time,
            int frame)
{
    const loa_segment_t *segment = loa_buffer_get_segment(&h->loa, 0);
    const loa_point_t *point = loa_buffer_get_point(&h->loa, 0);
    int i;
    cairo_save(cr);
    cairo_scale(cr, 1, -1);
    cairo_translate(cr, 0, -h->info->height);
    cairo_scale(cr, 1.0 / h->scale_factor, 1.0 / h->scale_factor);
    if(h->background >= 0){
        //cairo_set_source_rgb(cr, h->palette[h->].r, h->palette[obj->color].g, h->palette[obj->color].b);
        _set_color(cr, &h->palette[h->background]);
        cairo_paint(cr);
    }
    for(i=0; i<h->info->layer_count; ++i){
        layer_t *layer;
        int j;
        if(!layer_visible[i])
            continue;
        layer = &h->layer[i];
        for(j=0; j<loa_array_length(&layer->object_array); ++j){
            //const loa_object_t *obj = loa_buffer_get_object(&h->loa, i);
            const loa_object_t *obj = (loa_object_t*)loa_array_get(&layer->object_array, j);
            switch(obj->type){
            case SH_PATH:
                loa_cairo_draw_path(cr, &obj->path, segment, point);
                break;
            case SH_RECTANGLE:
                cairo_rectangle(cr, obj->rectangle.x, obj->rectangle.y,
                                    obj->rectangle.w, obj->rectangle.h);
                break;
            case SH_CIRCLE:
                cairo_arc(cr, obj->circle.x, obj->circle.y, obj->circle.r,
                              0, 2*M_PI);
                break;
            case SH_PIE:
                cairo_move_to(cr, obj->pie.x, obj->pie.y);
                cairo_arc(cr, obj->pie.x, obj->pie.y, obj->pie.r,
                              obj->pie.t0, obj->pie.t1);
            }
            //_set_color(cr, obj->color);
            //cairo_set_source_rgb(cr, h->palette[obj->color].r, h->palette[obj->color].g, h->palette[obj->color].b);
            _set_color(cr, &h->palette[obj->color]);
            if(obj->fill){
                cairo_fill(cr);
            }else{
                cairo_set_line_width(cr, obj->width * h->scale_factor);
                cairo_stroke(cr);
            }
        }
    }
    cairo_restore(cr);
}


static void
_svf_get_layer_visibility(abydos_plugin_handle_t *h, int page, cairo_bool_t *visibility)
{
    int i;
    for(i=0; i<h->info->layer_count; ++i)
        visibility[i] = h->layer[i].visible;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _svf_get_mime_types,
        .new = _svf_new,
        .free = _svf_free,
        .create_from_data = _svf_create_from_data,
        .render = _svf_render,
        .get_layer_visibility = _svf_get_layer_visibility
    };
    return &p;
}
