/*
 * Copyright (C) 2021 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>
#include <string.h>

#include "tribble.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    tribble_document_t *doc;
};


static const char **
_gs_get_mime_types()
{
    static const char *mime_types[] = {
        "application/postscript",
        "application/eps", "application/x-eps", "image/eps", "image/x-eps",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->scalable_size = 1;
    info->threadsafe = 1;
    h->info = info;
    h->doc = NULL;
    return h;
}


static void
_free(abydos_plugin_handle_t *h)
{
    if(h->doc)
        tribble_document_destroy(h->doc);
    free(h);
}


static int
_gs_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    h->doc = tribble_document_create_from_data(data, len);
    return h->doc ? 0 : -1;
}


static int
_gs_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    h->doc = tribble_document_create_from_file(filename);
    return h->doc ? 0 : -1;
}


static void
_render(abydos_plugin_handle_t *h,
        cairo_t *cr,
        int pagenr,
        cairo_bool_t *layer_visibility,
        int variant,
        double time,
        int frame)
{
    const tribble_page_t *page = tribble_document_get_page(h->doc, pagenr);
    tribble_page_render(page, cr);
}


static void
_get_page_info(abydos_plugin_handle_t *h, int pagenr, abydos_plugin_page_info_t *info)
{
    double width;
    double height;
    const tribble_page_t *page = tribble_document_get_page(h->doc, pagenr);
    tribble_page_get_size(page, &width, &height);
    DEBUG_p("size: %fx%f", width, height);
    info->width = width;
    info->height = height;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gs_get_mime_types,
        .new = _new,
        .free = _free,
        .create_from_data = _gs_create_from_data,
        //.create_from_file = _gs_create_from_file,
        .render = _render,
        .get_page_info = _get_page_info
    };
    return &p;
}
