/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <math.h>
#include <stddef.h>

#include "emfcairo.h"
#include "uemf.h"
#include "uemf_safe.h"


#define UM2PX (0.00370370370369)
#define SCALE (0.07988017973)


#if 0
static uint32_t lu_crc32(const uint8_t *message, uint32_t len){
   uint32_t i, j;
   uint32_t crc, mask;

   crc = 0xFFFFFFFF;
   for(i=0;i<len;i++){     // over all bytes
      crc = crc ^ *message++;
      for (j = 0; j < 8; j++) { // over all bits
         mask = -(crc & 1);
         crc = (crc >> 1) ^ (0xEDB88320 & mask);
      }
   }
   return ~crc;
}
#endif


static void on_header(U_EMRHEADER *emr, cairo_t *cr){
    double scale = UM2PX*emr->szlMicrometers.cx/emr->rclBounds.right;
    DEBUG_p("bounds %dx%d - %dx%d",emr->rclBounds.left,emr->rclBounds.top,emr->rclBounds.right,emr->rclBounds.bottom);
    DEBUG_p("size in mm %fx%f",(double)emr->szlMicrometers.cx/1000,(double)emr->szlMicrometers.cy/1000);
    scale = SCALE;
    DEBUG_p("scale: %f",scale); // 0,07988017973
    cairo_scale(cr,scale,scale);
}


static void on_polybezier(){
    DEBUG_p("begin");
}


static void on_polygon(){
    DEBUG_p("begin");
}


static void on_polyline(){
    DEBUG_p("begin");
}


static void on_polybezierto(U_EMRPOLYBEZIERTO *emr,cairo_t *cr){
    //DEBUG_p("points: %d",emr->cptl);
    cairo_curve_to(cr,emr->aptl[0].x,emr->aptl[0].y,emr->aptl[1].x,emr->aptl[1].y,emr->aptl[2].x,emr->aptl[2].y);
}


static void on_eof(U_EMREOF *emr,cairo_t *cr){
    DEBUG_p("begin");
}


static void on_setmapmode(U_EMRSETMAPMODE *emr,cairo_t *cr){
    //DEBUG_p("begin");
}


static void on_setbkmode(U_EMRSETBKMODE *emr,cairo_t *cr){
    //DEBUG_p("begin");
}


static void on_setpolyfillmode(U_EMRSETPOLYFILLMODE *emr,cairo_t *cr){
    //DEBUG_p("begin");
}


static void on_setrop2(){
    DEBUG_p("begin");
}


static void on_settextalign(){
    DEBUG_p("begin");
}


static void on_settextcolor(U_EMRSETTEXTCOLOR *emr,cairo_t *cr){
    cairo_set_source_rgb(cr,(double)emr->crColor.Red/255,(double)emr->crColor.Green/255,(double)emr->crColor.Blue/255);
}


static void on_movetoex(U_EMRMOVETOEX *emr,cairo_t *cr){
    cairo_move_to(cr,emr->ptl.x,emr->ptl.y);
}


static void on_modifyworldtransform(U_EMRMODIFYWORLDTRANSFORM *emr,cairo_t *cr){
    DEBUG_p("begin");
}


static void on_selectobject(){
    DEBUG_p("begin");
}


static void on_createbrushindirect(U_EMRCREATEBRUSHINDIRECT *emr,cairo_t *cr){
    cairo_set_source_rgb(cr,(double)emr->lb.lbColor.Red/255,(double)emr->lb.lbColor.Green/255,(double)emr->lb.lbColor.Blue/255);
}


static void on_deleteobject(){
    DEBUG_p("begin");
}


static void on_lineto(U_EMRLINETO *emr,cairo_t *cr){
    cairo_line_to(cr,emr->ptl.x,emr->ptl.y);
}


static void on_beginpath(U_EMRBEGINPATH *emr,cairo_t *cr){
    DEBUG_p("begin");
}


static void on_endpath(U_EMRENDPATH *emr,cairo_t *cr){
    DEBUG_p("begin");
}


static void on_closefigure(){
    DEBUG_p("begin");
}


static void on_fillpath(U_EMRFILLPATH *emr,cairo_t *cr){
    DEBUG_p("bounds: %dx%d - %dx%d",emr->rclBounds.left,emr->rclBounds.top,emr->rclBounds.right,emr->rclBounds.bottom);
    cairo_fill(cr);
}


static void on_comment(){
    DEBUG_p("begin");
}


static ssize_t
handle_rec(U_ENHMETARECORD *emr, cairo_t *cr)
{
#if 0
    uint32_t crc;
#if __BYTE_ORDER == __BIG_ENDIAN
    {
        char *swapbuf = malloc(size);
        memcpy(swapbuf,record,size);
        U_emf_endian(swapbuf,size,1);
        crc = lu_crc32(swapbuf,size);
        free(swapbuf);
    }
#else
    crc = lu_crc32((const uint8_t*)emr,emr->nSize);
#endif
    //DEBUG_p("crc: %d",crc);
#endif

    switch(emr->iType){
    case U_EMR_HEADER:
        on_header((U_EMRHEADER*)emr,cr);
        break;
    case U_EMR_POLYBEZIER:
        on_polybezier();
        break;
    case U_EMR_POLYGON:
        on_polygon();
        break;
    case U_EMR_POLYLINE:
        on_polyline();
        break;
    case U_EMR_POLYBEZIERTO:
        on_polybezierto((U_EMRPOLYBEZIERTO*)emr,cr);
        break;

    case U_EMR_EOF:
        on_eof((U_EMREOF*)emr,cr);
        return 0;

    case U_EMR_SETMAPMODE:
        on_setmapmode((U_EMRSETMAPMODE*)emr,cr);
        break;
    case U_EMR_SETBKMODE:
        on_setbkmode((U_EMRSETBKMODE*)emr,cr);
        break;
    case U_EMR_SETPOLYFILLMODE:
        on_setpolyfillmode((U_EMRSETPOLYFILLMODE*)emr,cr);
        break;
    case U_EMR_SETROP2:
        on_setrop2();
        break;

    case U_EMR_SETTEXTALIGN:
        on_settextalign();
        break;

    case U_EMR_SETTEXTCOLOR:
        on_settextcolor((U_EMRSETTEXTCOLOR*)emr,cr);
        break;

    case U_EMR_MOVETOEX:
        on_movetoex((U_EMRMOVETOEX*)emr,cr);
        break;

    case U_EMR_MODIFYWORLDTRANSFORM:
        on_modifyworldtransform((U_EMRMODIFYWORLDTRANSFORM*)emr,cr);
        break;
    case U_EMR_SELECTOBJECT:
        on_selectobject();
        break;

    case U_EMR_CREATEBRUSHINDIRECT:
        on_createbrushindirect((U_EMRCREATEBRUSHINDIRECT*)emr,cr);
        break;
    case U_EMR_DELETEOBJECT:
        on_deleteobject();
        break;

    case U_EMR_LINETO:
        on_lineto((U_EMRLINETO*)emr,cr);
        break;

    case U_EMR_BEGINPATH:
        on_beginpath((U_EMRBEGINPATH*)emr,cr);
        break;
    case U_EMR_ENDPATH:
        on_endpath((U_EMRENDPATH*)emr,cr);
        break;
    case U_EMR_CLOSEFIGURE:
        on_closefigure();
        break;
    case U_EMR_FILLPATH:
        on_fillpath((U_EMRFILLPATH*)emr,cr);
        break;

    case U_EMR_COMMENT:
        on_comment();
        break;

    default:
        DEBUG_p("unhandled type: %d",emr->iType);
        break;
    }
    return emr->nSize;
}


int emf_cairo_scan(const char* p,size_t len,int *width,int *height)
{
    const char *end = p+len;
    while(p<end){
        U_ENHMETARECORD *emr = (U_ENHMETARECORD*)p;
        if(U_EMR_HEADER==emr->iType){
            U_EMRHEADER *h = (U_EMRHEADER*)emr;
#if 0
            *width = UM2PX*h->szlMicrometers.cx;
            *height = UM2PX*h->szlMicrometers.cy;
#else
            *width = ceil(SCALE*h->rclBounds.right);
            *height = ceil(SCALE*h->rclBounds.bottom);
#endif
            return 0;
        }
        p += emr->nSize;
    }
    return -1;
}


int emf_cairo_render(const char *p, size_t len, cairo_t *cr)
{
    const char *end = p+len;
    while(p<end){
        U_ENHMETARECORD *emr = (U_ENHMETARECORD*)p;
        uint32_t dummy1;
        uint32_t dummy2;
        ssize_t result;
        /*
        if(!recnum && (emr->iType != U_EMR_HEADER)){
        }
        */
        if(!U_emf_record_sizeok(p, end, &dummy1, &dummy2, 1)){
            DEBUG_p("bogus size");
            return -1;
        }
        if(!U_emf_record_safe(p)){
            DEBUG_p("corrupt record");
            return -1;
        }

        result = handle_rec(emr, cr);
        if(result < 0)
            return -1;
        if(!result)
            return 0;

        p += result;
        //++recnum;
    }
    return 0;
}
