/*
 * Copyright (C) 2019 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
//#include <math.h>

#include "emfcairo.h"
#include "uemf.h"
#include "wmfcairo.h"

#define ENABLE_WMF

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
#ifdef ENABLE_WMF
        T_WMF,
#endif
        T_EMF,
        T_EMFP
    }type;
    char *contents;
    size_t len;
};


static const char **
_uemf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/emf", "image/x-emf",
#ifdef ENABLE_WMF
        "image/wmf", "image/x-wmf",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_uemf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_WMF
    if(!strcasecmp(mime_type,"image/wmf") ||
       !strcasecmp(mime_type,"image/x-wmf"))
    {
        h->type = T_WMF;
    }else
#endif
    {
        h->type = T_EMF;
    }
    h->info = info;
    info->scalable_size = 1;
    info->threadsafe = 1;
    h->contents = NULL;
    return h;
}


static void
_uemf_free(abydos_plugin_handle_t *h)
{
    if(h->contents)
        free(h->contents);
    free(h);
}


static int
_uemf_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    h->contents = malloc(len);
    memcpy(h->contents,data,len);
    h->len = len;
    /* FIXME!! do endian swap */
    switch(h->type){
#ifdef ENABLE_WMF
    case T_WMF:
        if(wmf_cairo_scan(h->contents,h->len,&h->info->width,&h->info->height)<0){
            h->info->error = "Bogus file";
            return -1;
        }
        break;
#endif
    case T_EMF:
        if(emf_cairo_scan(h->contents,h->len,&h->info->width,&h->info->height)<0){
            h->info->error = "Bogus file";
            return -1;
        }
        break;
    case T_EMFP:
        /* FIXME!! */
        //if(emfp_cairo_scal(...
        break;
    }
    return 0;
}


static void
_uemf_render(abydos_plugin_handle_t *h, cairo_t *cr, int page, cairo_bool_t *visible_layers,int prefered_variant, double time, int frame)
{
    //cairo_save(cr);
    switch(h->type){
#ifdef ENABLE_WMF
    case T_WMF:
        wmf_cairo_render(h->contents,h->len,cr);
        break;
#endif
    case T_EMF:
        emf_cairo_render(h->contents,h->len,cr);
        break;
    case T_EMFP:
        /* FIXME!! */
        break;
    }
    //cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _uemf_get_mime_types,
        .new = _uemf_new,
        .free = _uemf_free,
        .create_from_data = _uemf_create_from_data,
        .render = _uemf_render
    };
    return &p;
}
