/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memset() */

#include "dyncanvas.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#define MINIMUM_SIZE 64


static size_t
greedy_size(size_t request_size)
{
    size_t actual_size;
    for(actual_size=MINIMUM_SIZE;actual_size<request_size;actual_size<<=1);
    DEBUG_p("%d >= %d",(int)actual_size,(int)request_size);
    return actual_size;
}


dynamic_canvas_t *
dynamic_canvas_init(dynamic_canvas_t *c, int w, int h)
{
    c->stride = greedy_size(w);
    c->size = c->stride*greedy_size(h);
    c->data = (uint32_t*)malloc(c->size*sizeof(uint32_t));
    memset(c->data, 0, c->size);
    c->width = w;
    c->height = h;
    return c;
}


void
dynamic_canvas_done(dynamic_canvas_t *c){
    free(c->data);
}


dynamic_canvas_t *
dynamic_canvas_new(int w, int h)
{
    return dynamic_canvas_init((dynamic_canvas_t*)malloc(sizeof(dynamic_canvas_t)),w,h);
}


void
dynamic_canvas_free(dynamic_canvas_t *c)
{
    dynamic_canvas_done(c);
    free(c);
}


void
dynamic_canvas_resize(dynamic_canvas_t *c, int w, int h)
{
    uint32_t *dst;
    int y;

    if(c->width < w || c->height < h){
        if(c->stride < w){
            /* allocate a whole new canvas */
            int dst_stride = greedy_size(w);
            uint32_t *data;
            const uint32_t *src = c->data;
            DEBUG_p("resizing from %dx%d to %dx%d",c->width,c->height,w,h);
            c->size = dst_stride*greedy_size(h);
            data = dst = (uint32_t*)malloc(c->size*sizeof(uint32_t));
            for(y=0;y<MIN(c->height,h);++y){
                memcpy(dst, src, MIN(c->width, w)*sizeof(uint32_t));
                if(c->width < w)
                    memset(dst+c->width, 0, (w-c->width)*sizeof(uint32_t));
                src += c->stride;
                dst += dst_stride;
            }
            for(y=c->height;y<h;++y){
                memset(dst, 0, w*sizeof(uint32_t));
                dst += dst_stride;
            }
            free(c->data);
            c->stride = dst_stride;
            c->data = data;
        }else{
            if(c->width < w){
                /* clear the right margin */
                dst = c->data + c->width;
                for(y=0;y<c->height;++y){
                    memset(dst+c->width, 0, w-c->width);
                    dst += c->stride;
                }
            }
            if(c->size < c->stride*h){
                c->size = c->stride*greedy_size(h);
                c->data = (uint32_t*)realloc(c->data,c->size*sizeof(uint32_t));
            }
            dst = c->data + c->stride*c->height;
            /* clear the bottom margin */
            for(y=c->height;y<h;++y){
                memset(dst, 0, w*sizeof(uint32_t));
                dst += w;
            }
        }
    }
    c->width = w;
    c->height = h;
}


cairo_surface_t *
dynamic_canvas_create_surface(dynamic_canvas_t *c)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, c->width, c->height);
    const uint32_t *src = c->data;
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int y;
    for(y=0;y<c->height;++y){
        memcpy(dst, src, c->width*sizeof(uint32_t));
        src += c->stride;
        dst += stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}
