/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "dcs.h"
#include "regisdec.h"

/* The maximum angle of an arc to convert to one curve. */
#define MAX_ANGLE (M_PI/2)


typedef struct{
    regis_decoder_t *regis_dec;
    char dcs_command;
    dcs_decoder_t *dcs_dec;
    cairo_t *cr;
    int got_path;
    int got_pen;
    double pen_x;
    double pen_y;
    int after_arc;
}read_context_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
    read_context_t *context;
};


static void _read_context_free(read_context_t *);


static const char **
_regis_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-regis",
        NULL
    };
    return mime_types;
}


static void
_regis_on_begin(void *userdata, int command)
{
    read_context_t *c = (read_context_t*)userdata;
    DEBUG_p("command: %c",command);
    c->dcs_command = command;
}


static void
_regis_on_end(void *userdata, int command)
{
    read_context_t *c = (read_context_t*)userdata;
    DEBUG_p("command: %c",command);
    c->dcs_command = '\0';
}


static void
_regis_on_param(void *userdata, const char *p)
{
    //read_context_t *c = (read_context_t*)userdata;
    DEBUG_p("param: '%s'",p);
}


static void
_regis_on_data(void *userdata, int ch)
{
    read_context_t *c = (read_context_t*)userdata;
    if('p'==c->dcs_command)
        regis_decoder_data(c->regis_dec, ch);
}


static void
_regis_on_position(void *userdata, double x, double y)
{
    read_context_t *c = (read_context_t*)userdata;
    //DEBUG_p("%f %f",x,y);
    if(c->after_arc){
        DEBUG_p("arc end: %fx%f",x,y);
        cairo_line_to(c->cr, x, y);
        c->after_arc = 0;
    }else{
        c->got_pen = 1;
        c->pen_x = x;
        c->pen_y = y;
    }
}


static void
_regis_on_vector(void *userdata, double x, double y)
{
    read_context_t *c = (read_context_t*)userdata;
    if(c->got_pen){
        //DEBUG_p("move to: %fx%f",c->pen_x,c->pen_y);
        cairo_move_to(c->cr, c->pen_x, c->pen_y);
        c->got_pen = 0;
    }
    //DEBUG_p("line to: %fx%f",x,y);
    cairo_line_to(c->cr, x, y);
    c->got_path = 1;
}


static void
_regis_on_curve(void *userdata, double x1, double y1, double da)
{
    read_context_t *c = (read_context_t*)userdata;
    double cx;
    double cy;
    double a;
    double r;

    //DEBUG_p("arc begin: %fx%f",x1,y1);
    if(!c->got_path)
        cairo_move_to(c->cr, x1, y1);
    //cairo_line_to(c->cr, x1, y1);
    if(c->got_pen){
        cx = c->pen_x;
        cy = c->pen_y;
        c->got_pen = 0;
    }else{
        return;
        cairo_get_current_point(c->cr, &cx, &cy);
    }
    {
        double dx = x1-cx;
        double dy = y1-cy;
        a = atan2(dy,dx);
        r = sqrt(dx*dx+dy*dy);
    }
    //DEBUG_p("start angle: %f",180*a/M_PI);
    da *= -M_PI/180;
    //DEBUG_p("arc angle: %f",180*da/M_PI);
    //DEBUG_p("radius: %f",r);
    if(r > 800)
        return;
#if 1
    //cairo_line_to(h->cr, cx + r*cos(a), cy + r*sin(a));
    if(da < 0){
        cairo_arc_negative(c->cr, cx, cy, r, a, a+da);
    }else{
        cairo_arc(c->cr, cx, cy, r, a, a+da);
    }
#else
    {
        int steps = ceil(fabs(da) / MAX_ANGLE);
        double r_c;
        int i;
        da /= steps;
        r_c = r*4.0/3*tan(da/4);
        for(i=0;i<steps;++i){
            double x2 = x1 - r_c*sin(a);
            double y2 = y1 + r_c*cos(a);
            a += da;
            x1 = cx + r*cos(a);
            y1 = cy + r*sin(a);
            cairo_curve_to(h->cr, x2, y2, x1 + r_c*sin(a), y1 - r_c*cos(a), x1, y1);
        }
    }
#endif
    c->got_path = 1;
    c->after_arc = 1;
}


static void
_regis_on_fill(void *userdata)
{
    read_context_t *c = (read_context_t*)userdata;
    //DEBUG_p("begin");
    cairo_fill(c->cr);
    c->got_path = 0;
}


static void
_regis_on_write(void *userdata, int color)
{
    read_context_t *c = (read_context_t*)userdata;
    switch(color){
    case 0:
        DEBUG_p("black");
        cairo_set_source_rgb(c->cr, 0, 0, 0);
        break;
    case 1:
        DEBUG_p("blue");
        cairo_set_source_rgb(c->cr, 0, 0, 1);
        break;
    case 2:
        DEBUG_p("red");
        cairo_set_source_rgb(c->cr, 1, 0, 0);
        break;
    case 3:
        DEBUG_p("green");
        cairo_set_source_rgb(c->cr, 0, 1, 0);
        break;
    case 4:
        DEBUG_p("magenta");
        cairo_set_source_rgb(c->cr, 1, 0, 1);
        break;
    case 5:
        DEBUG_p("cyan");
        cairo_set_source_rgb(c->cr, 0, 1, 1);
        break;
    case 6:
        DEBUG_p("yellow");
        cairo_set_source_rgb(c->cr, 1, 1, 0);
        break;
    case 7:
        cairo_set_source_rgb(c->cr, 0.5, 0.5, 0.5);
        break;
    case 8:
        cairo_set_source_rgb(c->cr, 0.25, 0.25, 0.25);
        break;
    case 9:
        cairo_set_source_rgb(c->cr, 0, 0, 1);
        break;
    case 10:
        cairo_set_source_rgb(c->cr, 1, 0, 0);
        break;
    case 11:
        cairo_set_source_rgb(c->cr, 0, 1, 0);
        break;
    case 12:
        cairo_set_source_rgb(c->cr, 0, 1, 1);
        break;
    case 13:
        cairo_set_source_rgb(c->cr, 1, 1, 0);
        break;
    case 14:
        cairo_set_source_rgb(c->cr, 1, 1, 0);
        break;
    case 15:
        DEBUG_p("white");
        cairo_set_source_rgb(c->cr, 1, 1, 1);
        break;
    }
}


static read_context_t *
_read_context_new(cairo_surface_t *surface)
{
    static const dcs_interface_t regis_dcs_vt = {
        .begin = _regis_on_begin,
        .end = _regis_on_end,
        .param = _regis_on_param,
        .data = _regis_on_data,
        .write = _regis_on_data
    };
    static const regis_interface_t regis_vt = {
        .position = _regis_on_position,
        .vector = _regis_on_vector,
        .curve = _regis_on_curve,
        .fill = _regis_on_fill,
        .write = _regis_on_write
    };
    read_context_t *c = malloc(sizeof(read_context_t));
    c->regis_dec = regis_decoder_new(&regis_vt, c);
    c->dcs_dec = dcs_decoder_new(&regis_dcs_vt, c);
    c->cr = cairo_create(surface);
    c->got_path = 0;
    c->got_pen = 0;
    c->after_arc = 0;
    return c;
}


static void
_read_context_free(read_context_t *c)
{
    regis_decoder_free(c->regis_dec);
    dcs_decoder_free(c->dcs_dec);
    cairo_destroy(c->cr);
}


static abydos_plugin_handle_t *
_regis_new(const char *mime_type, abydos_plugin_info_t *info)
{
    static const cairo_rectangle_t ext = {0, 0, 256*256, 256*256};
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    info->scalable_size = 1;
    h->surface = cairo_recording_surface_create(CAIRO_CONTENT_COLOR_ALPHA, &ext);
    h->context = NULL;
    return h;
}


static void
_regis_free(abydos_plugin_handle_t *h)
{
    if(h->context)
        _read_context_free(h->context);
    cairo_surface_destroy(h->surface);
    free(h);
}


static void
_regis_progressive_begin(abydos_plugin_handle_t *h,
                         abydos_plugin_info_func_t info_func,
                         abydos_plugin_progress_func_t progress_func,
                         void *userdata)
{
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
    h->context = _read_context_new(h->surface);
}


static int
_regis_progressive_feed(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    dcs_decoder_writes(h->context->dcs_dec, data, len);
    return 0;
}


static int
_regis_progressive_end(abydos_plugin_handle_t *h)
{
    cairo_rectangle_t size;
    _read_context_free(h->context);
    h->context = NULL;
    cairo_surface_mark_dirty(h->surface);

    cairo_recording_surface_ink_extents(h->surface, &size.x, &size.y, &size.width, &size.height);
#if 0
    {
        cairo_rectangle_t ext;
        double x;
        double y;
        cairo_bool_t ok = cairo_recording_surface_get_extents(h->surface, &ext);
        DEBUG_p("ok: %s, ext %fx%f %fx%f",ok?"yes":"no",ext.x,ext.y,ext.width,ext.height);
        cairo_surface_get_device_offset(h->surface,&x,&y);
        DEBUG_p("device offset: %fx%f",x,y);
        cairo_surface_get_device_scale(h->surface,&x,&y);
        DEBUG_p("device scale: %fx%f",x,y);
    }
#endif

    h->info->width = size.width;
    h->info->height = size.height;
    /* This can't be called earlier since the size can grow at any time. */
    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t r;
        r.x = 0;
        r.y = 0;
        r.width = h->info->width;
        r.height = h->info->height;
        h->progress_func(h->userdata, &r);
    }
    return 0;
}


static void
_regis_render(abydos_plugin_handle_t *h,
              cairo_t *cr,
              int page,
              cairo_bool_t *layer_visibility,
              int variant,
              double time,
              int frame)
{
    cairo_set_source_surface(cr, h->surface, 0, 0);
    cairo_paint(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _regis_get_mime_types,
        .new = _regis_new,
        .free = _regis_free,
        .progressive_begin = _regis_progressive_begin,
        .progressive_feed = _regis_progressive_feed,
        .progressive_end = _regis_progressive_end,
        .render = _regis_render
    };
    return &p;
}
