/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "dcs.h"
#include "dyncanvas.h"
#include "sixeldec.h"

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    sixel_decoder_t *sixel_dec;
    char dcs_command;
    dcs_decoder_t *dcs_dec;
    dynamic_canvas_t canvas;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};


static const char **
_sixel_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-sixel",
        NULL
    };
    return mime_types;
}


static void
_sixel_on_begin(void *userdata, int command)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)userdata;
    DEBUG_p("command: %c",command);
    h->dcs_command = command;
}


static void
_sixel_on_end(void *userdata, int command)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)userdata;
    DEBUG_p("command: %c",command);
    h->dcs_command = '\0';
}


static void
_sixel_on_param(void *userdata, const char *p)
{
    DEBUG_p("param: '%s'",p);
}


static void
_sixel_on_data(void *userdata, int ch)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)userdata;
    if('q'==h->dcs_command)
        sixel_decoder_data(h->sixel_dec, ch);
}


static void
_on_sixel(void *userdata, int x, int y, sixel_color_t color, int bits, int repeat)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)userdata;
    sixel_color_t *dst;
    y *= 6;
    if(x+repeat > h->canvas.width || y+6 > h->canvas.height)
        dynamic_canvas_resize(&h->canvas, MAX(h->canvas.width,x+repeat), MAX(h->canvas.height,y+6));
    dst = h->canvas.data + x + y*h->canvas.stride;
    for(y=0;y<6;++y){
        if(bits&1)
            for(x=0;x<repeat;++x)
                dst[x] = color;
        bits >>= 1;
        dst += h->canvas.stride;
    }
}


static abydos_plugin_handle_t *
_sixel_new(const char *mime_type, abydos_plugin_info_t *info)
{
    static const dcs_interface_t sixel_vt = {
        .begin = _sixel_on_begin,
        .end = _sixel_on_end,
        .param = _sixel_on_param,
        .data = _sixel_on_data,
        .write = _sixel_on_data
    };
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->sixel_dec = sixel_decoder_new(_on_sixel, h);
    h->dcs_command = 0;
    h->dcs_dec = dcs_decoder_new(&sixel_vt, h);
    dynamic_canvas_init(&h->canvas,0,0);
    return h;
}


static void
_sixel_free(abydos_plugin_handle_t *h)
{
    sixel_decoder_free(h->sixel_dec);
    dcs_decoder_free(h->dcs_dec);
    dynamic_canvas_done(&h->canvas);
    free(h);
}


static void
_sixel_progressive_begin(abydos_plugin_handle_t *h,
                         abydos_plugin_info_func_t info_func,
                         abydos_plugin_progress_func_t progress_func,
                         void *userdata)
{
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
_sixel_progressive_feed(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    dcs_decoder_writes(h->dcs_dec, data, len);
    return 0;
}


static int
_sixel_progressive_end(abydos_plugin_handle_t *h)
{
    h->info->width = h->canvas.width;
    h->info->height = h->canvas.height;
    /* This can't be called earlier since the size can grow at any time. */
    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t r;
        r.x = 0;
        r.y = 0;
        r.width = h->info->width;
        r.height = h->info->height;
        h->progress_func(h->userdata,&r);
    }
    return 0;
}


static cairo_surface_t *
_sixel_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return dynamic_canvas_create_surface(&h->canvas);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _sixel_get_mime_types,
        .new = _sixel_new,
        .free = _sixel_free,
        .progressive_begin = _sixel_progressive_begin,
        .progressive_feed = _sixel_progressive_feed,
        .progressive_end = _sixel_progressive_end,
        .get_image_surface = _sixel_get_image_surface
    };
    return &p;
}
