/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <nil-cairo.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <webp/demux.h>


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_webp_get_mime_types()
{
    static const char *mime_types[] = {
        "image/webp",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_webp_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_webp_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_webp_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    WebPData data;
    WebPAnimDecoder *dec;
    WebPAnimInfo info;
    int last_timestamp = 0;
    int i;

    data.bytes = (const unsigned char*)bytes;
    data.size = len;

    dec = WebPAnimDecoderNew(&data, NULL);
    if(!dec)
        goto error;

    if(!WebPAnimDecoderGetInfo(dec, &info))
        goto error;
    h->info->width = info.canvas_width;
    h->info->height = info.canvas_height;
    h->info->frame_count = info.frame_count;

    h->frame = (frame_t*)malloc(info.frame_count * sizeof(frame_t));

    for(i=0;i<info.frame_count;++i){
        uint8_t* rgba;
        int timestamp;

        if (!WebPAnimDecoderGetNext(dec, &rgba, &timestamp))
            goto error;

        h->frame[i].surface = nil_cairo_surface_create_from_direct(rgba,info.canvas_width,info.canvas_height,-1,4,NILE_FORMAT_r8g8b8a8,0);
        h->frame[i].duration = (double)(timestamp - last_timestamp)/1000;
        last_timestamp = timestamp;
    }
    WebPAnimDecoderDelete(dec);
    return 0;

error:
    if(dec)
        WebPAnimDecoderDelete(dec);
    h->info->error = "failed to decode the image";
    return -1;
}


static cairo_surface_t *
_webp_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_webp_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _webp_get_mime_types,
        .new = _webp_new,
        .free = _webp_free,
        .create_from_data = _webp_create_from_data,
        .get_image_surface = _webp_get_image_surface,
        .get_duration = _webp_get_duration
    };
    return &p;
}
