/*
 * Copyright (C) 2000  Caolán McNamara (and others of the libwmf project)
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*
 * Originally based on ipa/svg.c from libwmf.
 */
#include <cairo.h>
#include <debug.h>
#include <libwmf/api.h>
#include <libwmf/defs.h>
#include <libwmf/ipa.h>
#include <libwmf/macro.h>
#include <math.h>
#include <string.h>                     /* memcpy() */

#define USE_NOOP

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif


typedef enum{
    ARC_OPEN,
    ARC_PIE,
    ARC_CHORD
}arc_t;


static void
_cairo_draw_ellipse(cairo_t *cr,wmfDrawArc_t *da)
{
    double rx = (da->BR.x-da->TL.x)/2;
    double ry = (da->BR.y-da->TL.y)/2;
    cairo_save (cr);
    cairo_translate (cr, da->TL.x+rx, da->TL.y+ry);
    cairo_scale (cr, rx, ry);
    cairo_arc (cr, 0, 0, 1, 0, 2*M_PI);
    cairo_restore (cr);
}


static void
_cairo_draw_poly(cairo_t *cr, wmfPolyLine_t *pl)
{
    int i;
    cairo_move_to(cr,pl->pt[0].x,pl->pt[0].y);
    for(i=1;i<pl->count;++i)
        cairo_line_to(cr,pl->pt[i].x,pl->pt[i].y);
}


static void
_cairo_set_rgb(cairo_t *cr, wmfRGB *rgb)
{
    //DEBUG_p("{%d,%d,%d}",rgb->r,rgb->g,rgb->b);
    cairo_set_source_rgb(cr,(double)rgb->r/255,(double)rgb->g/255,(double)rgb->b/255);
}


static void
_cairo_set_stroke_style(cairo_t *cr, wmfDC *dc)
{
    wmfPen* pen = WMF_DC_PEN(dc);
    float pen_width = (float)(WMF_PEN_WIDTH(pen)+WMF_PEN_HEIGHT(pen))/2;
    int pen_style = WMF_PEN_STYLE(pen);

    if(PS_NULL==pen_style){
        DEBUG_p("penstyle null");
    }

    cairo_set_line_width(cr,pen_width);

    switch (WMF_PEN_ENDCAP(pen)){
    case PS_ENDCAP_SQUARE:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_SQUARE);
        break;
    case PS_ENDCAP_ROUND:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_ROUND);
        break;
    case PS_ENDCAP_FLAT:
    default:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_BUTT);
        break;
    }

    switch (WMF_PEN_JOIN(pen)){
    case PS_JOIN_BEVEL:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_BEVEL);
        break;
    case PS_JOIN_ROUND:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_ROUND);
        break;
    case PS_JOIN_MITER:
    default:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_MITER);
        break;
    }

#if 0
    switch (pen_style){
    case PS_DASH: /* DASH_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width*10,pen_width*10);
        break;

    case PS_ALTERNATE:
    case PS_DOT: /* DOTTED_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width,pen_width*2);
        break;

    case PS_DASHDOT: /* DASH_DOT_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_DASHDOTDOT: /* DASH_2_DOTS_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_INSIDEFRAME: /* There is nothing to do in this case... */
    case PS_SOLID:
    default:
        wmf_stream_printf (API,out,"stroke-dasharray:none; ");
        break;
    }
#endif

    _cairo_set_rgb(cr,WMF_PEN_COLOR(pen));
}
#if 0
static void
cairo_style_stroke(wmfAPI *api,wmfDC *dc)
{
    cairo_t *cr = (cairo_t*)api->device_data;

    wmfPen* pen;
    wmfRGB* pen_color;
    float pen_width;

    unsigned int pen_style;
    unsigned int pen_endcap;
    unsigned int pen_join;

    pen = WMF_DC_PEN(dc);
    pen_color = WMF_PEN_COLOR(pen);

#if 0
    pen_width = ( svg_width  (api,(float) WMF_PEN_WIDTH  (pen))
                + svg_height (api,(float) WMF_PEN_HEIGHT (pen)) ) / 2;
#endif

    pen_style  = (unsigned int) WMF_PEN_STYLE(pen);
    pen_endcap = (unsigned int) WMF_PEN_ENDCAP(pen);
    pen_join   = (unsigned int) WMF_PEN_JOIN(pen);

#if 0
    if (pen_style == PS_NULL){
        wmf_stream_printf (api,out,"stroke:none");
        return;
    }

    wmf_stream_printf (API,out,"stroke-width:%f; ",MAX (0,pen_width));

    switch (pen_endcap){
    case PS_ENDCAP_SQUARE:
        wmf_stream_printf (API,out,"stroke-linecap:square; ");
        break;
    case PS_ENDCAP_ROUND:
        wmf_stream_printf (API,out,"stroke-linecap:round; ");
        break;
    case PS_ENDCAP_FLAT:
    default:
        wmf_stream_printf (API,out,"stroke-linecap:butt; ");
        break;
    }

    switch (pen_join){
    case PS_JOIN_BEVEL:
        wmf_stream_printf (API,out,"stroke-linejoin:bevel; ");
        break;
    case PS_JOIN_ROUND:
        wmf_stream_printf (API,out,"stroke-linejoin:round; ");
        break;
    case PS_JOIN_MITER:
    default:
        wmf_stream_printf (API,out,"stroke-linejoin:miter; ");
        break;
    }

    switch (pen_style){
    case PS_DASH: /* DASH_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width*10,pen_width*10);
        break;

    case PS_ALTERNATE:
    case PS_DOT: /* DOTTED_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width,pen_width*2);
        break;

    case PS_DASHDOT: /* DASH_DOT_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_DASHDOTDOT: /* DASH_2_DOTS_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_INSIDEFRAME: /* There is nothing to do in this case... */
    case PS_SOLID:
    default:
        wmf_stream_printf (API,out,"stroke-dasharray:none; ");
        break;
    }
#endif
    _cairo_set_rgb(cr,pen_color);
    cairo_stroke(cr);
}
#endif


static void
_cairo_set_fill_style(cairo_t *cr, wmfDC *dc)
{
    wmfBrush *brush = WMF_DC_BRUSH(dc);
    switch(WMF_DC_POLYFILL(dc)){
    case WINDING:
        cairo_set_fill_rule(cr,CAIRO_FILL_RULE_WINDING);
        break;
    case ALTERNATE:
    default:
        cairo_set_fill_rule(cr,CAIRO_FILL_RULE_EVEN_ODD);
        break;
    }
    _cairo_set_rgb(cr,WMF_BRUSH_COLOR(brush));
}
#if 0
static void
_cairo_style_fill(wmfAPI *api,wmfDC *dc)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    wmfRGB *brush_color;
    wmfBMP *brush_bmp;
    wmfBrush *brush;

    unsigned int fill_opaque;
    unsigned int fill_polyfill;

    unsigned int brush_style;

    DEBUG_p("begin");

    fill_opaque   = (unsigned int) WMF_DC_OPAQUE (dc);
    fill_polyfill = (unsigned int) WMF_DC_POLYFILL (dc);

    brush = WMF_DC_BRUSH (dc);

    brush_style = (unsigned int) WMF_BRUSH_STYLE (brush);

    brush_color = WMF_BRUSH_COLOR (brush);

    brush_bmp = WMF_BRUSH_BITMAP (brush);

    if (brush_style == BS_NULL){
        //wmf_stream_printf(api,out,"fill:none");
        return;
    }

#if 0
    if (fill_opaque){
        wmf_stream_printf(api,out,"fill-opacity:1.0; ");
    }else{
        wmf_stream_printf(api,out,"fill-opacity:0.5; "); /* semi-transparent... ?? */
    }
#endif

#if 0
    /* Is this correct ?? */
    switch (fill_polyfill){
    case WINDING:
        wmf_stream_printf (API,out,"fill-rule:nonzero; ");
        break;
    case ALTERNATE:
    default:
        wmf_stream_printf (API,out,"fill-rule:evenodd; ");
        break;
    }

    switch (brush_style){
#ifdef WRITE_EPS_NOT_SVG
    /* float side; */

	case BS_HATCHED:
		fputs ("clip ",out);

		if (dc->bgmode != TRANSPARENT)
		{	rgb = dc->bgcolor;

			red   = (float) ((int) rgb.r) / 255;
			green = (float) ((int) rgb.g) / 255;
			blue  = (float) ((int) rgb.b) / 255;

			fprintf (out,"%f %f %f setrgbcolor ",red,green,blue);

			fputs ("fill ",out);
		}

		fputs ("\n",out);
		fputs ("1 setlinewidth ",out);
		fputs ("[] 0 setdash ",out);

		rgb = brush->lbColor;

		red   = (float) ((int) rgb.r) / 255;
		green = (float) ((int) rgb.g) / 255;
		blue  = (float) ((int) rgb.b) / 255;

		fprintf (out,"%f %f %f setrgbcolor\n",red,green,blue);

		switch (brush->lbHatch)
		{
		case HS_HORIZONTAL:
			fprintf (out,"%f 5 %f { newpath dup %f exch moveto %f exch lineto stroke } for\n",
			         bbox->TL.y,bbox->BR.y,bbox->TL.x,bbox->BR.x);
		break;

		case HS_VERTICAL:
			fprintf (out,"%f 5 %f { newpath dup %f moveto %f lineto stroke } for\n",
			         bbox->TL.x,bbox->BR.x,bbox->TL.y,bbox->BR.y);
		break;

		case HS_FDIAGONAL:
			fputs ("gsave % HS_FDIAGONAL\n",out);
			fprintf (out,"%f %f translate -45 rotate ",
			         bbox->TL.x-(bbox->BR.y-bbox->TL.y)/2,(bbox->TL.y+bbox->BR.y)/2);
			side = ((bbox->BR.x-bbox->TL.x) + (bbox->BR.y-bbox->TL.y)) / 1.41421356237309504880;
			fprintf (out,"0 5 %f { newpath dup 0 moveto %f lineto stroke } for ",
			         side,side);
			fputs ("grestore\n",out);
		break;

		case HS_BDIAGONAL:
			fputs ("gsave % HS_BDIAGONAL\n",out);
			fprintf (out,"%f %f translate -45 rotate ",
			         bbox->TL.x-(bbox->BR.y-bbox->TL.y)/2,(bbox->TL.y+bbox->BR.y)/2);
			side = ((bbox->BR.x-bbox->TL.x) + (bbox->BR.y-bbox->TL.y)) / 1.41421356237309504880;
			fprintf (out,"0 5 %f { newpath dup 0 exch moveto %f exch lineto stroke } for ",
			         side,side);
			fputs ("grestore\n",out);
		break;

		case HS_CROSS:
			fprintf (out,"%f 5 %f { newpath dup %f exch moveto %f exch lineto stroke } for\n",
			         bbox->TL.y,bbox->BR.y,bbox->TL.x,bbox->BR.x);
			fprintf (out,"%f 5 %f { newpath dup %f moveto %f lineto stroke } for\n",
			         bbox->TL.x,bbox->BR.x,bbox->TL.y,bbox->BR.y);
		break;

		case HS_DIAGCROSS:
			fputs ("gsave % HS_DIAGCROSS\n",out);
			fprintf (out,"%f %f translate -45 rotate ",
			         bbox->TL.x-(bbox->BR.y-bbox->TL.y)/2,(bbox->TL.y+bbox->BR.y)/2);
			side = ((bbox->BR.x-bbox->TL.x) + (bbox->BR.y-bbox->TL.y)) / 1.41421356237309504880;
			fprintf (out,"0 5 %f { newpath dup 0 moveto %f lineto stroke } for ",
			         side,side);
			fprintf (out,"0 5 %f { newpath dup 0 exch moveto %f exch lineto stroke } for ",
			         side,side);
			fputs ("grestore\n",out);
		break;

		default:
			if (API->flags & WMF_OPT_IGNORE_NONFATAL)
			{	DEBUG_p("Unsupported brush/hatch style!");
			}
			else
			{	WMF_ERROR (API,"Unsupported brush/hatch style!");
				API->err = wmf_E_Glitch;
			}
		break;
		}
	break;
#endif /* WRITE_EPS_NOT_SVG */

    case BS_DIBPATTERN:
        if (brush_bmp->data == 0){
            if (API->flags & WMF_OPT_IGNORE_NONFATAL){
                DEBUG_p("Attempt to fill with non-existent pattern!");
            }else{
                WMF_ERROR (API,"Attempt to fill with non-existent pattern!");
                API->err = wmf_E_Glitch;
                break;
            }
        }
        /* fallthrough - TODO: implement bitmap fill */
    default:
        if (API->flags & WMF_OPT_IGNORE_NONFATAL)
        {
            DEBUG_p(API,"Unsupported brush style!");
            /* no break here */
        }else{
            WMF_ERROR (API,"Unsupported brush style!");
			API->err = wmf_E_Glitch;
			break;
		}
		/* fallthrough - no break here */
    case BS_SOLID:
        break;
    }
#endif

    _cairo_set_rgb(cr,brush_color);
    cairo_fill(cr);
}
#endif


static void
_cairo_fill_and_stroke(cairo_t *cr, wmfDC *dc)
{
    _cairo_set_fill_style(cr,dc);
    cairo_fill_preserve(cr);
    _cairo_set_stroke_style(cr,dc);
    cairo_stroke(cr);
}


static void
_cairo_draw_arc (wmfAPI *api,wmfDrawArc_t *da,arc_t finish)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");

    if(TO_FILL (da) || TO_DRAW(da)){
        if ((da->start.x == da->end.x) || (da->start.y == da->end.y)){
            _cairo_draw_ellipse(cr,da);
        }else{
            double cx = (da->TL.x+da->BR.x)/2;
            double cy = (da->TL.y+da->BR.y)/2;
            double rx = (da->BR.x-da->TL.x)/2;
            double ry = (da->BR.y-da->TL.y)/2;
            double start = atan2(da->start.y,da->start.x);
            double end = atan2(da->end.y,da->end.x);
            cairo_save(cr);
            cairo_translate(cr, cx, cy);
            cairo_scale(cr, rx, ry);
            cairo_arc(cr, 0, 0, 1, start, end);
            cairo_restore(cr);
            if(ARC_PIE==finish){
                cairo_line_to(cr,cx,cy);
                cairo_close_path(cr);
            }
        }
        if(ARC_OPEN==finish){
            _cairo_set_stroke_style(cr,da->dc);
            cairo_stroke(cr);
        }else{
            _cairo_fill_and_stroke(cr,da->dc);
        }
    }
}


#ifdef USE_NOOP
static void
_cairo_noop(){}
#endif


#ifndef USE_NOOP
/* This is called by wmf_play() the *first* time the meta file is played
 */
static void
_wmf_cairo_device_open(wmfAPI *api)
{
    DEBUG_p("begin");
}

/* This is called by wmf_api_destroy()
 */
static void
_wmf_cairo_device_close (wmfAPI* api)
{
    DEBUG_p("begin");
}

/* This is called from the beginning of each play for initial page setup
 */
static void
_wmf_cairo_device_begin (wmfAPI* api)
{
    DEBUG_p("begin");
}

/* This is called from the end of each play for page termination
 */
static void
_wmf_cairo_device_end (wmfAPI* api)
{
    DEBUG_p("begin");
}
#endif


#if 0
static svgPoint
_cairo_translate (wmfAPI* API,wmfD_Coord d_pt)
{
    wmf_svg_t* ddata = WMF_SVG_GetData (API);

	svgPoint g_pt;

	double x;
	double y;

	x = ((double) d_pt.x - (double) ddata->bbox.TL.x);
	x /= ((double) ddata->bbox.BR.x - (double) ddata->bbox.TL.x);
	x *= (double) ddata->width;

	y = ((double) d_pt.y - (double) ddata->bbox.TL.y);
	y /= ((double) ddata->bbox.BR.y - (double) ddata->bbox.TL.y);
	y *= (double) ddata->height;

	g_pt.x = (float) x;
	g_pt.y = (float) y;

	return (g_pt);
}
#endif

#if 0
static float
_cairo_width (wmfAPI* API,float wmf_width)
{
    wmf_svg_t* ddata = WMF_SVG_GetData (API);

    double width;

    width = (double) wmf_width * (double) ddata->width;
    width /= ((double) ddata->bbox.BR.x - (double) ddata->bbox.TL.x);

    return ((float) width);
}
#endif

#if 0
static float
_cairo_height (wmfAPI* API,float wmf_height)
{
    wmf_svg_t* ddata = WMF_SVG_GetData (API);

    double height;

    height = (double) wmf_height * (double) ddata->height;
    height /= ((double) ddata->bbox.BR.y - (double) ddata->bbox.TL.y);

    return ((float) height);
}
#endif

static void
_wmf_cairo_flood_interior (wmfAPI* api,wmfFlood_t* flood)
{
    DEBUG_p("begin");
    if(api->flags & WMF_OPT_IGNORE_NONFATAL){
        DEBUG_p("flood_interior unsupported.");
    }else{
        WMF_ERROR(api,"flood_interior unsupported.");
        api->err = wmf_E_Glitch;
    }
}

static void
_wmf_cairo_flood_exterior (wmfAPI* api,wmfFlood_t* flood)
{
    DEBUG_p("begin");
    if(api->flags & WMF_OPT_IGNORE_NONFATAL){
        DEBUG_p("flood_exterior unsupported.");
    }else{
        WMF_ERROR(api,"flood_exterior unsupported.");
        api->err = wmf_E_Glitch;
    }
}

static void
_wmf_cairo_draw_pixel (wmfAPI* api,wmfDrawPixel_t* draw_pixel)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");
    cairo_rectangle(cr,draw_pixel->pt.x,draw_pixel->pt.y,draw_pixel->pixel_width,draw_pixel->pixel_height);
    _cairo_set_rgb(cr,&draw_pixel->color);
    cairo_fill(cr);
}

static void
_wmf_cairo_draw_pie (wmfAPI* API,wmfDrawArc_t* draw_arc)
{
    DEBUG_p("begin");
    _cairo_draw_arc(API,draw_arc,ARC_PIE);
}

static void
_wmf_cairo_draw_chord (wmfAPI* API,wmfDrawArc_t* draw_arc)
{
    DEBUG_p("begin");
    _cairo_draw_arc(API,draw_arc,ARC_CHORD);
}

static void
_wmf_cairo_draw_arc (wmfAPI* API,wmfDrawArc_t* draw_arc)
{
    DEBUG_p("begin");
    _cairo_draw_arc(API,draw_arc,ARC_OPEN);
}

static void
_wmf_cairo_draw_ellipse(wmfAPI *api,wmfDrawArc_t* draw_arc)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");
    if (TO_FILL (draw_arc) || TO_DRAW (draw_arc)){
        _cairo_draw_ellipse(cr,draw_arc);
        _cairo_fill_and_stroke(cr,draw_arc->dc);
    }
}

static void
_wmf_cairo_draw_line (wmfAPI *api,wmfDrawLine_t *draw_line)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");
    if (TO_DRAW (draw_line)){
        cairo_move_to(cr,draw_line->from.x,draw_line->from.y);
        cairo_line_to(cr,draw_line->to.x,draw_line->to.y);
        _cairo_set_stroke_style(cr,draw_line->dc);
        cairo_stroke(cr);
    }
}

static void
_wmf_cairo_poly_line (wmfAPI *api,wmfPolyLine_t *poly_line)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");
    if (poly_line->count <= 1)
        return;

    if (TO_DRAW (poly_line)){
        _cairo_draw_poly(cr, poly_line);
        _cairo_set_stroke_style(cr,poly_line->dc);
        cairo_stroke(cr);
    }
}

static void
_wmf_cairo_draw_polygon (wmfAPI *api,wmfPolyLine_t *poly_line)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");
    if(poly_line->count <= 2)
        return;

    if (TO_FILL (poly_line) || TO_DRAW (poly_line)){
        _cairo_draw_poly(cr,poly_line);
        _cairo_fill_and_stroke(cr,poly_line->dc);
    }
}

static void
_wmf_cairo_draw_polypolygon(wmfAPI *api, wmfPolyPoly_t *polypolygon)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");

    if (TO_FILL(polypolygon) || TO_DRAW(polypolygon)){
        int valid = 0;
        int i;
        for(i=0;i<polypolygon->npoly;++i){
            wmfPolyLine_t pl;
            pl.count = polypolygon->count[i];
            if(pl.count>2){
                pl.pt = polypolygon->pt[i];
                _cairo_draw_poly(cr,&pl);
                valid = 1;
            }
        }
        if(valid)
            _cairo_fill_and_stroke(cr,polypolygon->dc);
    }
}

static void
_wmf_cairo_draw_rectangle (wmfAPI *api,wmfDrawRectangle_t *draw_rect)
{
    cairo_t *cr = (cairo_t*)api->device_data;
    DEBUG_p("begin");

    if (TO_FILL (draw_rect) || TO_DRAW (draw_rect)){
        cairo_rectangle(cr,draw_rect->TL.x,draw_rect->TL.y,draw_rect->BR.x-draw_rect->TL.x,draw_rect->BR.y-draw_rect->TL.y);
        _cairo_fill_and_stroke(cr,draw_rect->dc);
    }
}


static void
_wmf_cairo_rop_draw (wmfAPI* API,wmfROP_Draw_t* rop_draw)
{
    DEBUG_p("begin");
}


/* TODO ?? Care about bmp_draw->type
 */
static void
_wmf_cairo_bmp_draw (wmfAPI *api,wmfBMP_Draw_t* bmp_draw)
{
    //cairo_t *cr = (cairo_t*)api->device_data;
    //float width;
    //float height;
    //char* name = 0;
    DEBUG_p("begin");
#if 0
    if(!bmp_draw->bmp.data)
        return;

	if ((ddata->flags & WMF_SVG_INLINE_IMAGES) == 0)
	{	if (ddata->image.name == 0) return;

		name = ddata->image.name (ddata->image.context);

		if (name == 0) return;

		wmf_ipa_bmp_png (API,bmp_draw,name);

		if (ERR (API))
		{	DEBUG_p("bailing...");
			return;
		}

		/* Okay, if we've got this far then "name" is the filename of an png (cropped) image */
	}

	pt = svg_translate (API,bmp_draw->pt);

	width  = (float) (bmp_draw->pixel_width  * (double) bmp_draw->crop.w);
	height = (float) (bmp_draw->pixel_height * (double) bmp_draw->crop.h);

#if 0
	width  = svg_width  (API,width);
	height = svg_height (API,height);

	width  = ABS (width);
	height = ABS (height);
#endif

	wmf_stream_printf (API,out,"<image ");

	wmf_stream_printf (API,out,"x=\"%f\" ",pt.x);
	wmf_stream_printf (API,out,"y=\"%f\" ",pt.y);

	wmf_stream_printf (API,out, "width=\"%f\" ",width);
	wmf_stream_printf (API,out,"height=\"%f\"\n",height);

	width  /= (float) bmp_draw->crop.w;
	height /= (float) bmp_draw->crop.h;

	wmf_stream_printf (API,out,"\ttransform=\"matrix(");
	wmf_stream_printf (API,out,"%f 0 0 %f %f %f)\"\n",width,height,pt.x,pt.y);

	if ((ddata->flags & WMF_SVG_INLINE_IMAGES) == 0)
	{	wmf_stream_printf (API,out,"\tsodipodi:absref=\"%s\"\n",name);
		wmf_stream_printf (API,out,"\txlink:href=\"%s\"/>\n",name);
	}
	else
	{	wmf_stream_printf (API,out,"\txlink:href=\"data:image/png;base64,");

		wmf_ipa_bmp_b64 (API,bmp_draw,out);

		if (ERR (API))
		{
            DEBUG_p("bailing...");
			return;
		}

		wmf_stream_printf (API,out,"\"/>\n");
	}
#endif
}

static void
_wmf_cairo_bmp_read(wmfAPI *api,wmfBMP_Read_t* bmp_read)
{
    DEBUG_p("begin");
    wmf_ipa_bmp_read(api,bmp_read);
}

static void
_wmf_cairo_bmp_free (wmfAPI* api,wmfBMP* bmp)
{
    DEBUG_p("begin");
    wmf_ipa_bmp_free(api,bmp);
}


static void
_wmf_cairo_region_frame(wmfAPI *api,wmfPolyRectangle_t* poly_rect)
{
    DEBUG_p("begin");
}

static void
_wmf_cairo_region_paint (wmfAPI* api,wmfPolyRectangle_t* poly_rect)
{
    DEBUG_p("begin");
}

static void
_wmf_cairo_region_clip (wmfAPI* API,wmfPolyRectangle_t* pr)
{
    int i;
    DEBUG_p("rectangles: %d",pr->count);
    for(i=0;i<pr->count;++i){
        DEBUG_p("clip: %fx%f -> %fx%f",pr->TL[i].x,pr->TL[i].y,pr->BR[i].x,pr->BR[i].y);
    }
}


static void
_wmf_cairo_draw_text (wmfAPI* api,wmfDrawText_t* draw_text)
{
#if 0
    cairo_t *cr = (cairo_t *)api->device_data;

    svgFont font;

    svgPoint pt;

    float font_height;

    float sin_theta;
    float cos_theta;

    double theta;

    wmfStream* out = ddata->out;

    DEBUG_p("begin");

    if (out == 0)
        return;

    pt = svg_translate (API,draw_text->pt);

    font_height = svg_height (API,(float)  draw_text->font_height);

    theta = - WMF_TEXT_ANGLE (draw_text->dc->font);

    sin_theta = (float) sin (theta);
    cos_theta = (float) cos (theta);

    /* TODO: Scaling? */
    /* TODO: Background color? */
    /* TODO: string-length?? */

    wmf_stream_printf (API,out,"<text ");

/*  wmf_stream_printf (API,out,"x=\"%f\" ",pt.x);
    wmf_stream_printf (API,out,"y=\"%f\" ",pt.y);
 */ wmf_stream_printf (API,out,"x=\"0\" ");
    wmf_stream_printf (API,out,"y=\"0\" ");

    wmf_stream_printf (API,out,"style=\"");

    font = svg_font (WMF_FONT_PSNAME (WMF_DC_FONT (draw_text->dc)));

    wmf_stream_printf (API,out,"font-family:%s; ",font.family);
    wmf_stream_printf (API,out,"font-style:%s; ", font.style);
    wmf_stream_printf (API,out,"font-weight:%s; ",font.weight);

    wmf_stream_printf (API,out,"font-size:%f; ",font_height);

    wmf_stream_printf (API,out,"fill:%s",svg_color_closest (WMF_DC_TEXTCOLOR (draw_text->dc)));

/*  wmf_stream_printf (API,out,"\"><tspan rotate=\"%f\"\n\t>",(float) (theta * 180 / PI));
 */ wmf_stream_printf (API,out,"\" transform=\"matrix(");
    wmf_stream_printf (API,out,"%f %f %f %f ",cos_theta,sin_theta,-sin_theta,cos_theta);
    wmf_stream_printf (API,out,"%f %f)\"\n\t>",pt.x,pt.y);

    wmf_stream_printf (API,out,"%s",draw_text->str);

/*  wmf_stream_printf (API,out,"</tspan></text>\n");
 */ wmf_stream_printf (API,out,"</text>\n");
#endif
}

#ifndef USE_NOOP
static void wmf_cairo_udata_init (wmfAPI* API,wmfUserData_t* user_data)
{
    DEBUG_p("begin");
}

static void wmf_cairo_udata_copy (wmfAPI* API,wmfUserData_t* user_data)
{
    DEBUG_p("begin");
}

static void wmf_cairo_udata_set (wmfAPI* API,wmfUserData_t* user_data)
{
    DEBUG_p("begin");
}

static void wmf_cairo_udata_free (wmfAPI* API,wmfUserData_t* user_data)
{
    DEBUG_p("begin");
}
#endif /* ! USE_NOOP */


#if 0
static svgFont
_cairo_font (char* ps_name)
{
#if 0
    svgFont* font = svg_known_fonts;
    while (font->ps_name){
        if(!strcmp (ps_name,font->ps_name) == 0)
            break;
        ++font++;
    }
    return (*font);
#endif
}
#endif


void
wmf_cairo_function(wmfAPI *api)
{
    static const wmfFunctionReference vt = {
#ifdef USE_NOOP
        .device_open      = _cairo_noop,
        .device_close     = _cairo_noop,
        .device_begin     = _cairo_noop,
        .device_end       = _cairo_noop,
#else
        .device_open      = _wmf_cairo_device_open,
        .device_close     = _wmf_cairo_device_close,
        .device_begin     = _wmf_cairo_device_begin,
        .device_end       = _wmf_cairo_device_end,
#endif
        .flood_interior   = _wmf_cairo_flood_interior,
        .flood_exterior   = _wmf_cairo_flood_exterior,
        .draw_pixel       = _wmf_cairo_draw_pixel,
        .draw_pie         = _wmf_cairo_draw_pie,
        .draw_chord       = _wmf_cairo_draw_chord,
        .draw_arc         = _wmf_cairo_draw_arc,
        .draw_ellipse     = _wmf_cairo_draw_ellipse,
        .draw_line        = _wmf_cairo_draw_line,
        .poly_line        = _wmf_cairo_poly_line,
        .draw_polygon     = _wmf_cairo_draw_polygon,
        .draw_polypolygon = _wmf_cairo_draw_polypolygon,
        .draw_rectangle   = _wmf_cairo_draw_rectangle,
        .rop_draw         = _wmf_cairo_rop_draw,
        .bmp_draw         = _wmf_cairo_bmp_draw,
        .bmp_read         = _wmf_cairo_bmp_read,
        .bmp_free         = _wmf_cairo_bmp_free,
        .draw_text        = _wmf_cairo_draw_text,
#ifdef USE_NOOP
        .udata_init       = _cairo_noop,
        .udata_copy       = _cairo_noop,
        .udata_set        = _cairo_noop,
        .udata_free       = _cairo_noop,
#else
        .udata_init       = _wmf_cairo_udata_init,
        .udata_copy       = _wmf_cairo_udata_copy,
        .udata_set        = _wmf_cairo_udata_set,
        .udata_free       = _wmf_cairo_udata_free,
#endif
        .region_frame     = _wmf_cairo_region_frame,
        .region_paint     = _wmf_cairo_region_paint,
        .region_clip      = _wmf_cairo_region_clip,
    };
    memcpy(api->function_reference,&vt,sizeof(wmfFunctionReference));
}
