/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdlib.h>
#include <string.h>


#define WUFFS_IMPLEMENTATION
#define WUFFS_CONFIG__MODULES
#define WUFFS_CONFIG__MODULE__BASE
#define WUFFS_CONFIG__MODULE__LZW
#define WUFFS_CONFIG__MODULE__GIF
#include "wuffs-unsupported-snapshot.c"

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_gifw_get_mime_types()
{
    static const char *mime_types[] = {
        "image/gif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gifw_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->frame = NULL;
    return h;
}


static void
_gifw_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


#define TRY(a) do{status = a ; if(!wuffs_base__status__is_ok(&status)) goto out;}while(0)

static int
_gifw_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    wuffs_gif__decoder dec;
    wuffs_base__image_config ic = {0};
    wuffs_base__io_buffer src;
    wuffs_base__pixel_buffer pb = {0};
    wuffs_base__slice_u8 workbuf = {0};
    wuffs_base__status status;

    TRY(wuffs_gif__decoder__initialize(&dec, sizeof dec, WUFFS_VERSION, 0));

    src.data.ptr = (uint8_t*)data;
    src.data.len = len;
    src.meta.wi = len;
    src.meta.ri = 0;
    src.meta.pos = 0;
    src.meta.closed = true;

    TRY(wuffs_gif__decoder__decode_image_config(&dec, &ic, &src));

    h->info->width = wuffs_base__pixel_config__width(&ic.pixcfg);
    h->info->height = wuffs_base__pixel_config__height(&ic.pixcfg);
    wuffs_base__pixel_config__set(&ic.pixcfg,
                                  WUFFS_BASE__PIXEL_FORMAT__BGRA_PREMUL,
                                  WUFFS_BASE__PIXEL_SUBSAMPLING__NONE,
                                  h->info->width,
                                  h->info->height);

    {
        uint64_t buflen = wuffs_gif__decoder__workbuf_len(&dec).max_incl;
        if(buflen > 0)
            workbuf = wuffs_base__malloc_slice_u8(malloc, buflen);
    }

    h->info->frame_count = 0;
    for(;;){
        wuffs_base__frame_config fc;
        frame_t *frame;
        TRY(wuffs_gif__decoder__decode_frame_config(&dec, &fc, &src));
        ++h->info->frame_count;
        if(h->frame){
            h->frame = (frame_t*)realloc(h->frame, h->info->frame_count * sizeof(frame_t));
        }else{
            h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
        }
        frame = &h->frame[h->info->frame_count - 1];
        frame->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, h->info->width, h->info->height);
        frame->duration = (double)wuffs_base__frame_config__duration(&fc) / (WUFFS_BASE__FLICKS_PER_MILLISECOND * 1000);

        TRY(wuffs_base__pixel_buffer__set_from_slice(&pb, &ic.pixcfg,
                                                     wuffs_base__make_slice_u8(cairo_image_surface_get_data(frame->surface),
                                                                               h->info->width * h->info->height * 4)));

        TRY(wuffs_gif__decoder__decode_frame(&dec, &pb, &src, WUFFS_BASE__PIXEL_BLEND__SRC_OVER, workbuf, NULL));
        cairo_surface_mark_dirty(frame->surface);
    }
out:
    if(workbuf.ptr)
        free(workbuf.ptr);
    if(wuffs_base__note__end_of_data == status.repr){
        return 0;
    }else{
        h->info->error = wuffs_base__status__message(&status);
        return -1;
    }
}


static cairo_surface_t *
_gifw_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_gifw_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gifw_get_mime_types,
        .new = _gifw_new,
        .free = _gifw_free,
        .create_from_data = _gifw_create_from_data,
        .get_image_surface = _gifw_get_image_surface,
        .get_duration = _gifw_get_duration
    };
    return &p;
}
