/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <ctype.h>
#include <debug.h>
#include <loa-buffer.h>
#include <stdlib.h>
#include <string.h>

#include "transform.h"
#include "svgcolor.h"
#include "xmlutil.h"


struct _loa_object_t{
    enum{
        OT_RECT,
        OT_PATH
    }type;
    color_t color;
    union{
        struct{
            double x;
            double y;
            double width;
            double height;
        }rect;
        loa_path_t path;
    };
};

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    loa_point_t *point;
    int point_count;
    loa_segment_t *segment;
    int segment_count;
    loa_object_t *shape;
    int shape_count;
};


static const char **
_fxg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-flash-xml-graphics",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_fxg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->scalable_size = 1;
    h->info->threadsafe = 1;
    h->point = NULL;
    h->segment = NULL;
    h->shape = NULL;
    return h;
}


static void
_fxg_free(abydos_plugin_handle_t *h)
{
    if(h->point)
        free(h->point);
    if(h->segment)
        free(h->segment);
    if(h->shape)
        free(h->shape);
    free(h);
}


typedef struct{
    abydos_plugin_handle_t *h;
    loa_buffer_t vb;
    svg_color_parser_t color_parser;
    int in_mask;
    transform_t transform;
    loa_object_t *shape;
    int blindlevel;
}parse_context_t;


static void
_path_set_string(loa_buffer_t *vb, transform_t *t, loa_path_t *path, const char *s)
{
    int op;
    loa_point_t start;
    loa_point_t last;
    path->segment = loa_buffer_segment_count(vb);
    path->segment_count = 0;

    op = 'M';
    start.x = 0;
    start.y = 0;
    last.x = 0;
    last.y = 0;
    for(;;){
        loa_point_t *p;
        SKIP_SPACE(s);
        if(!*s)
            return;
        if(isalpha(*s))
            op = *s++;
        switch(toupper(op)){
        case 'C':
            p = loa_buffer_push_segment(vb, path, LOA_CURVE);
            _get_point(&p[0], &s, &last, op);
            _get_point(&p[1], &s, &last, op);
            _get_point(&p[2], &s, &last, op);
            last = p[2];
            transform_apply_pos(t, &p[0].x, &p[0].y);
            transform_apply_pos(t, &p[1].x, &p[1].y);
            transform_apply_pos(t, &p[2].x, &p[2].y);
            break;
        case 'H':
            //DEBUG_p("hline %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            p[0].x = _get_double(&s);
            if(islower(op))
                p[0].x += last.x;
            p[0].y = last.y;
            last.x = p[0].x;
            transform_apply_pos(t, &p[0].x, &p[0].y);
            break;
        case 'L':
            //DEBUG_p("line %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            _get_point(&p[0], &s, &last, op);
            last = p[0];
            transform_apply_pos(t, &p[0].x, &p[0].y);
            break;
        case 'M':
            //DEBUG_p("move %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_GOTO);
            _get_point(&p[0], &s, &last, op);
            start = last = p[0];
            transform_apply_pos(t, &p[0].x, &p[0].y);
            op += 'L'-'M';
            break;
        case 'V':
            //DEBUG_p("vline %c",op);
            p = loa_buffer_push_segment(vb, path, LOA_LINE);
            p[0].y = _get_double(&s);
            if(islower(op))
                p[0].y += last.y;
            p[0].x = last.x;
            last.y = p[0].y;
            transform_apply_pos(t, &p[0].x, &p[0].y);
            break;
        case 'Z':
            //DEBUG_p("vline %c",op);
            loa_buffer_push_segment(vb, path, LOA_CLOSE);
            last = start;
            break;
        default:
            DEBUG_p("UNKNOWN OP: %c",op);
            break;
        }
    }
}



static void
_on_element_begin(void *userdata, const char *e, xml_attributes_t *attr)
{
    parse_context_t *c = (parse_context_t*)userdata;
    if(c->blindlevel > 0){
        ++c->blindlevel;
        return;
    }
    DEBUG_p("begin: %s",e);
    if(!strcasecmp(e, "fill")){
    }else if(!strcasecmp(e, "graphic")){
    }else if(!strcasecmp(e, "group")){
        DEBUG_p("id: %s",xml_parser_attr_string(attr, "id"));
        transform_push_translate(&c->transform, xml_parser_attr_int(attr, "x", 0), xml_parser_attr_int(attr, "y", 0));
        //DEBUG_p("(group) x offset: %f, y offset: %f", m->x0, m->y0);
    }else if(!strcasecmp(e, "mask")){
        ++c->in_mask;
    }else if(!strcasecmp(e, "matrix")){
        cairo_matrix_t m;
        m.xx = xml_parser_attr_double(attr, "a", 1);
        m.yx = xml_parser_attr_double(attr, "b", 0);
        m.xy = xml_parser_attr_double(attr, "c", 0);
        m.yy = xml_parser_attr_double(attr, "d", 1);
        m.x0 = xml_parser_attr_double(attr, "tx", 0);
        m.y0 = xml_parser_attr_double(attr, "ty", 0);
        transform_multiply(&c->transform, &m);
    }else if(!strcasecmp(e, "path")){
        const char *s = xml_parser_attr_string(attr, "data");
        if(s){
            c->shape = loa_buffer_add_object(&c->vb);
            c->shape->type = OT_PATH;
            _path_set_string(&c->vb, &c->transform, &c->shape->path, s);
        }else{
            c->blindlevel = 1;
        }
    }else if(!strcasecmp(e, "rect")){
        if(c->in_mask){
            c->h->info->width = xml_parser_attr_int(attr, "width", 0);
            c->h->info->height = xml_parser_attr_int(attr, "height", 0);
            c->blindlevel = 1;
        }else{
            c->shape = loa_buffer_add_object(&c->vb);
            c->shape->type = OT_RECT;
            c->shape->rect.x = xml_parser_attr_int(attr, "x", 0);
            c->shape->rect.y = xml_parser_attr_int(attr, "y", 0);
            c->shape->rect.width = xml_parser_attr_int(attr, "width", 0);
            c->shape->rect.height = xml_parser_attr_int(attr, "height", 0);
            transform_apply_pos(&c->transform, &c->shape->rect.x, &c->shape->rect.y);
            transform_apply_size(&c->transform, &c->shape->rect.width, &c->shape->rect.height);
            DEBUG_p("rect: %fx%f %fx%f", c->shape->rect.x, c->shape->rect.y, c->shape->rect.width, c->shape->rect.height);
        }
    }else if(!strcasecmp(e, "solidcolor")){
        svg_color_parse(&c->color_parser, &c->shape->color, xml_parser_attr_string(attr, "color"), -1);
        c->shape->color.a = xml_parser_attr_double(attr, "alpha", 1);
    }else if(!strcasecmp(e, "transform")){
    }else{
        c->blindlevel = 1;
    }
}


static void
_on_element_end(void *userdata, const char *e)
{
    parse_context_t *c = (parse_context_t*)userdata;
    if(c->blindlevel > 0){
        --c->blindlevel;
        return;
    }
    DEBUG_p("end: %s",e);
    if(!strcasecmp(e, "group")){
        transform_pop(&c->transform);
        DEBUG_p("transform stack size: %d",transform_size(&c->transform));
    }else if(!strcasecmp(e, "mask")){
        --c->in_mask;
    }
}


static int
_fxg_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    xml_handle_t *xml;
    parse_context_t c;
    c.h = h;
    loa_buffer_init(&c.vb, sizeof(loa_object_t), 0, 0, 0);
    c.in_mask = 0;
    transform_init(&c.transform);
    svg_color_parser_init(&c.color_parser);
    c.blindlevel = 0;

    xml = xml_parser_begin(_on_element_begin, _on_element_end, &c);
    xml_parser_feed(xml, data, len);
    xml_parser_end(xml);

    h->point = loa_buffer_copy_points(&c.vb);
    h->point_count = loa_buffer_point_count(&c.vb);
    h->segment = loa_buffer_copy_segments(&c.vb);
    h->segment_count = loa_buffer_segment_count(&c.vb);
    h->shape = loa_buffer_copy_objects(&c.vb);
    h->shape_count = loa_buffer_object_count(&c.vb);

    loa_buffer_done(&c.vb);
    transform_done(&c.transform);
    svg_color_parser_done(&c.color_parser);
    return 0;
}


static void
_fxg_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visibility,
            int variant,
            double time,
            int frame)
{
    int i;
    for(i=0;i<h->shape_count;++i){
        const loa_object_t *shape = &h->shape[i];
        int j;
        cairo_save(cr);
        switch(shape->type){
        case OT_RECT:
            cairo_rectangle(cr, shape->rect.x, shape->rect.y, shape->rect.width, shape->rect.height);
            break;
        case OT_PATH:
            for(j=0;j<shape->path.segment_count;++j){
                const loa_segment_t *seg = &h->segment[shape->path.segment+j];
                loa_point_t *p = &h->point[seg->point];
                switch(seg->type){
                case LOA_GOTO:
                    cairo_move_to(cr, p[0].x, p[0].y);
                    break;
                case LOA_LINE:
                    cairo_line_to(cr, p[0].x, p[0].y);
                    break;
                case LOA_CURVE:
                    cairo_curve_to(cr, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y);
                    break;
                case LOA_CLOSE:
                    cairo_close_path(cr);
                    break;
               }
            }
            break;
        }
        cairo_set_source_rgb(cr, shape->color.r, shape->color.g, shape->color.b);
        cairo_fill(cr);
        cairo_restore(cr);
    }
}


static void
_fxg_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _fxg_get_mime_types,
        .new = _fxg_new,
        .free = _fxg_free,
        .create_from_data = _fxg_create_from_data,
        .render = _fxg_render,
        .get_duration = _fxg_get_duration
    };
    return &p;
}
