/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <ctype.h>
#include <stdlib.h>                     /* strtod() */
#include <string.h>                     /* strcmp() */
#ifdef XML_USE_EXPAT
# include <expat.h>
#elif defined(XML_USE_GLIB)
# include <glib.h>
#elif defined(XML_USE_LIBXML)
# include <libxml/SAX.h>
#endif

#include "xmlutil.h"


int
_get_int(const char **s)
{
    SKIP_SEP(*s);
    return strtol(*s, (char**)s,10);
}


double
_get_double(const char **s)
{
    SKIP_SEP(*s);
    return strtod(*s, (char**)s);
}


void
_get_point(loa_point_t *p, const char **s, loa_point_t *last, int op)
{
    p->x = _get_double(s);
    p->y = _get_double(s);
    if(islower(op)){
        p->x += last->x;
        p->y += last->y;
    }
}


static int
_hexnum(const char* s, int len)
{
    int num = 0;
    while(len-->0){
        num *= 16;
        if(*s >= '0' && *s <= '9'){
            num += *s-'0';
        }else if(*s >= 'A' && *s <= 'F'){
            num += 10 + *s-'A';
        }else if(*s >= 'a' && *s <= 'f'){
            num += 10 + *s-'a';
        }
    }
    return num;
}


void
_color_set_hexstring(color_t *c, const char *s)
{
    if('#'==*s)
        ++s;
    if(s[3]){
        c->r = (double)_hexnum(s+0, 2)/255;
        c->g = (double)_hexnum(s+2, 2)/255;
        c->b = (double)_hexnum(s+4, 2)/255;
    }else{
        c->r = (double)_hexnum(s+0, 1)/15;
        c->g = (double)_hexnum(s+1, 1)/15;
        c->b = (double)_hexnum(s+3, 1)/15;
    }
}



struct _xml_handle_t{
#ifdef XML_USE_EXPAT
    XML_Parser context;
#elif defined(XML_USE_GLIB)
    GMarkupParseContext *context;
#elif defined(XML_USE_LIBXML)
    xmlParserCtxtPtr context;
#endif
    xml_begin_func_t begin_func;
    xml_end_func_t end_func;
    void *userdata;
};


struct _xml_attributes_t{
#ifdef XML_USE_GLIB
    const char **attr_name;
    const char **attr_value;
#else
    const char **attr;
#endif
};

static void
_on_element_begin(
#ifdef XML_USE_EXPAT
    void *userdata,
    const char *e,
    const char **attr
#elif defined(XML_USE_GLIB)
    GMarkupParseContext *context
    const char *e
    const char **attr_name,
    const char **attr_value,
    void *userdata,
    GError **error
#elif defined(XML_USE_LIBXML)
    void *userdata,
    const char *e,
    const char **attr
#endif
)
{
    xml_handle_t *h = (xml_handle_t*)userdata;
    xml_attributes_t a;
#ifdef XML_USE_GLIB
    a.attr_name = attr_names;
    a.attr_values = attr_values;
#else
    a.attr = attr;
#endif
    h->begin_func(h->userdata, e, &a);
}


static void
_on_element_end(
#ifdef XML_USE_EXPAT
    void *userdata,
    const char *e
#elif defined(XML_USE_GLIB)
    GMarkupParseContext *context,
    const char *e,
    void *userdata,
    GError **error
#elif defined(XML_USE_LIBXML)
    void *userdata,
    const char *e
#endif
)
{
    xml_handle_t *h = (xml_handle_t*)userdata;
    h->end_func(h->userdata, e);
}


#if 0
#ifdef USE_DEBUG
static void
_on_error(
#ifdef XML_USE_GLIB
    GMarkupParseContext *context,
    GError **error,
    void *userdata
#endif
)
{
}
#endif
#endif


xml_handle_t *
xml_parser_begin(xml_begin_func_t begin, xml_end_func_t end, void *userdata)
{
#ifdef XML_USE_GLIB
    static GMarkupParser xml_parser = {
        _on_element_begin,
        _on_element_end,
        NULL,                           /* text */
        NULL,                           /* passthrough */
#ifdef USE_DEBUG
        _on_error
#else
        NULL
#endif
    };
#elif defined(XML_USE_LIBXML)
    static xmlSAXHandler xml_parser = {
        .startElement = (startElementSAXFunc)_on_element_begin,
        .endElement = (endElementSAXFunc)_on_element_end
#ifdef USE_DEBUG
        ,.error = _on_error
#endif
    };
#endif
    xml_handle_t *h = malloc(sizeof(xml_handle_t));
#ifdef XML_USE_EXPAT
    h->context = XML_ParserCreate(NULL);
    XML_SetUserData(h->context, h);
    XML_SetElementHandler(h->context, _on_element_begin, _on_element_end);
#elif defined(XML_USE_GLIB)
    h->context = g_markup_parse_context_new(&xml_parser, 0, h, NULL);
#elif defined(XML_USE_LIBXML)
    h->context = xmlCreatePushParserCtxt(&xml_parser, h, NULL, 0, NULL);
#endif
    h->begin_func = begin;
    h->end_func = end;
    h->userdata = userdata;
    return h;
}


int
xml_parser_feed(xml_handle_t *h, const char *data, ssize_t len)
{
    if(len<0)
        len = strlen(data);
#ifdef XML_USE_EXPAT
    XML_Parse(h->context, data, len, 0);
#elif defined(XML_USE_GLIB)
    g_markup_parse_context_parse(h->context, data, len, NULL);
#elif defined(XML_USE_LIBXML)
    xmlParseChunk(h->context, data, len, 0);
#endif
    return 0;
}


int
xml_parser_end(xml_handle_t *h)
{
#ifdef XML_USE_EXPAT
    XML_ParserFree(h->context);
#elif defined(XML_USE_GLIB)
    g_markup_parse_context_end_parse(h->context, NULL);
    g_markup_parse_context_free(h->context);
#elif defined(XML_USE_LIBXML)
    xmlFreeParserCtxt(h->context);
#endif
    return 0;
}


const char *
xml_parser_attr_string(const xml_attributes_t *attr, const char *name)
{
    int i;
    for(i=0;;++i){
#ifdef XML_USE_GLIB
        if(!attr->attr_name[i])
            return NULL;
        if(!strcmp(!attr->attr_name[i], name))
            return attr->attr_value[i];
#else
        if(!attr->attr[i*2])
            return NULL;
        if(!strcmp(attr->attr[i*2], name))
            return attr->attr[i*2+1];
#endif
    }
}


int
xml_parser_attr_int(const xml_attributes_t *attr, const char *name, int fallback)
{
    const char *value = xml_parser_attr_string(attr, name);
    return value ? atoi(value) : fallback;
}


double
xml_parser_attr_double(const xml_attributes_t *attr, const char *name, int fallback)
{
    const char *value = xml_parser_attr_string(attr, name);
    return value ? strtod(value, NULL) : fallback;
}
