/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
extern "C"{
#include <debug.h>
}

#include "dec.h" // for exceptions
#include "streami.h"

ForwardInputStream::ForwardInputStream(const Buffer &buffer,size_t startOffset,size_t endOffset,bool allowOverrun) :
    _bufPtr(buffer.data()),
    _currentOffset(startOffset),
    _endOffset(endOffset),
    _allowOverrun(allowOverrun)
{
    if (_currentOffset>_endOffset || _currentOffset>buffer.size() || _endOffset>buffer.size())
        throw Decompressor::DecompressionError();
}

uint8_t ForwardInputStream::readByte()
{
	if (_currentOffset>=_endOffset) {
		if (_allowOverrun)
		{
			_currentOffset++;
			return 0;
		}
		DEBUG_p("read past end of data");
		throw Decompressor::DecompressionError();
	}
	uint8_t ret=_bufPtr[_currentOffset++];
	if (_linkedInputStream) _linkedInputStream->setOffset(_currentOffset);
	return ret;
}

const uint8_t *ForwardInputStream::consume(size_t bytes,uint8_t *buffer)
{
    if (_currentOffset+bytes>_endOffset){
		if (_allowOverrun && buffer){
			for (size_t i=0;i<bytes;i++){
				buffer[i]=(_currentOffset<_endOffset)?_bufPtr[_currentOffset]:0;
				_currentOffset++;
			}
			return buffer;
		}
		DEBUG_p("read past end of data");
		throw Decompressor::DecompressionError();
	}
    const uint8_t *ret=&_bufPtr[_currentOffset];
    _currentOffset+=bytes;
    if (_linkedInputStream)
        _linkedInputStream->setOffset(_currentOffset);
    return ret;
}

BackwardInputStream::BackwardInputStream(const Buffer &buffer,size_t startOffset,size_t endOffset,bool allowOverrun) :
    _bufPtr(buffer.data()),
    _currentOffset(endOffset),
    _endOffset(startOffset),
    _allowOverrun(allowOverrun)
{
    if (_currentOffset<_endOffset || _currentOffset>buffer.size() || _endOffset>buffer.size())
        throw Decompressor::DecompressionError();
}

uint8_t BackwardInputStream::readByte()
{
	if (_currentOffset<=_endOffset){
		if (_allowOverrun){
			--_currentOffset;
			return 0;
		}
		DEBUG_p("read past end of data");
		throw Decompressor::DecompressionError();
	}
	uint8_t ret=_bufPtr[--_currentOffset];
	if (_linkedInputStream) _linkedInputStream->setOffset(_currentOffset);
	return ret;
}

const uint8_t *BackwardInputStream::consume(size_t bytes,uint8_t *buffer){
	if (_currentOffset<_endOffset+bytes){
		if (_allowOverrun && buffer){
			for (size_t i=bytes;i;i--){
				buffer[i-1]=(_currentOffset>_endOffset)?_bufPtr[_currentOffset-1]:0;
				--_currentOffset;
			}
			return buffer;
		}
		DEBUG_p("read past end of data");
		throw Decompressor::DecompressionError();
	}
	_currentOffset-=bytes;
	if (_linkedInputStream) _linkedInputStream->setOffset(_currentOffset);
	return &_bufPtr[_currentOffset];
}
