/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#include <algorithm>
#include <cstring>
extern "C"{
#include <debug.h>
}

#include "dec.h" // for exceptions
#include "streamo.h"

OutputStream::OutputStream(uint8_t *bufPtr, size_t startOffset,size_t currentOffset,size_t endOffset) :
    Stream(startOffset, currentOffset, endOffset),
    _bufPtr(bufPtr)
{}


OutputStream::OutputStream(Buffer &buffer,size_t startOffset,size_t currentOffset,size_t endOffset) :
    OutputStream(buffer.data(), startOffset, startOffset, endOffset)
{
    if (_currentOffset>buffer.size() || _endOffset>buffer.size())
        throw OverrunError();
}

void ForwardOutputStream::writeByte(uint8_t value)
{
    if (_currentOffset>=_endOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
    _bufPtr[_currentOffset++]=value;
}

uint8_t ForwardOutputStream::copy(size_t distance,size_t count)
{
    if (!distance || _startOffset+distance>_currentOffset){
        DEBUG_p("incorrect distance: start: %d, pos: %d, dist: %d, count: %d",(int)_startOffset,(int)_currentOffset,(int)distance,(int)count);
        throw Decompressor::DecompressionError();
    }
    if (_currentOffset+count>_endOffset){
        DEBUG_p("write overrun: start: %d, pos: %d, dist: %d, count: %d",(int)_startOffset,(int)_currentOffset,(int)distance,(int)count);
        throw OverrunError();
    }
    uint8_t ret=0;
    for (size_t i=0;i<count;i++,_currentOffset++)
        ret=_bufPtr[_currentOffset]=_bufPtr[_currentOffset-distance];
    return ret;
}

uint8_t ForwardOutputStream::copy(size_t distance,size_t count,const Buffer &prevBuffer)
{
    if (!distance || _currentOffset+count>_endOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
	size_t prevCount=0;
	uint8_t ret=0;
	if (_startOffset+distance>_currentOffset)
	{
		size_t prevSize=prevBuffer.size();
		if (_startOffset+distance>_currentOffset+prevSize) throw Decompressor::DecompressionError();
		size_t prevDist=_startOffset+distance-_currentOffset;
		prevCount=std::min(count,prevDist);
		const uint8_t *prev=&prevBuffer[prevSize-prevDist];
		for (size_t i=0;i<prevCount;i++,_currentOffset++)
			ret=_bufPtr[_currentOffset]=prev[i];
	}
	for (size_t i=prevCount;i<count;i++,_currentOffset++)
		ret=_bufPtr[_currentOffset]=_bufPtr[_currentOffset-distance];
	return ret;
}

uint8_t ForwardOutputStream::copy(size_t distance,size_t count,uint8_t defaultChar)
{
	if (!distance || _currentOffset+count>_endOffset) throw Decompressor::DecompressionError();
	size_t prevCount=0;
	uint8_t ret=0;
	if (_startOffset+distance>_currentOffset)
	{
		prevCount=std::min(count,_startOffset+distance-_currentOffset);
		for (size_t i=0;i<prevCount;i++,_currentOffset++)
			ret=_bufPtr[_currentOffset]=defaultChar;
	}
	for (size_t i=prevCount;i<count;i++,_currentOffset++)
		ret=_bufPtr[_currentOffset]=_bufPtr[_currentOffset-distance];
	return ret;
}

const uint8_t *ForwardOutputStream::history(size_t distance) const
{
    if (distance>_currentOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
    return &_bufPtr[_currentOffset-distance];
}

void ForwardOutputStream::produce(const uint8_t *src,size_t bytes)
{
    if (_currentOffset+bytes>_endOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
    std::memcpy(&_bufPtr[_currentOffset],src,bytes);
    _currentOffset+=bytes;
}

void BackwardOutputStream::writeByte(uint8_t value)
{
    if (_currentOffset<=_startOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
    _bufPtr[--_currentOffset]=value;
}

uint8_t BackwardOutputStream::copy(size_t distance,size_t count)
{
    if (!distance || _startOffset+count>_currentOffset || _currentOffset+distance>_endOffset){
        DEBUG_p("write overrun");
        throw Decompressor::DecompressionError();
    }
    uint8_t ret=0;
    for (size_t i=0;i<count;i++,--_currentOffset)
        ret=_bufPtr[_currentOffset-1]=_bufPtr[_currentOffset+distance-1];
    return ret;
}
