/*
 * Copyright (C) 2020 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-amiga.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#include "xpk-dec.h"

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |   18 | "SVG Graphics File\0"
    18 |    2 | Version (always 1)
    20 |    4 | Header size
    24 |    4 | X offset
    28 |    4 | Y offset
    32 |    4 | Width
    36 |    4 | Height
    40 |    4 | Depth
    44 |    4 | View mode
    48 |    1 | Bits
    49 |    1 | Planes
    50 |    4 | Stride
    54 |  768 | Palette

*/

#define ID_XPKF ID('X','P','K','F')
#define ID_PP20 ID('P','P','2','0')
#define ID_LZ77 ID('L','Z','7','7')

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static const char **
_sview_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-superview",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_sview_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_sview_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_sview_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    nile_t ni;
    int bits;
    int planes;
    int depth;
    uint32_t bplcon0;
    int body_offs;

    if(len<58)
        return -1;
    if(memcmp(data, "SVG Graphics File\0", 18) && memcmp(data, "SGX Graphics File\0", 18))
        return -1;
    if(get16(data+18)!=1)
        return -1;

    nile_init(&ni);

    bits = get8(data+48);
    planes = get8(data+49);
    if(bits > 1 && planes > 1)
        return -1;

    depth = get32(data+40);
    DEBUG_p("bits: %d, planes: %d, depth: %d", bits, planes, depth);

    switch(bits){
    case 1:
        if(planes > 8 || depth > planes)
            return -1;
        break;
    case 8:
        if(depth > 8)
            return -1;
        break;
    case 24:
    case 32:
    case 48:
    case 64:
        if(bits != depth)
            return -1;
        break;
    default:
        return -1;
    }

    if(depth <= 8){
        if(54 + (1<<depth) > len)
            return -1;
        ni.palette = (uint8_t*)data+54;
        //ni.colors = 1<<depth;
        //_expand_palette((uint8_t*)data+54,palette,1<<depth);
    }

    ni.width = h->info->width = get32(data+32);
    ni.height = h->info->height = get32(data+36);
    ni.stride = get32(data+50);
    DEBUG_p("size: %dx%d, stride: %d", ni.width, ni.height, ni.stride);

    body_offs = get32(data+20);
    DEBUG_p("data offset: %d", body_offs);
    ni.pixels_allocation = ni.stride * h->info->height * planes;

    bplcon0 = get32(data+44);
    ni.pixels = (uint8_t*)malloc(ni.pixels_allocation);
    if(xpk_decompress((char*)ni.pixels, ni.pixels_allocation, data + body_offs, len-body_offs) < 0){
        nile_done(&ni);
        return -1;
    }

    DEBUG_p("success");

    /* create surface from body */
#if 0
    {
        nile_t dst_ni;
        nile_t src_ni;
        nil_cairo_from_surface(&dst_ni, h->surface);
        dst_ni.width = h->info->width;
        dst_ni.height = h->info->height;
        if(bits < 8){
            /* bitplanes */
            uint8_t *bitplane[8];
            int i;
            nile_ensure_bitplanes(&dst_ni, -1, -1, depth, 1);
        for(i=0;i<planes;++i)
            bitplane[i] = &body[src_stride*h->info->height*i];
        }else if(8==bits){
        }else{
            if(depth < 8
        }
    }
#else
    if(1==bits){
        /* bitplanes */
        uint8_t *dst;
        int dst_stride;
        h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, h->info->width, h->info->height);
        dst = cairo_image_surface_get_data(h->surface);
        dst_stride = cairo_image_surface_get_stride(h->surface);
        uint8_t *bitplane[8];
        int i;
        for(i=0;i<planes;++i)
            bitplane[i] = &ni.pixels[ni.stride * ni.height * i];
        nil_amiga_blit_direct_from_bitplanes(dst, dst_stride,
                                             4, NIL_CAIRO_FORMAT_RGB24,
                                             bitplane, nil_amiga_plane_stride(ni.width), depth,
                                             ni.palette, 1<<depth, 3, NILE_FORMAT_r8g8b8,
                                             ni.width, ni.height,
                                             bplcon0, 0, 0);
    }else if(8==bits){
        ni.type = NILE_TYPE_INDEXED;
        ni.colors = 1<<depth;
        ni.bpc = 3;
        ni.format = NILE_FORMAT_r8g8b8;
        h->surface = nil_cairo_surface_create(&ni, 0);
    }else{
        ni.type = NILE_TYPE_DIRECT;
        switch(depth){
        case 24:
            ni.format = NILE_FORMAT_r8g8b8;
            ni.bpc = 3;
            break;
        case 32:
            ni.format = NILE_FORMAT_r8g8b8a8;
            ni.bpc = 4;
            break;
        case 48:
            ni.format = NILE_FORMAT_be_r16g16b16;
            ni.bpc = 3;
            break;
        case 64:
            ni.format = NILE_FORMAT_be_r16g16b16a16;
            ni.bpc = 4;
            break;
        }
        h->surface = nil_cairo_surface_create(&ni, 0);
    }
#endif
    nile_done(&ni);
    return 0;
}


static cairo_surface_t *
_sview_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _sview_get_mime_types,
        .new = _sview_new,
        .free = _sview_free,
        .create_from_data = _sview_create_from_data,
        .get_image_surface = _sview_get_image_surface
    };
    return &p;
}
