/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <cstring>
#include <endian.h>

#include "xpk.h"

extern "C"{
#include <debug.h>
#include <nil-iff.h>

#include "xpk-dec.h"
}

#define ID_XPKF IFF_ID('X','P','K','F')
#define ID_PP20 IFF_ID('P','P','2','0')
#define ID_LZ77 IFF_ID('L','Z','7','7')

#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static int
_unpack_xpk_chunk(xpk_compression_t comp, const char *src, size_t src_len, const char *dst_start, char *dst, size_t dst_len)
{
    ConstStaticBuffer src_buf((uint8_t*)src, src_len);
    MemoryBuffer dst_buf((uint8_t*)dst, dst_len);
    ConstStaticBuffer prev_buf((uint8_t*)dst_start, dst-dst_start);
    std::unique_ptr<XPKDecompressor::State> state;
    XPKDecompressor *dec = xpk_lookup_decompressor(comp, src_buf, state);
    if(!dec)
        return -1;
    int ret = 0;
    try{
        dec->decompressImpl(dst_buf, prev_buf, true);
    }catch(const std::exception&){
        DEBUG_p("chunk decompression failed");
        ret = -1;
    }
    delete dec;
    return ret;
}


int
_unpack_xpk(char *dst, size_t dst_len, const char *src, const char *src_end)
{
    const char *dst_start = dst;
    const char *dst_end = dst + dst_len;
    DEBUG_p("XPKF compression (%d bytes)",(int)(src_end - src));

    if(src+40 > src_end)
        return -1;

    size_t packed_size = get32(src+0);
    if(src+4+packed_size > src_end)
        return -1;

    size_t unpacked_size = get32(src+8);
    if(dst_len > unpacked_size)
        return -1;

    int flags = get8(src+28);
    if(flags&2){
        DEBUG_p("password protected");
        return -1;
    }
    int header_len = flags&1 ? 12 : 8;

    xpk_compression_t comp = (xpk_compression_t)*(uint32_t*)(src+4);

    if(flags&4)
        src += 2+get16(src+32);
    src += 32;

    for(;;){
        int chunk_type = get8(src);
        int packed_chunk_size;
        int unpacked_chunk_size;
        if(flags&1){
            packed_chunk_size = get32(src+4);
            unpacked_chunk_size = get32(src+8);
        }else{
            packed_chunk_size = get16(src+4);
            unpacked_chunk_size = get16(src+6);
        }

        if(src+header_len+packed_chunk_size > src_end)
            goto error;
        if(dst+unpacked_chunk_size > dst_end)
            goto error;

        src += header_len;
        if(0==chunk_type){
            /* uncompressed data */
            if(packed_chunk_size != unpacked_chunk_size){
                DEBUG_p("packed chunk size (%d) differs from unpacked chunk size (%d)",(int)packed_chunk_size,(int)unpacked_chunk_size);
                goto error;
            }
            memcpy(dst, src, unpacked_chunk_size);
        }else if(1==chunk_type){
            /* compressed data */
            if(_unpack_xpk_chunk(comp, src, packed_chunk_size, dst_start, dst, unpacked_chunk_size)<0)
                goto error;
        }else if(15==chunk_type){
            /* explicit end */
            break;
        }else{
            /* invalid chunk type */
            DEBUG_p("invalid chunk type");
            goto error;
        }

        src += (packed_chunk_size+3)&~3;
        if(src >= src_end)
            break;

        DEBUG_p("bytes left: %d",(int)(src_end-src));

        dst += unpacked_chunk_size;
        if(dst >= dst_end)
            break;
    }
    return 0;
error:
    DEBUG_p("unpack error");
    return -1;
}


static int
_copy_uncompressed(char *dst, size_t dst_len, const char *src, size_t src_len)
{
    if(dst_len > src_len)
        std::memset(dst + src_len, 0, dst_len-src_len);
    std::memcpy(dst, src, src_len);
    return 0;
}


extern "C" int
xpk_decompress(char *dst, size_t dst_len, const char *src, size_t src_len)
{
    const char *end = src + src_len;
    DEBUG_p("packed body is %d bytes", (int)src_len);
    if(src+8 > end){
        DEBUG_p("data short");
        return -1;
    }
    switch(*(uint32_t*)src){
    case ID_XPKF:
        DEBUG_p("PKF compression");
        src += 4;
        return _unpack_xpk(dst, dst_len, src, end);
    case ID_PP20:
        DEBUG_p("PP20 compression");
        return -1;
    case ID_LZ77:
        DEBUG_p("LZ77 compression");
#if 0
        src += 4;
        return _unpack_lz77(src, end, dst_len);
#endif
    default:
        DEBUG_p("unknown compression, assume uncompressed");
        return _copy_uncompressed(dst, dst_len, src, src_len);
    }
}
