/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nile.h>
#include <nil-cairo.h>
#include <nil-iff.h>
#include <stdlib.h>
#include <string.h>

#include "xpk-dec.h"


#define FPS 55

#define IFF_YAFA IFF_ID('Y','A','F','A')
#define YAFA_INFO IFF_ID('I','N','F','O')
#define YAFA_DRGB IFF_ID('D','R','G','B')
#define YAFA_DLTX IFF_ID('D','L','T','X')
#define YAFA_PROF IFF_ID('P','R','O','F')
#define YAFA_TTBL IFF_ID('T','T','B','L')
#define YAFA_ANNO IFF_ID('A','N','N','O')
#define YAFA_BODY IFF_ID('B','O','D','Y')

#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_yafa_get_mime_types()
{
    static const char *mime_types[] = {
        "video/x-yafa",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_yafa_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->frame = NULL;
    return h;
}


static void
_yafa_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_unpack(char *dst, size_t dst_len, const char *src, size_t src_len)
{
    return xpk_decompress(dst, dst_len, src, src_len);
}


static int
_set_palette(nile_t *ni, const char *data, size_t len)
{
    int count;
    int first;
    int i;
    if(4 > len)
        return -1;
    count = get16(data);
    DEBUG_p("%d < %d", count*12, (int)len);
    if(4 + count * 12 > len)
        return -1;
    first = get16(data+2);
    if(first + count > ni->colors)
        count = ni->colors - first;
    for(i=0;i<count;++i){
        ni->palette[i*3 + 0] = get8(data + i*12 + 0);
        ni->palette[i*3 + 1] = get8(data + i*12 + 4);
        ni->palette[i*3 + 2] = get8(data + i*12 + 8);
    }
    return 0;
}


typedef struct{
    abydos_plugin_handle_t *h;
    nile_t ni;
    int image_size;
    int speed;
    int type;
    int ham;
    int multipal;
    int delta;
    size_t frame_size;
    char *allocation;
    size_t *frame_len;
}parse_context_t;


static int
_on_chunk(void *userdata, uint32_t id, const char *data, size_t len)
{
    parse_context_t *pc = (parse_context_t*)userdata;
    int i;
    DEBUG_p("id: %.4s, length: %d", (char*)&id, (int)len);
    switch(id){
    case YAFA_INFO:
        if(14 > len)
            return -1;
        pc->ni.type = NILE_TYPE_INDEXED;
        pc->h->info->width = pc->ni.width = get16(data);
        pc->h->info->height = pc->ni.height = get16(data+2);
        pc->image_size = pc->ni.width * pc->ni.height;
        pc->ni.depth = get16(data+4);
        nile_ensure_palette(&pc->ni, 1 << pc->ni.depth, 3);
        pc->ni.format = NILE_FORMAT_r8g8b8;
        pc->speed = get16(data+6);
        pc->h->info->frame_count = get16(data+8);
        pc->type = get16(data+10);
        // flags = get16(data+12);
        /* This is the expected size of a frame if there is no global palette. */
        pc->frame_size = pc->image_size + 4 + pc->ni.colors * 12 + 4;
        pc->allocation = malloc(pc->frame_size);
        pc->ni.pixels = (uint8_t*)pc->allocation;
        pc->h->frame = malloc(pc->h->info->frame_count * sizeof(frame_t));
        for(i=0;i<pc->h->info->frame_count;++i){
            pc->h->frame[i].surface = NULL;
            pc->h->frame[i].duration = (double)pc->speed / FPS;
        }
        break;
    case YAFA_DRGB:
        if(!pc->ni.palette)
            return -1;
        if(_set_palette(&pc->ni, data, len) < 0)
            return -1;
        /* Expect the frame to only contain an image. */
        pc->frame_size = pc->image_size;
        break;
    case YAFA_DLTX:
        /* FIXME!! */
        break;
    case YAFA_PROF:
        if(!pc->h->frame)
            return -1;
        if(pc->h->info->frame_count * 4 > len)
            return -1;
        pc->frame_len = (size_t*)malloc(sizeof(size_t) * pc->h->info->frame_count);
        for(i=0;i<pc->h->info->frame_count;++i)
            pc->frame_len[i] = get32(data + i * 4);
        break;
    case YAFA_TTBL:
        if(!pc->h->frame)
            return -1;
        if(pc->h->info->frame_count * 2 > len)
            return 0; /* No big deal if the time codes are invalid. */
        for(i=0;i<pc->h->info->frame_count;++i)
            pc->h->frame[i].duration = (double)get16(data + i*2) / FPS;
        break;
    case YAFA_ANNO:
        break;
    case YAFA_BODY:
        if(!pc->h->frame)
            return -1;
        DEBUG_p("unpacking body type %d", pc->type);
        switch(pc->type){
        case 3: /* chunky XPK */
            if(!pc->frame_len){
                /* FIXME!! read just one frame */
                return -1;
            }
            nile_ensure_indexed(&pc->ni, pc->ni.width, pc->ni.height);
            {
                const char *end = data + len;
                const char *p1 = data;
                for(i=0;i<pc->h->info->frame_count;++i){
                    const char *p2 = data + pc->frame_len[i];
                    if(p2 > end)
                        return -1;
                    _unpack((char*)pc->allocation, pc->frame_size, p1, p2-p1);
                    if(pc->frame_size > pc->image_size)
                        if(_set_palette(&pc->ni, pc->allocation + pc->image_size, pc->frame_size-pc->image_size) < 0)
                            return -1;
                    pc->h->frame[i].surface = nil_cairo_surface_create(&pc->ni, 0);
                    p1 = p2;
                }
            }
            break;
        }
        break;
    }
    return 0;
}


static int
_yafa_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    parse_context_t pc;
    int res;
    pc.h = h;
    nile_init(&pc.ni);
    pc.allocation = NULL;
    pc.frame_len = NULL;
    res = nil_iff_parse(IFF_YAFA, 0, data, len, _on_chunk, &pc);
    nile_done(&pc.ni);
    if(pc.allocation)
        free(pc.allocation);
    if(pc.frame_len)
        free(pc.frame_len);
    return res;
}


static cairo_surface_t *
_yafa_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_yafa_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _yafa_get_mime_types,
        .new = _yafa_new,
        .free = _yafa_free,
        .create_from_data = _yafa_create_from_data,
        .get_image_surface = _yafa_get_image_surface,
        .get_duration = _yafa_get_duration
    };
    return &p;
}
