/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef abydos_internal_h
#define abydos_internal_h

#include <stdio.h>
#include <pthread.h>

#include "abydos.h"
#include "abydos-plugin.h"
#include "atomic.h"

typedef struct{
    int refcount;
    char *filename;                     /* Store the filename in order to load it on demand. */
    void *handle;
    pthread_mutex_t lock;
    const abydos_plugin_interface_t *vt;
} plugin_t;

const char *_abydos_unalias(const char *mime_type);
plugin_t * _abydos_get_plugin(const char *mime_type);
plugin_t * _abydos_ref_plugin(plugin_t *);
void _abydos_unref_plugin(plugin_t *);

#define PLUGIN_CALL(p,func,arg,...) \
  (p)->vt->func(arg,##__VA_ARGS__)

#define PLUGIN_LOCK_CALL(p,unsafe,func,arg,...) \
  { \
    if(unsafe) pthread_mutex_lock(&(p)->lock); \
    (p)->vt->func(arg,##__VA_ARGS__); \
    if(unsafe) pthread_mutex_unlock(&(p)->lock); \
  }

#define PLUGIN_LOCK_CALL_RET(p,unsafe,func,type,arg,...) \
  ({ \
    type ret; \
    if(unsafe) pthread_mutex_lock(&(p)->lock); \
    ret = (p)->vt->func(arg,##__VA_ARGS__); \
    if(unsafe) pthread_mutex_unlock(&(p)->lock); \
    ret; \
  })

typedef struct{
    abydos_plugin_page_info_t info;
    /* layer */
    /* offset to the first (togglable) layer of this page in the list of all layers */
    int layer_offset;
    /* variant */
    abydos_size_t *variant_size;
    int default_variant;
    /* animation */
    double *frame_duration;
    double *frame_time;
    double total_duration;
} page_t;

void _abydos_page_get_size_int(page_t *page, int *r_width, int *r_height, int flags);

typedef struct{
    AtomicInt refcount;
    plugin_t *plugin;
    abydos_plugin_handle_t *handle;
    abydos_plugin_info_t info;
    pthread_mutex_t lock;
    /* progressive loading */
    struct{
        abydos_info_callback_t info_func;
        abydos_update_callback_t update_func;
        void *userdata;
        FILE *f;
        union{
            char *buffer;
            char *tmpfile;
        };
        size_t buffer_len;
    }load;
    struct{
        page_t *page;
        /* Total number of layers (that can be toggled */
        int total_layer_count;
        int square_pixels;
    }done;
} immutable_t;

immutable_t *_abydos_immutable_create(void);
immutable_t *_abydos_immutable_reference(immutable_t *im);
void _abydos_immutable_destroy(immutable_t *im);
void _abydos_render_internal(immutable_t *im,cairo_t *cr,page_t *page,int pagenr,cairo_bool_t *visible_layer,int variant,double time,int frame);
cairo_surface_t *_abydos_get_image_surface_internal(immutable_t *im_,int flags,int pagenr,cairo_bool_t *visible_layer,int variant,double time,int frame);

struct _abydos_t{
    /*< private >*/
    AtomicInt refcount;
    immutable_t *im;
    int options;
    cairo_surface_t* cached_surface;
    int page;
    cairo_bool_t *layer_visibility;
    int prefered_variant;
    double time;
    int frame;
    double max_adjusted_width;
    double max_adjusted_height;
};

void _abydos_setup(abydos_t *ar);

#endif /* abydos_internal_h */
