/*
 * Copyright (C) 2021 - 2024  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdlib.h>                     /* malloc(), free() */

#include "abydos-internal.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    immutable_t **im;
};


static abydos_plugin_handle_t *
_variants_new(abydos_plugin_info_t *info, abydos_t **variants)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    int i;
    while(variants[info->variant_count])
        ++info->variant_count;
    h->info = info;
    h->im = (immutable_t**)malloc(info->variant_count * sizeof(immutable_t*));
    for(i=0;i<info->variant_count;++i){
        h->im[i] = _abydos_immutable_reference(variants[i]->im);
    }
    return h;
}


static void
_variants_free(abydos_plugin_handle_t *h)
{
    int i;
    for(i=0; i<h->info->variant_count; ++i)
        _abydos_immutable_destroy(h->im[i]);
    free(h->im);
    free(h);
}


static void
_variants_render(abydos_plugin_handle_t *h,
                 cairo_t *cr,
                 int page,
                 cairo_bool_t *visible_layers,
                 int prefered_variant,
                 double time,
                 int frame)
{
    immutable_t *im = h->im[prefered_variant];
    _abydos_render_internal(im, cr, &im->done.page[0], 0, NULL, 0, time, frame);
}


static cairo_surface_t *
_variants_get_image_surface(abydos_plugin_handle_t *h,
                            int page,
                            cairo_bool_t *visible_layers,
                            int prefered_variant,
                            double time,
                            int frame)
{
    immutable_t *im = h->im[prefered_variant];
    return _abydos_get_image_surface_internal(im, 0, 0, NULL, 0, time, frame);
}


static void
_variants_get_variant_size(abydos_plugin_handle_t *h, int page, abydos_size_t *size)
{
    int i;
    for(i=0;i<h->info->variant_count;++i){
        immutable_t *im = h->im[i];
        _abydos_page_get_size_int(&im->done.page[0], &size[i].width, &size[i].height, 0);
    }
}


abydos_t *
abydos_create_from_variants(abydos_t **variants)
{
    static const abydos_plugin_interface_t vt = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .free = _variants_free,
        .render = _variants_render,
        .get_image_surface = _variants_get_image_surface,
        .get_variant_size = _variants_get_variant_size
    };
    static plugin_t p = {
        .vt = &vt
    };
    immutable_t *im = _abydos_immutable_create();
    abydos_t *ar = (abydos_t*)malloc(sizeof(abydos_t));

    im->plugin = &p;
    im->handle = _variants_new(&im->info, variants);

    ar->refcount = 1;
    ar->im = im;
    ar->options = 0;
    ar->cached_surface = NULL;
    ar->page = 0;
    ar->layer_visibility = NULL;
    ar->prefered_variant = -1;
    ar->time = 0;
    ar->frame = 0;

    /* Special case if there is only one variant. */
    if(1 == im->info.variant_count){
        im->info.width = im->handle->im[0]->info.width;
        im->info.height = im->handle->im[0]->info.height;
    }

    _abydos_setup(ar);
    return ar;
}
