/*
 * Copyright (C) 2018 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <dirent.h>
#include <dlfcn.h>                      /* dlopen */
#include <errno.h>
#include <fcntl.h>                      /* open() */
#include <math.h>                       /* fmod() */
#include <pthread.h>
#include <stdio.h>                      /* readdir() */
#include <stdlib.h>                     /* free() */
#include <string.h>                     /* memcpy(), strlen() */
#include <sys/mman.h>                   /* mmap() */
#include <sys/stat.h>
#include <unistd.h>                     /* unlink() */

#include "abydos.h"
#include "abydos-internal.h"
#include "abydos-plugin.h"
#include "atomic.h"
#include "debug.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define MINIMUM_DURATION 0.02

/**
 * SECTION: abydos
 * @short_description: The basic set of the application API.
 * @title: Basics
 * @include: abydos.h
 *
 * There are three steps for using abydos. The first step is to create an
 * #abydos_t for a specific MIME type using abydos_create(). Then you are
 * done using it you should destroy it using abydos_destroy(). Determining
 * the correct MIME type for your data is outside the scope of abydos. Read
 * more about [image format detection](../detection.html).
 *
 * The second step is to load the data. There are three different methods
 * to choose from for doing this. You can load it from a file using
 * abydos_from_file(). Or you can load it from memory using
 * abydos_from_data(). Or you can use abydos_load_begin(),
 * abydos_load_feed() and abydos_load_end() to progressively load the image.
 * With the last option you may get the opertunity to display a partially
 * decoded image. But this is not always possible since some images store
 * crucial information at the end of the file.
 *
 * The third step is to get the image data. There are two different ways to
 * do this. Either you can render it to a #cairo_t using abydos_render().
 * This is the recommended approch if you intend to display the image since
 * vector graphics will always be rendered in full quality intependant of
 * scale. Or you can get a cairo image surface with
 * abydos_get_image_surface().
 * |[
#include <abydos.h>
#include <cairo.h>

int main (int argc, char **argv)
{
    abydos_t *ar;
    cairo_surface_t *surface;

    ar = abydos_create ("image/x-ilbm");
    if (!ar)
        return 1;
    if (abydos_from_file (ar, "example.ilbm") < 0)
        return 1;
    surface = abydos_get_image_surface (ar);
    abydos_destroy (ar);
    cairo_surface_write_to_png (surface, "example.png");
    cairo_surface_destroy (surface);
    return 0;
}
 * ]|
 **/


static cairo_bool_t
_open_tmpfile(char **r_filename,FILE** r_f)
{
    static const char template[] = PACKAGE "-XXXXXX";
    const char *tmpdir;
    size_t len;
    char *filename;
    int fd;
    tmpdir = getenv("XDG_RUNTIME_DIR");
    if(!tmpdir)
        tmpdir = getenv("TMPDIR");
    if(!tmpdir)
        tmpdir = "/tmp";
    len = strlen(tmpdir);
    filename = (char*)malloc(len+1+sizeof(template));
    memcpy(filename,tmpdir,len);
    filename[len] = '/';
    memcpy(filename+len+1,template,sizeof(template));
    fd = mkstemp(filename);
    if(fd<0){
        free(filename);
        return 0;
    }
    *r_filename = filename;
    *r_f = fdopen(fd,"w+");
    return 1;
}


static page_t *
_get_page(immutable_t *im, int pagenr)
{
    if(pagenr >= im->info.page_count)
        pagenr = im->info.page_count-1;
    return &im->done.page[pagenr];
}


static void
_adjust_size(page_t *page, double *width, double *height, int flags)
{
    switch(flags & ABYDOS_ASPECT_RATIO_MASK){
    case ABYDOS_IGNORE_ASPECT_RATIO:
        break;
    case ABYDOS_SHRINK_TO_ASPECT_RATIO:
        if(page->info.pixel_ratio < 1){
            /* high pixels */
            *width *= page->info.pixel_ratio;
        }else if(page->info.pixel_ratio > 1){
            /* wide pixels */
            *height /= page->info.pixel_ratio;
        }
        break;
    case ABYDOS_GROW_TO_ASPECT_RATIO:
    default:
        if(page->info.pixel_ratio < 1){
            /* high pixels */
            *height /= page->info.pixel_ratio;
        }else if(page->info.pixel_ratio > 1){
            /* wide pixels */
            *width *= page->info.pixel_ratio;
        }
        break;
    }
}


static void
_get_page_size(page_t *page, double *width, double *height, int flags)
{
    *width = page->info.width;
    *height = page->info.height;
    _adjust_size(page, width, height, flags);
}


void
_abydos_page_get_size_int(page_t *page, int *r_width, int *r_height, int flags)
{
    double width;
    double height;
    _get_page_size(page, &width, &height, flags);
    if(r_width)
        *r_width = width+0.5;
    if(r_height)
        *r_height = height+0.5;
}


/*
 * FIXME!! this value could be cached, but that requires locking of the
 * abydos_t struct.
 * If the image has square pixels it never needs to be recalculated.
 */
static void
_get_max_size(abydos_t *ar, double *r_width, double *r_height)
{
    double width;
    double height;
    int i;
    *r_width = 0;
    *r_height = 0;
    for(i=0;i<ar->im->info.page_count;++i){
        _get_page_size(&ar->im->done.page[i], &width, &height, ar->options);
        if(width > *r_width)
            *r_width = width;
        if(height > *r_height)
            *r_height = height;
    }
}


static void
_get_max_size_int(abydos_t *ar, int *r_width, int *r_height)
{
    double width;
    double height;
    _get_max_size(ar, &width, &height);
    if(r_width)
        *r_width = width+0.5;
    if(r_height)
        *r_height = height+0.5;
}


immutable_t *
_abydos_immutable_create()
{
    immutable_t *im = (immutable_t*)malloc(sizeof(immutable_t));
    im->refcount = 1;
    im->info.version = ABYDOS_PLUGIN_INFO_VERSION;
    im->info.error = NULL;

    /* setup default values */
    im->info.width = 0;
    im->info.height = 0;
    im->info.pixel_ratio = 1;
    im->info.page_count = 1;
    im->info.default_page = 0;
    im->info.layer_count = 1;
    im->info.variant_count = 1;
    im->info.frame_count = 1;
    im->info.default_frame = 0;
    im->info.scalable_size = 0;
    im->info.scalable_time = 0;
    im->info.threadsafe = 0;
    im->info.threadsafe_render = 1;

    pthread_mutex_init(&im->lock,NULL);
    return im;
}


static immutable_t *
_immutable_create_with_mime_type(const char *mime_type)
{
    plugin_t *p = _abydos_get_plugin(mime_type);
    immutable_t *im;
    if(!p){
        DEBUG_p("no loader found for mime type '%s'",mime_type);
        return NULL;
    }
    im =  _abydos_immutable_create();

    im->plugin = p;
    im->handle = PLUGIN_LOCK_CALL_RET(p,1,new,abydos_plugin_handle_t*,mime_type,&im->info);
    if(!im->handle){
        free(im);
        return NULL;
    }
    if(im->info.threadsafe)
        im->info.threadsafe_render = 1;

    im->done.page = NULL;
    //im->done.unique_page_count = 0;

    return im;
}


immutable_t *
_abydos_immutable_reference(immutable_t *im)
{
    atomic_int_inc(&im->refcount);
    return im;
}


void
_abydos_immutable_destroy(immutable_t *im)
{
    if(atomic_int_dec_and_test(&im->refcount)){
        PLUGIN_LOCK_CALL(im->plugin,!im->info.threadsafe,free,im->handle);
        //ar->im->vt->free(ar->im->handle);
        _abydos_unref_plugin(im->plugin);
        if(im->done.page){
            int i;
            for(i=0;i<im->info.page_count;++i){
                if(im->done.page[i].variant_size)
                    free(im->done.page[i].variant_size);
                if(im->done.page[i].frame_duration){
                    free(im->done.page[i].frame_duration);
                    free(im->done.page[i].frame_time);
                }
            }
            free(im->done.page);
        }
        pthread_mutex_destroy(&im->lock);
        free(im);
    }
}

/**
 * abydos_create:
 * @mime_type: MIME type
 *
 * Creates a new #abydos_t object to load an image of the type specified by @mime_type.
 *
 * returns: An abydos_t capable of handling @mime_type. Or NULL if none is
 *          found. Call abydos_destroy() then you are done using it.
 **/
abydos_t *
abydos_create(const char *mime_type)
{
    immutable_t *im = _immutable_create_with_mime_type(_abydos_unalias(mime_type));
    abydos_t *ar;
    if(!im)
        return NULL;
    ar = (abydos_t*)malloc(sizeof(abydos_t));
    ar->refcount = 1;
    ar->im = im;
    ar->options = 0;
    ar->cached_surface = NULL;
    ar->page = 0;
    ar->layer_visibility = NULL;
    ar->prefered_variant = -1;
    ar->time = 0;
    ar->frame = 0;

    return ar;
}


/**
 * abydos_clone
 * @src: An #abydos_t to clone
 *
 * Make a clone of @src, which is cheap and effectively only copies the
 * internal state. This is useful for maintaining multiple states at once.
 * Such as rendering several pages of a multi page image in parallell.
 *
 * returns: A clone of @src. Call abydos_destroy() then you are done using
            it.
 **/
abydos_t *
abydos_clone(abydos_t *src)
{
    abydos_t *dst = (abydos_t*)malloc(sizeof(abydos_t));
    memcpy(dst,src,sizeof(abydos_t));
    dst->im = _abydos_immutable_reference(src->im);
    if(dst->layer_visibility){
        dst->layer_visibility = (cairo_bool_t*)malloc(dst->im->info.layer_count*sizeof(cairo_bool_t));
        memcpy(dst->layer_visibility,src->layer_visibility,dst->im->info.layer_count*sizeof(cairo_bool_t));
    }
    return dst;
}


/**
 * abydos_reference
 * @ar: an #abydos_t
 *
 * Increases the reference count on @ar by one. This prevents ar from being
 * destroyed until a matching call to abydos_destroy() is made.
 *
 * returns: The same #abydos_t. Call abydos_destroy() then you are done
            using it.
 **/
abydos_t *
abydos_reference(abydos_t *ar)
{
    atomic_int_inc(&ar->refcount);
    return ar;
}


/**
 * abydos_destroy
 * @ar: an #abydos_t
 *
 * Decreases the reference count on @ar by one. If the result is zero, then
 * @ar and all associated resources are freed.
 */
void
abydos_destroy(abydos_t *ar)
{
    if(atomic_int_dec_and_test(&ar->refcount)){
        _abydos_immutable_destroy(ar->im);
        if(ar->layer_visibility)
            free(ar->layer_visibility);
        free(ar);
    }
}


/**
 * abydos_get_options
 * @ar: an #abydos_t
 *
 * Gets the options currently in effect.
 *
 * returns: Current effective options.
 */
int abydos_get_options(abydos_t *ar){
    return ar->options;
}


/**
 * abydos_set_options
 * @ar: an #abydos_t
 * @options: Options
 *
 * The @options is a combination of the following (ored together):
 * - #ABYDOS_PREFER_SHARE
 * - #ABYDOS_PREFER_COPY
 * - #ABYDOS_PREFER_TRANSFER
 * - zero for the default
 *
 * and one of the following:
 * - #ABYDOS_IGNORE_ASPECT_RATIO
 * - #ABYDOS_SHRINK_TO_ASPECT_RATIO
 * - #ABYDOS_GROW_TO_ASPECT_RATIO
 * - zero for the default
 */
void abydos_set_options(abydos_t *ar,int options){
    ar->options = options;
}

/* --- loading ---------------------------------------------------------- */

static void
_setup_variants(page_t *page)
{
    unsigned long best_area = 0;
    int i;
    for(i=0;i<page->info.variant_count;++i){
        unsigned long area = page->variant_size[i].width * page->variant_size[i].height;
        if(area > best_area){
            page->default_variant = i;
            best_area = area;
        }
        if(page->variant_size[i].width > page->info.width)
            page->info.width = page->variant_size[i].width;
        if(page->variant_size[i].height > page->info.height)
            page->info.height = page->variant_size[i].height;
    }
}


static void
_setup_page(page_t *page,int num,plugin_t *p,int unsafe,abydos_plugin_handle_t *handle){
    int i;

    /* setup variants */
    if(page->info.variant_count > 1){
        page->variant_size = (abydos_size_t*)malloc(page->info.variant_count*sizeof(abydos_size_t));
        if(ABYDOS_PLUGIN_INTERFACE_HAS(p->vt,get_variant_size) && p->vt->get_variant_size){
            PLUGIN_LOCK_CALL(p,unsafe,get_variant_size,handle,num,page->variant_size);
        }else if(p->vt->get_variant_size_OLD){
            PLUGIN_LOCK_CALL(p,unsafe,get_variant_size_OLD,handle,page->variant_size);
        }else{
            for(i=0;i<page->info.variant_count;++i){
                page->variant_size[i].width = page->info.width;
                page->variant_size[i].height = page->info.height;
            }
            page->default_variant = 0;
            goto variants_done;
        }
        _setup_variants(page);
    }else{
        page->variant_size = NULL;
        page->default_variant = 0;
    }
variants_done:

    /* setup animation */
    page->total_duration = 0;
    if(page->info.frame_count > 1 || page->info.scalable_time){
        page->frame_duration = (double*)malloc(page->info.frame_count*sizeof(double));
        page->frame_time = (double*)malloc(page->info.frame_count*sizeof(double));
        if(ABYDOS_PLUGIN_INTERFACE_HAS(p->vt,get_duration) && p->vt->get_duration){
            PLUGIN_LOCK_CALL(p,unsafe,get_duration,handle,num,page->frame_duration);
        }else if(p->vt->get_duration_OLD){
            PLUGIN_LOCK_CALL(p,unsafe,get_duration_OLD,handle,page->frame_duration);
        }else{
            for(i=0;i<page->info.frame_count;++i){
                page->frame_duration[i] = 50;
                page->frame_time[i] = 50 * i;
            }
            page->total_duration = 50 * page->info.frame_count;
            goto animation_done;
        }
        for(i=0;i<page->info.frame_count;++i){
            if(page->frame_duration[i] < MINIMUM_DURATION)
                page->frame_duration[i] = MINIMUM_DURATION;
            page->frame_time[i] = page->total_duration;
            page->total_duration += page->frame_duration[i];
        }
    }else{
        page->frame_duration = NULL;
        page->frame_time = NULL;
    }
animation_done:
    return;
}


static void
_setup_immutable(immutable_t *im)
{
    int i;
    /* setup pages */
    im->done.page = (page_t*)malloc(im->info.page_count*sizeof(page_t));
    im->done.total_layer_count = 0;
    im->done.square_pixels = 1;
    for(i=0;i<im->info.page_count;++i){
        /* set defaults */
        im->done.page[i].info.version = ABYDOS_PLUGIN_PAGE_INFO_VERSION;
        im->done.page[i].info.width = im->info.width;
        im->done.page[i].info.height = im->info.height;
        im->done.page[i].info.pixel_ratio = im->info.pixel_ratio;
        im->done.page[i].info.layer_count = im->info.layer_count;
        im->done.page[i].info.variant_count = im->info.variant_count;
        im->done.page[i].info.frame_count = im->info.frame_count;
        im->done.page[i].info.default_frame = 0;
        im->done.page[i].info.scalable_size = im->info.scalable_size;
        im->done.page[i].info.scalable_time = im->info.scalable_time;
        /* get page specific data */
        if(ABYDOS_PLUGIN_INTERFACE_HAS(im->plugin->vt,get_page_info) && im->plugin->vt->get_page_info)
            PLUGIN_LOCK_CALL(im->plugin,!im->info.threadsafe,get_page_info,im->handle,i,&im->done.page[i].info);
        im->done.page[i].layer_offset = im->done.total_layer_count;
        if(im->done.page[i].info.layer_count > 1)
            im->done.total_layer_count += im->done.page[i].info.layer_count;
        _setup_page(&im->done.page[i],i,im->plugin,!im->info.threadsafe,im->handle);
        if(im->done.page[i].info.pixel_ratio != 1)
            im->done.square_pixels = 0;
    }
}


void
_abydos_setup(abydos_t *ar)
{
    _setup_immutable(ar->im);
    if(ar->im->done.total_layer_count > 1){
        int i;
        ar->layer_visibility = (cairo_bool_t*)malloc(ar->im->done.total_layer_count * sizeof(cairo_bool_t));
        /* set every layer visible by default */
        for(i=0;i<ar->im->done.total_layer_count;++i)
            ar->layer_visibility[i] = 1;
        if(ABYDOS_PLUGIN_INTERFACE_HAS(ar->im->plugin->vt,get_layer_visibility) && ar->im->plugin->vt->get_layer_visibility){
            /* FIXME!! have a layar visibility array for each page? */
            for(i=0;i<ar->im->info.page_count;++i)
                PLUGIN_LOCK_CALL(ar->im->plugin,!ar->im->info.threadsafe,get_layer_visibility,ar->im->handle,i,&ar->layer_visibility[ar->im->done.page[i].layer_offset]);
        }else if(ar->im->plugin->vt->get_layer_visibility_OLD){
            PLUGIN_LOCK_CALL(ar->im->plugin,!ar->im->info.threadsafe,get_layer_visibility_OLD,ar->im->handle,ar->layer_visibility);
        }
    }else{
        ar->layer_visibility = NULL;
    }
}


/**
 * abydos_from_data
 * @ar: an #abydos_t
 * @data: the data containing the image
 * @len: length of the data
 *
 * Create the image from a single chunk of data. If you prefer to feed the
 * data a little at a time, use progressive interface abydos_load_begin(),
 * abydos_load_feed() and abydos_load_end(). If the operation fails you can
 * call abydos_error() to get an error message. Apart fram that nothing more
 * can be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_from_data(abydos_t *ar, const char *data, size_t len)
{
    immutable_t *im = ar->im;
    if(im->plugin->vt->create_from_data){
        if(PLUGIN_LOCK_CALL_RET(im->plugin,!im->info.threadsafe,create_from_data,int,im->handle,data,len)<0)
            return -1;
    }else if(im->plugin->vt->progressive_begin && im->plugin->vt->progressive_feed && im->plugin->vt->progressive_end){
        PLUGIN_LOCK_CALL(im->plugin,!im->info.threadsafe,progressive_begin,im->handle,NULL,NULL,NULL);
        if(PLUGIN_LOCK_CALL_RET(im->plugin,!im->info.threadsafe,progressive_feed,int,im->handle,data,len) < 0)
            return -1;

        if(PLUGIN_LOCK_CALL_RET(im->plugin,!im->info.threadsafe,progressive_end,int,im->handle)<0)
            return -1;
    }else{
        FILE *f;
        char *tmpfile;
        int res;
        _open_tmpfile(&tmpfile,&f);
        res = fwrite(data,1,len,f) < len ? -1 : 0;
        fclose(f);
        if(res >= 0)
            res = abydos_from_file(ar,tmpfile);
        unlink(tmpfile);
        free(tmpfile);
        if(res < 0)
            return res;
    }
    _abydos_setup(ar);
    return 0;
}


/**
 * abydos_from_file
 * @ar: an #abydos_t
 * @filename: file to load the image from
 *
 * Create the image from a file. You may want to use the convenience
 * function abydos_create_from_file(), which is identical to calling
 * abydos_create() followed by abydos_from_file(). If the operation fails
 * you can call abydos_error() to get an error message. Apart fram that
 * nothing more can be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_from_file(abydos_t *ar, const char *filename)
{
    immutable_t *im = ar->im;
    struct stat st;
    int fd;
    void *data;
    int res;

    if(im->plugin->vt->create_from_file){
        if(PLUGIN_LOCK_CALL_RET(im->plugin,!im->info.threadsafe,create_from_file,int,im->handle,filename)<0)
            return -1;
        _abydos_setup(ar);
        return 0;
    }

    if(stat(filename,&st)<0){
        im->info.error = strerror(errno);
        return -1;
    }
    fd = open(filename,O_RDONLY);
    if(fd<0){
        im->info.error = strerror(errno);
        return -1;
    }
    data = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE | MAP_POPULATE, fd, 0);
    if(MAP_FAILED == data){
        close(fd);
        im->info.error = strerror(errno);
        return -1;
    }
    res = abydos_from_data(ar, (const char*)data, st.st_size);
    munmap(data,st.st_size);
    close(fd);
    return res;
}


static void
_on_info(void *userdata)
{
    abydos_t *ar = (abydos_t*)userdata;
    _abydos_setup(ar);
    if(ar->im->load.info_func)
        ar->im->load.info_func(ar->im->load.userdata,ar);
}


static void
_on_update(void *userdata, cairo_rectangle_int_t *r)
{
    abydos_t *ar = (abydos_t*)userdata;
    if(ar->im->load.update_func)
        ar->im->load.update_func(ar->im->load.userdata,ar,r);
}


/**
 * abydos_load_begin
 * @ar: an #abydos_t
 * @info_func: a function to be called as soon as abydos_get_info() can be
 *             called, or NULL
 * @update_func: a function to be called every time the image has been
 *               updated, or NULL
 * @userdata: user data to be passed to @info_func and @update_func
 *
 * Begin progressive loading of an image. If you pass @info_func and
 * @update_func they are guaranteed to be called before abydos_load_end()
 * returns successfully. @info_func() will be called first and exactly once.
 * @update_func() will be called at least once. If abydos_load_feed() or
 * abydos_load_end() fails @info_func() and @update_func() may or may not be
 * called.
 **/
void
abydos_load_begin(abydos_t *ar,abydos_info_callback_t info_func,abydos_update_callback_t update_func,void *userdata)
{
    immutable_t *im = ar->im;
    im->load.info_func = info_func;
    im->load.update_func = update_func;
    im->load.userdata = userdata;
    if(im->plugin->vt->progressive_begin && im->plugin->vt->progressive_feed && im->plugin->vt->progressive_end){
        PLUGIN_LOCK_CALL(im->plugin,!im->info.threadsafe,progressive_begin,im->handle,_on_info,_on_update,ar);
    }else if(im->plugin->vt->create_from_data){
        im->load.f = open_memstream(&im->load.buffer,&im->load.buffer_len);
    }else{
        _open_tmpfile(&im->load.tmpfile,&im->load.f);
    }
}


/**
 * abydos_load_feed
 * @ar: an #abydos_t
 * @data: the data containing a part of the image
 * @len: length of the data
 *
 * Feed bytes of the image. If the operation fails you can call
 * abydos_error() to get an error message. Apart from that nothing more can
 * be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 */
int
abydos_load_feed(abydos_t *ar, const char* data, size_t len)
{
    immutable_t *im = ar->im;
    if(im->info.error)
        return -1;

    if(im->plugin->vt->progressive_begin && im->plugin->vt->progressive_feed && im->plugin->vt->progressive_end){
        return PLUGIN_LOCK_CALL_RET(im->plugin,im->info.threadsafe,progressive_feed,int,im->handle,data,len);
    }else{
        return fwrite(data,1,len,im->load.f) < len ? -1 : 0;
    }
    return -1;
}


/**
 * abydos_load_end
 * @ar: an #abydos_t
 *
 * End progressive loading of an image. If the operation fails you can call
 * abydos_error() to get an error message. Apart fram that nothing more can
 * be done with @ar and it must be destroyed.
 *
 * returns: 0 on success. Otherwise the operation has failed.
 **/
int
abydos_load_end(abydos_t *ar)
{
    immutable_t *im = ar->im;
    if(im->info.error)
        return -1;

    if(im->plugin->vt->progressive_begin && im->plugin->vt->progressive_feed && im->plugin->vt->progressive_end){
        if(im->plugin->vt->progressive_end(im->handle)<0)
            return -1;
        _abydos_setup(ar);
    }else{
        int res;
        fclose(im->load.f);
        if(im->plugin->vt->create_from_data){
            res = abydos_from_data(ar,im->load.buffer,im->load.buffer_len);
            free(im->load.buffer);
        }else{
            res = abydos_from_file(ar,im->load.tmpfile);
            unlink(im->load.tmpfile);
            free(im->load.tmpfile);
        }
        if(res < 0)
            return res;
        if(im->load.info_func)
            im->load.info_func(im->load.userdata,ar);
        if(im->load.update_func){
            cairo_rectangle_int_t rect;
            rect.x = 0;
            rect.y = 0;
            _get_max_size_int(ar, &rect.width, &rect.height);
            im->load.update_func(im->load.userdata,ar,&rect);
        }
    }
    return 0;
}

/* --- info ------------------------------------------------------------- */

void
abydos_get_info_V(abydos_t *ar,abydos_info_t *info,int info_version)
{
    immutable_t *im = ar->im;
    page_t *page = _get_page(im,ar->page);
    _abydos_page_get_size_int(page, &info->width, &info->height, ar->options);
    info->features = (im->done.square_pixels?ABYDOS_FEATURE_SQUARE_PIXELS:0)
                    |(im->info.scalable_size?ABYDOS_FEATURE_SCALABLE_SIZE:0)
                    |(im->info.scalable_time?ABYDOS_FEATURE_SCALABLE_TIME:0);
    info->pixel_ratio = im->info.pixel_ratio;
}


/**
 * abydos_get_size
 * @ar: an #abydos_t
 * @height: where to store the width (or NULL)
 * @width: where to store the height (or NULL)
 *
 * Get the size of a loaded image. Note that this is the size of the image
 * as rendered and may be affected by the currently set options
 * (see abydos_set_options()) in the case pixels are not square.
 */
void
abydos_get_size(abydos_t *ar,int *width,int *height)
{
    page_t *page = _get_page(ar->im,ar->page);
    _abydos_page_get_size_int(page, width, height, ar->options);
}


/**
 * abydos_get_max_size
 * @ar: an #abydos_t
 * @width: where to store the width (or NULL)
 * @height: where to store_the height (or NULL)
 *
 * Get the maximum size of a loaded image, a size large enough to hold any
 * variant of any page.
 */
void
abydos_get_max_size(abydos_t *ar,int *width,int *height)
{
    _get_max_size_int(ar, width, height);
}


/**
 * abydos_error
 * @ar: an #abydos_t
 *
 * If an error has occured, this function can be used to get a description
 * of the error.
 *
 * returns: A string describing an error or NULL.
 */
const char *
abydos_error(abydos_t *ar)
{
    return ar->im->info.error;
}

/* --- drawing ---------------------------------------------------------- */

void
_abydos_render_internal(immutable_t *im,cairo_t *cr,page_t *page,int pagenr,cairo_bool_t *visible_layer,int variant,double time,int frame)
{
    static cairo_bool_t visible = 1;
    /* Set the only layer as visible */
    if(!visible_layer)
        visible_layer = &visible;
    if(variant<0 && page->info.variant_count>1){
        cairo_matrix_t matrix;
        double scale;
        int best_variant = page->default_variant;
        double best_scale = 1;
        int i;
        cairo_get_matrix(cr,&matrix);
        scale = MAX(sqrt(matrix.xx*matrix.xx+matrix.yx*matrix.yx),
                    sqrt(matrix.xy*matrix.xy+matrix.yy*matrix.yy));
        DEBUG_p("scale: %f",scale);
        for(i=0;i<page->info.variant_count;++i){
            double this_scale_x = (double)page->variant_size[i].width/page->info.width;
            double this_scale_y = (double)page->variant_size[i].height/page->info.height;
            double this_scale = MIN(this_scale_x,this_scale_y);
            DEBUG_p("trying %f",this_scale);
            if(this_scale >= scale && this_scale < best_scale){
                DEBUG_p("this is better");
                best_variant = i;
                best_scale = this_scale;
            }
        }
        variant = best_variant;
        cairo_scale(cr,1/best_scale,1/best_scale);
    }
    if(variant>=page->info.variant_count)
        variant = page->info.variant_count-1;
    if(variant<0)
        variant = 0;
    if(im->plugin->vt->render){
        PLUGIN_LOCK_CALL(im->plugin,!im->info.threadsafe_render,render,im->handle,cr,pagenr,visible_layer,variant,time,frame);
    }else{
        cairo_surface_t *src = PLUGIN_LOCK_CALL_RET(im->plugin,!im->info.threadsafe_render,get_image_surface,cairo_surface_t*,im->handle,pagenr,visible_layer,variant,time,frame);
        cairo_pattern_t *pattern = cairo_pattern_create_for_surface(src);
        cairo_set_source(cr, pattern);
        cairo_pattern_set_filter(pattern, CAIRO_FILTER_BEST);
        cairo_paint(cr);
        cairo_pattern_destroy(pattern);
        cairo_surface_destroy(src);
    }
}


/**
 * abydos_render
 * @ar: an #abydos_t
 * @cr: a cairo context
 *
 * Render the image to a #cairo_t. Use ordinary cairo trasformations to
 * scale the image. The result may also be affected by the currently set
 * options (see abydos_set_options()) if the pixels are not square.
 */
void
abydos_render(abydos_t *ar,cairo_t *cr)
{
    immutable_t *im = ar->im;
    page_t *page = _get_page(im,ar->page);
    double width;
    double height;
    _get_page_size(page, &width, &height, ar->options);
    cairo_save(cr);
    if(width != page->info.width || height != page->info.height)
        cairo_scale(cr,width/page->info.width,height/page->info.height);
    _abydos_render_internal(ar->im, cr, page, ar->page, ar->layer_visibility, ar->prefered_variant, ar->time, ar->frame);
    cairo_restore(cr);
}


cairo_surface_t *
_abydos_get_image_surface_internal(immutable_t *im_,
                                   int flags,
                                   int pagenr,
                                   cairo_bool_t *visible_layer,
                                   int variant,
                                   double time,
                                   int frame)
{
    static cairo_bool_t visible = 1;
    page_t *page = _get_page(im_,pagenr);
    cairo_t *cr;
    double width;
    double height;
    /* Set the only layer as visible */
    if(!visible_layer)
        visible_layer = &visible;
    /* Always fall back to the default variant. */
    if(variant<0)
        variant = page->default_variant;
    _get_page_size(page, &width, &height, flags);
    if(width==page->info.width && height==page->info.height)
        if(im_->plugin->vt->get_image_surface)
            return PLUGIN_LOCK_CALL_RET(im_->plugin,!im_->info.threadsafe_render,get_image_surface,cairo_surface_t*,im_->handle,pagenr,visible_layer,variant,time,frame);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    cr = cairo_create(surface);
    if(width != page->info.width || height != page->info.height)
        cairo_scale(cr,width/page->info.width,height/page->info.height);
    _abydos_render_internal(im_,cr,page,pagenr,visible_layer,variant,time,frame);
    cairo_destroy(cr);
    return surface;
}


/**
 * abydos_get_image_surface
 * @ar: an #abydos_t
 *
 * Get a #cairo_surface_t containing the image. This may or may not be a
 * reference to an internally cached image surface. If you don't intend to
 * modify it or if you intend to get it and modify it only once, then this
 * is no problem. But you intend to get several copies and modify them you
 * need to pass one of the flags %ABYDOS_PREFER_COPY or
 * %ABYDOS_PREFER_TRANSFER. The @flags may also affect the size of the image
 * if it has non square pixels. Valid flags are one of:
 * - %ABYDOS_PREFER_SHARE
 * - %ABYDOS_PREFER_COPY
 * - %ABYDOS_PREFER_TRANSFER
 *
 * and one of:
 * - %ABYDOS_IGNORE_ASPECT_RATIO
 * - %ABYDOS_SHRINK_TO_ASPECT_RATIO
 * - %ABYDOS_GROW_TO_ASPECT_RATIO
 *
 * The default is %ABYDOS_PREFER_SHARE and %ABYDOS_GROW_TO_ASPECT_RATIO, but
 * it may change in the future.
 *
 * For the common simple cases you might want to use the convinience
 * function abydos_load() to quickly get an image surface from a file.
 *
 * returns: If the image loading has suceeded a valid @cairo_surface_t will
 *          be returned. Call cairo_surface_destroy() then you are done
 *          using it. If the image loading hasn't succeeded the return value
 *          is undefined.
 */
cairo_surface_t *
abydos_get_image_surface(abydos_t *ar)
{
    cairo_surface_t *surface;
    if(ar->cached_surface)
        return cairo_surface_reference(ar->cached_surface);
    surface = _abydos_get_image_surface_internal(ar->im,ar->options,ar->page,ar->layer_visibility,ar->prefered_variant,ar->time,ar->frame);
    if(!ar->im->plugin->vt->get_image_surface)
        ar->cached_surface = cairo_surface_reference(surface);
    return surface;
}

/* --- multi page ------------------------------------------------------- */
/**
 * SECTION: abydos-pages
 * @short_description: Functions for handling images with multiple pages.
 * @title: Pages
 * @include: abydos.h
 *
 * Some image formats can caontain multiple pages. The definition of pages
 * in this context is if the image contains more than one sub image, which
 * are not frames of an animation, layers of the same image nor the same
 * image in different sizes. It could to subimages which are defined as
 * pages by the format (such as in the case of PDF documents) but it could
 * also be formats that simply stores severaral images together in a file in
 * a more loosely defined manner. Some formats may have an internal
 * organization of the sub images in the form of a tree. For example icons
 * formats can contain several icons, each with it's own set of states.
 * But abydos will in that case always flatten the tree and index the pages
 * sequentially.
 **/

/**
 * abydos_get_page_count
 * @ar: an #abydos_t
 *
 * Get the number of pages of an image.
 **/
int
abydos_get_page_count(abydos_t *ar)
{
    return ar->im->info.page_count;
}


/**
 * abydos_get_page
 * @ar: an #abydos_t
 *
 * Returns the index of the currently selected page.
 *
 * returns: The currently selected page, starting from 0.
 **/
int
abydos_get_page(abydos_t *ar)
{
    return ar->page;
}


/**
 * abydos_set_page
 * @ar: an #abydos_t
 * @page: Page starting at 0
 *
 * Selects a page. If the page is out of range, a valid, but otherwise
 * undefined page will be selected.
 */
void
abydos_set_page(abydos_t *ar,int page)
{
    if(page<0){
        page = 0;
    }else if(page >= ar->im->info.page_count){
        page = ar->im->info.page_count - 1;
    }
    ar->page = page;
}


/**
 * abydos_set_page_default
 * @ar: an #abydos_t
 *
 * Selectes the page that is defined as the default page (or the first page
 * if no default page is defined). Call this function if you intend to show
 * only one page. (It is currently assumed that you are going to show all
 * one by one so the first page is initially selected, but that may change.)
 */
void abydos_set_page_default(abydos_t *ar)
{
    ar->page = 0;
}

/* --- layers ----------------------------------------------------------- */
/**
 * SECTION: abydos-layers
 * @short_description: Functions for handling image layers.
 * @title: Layers
 * @include: abydos.h
 *
 * Some formats may compose an image out of several layers. It could simply
 * be a set of partly transparent pixmaps placed on top of each other. But
 * layers could also be more like filter operations. Sometimes the internal
 * organization of the layers may be in the form of a tree. But abydos will
 * in that case always flatten the tree and index the layers sequentially.
 **/

/**
 * abydos_get_layer_count
 * @ar: an #abydos_t
 *
 * Get the number of layers of the image.
 */
int
abydos_get_layer_count(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    return page->info.layer_count;
}


/**
 * abydos_get_layer_visibility
 * @ar: an #abydos_t
 * @layer: layer number starting from 0
 *
 * Check if a layer is currently visible.
 *
 * returns: True if the layer is visible
 */
cairo_bool_t
abydos_get_layer_visibility(abydos_t *ar,int layer)
{
    page_t *page = _get_page(ar->im,ar->page);
    if(layer < 0 || layer >= page->info.layer_count)
        return 0;
    if(!ar->layer_visibility)
        return 1;
    return ar->layer_visibility[page->layer_offset + layer];
}


/**
 * abydos_set_layer_visibility
 * @ar: an #abydos_t
 * @layer: layer number staring from 0
 * @visibility: True show the layer, false to hide it
 *
 * Sets the visibility of a layer. Currently this has no effect for images
 * with only one layer.
 */
void
abydos_set_layer_visibility(abydos_t *ar, int layer, cairo_bool_t visible)
{
    page_t *page = _get_page(ar->im,ar->page);
    if(layer < 0 || layer >= page->info.layer_count || !ar->layer_visibility)
        return;
    ar->layer_visibility[page->layer_offset + layer] = visible;
}

/* --- variants --------------------------------------------------------- */
/**
 * SECTION: abydos-variants
 * @short_description: Functions for handling different size variants.
 * @title: Variants
 * @include: abydos.h
 *
 * Some formats may contain different varaints of the same image. The
 * definition of variant in this context is different versions of the same
 * image in different sizes. It could be embedded thumbnails. The thumbnail
 * might not just be a scaled down version of the full size image, it could
 * be some intelligent cropping involved. Or it could be a set of icons with
 * different sizes and different detail level. Or it could be mipmap
 * textures.
 **/

/**
 * abydos_get_variant_count
 * @ar: an #abydos_t
 *
 * Get the number of variants of the image.
 *
 * returns: The number of variants.
 **/
int
abydos_get_variant_count(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    return page->info.variant_count;
}


/**
 * abydos_get_variant_size
 * @ar: an #abydos_t
 * @variant: variant number starting from 0
 * @width: where to store the width (or NULL)
 * @height: where to store the height (or NULL)
 *
 * Get the size of a certain variant.
 */
void
abydos_get_variant_size(abydos_t *ar,int variant,int *width, int *height)
{
    page_t *page = _get_page(ar->im,ar->page);
    if(page->info.variant_count > 1){
        variant %= page->info.variant_count;
        if(width)
            *width = page->variant_size[variant].width;
        if(height)
            *height = page->variant_size[variant].height;
    }else{
        if(width)
            *width = page->info.width;
        if(height)
            *height = page->info.height;
    }
}


int abydos_get_variant_at_size_min(abydos_t *ar,int width, int height)
{
    /* FIXME!! */
    return 0;
}


int
abydos_get_variant_at_size_max(abydos_t *ar,int width, int height)
{
    /* FIXME!! */
    return 0;
}


/**
 * abydos_set_variant
 * @ar: an #abydos_t
 * @variant: variant number starting from 0
 *
 * Force the use of a specified variant. If the variant is out of range,
 * a valid, but otherwise unspecified variant will be used.
 */
void
abydos_set_variant(abydos_t *ar,int variant)
{
    ar->prefered_variant = variant>=0 ? variant%ar->im->info.variant_count : -1;
}


void
abydos_prefer_variant(abydos_t *ar,int width,int height)
{
    /* FIXME!! */
}

/* --- animation -------------------------------------------------------- */
/**
 * SECTION: abydos-animation
 * @short_description: Functions for handling animation.
 * @title: Animation
 * @include: abydos.h
 *
 * Two approaches can be used for handling animations. Either by frame or by
 * time. The latter is usually more convinient for displaying the animation.
 * Abydos can tell you for how long you should display a frame, or how long
 * you should wait before updating the time of the animation and redraw it.
 * But measuring time is beyond the scope of abydos. Basically you do
 * something like this.
 *|[
    last_time = now();
    for(;;){
        abydos_render(ar,cr,0);
        delay(abydos_get_delay(ar));
        this_time = now();
        abydos_add_time(ar,this_time - last_time)
        last_time = this_time;
    }
 *]|
 * Of course you don't want to use a simple delay in a real application.
 * Instead you should use some kind of event scheduler to get a callback.
 * Please also note that if you need to round the delay time you get from
 * abydos_get_delay() you must round it up. Otherwise the time of the
 * animation will be set to right before the frame change (and the next time
 * it will be rounded down to zero and you will get nowere).
 **/

/**
 * abydos_get_frame_count
 * @ar: an #abydos_t
 *
 * Get the number of frames of an animation. Some animations may not be
 * based around the concept of frames but are continous to their nature. In
 * this case %ABYDOS_FEATURE_SCALABLE_TIME will be set and frames will be
 * placed at reasonable intervals. It is also possible that
 * %ABYDOS_FEATURE_SCALABLE_TIME is set for singe frame images (and duration
 * of zero). In that case the image will still be animated somehow, but
 * depending on some external variable (such as the clock time).
 *
 * retutns: The number of frames.
 **/
int
abydos_get_frame_count(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    return page->info.frame_count;
}


/**
 * abydos_get_total_duration
 * @ar: an #abydos_t
 *
 * Get the total duration of an animation.
 *
 * returns: The total duration in seconds.
 */
double
abydos_get_total_duration(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    return page->total_duration;
}


/**
 * abydos_get_frame_at_time
 * @ar: an #abydos_t
 * @sec: time in seconds
 *
 * Get the index if the frame at @sec. Note that this cannot simply be
 * calculated by dividing @sec with the duration of the first frame, since
 * frames may have different duration.
 *
 * returns: The frame index starting at 0.
 **/
int
abydos_get_frame_at_time(abydos_t *ar,double sec)
{
    page_t *page = _get_page(ar->im,ar->page);
    int frame;
    sec = fmod(sec,page->total_duration);
    for(frame=0;page->frame_time[frame]+page->frame_duration[frame] <= sec && frame < page->info.frame_count;++frame);
    return frame;
}


/**
 * abydos_get_frame_duration
 * @ar: an #abydos_t
 * @frame: frame index starting from 0
 *
 * Get the time (into the animation) at which @frame should appear.
 *
 * returns: The time in seconds.
 **/
double
abydos_get_frame_duration(abydos_t *ar,int frame)
{
    page_t *page = _get_page(ar->im,ar->page);
    return page->frame_duration[frame];
}


/**
 * abydos_get_frame
 * @ar: an #abydos_t
 *
 * Get the index of the current frame.
 *
 * returns: The frame index staring at 0.
 **/
int
abydos_get_frame(abydos_t *ar)
{
    return ar->frame;
}


/**
 * abydos_get_time
 * @ar: an #abydos_t
 *
 * Get the current time into the animation.
 *
 * returns: The time in seconds.
 **/
double
abydos_get_time(abydos_t *ar)
{
    return ar->time;
}


/**
 * abydos_set_frame
 * @ar: an #abydos_t
 * @frame: frame index starting at 0
 *
 * Select @frame which will also implicitly set the time of the animation to
 * the time for this frame.
 **/
void
abydos_set_frame(abydos_t *ar,int frame)
{
    page_t *page = _get_page(ar->im,ar->page);
    if(page->info.frame_count <= 1)
        return;
    frame %= page->info.frame_count;
    ar->time = page->frame_time[frame];
    ar->frame = frame;
}


/**
 * abydos_set_frame_default
 * @ar: an #abydos_t
 *
 * Selectes the frame that is defined as the default frame (or the first
 * frame if no default frame is defined). Call this function if you intend
 * to show only one frame. (It is currently assumed that you are going to
 * show the animation from beginning to end so the first frame is initially
 * selected, but that may change.)
 */
void
abydos_set_frame_default(abydos_t *ar)
{
    ar->time = 0;
    ar->frame = 0;
}


/**
 * abydos_set_time
 * @ar: an #abydos_t
 * @sec: time in seconds
 *
 * Set the time of the animation in seconds. It will also implicitly set the
 * corresponding frame.
 *
 * returns: True if the the frame has changed.
 **/
cairo_bool_t
abydos_set_time(abydos_t *ar,double sec)
{
    page_t *page;
    int new_frame;
    cairo_bool_t res;
    if(!abydos_can_animate(ar))
        return 0;
    page = _get_page(ar->im,ar->page);
    /* Add half a microsecond to compensate for rounding errors causing the
     * call to be made just a little bit too early. */
    sec += 0.0000005;
    DEBUG_p("requested time %f",sec);
    sec = fmod(sec,page->total_duration);
    DEBUG_p("current time %f, total duration = %f",sec,page->total_duration);
    new_frame = abydos_get_frame_at_time(ar,sec);
    DEBUG_p("old frame %d, new frame %d",ar->frame,new_frame);
    ar->time = sec;
    res = ar->frame != new_frame || page->info.scalable_time;
    ar->frame = new_frame;
    return res;
}


/**
 * abydos_next_frame
 * @ar: an #abydos_t
 *
 * Advance the animation one frame and increse the time accordingly.
 **/
void
abydos_next_frame(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    if(++ar->frame >= page->info.frame_count)
        ar->frame = 0;
    ar->time = page->frame_time[ar->frame];
}


/**
 * abydos_add_time
 * @ar: an #abydos_t
 * @sec: time in seconds
 *
 * Advance the animation @sec seconds. It will also implicitly set the
 * corresponding frame.
 *
 * returns: True if the the frame has changed.
 */
cairo_bool_t
abydos_add_time(abydos_t *ar,double sec)
{
    DEBUG_p("adding %f seconds",sec);
    return abydos_set_time(ar,ar->time+sec);
}


/**
 * abydos_get_delay
 * @ar: an #abydos_t
 *
 * Get the amount of time until the next frame in seconds. Note that if this
 * value needs to be rounded for some reason, it must be rounded up
 * (otherwise the next time advancement will land right before the frame
 * change).
 *
 * returns: Time in seconds.
 **/
double
abydos_get_delay(abydos_t *ar)
{
    page_t *page = _get_page(ar->im,ar->page);
    double sec;
    double delay;
    if(page->info.frame_count <= 1 && !ar->im->info.scalable_time)
        return -1;
    sec = ar->time - page->frame_time[ar->frame];
    DEBUG_p("current time : %f",ar->time);
    DEBUG_p("current_frame: %d",ar->frame);
    DEBUG_p("current_frame time: %f",page->frame_time[ar->frame]);
    DEBUG_p("current_frame duration: %f",page->frame_duration[ar->frame]);
    DEBUG_p("%f sec into current frame (%d)",sec,ar->frame);
    delay = page->frame_duration[ar->frame] - sec;
    if(ar->im->info.scalable_time && delay>MINIMUM_DURATION)
        delay = MINIMUM_DURATION;
    DEBUG_p("%f sec left of frame duration",delay);
    return delay;
}
