/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memset() */
#include <png.h>

#include "debug.h"
#include "nile-blit.h"
#include "nil-png.h"

//#define PNG_COLOR_FORMAT NILE_FORMAT_r8g8b8

static int
double_to_frac(double d, int *r_num, int *r_den)
{
    int i;
    for(i=1;i<10000;++i){
        double num = d*i;
        if(num - (int)num < 0.0000000001){
            *r_num = num;
            *r_den = i;
            return 1;
        }
    }
    return 0;
}

#if 0
static int
_convert_colors(png_color *rgb, uint8_t *alpha, const uint8_t *src, nile_format_t src_format, int src_bpc, int count)
{
    int o_r;
    int o_g;
    int o_b;
    int o_a;
    int i;
    if(src_bpc > 4)
        return -1;
    o_r = NILE_OFFSET_RED(src_format);
    o_g = NILE_OFFSET_GREEN(src_format);
    o_b = NILE_OFFSET_BLUE(src_format);
    o_a = NILE_OFFSET_ALPHA(src_format);
    for(i=0;i<count;++i){
        rgb[i].red = src[o_r];
        rgb[i].green = src[o_g];
        rgb[i].blue = src[o_b];
        if(NILE_HAS_ALPHA(src_format)){
            alpha[i] = src[o_a];
        }else if(alpha){
            alpha[i] = 255;
        }
        src += src_bpc;
    }
    return 0;
}
#endif


static int
_blit_alpha(uint8_t *alpha, const uint8_t *src, int bpc, nile_format_t src_format, int count)
{
    int o_a = NILE_OFFSET_ALPHA(src_format);
    while(count-->0){
        *alpha++ = src[o_a];
        src += bpc;
    }
    return 0;
}


static int
_write_indexed(nile_t *ni,png_structp png,png_infop info){
    png_color png_palette[256];
    uint8_t png_alpha[256];

    png_set_IHDR(png,info,
                 ni->width,
                 ni->height,
                 8,PNG_COLOR_TYPE_PALETTE,
                 PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_DEFAULT,
                 PNG_FILTER_TYPE_DEFAULT);

#if 1
    {
        nile_blit_t blit;
        nile_blit_init_rgb(&blit, 3, NILE_FORMAT_r8g8b8, ni->bpc, ni->format);
        nile_blit_line(&blit, (uint8_t*)png_palette, ni->palette, ni->colors);
        if(NILE_HAS_ALPHA(ni->format))
            _blit_alpha(png_alpha, ni->palette, ni->bpc, ni->format, ni->colors);
    }
#else
    _convert_colors(png_palette,png_alpha,ni->palette,ni->format,ni->bpc,ni->colors);
#endif
    png_set_PLTE(png,info,png_palette,ni->colors);
    if(NILE_HAS_ALPHA(ni->format))
        png_set_tRNS(png,info,png_alpha,ni->colors,NULL);

    png_write_info(png,info);
    {
        png_byte** rows = malloc(ni->height*sizeof(png_byte*));
        int stride = nile_get_stride(ni);
        int i;
        for(i=0;i<ni->height;++i)
            rows[i] = ni->pixels + i*stride;
        png_write_image(png, rows);
        free(rows);
    }
    return 0;
}


static int
_get_color_type(nile_t *ni, int *depth, int *transform)
{
    *depth = 8;
    *transform = PNG_TRANSFORM_IDENTITY;
    switch(ni->format){
    case NILE_FORMAT_r8g8b8:
        if(ni->bpc > 3)
            *transform |= PNG_TRANSFORM_STRIP_FILLER_AFTER;
        return PNG_COLOR_TYPE_RGB;
    case NILE_FORMAT_b8g8r8:
        *transform |= PNG_TRANSFORM_BGR;;
        if(ni->bpc > 3)
            *transform |= PNG_TRANSFORM_STRIP_FILLER_AFTER;
        return PNG_COLOR_TYPE_RGB;
    case NILE_FORMAT_x8r8g8b8:
        *transform |= PNG_TRANSFORM_STRIP_FILLER_BEFORE;
        return PNG_COLOR_TYPE_RGB;
    case NILE_FORMAT_x8b8g8r8:
        *transform |= PNG_TRANSFORM_BGR | PNG_TRANSFORM_STRIP_FILLER_BEFORE;
        return PNG_COLOR_TYPE_RGB;
    case NILE_FORMAT_b8g8r8a8:
        *transform |= PNG_TRANSFORM_BGR;;
        return PNG_COLOR_TYPE_RGBA;
    case NILE_FORMAT_r8g8b8a8:
        return PNG_COLOR_TYPE_RGBA;
    case NILE_FORMAT_a8r8g8b8:
        *transform |= PNG_TRANSFORM_SWAP_ALPHA;
        return PNG_COLOR_TYPE_RGBA;
    case NILE_FORMAT_a8b8g8r8:
        *transform |= PNG_TRANSFORM_BGR | PNG_TRANSFORM_SWAP_ALPHA;
        return PNG_COLOR_TYPE_RGBA;
    default:
        return -1;
    }
}


static int
_write_direct(nile_t *ni, png_structp png, png_infop info){
    int color_type;
    int depth;
    int transform;
    if(!ni->pixels)
        return -1;

    color_type = _get_color_type(ni, &depth, &transform);
    if(color_type >= 0){
        const uint8_t *src = ni->pixels;
        int src_stride = nile_get_stride(ni);
        uint8_t **rows = malloc(ni->height * sizeof(uint8_t*));
        int i;

        png_set_IHDR(png, info,
                     ni->width,
                     ni->height,
                     depth, color_type,
                     PNG_INTERLACE_NONE,
                     PNG_COMPRESSION_TYPE_DEFAULT,
                     PNG_FILTER_TYPE_DEFAULT);
        png_write_info(png,info);
        if(transform & PNG_TRANSFORM_STRIP_FILLER_BEFORE)
            png_set_filler(png, 0, PNG_FILLER_BEFORE);
        if(transform & PNG_TRANSFORM_STRIP_FILLER_AFTER)
            png_set_filler(png, 0, PNG_FILLER_AFTER);
        if(transform & PNG_TRANSFORM_BGR)
            png_set_bgr(png);
        if(transform & PNG_TRANSFORM_SWAP_ALPHA)
            png_set_swap_alpha(png);
        for(i=0;i<ni->height;++i){
            rows[i] = (uint8_t*)src;
            src += src_stride;
        }
        png_write_image(png, rows);
        free(rows);
        return 0;
    }else{
        nile_t dst;
        int res;
        nile_init(&dst);
#if 1
        nile_copy_identity(&dst, ni);
        if(NILE_HAS_ALPHA(ni->format)){
            dst.bpc = 4;
            dst.format = NILE_FORMAT_r8g8b8a8;
        }else{
            dst.bpc = 4;
            dst.format = NILE_FORMAT_r8g8b8;
        }
        nile_convert(&dst, ni, 0);
#else
        if(NILE_HAS_ALPHA(ni->format)){
            dst.bpc = 4;
            dst.format = NILE_FORMAT_r8g8b8a8;
        }else{
            dst.bpc = 4;
            dst.format = NILE_FORMAT_r8g8b8;
        }
        nile_ensure_direct(&dst, ni->width, ni->height, dst.bpc);
        nile_blit_direct_rgb(dst.pixels, nile_get_stride(&dst), dst.bpc, dst.format,
                             ni->pixels, nile_get_stride(ni), ni->bpc, ni->format,
                             dst.width, dst.height, 0);
#endif
        res = _write_direct(&dst, png, info);
        nile_done(&dst);
        return res;
    }
}


int
nil_png_write(nile_t *ni, double pixel_ratio, FILE *f)
{
    png_structp png;
    png_infop info;
    int res;

    png = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
    if(!png)
        return -1;
    info = png_create_info_struct(png);
    if(!info)
        goto out;
    png_init_io(png,f);
    {
        int num;
        int den;
        if(double_to_frac(pixel_ratio, &num, &den)){
            DEBUG_p("pixel ratio: %d/%d",num,den);
            if(num != den)
                png_set_pHYs(png,info,den,num,0);
        }
    }

    switch(ni->type){
    case NILE_TYPE_NONE:
        goto out;
    case NILE_TYPE_INDEXED:
        res = _write_indexed(ni,png,info);
        break;
    case NILE_TYPE_DIRECT:
        res = _write_direct(ni,png,info);
        break;
    case NILE_TYPE_BITPLANES:
        {
            nile_t tmp;
            nile_init(&tmp);
            nile_copy_identity(&tmp, ni);
            nile_copy_palette(&tmp, ni, 0);
            tmp.type = NILE_TYPE_INDEXED;
            nile_convert(&tmp, ni, 0);
            res = _write_indexed(&tmp, png, info);
            nile_done(&tmp);
        }
        break;
    case NILE_TYPE_BITPLANES_DIRECT:
        {
            nile_t tmp;
            nile_init(&tmp);
            nile_copy_identity(&tmp, ni);
            tmp.type = NILE_TYPE_DIRECT;
            nile_convert(&tmp, ni, 0);
            res = _write_direct(&tmp, png, info);
            nile_done(&tmp);
        }
    }

    if(0==res)
        png_write_end(png,info);
out:
    png_destroy_write_struct(&png,&info);
    return res;
}


int
nil_png_save(nile_t *ni, double pixel_ratio, const char *filename)
{
    FILE* f = fopen(filename,"wb");
    int res;
    if(!f)
        return -1;
    res = nil_png_write(ni, pixel_ratio, f);
    fclose(f);
    return res;
}
