/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <string.h>                     /* memset() */

#include "nile-blit.h"


static void
_blit_from_bitplanes(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    int bit;
    while(count-->0){
        uint32_t pix;
        int i;
        pix = 0;
        for(i=0; i<blit->src_depth ;++i)
            if(src[i][offs] & (128>>bit))
                pix |= 1<<i;
        pix = htobe32(pix);
        memcpy(dst, ((uint8_t*)&pix)+4-blit->dst_bpc, blit->dst_bpc);
        dst += blit->dst_bpc;
        if(++bit>=8){
            ++offs;
            bit = 0;
        }
    }
}

nile_blit_t *nile_blit_init_from_bitplanes_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, int depth, nile_format_t src_format)
{
    blit->dst_o[0] = NILE_OFFSET_CHANNEL(dst_format,0);
    blit->dst_o[1] = NILE_OFFSET_CHANNEL(dst_format,1);
    blit->dst_o[2] = NILE_OFFSET_CHANNEL(dst_format,2);
    blit->dst_o[4] = NILE_OFFSET_ALPHA(dst_format);
    blit->dst_bpc = dst_bpc;
    blit->src_depth = depth;
    blit->blit_from_bitplanes_func = _blit_from_bitplanes;
    return blit;
}


static void
_blit_from_bitplanes_unindex(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    int bit = 0;
    while(count-->0){
        int pix;
        int i;
        pix = 0;
        for(i=0; i<blit->src_depth;++i)
            if(src[i][offs] & (128>>bit))
                pix |= 1<<i;
        memcpy(dst, blit->palette + pix * blit->dst_bpc, blit->dst_bpc);
        dst += blit->dst_bpc;
        if(++bit >= 8){
            ++offs;
            bit = 0;
        }
    }
}

nile_blit_t *
nile_blit_init_from_bitplanes_unindex_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(nile_blit_setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format) < 0)
        return NULL;
    blit->dst_bpc = dst_bpc;
    blit->src_depth = depth;
    blit->blit_from_bitplanes_func = _blit_from_bitplanes_unindex;
    return blit;
}



void
nile_blit_from_bitplanes(nile_blit_t *blit, uint8_t *dst, int dst_stride, uint8_t **src, int src_stride, int width, int height, int flags)
{
    off_t offs = 0;
    if(flags & NILE_FLIPPED){
        dst += dst_stride * (height-1);
        dst_stride *= -1;
    }
    while(height-->0){
        nile_blit_from_bitplanes_line(blit, dst, src, offs, width);
        dst += dst_stride;
        offs += src_stride;
    }
}



void nile_blit_pixels_from_bitplanes_unindex(uint8_t *dst, int dst_stride, int bpc, uint8_t **src, int src_stride, int depth, const uint8_t *palette, int width, int height)
{
#if 1
    nile_blit_t blit;
    nile_blit_init_from_bitplanes_unindex_rgb(&blit, bpc, NILE_FORMAT_UNKNOWN, depth, palette, 256, bpc, NILE_FORMAT_UNKNOWN);
    nile_blit_from_bitplanes(&blit, dst, dst_stride, src, src_stride, width, height, 0);
#else
    off_t offs = 0;
    while(height-->0){
        _copy_from_bitplanes_unindex_line(dst, bpc, src, offs, depth, palette, width);
        dst += dst_stride;
        offs += src_stride;
    }
#endif
}



static void
_blit_line_from_bitplanes(uint8_t *dst, int bpc, uint8_t **src, off_t offs, int depth, int count)
{
#if 1
    int bit;
    while(count-->0){
        uint32_t pix;
        int i;
        pix = 0;
        for(i=0; i<depth ;++i)
            if(src[i][offs] & (128>>bit))
                pix |= 1<<i;
        pix = htobe32(pix);
        memcpy(dst, ((uint8_t*)&pix)+4-bpc, bpc);
        dst += bpc;
        if(++bit>=8){
            ++offs;
            bit = 0;
        }
    }
#else
    while(count>0){
        uint32_t b[8];
        int plane;
        int i;
        memset(b, 0, 8 * sizeof(uint32_t));
        for(plane=0; plane<depth; ++plane){
            for(i=0;i<8;++i)
                if(src[plane][offs] & (128>>i))
                    b[i] |= 1<<plane;
        }
        for(i=0;i<8;++i){
            if(count<=0)
                break;
            b[i] = htobe32(b[i]);
            memcpy(dst, ((uint8_t*)&b[i])+4-bpc, bpc);
            dst += bpc;
            --count;
        }
        ++offs;
    }
#endif
}

void
nile_blit_pixels_from_bitplanes(uint8_t *dst, int bpc, int dst_stride, uint8_t **src, int src_stride, int depth, int width, int height)
{
    off_t offs = 0;
    while(height-->0){
        _blit_line_from_bitplanes(dst, bpc, src, offs, depth, width);
        dst += dst_stride;
        offs += src_stride;
    }
}
