/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <byteswap.h>
#include <string.h>                     /* memcpy(), memset() */

#include "debug.h"
#include "nile-blit.h"


int
nile_blit_setup_palette(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(dst_bpc != src_bpc || dst_format != src_format){
        if(colors > 256){
            DEBUG_p("more that 256 colors is not supported");
            return -1;
        }
        nile_blit_colors_rgb(blit->tmp_palette, dst_bpc, dst_format, palette, src_bpc, src_format, colors);
        blit->palette = blit->tmp_palette;
    }else{
        blit->palette = palette;
    }
    return 0;
}


static void
_blit_copy(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    memcpy(dst, src, count * blit->src_bpc);
}


static void
_blit_8i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        dst[blit->dst_o[4]] = src[blit->src_o[4]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_fill(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        dst[blit->dst_o[4]] = 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int a = src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a*src[blit->src_o[0]]/255;
        dst[blit->dst_o[1]] = a*src[blit->src_o[1]]/255;
        dst[blit->dst_o[2]] = a*src[blit->src_o[2]]/255;
        dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_unmultiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int a = src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a ? 255*src[blit->src_o[0]]/a : 0;
        dst[blit->dst_o[1]] = a ? 255*src[blit->src_o[1]]/a : 0;
        dst[blit->dst_o[2]] = a ? 255*src[blit->src_o[2]]/a : 0;
        dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_r5g6b5(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int p = *(uint16_t*)src;
        int v;
        v = p >> 11;
        dst[blit->dst_o[0]] = v << 3 | v >> 2;
        v = (p>>5) & 63;
        dst[blit->dst_o[1]] = v << 2 | v >> 4;
        v = p & 31;
        dst[blit->dst_o[2]] = v << 3 | v >> 2;
        src += 2;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_6i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int v;
        v = src[blit->src_o[0]];
        dst[blit->dst_o[0]] = v << 2 | v >> 4;
        v = src[blit->src_o[1]];
        dst[blit->dst_o[1]] = v << 2 | v >> 4;
        v = src[blit->src_o[2]];
        dst[blit->dst_o[2]] = v << 2 | v >> 4;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_16i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] >> 8;
        dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] >> 8;
        dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] >> 8;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_16i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] >> 8;
        dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] >> 8;
        dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] >> 8;
        dst[blit->dst_o[4]] = *(uint16_t*)&src[blit->src_o[4]] >> 8;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_32f3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]] * 255;
        dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]] * 255;
        dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]] * 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_32f3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]] * 255;
        dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]] * 255;
        dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]] * 255;
        dst[blit->dst_o[4]] = *(float*)&src[blit->src_o[4]] * 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_32f3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = *(float*)&src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a * *(float*)&src[blit->src_o[0]]/255;
        dst[blit->dst_o[1]] = a * *(float*)&src[blit->src_o[1]]/255;
        dst[blit->dst_o[2]] = a * *(float*)&src[blit->src_o[2]]/255;
        dst[blit->dst_o[4]] = a * 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_32f3a_unmultiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = *(float*)&src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a ? *(float*)&src[blit->src_o[0]] * 255 / a : 0;
        dst[blit->dst_o[1]] = a ? *(float*)&src[blit->src_o[1]] * 255 / a : 0;
        dst[blit->dst_o[2]] = a ? *(float*)&src[blit->src_o[2]] * 255 / a : 0;
        dst[blit->dst_o[4]] = a * 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]];
        *(float*)&dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]];
        *(float*)&dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]];
        *(float*)&dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]];
        *(float*)&dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]];
        *(float*)&dst[blit->dst_o[4]] = *(float*)&src[blit->src_o[4]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = *(float*)&src[blit->src_o[4]];
        *(float*)&dst[blit->dst_o[0]] = a * *(float*)&src[blit->src_o[0]];
        *(float*)&dst[blit->dst_o[1]] = a * *(float*)&src[blit->src_o[1]];
        *(float*)&dst[blit->dst_o[2]] = a * *(float*)&src[blit->src_o[2]];
        *(float*)&dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3a_from_16i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[4]] = *(uint16_t*)&src[blit->src_o[4]] / UINT16_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3_from_16i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3_from_32i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = (float)*(uint32_t*)&src[blit->src_o[0]] / UINT32_MAX;
        *(float*)&dst[blit->dst_o[1]] = (float)*(uint32_t*)&src[blit->src_o[1]] / UINT32_MAX;
        *(float*)&dst[blit->dst_o[2]] = (float)*(uint32_t*)&src[blit->src_o[2]] / UINT32_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3_from_32i3_swap(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = (float)bswap_32(*(uint32_t*)&src[blit->src_o[0]]) / UINT32_MAX;
        *(float*)&dst[blit->dst_o[1]] = (float)bswap_32(*(uint32_t*)&src[blit->src_o[1]]) / UINT32_MAX;
        *(float*)&dst[blit->dst_o[2]] = (float)bswap_32(*(uint32_t*)&src[blit->src_o[2]]) / UINT32_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_32f3a_from_16i3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = (float)*(uint16_t*)&src[blit->src_o[4]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[0]] = a * *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = a * *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = a * *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

nile_blit_t *
nile_blit_init_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, int src_bpc, nile_format_t src_format)
{
    blit->dst_bpc = dst_bpc;
    blit->src_bpc = src_bpc;
    if(dst_bpc == src_bpc && dst_format == src_format){
        blit->blit_func = _blit_copy;
    }else{
        blit->dst_o[0] = NILE_OFFSET_CHANNEL(dst_format,0);
        blit->dst_o[1] = NILE_OFFSET_CHANNEL(dst_format,1);
        blit->dst_o[2] = NILE_OFFSET_CHANNEL(dst_format,2);
        blit->dst_o[4] = NILE_OFFSET_ALPHA(dst_format);
        if(NILE_FORMAT_ne_r5g6b5 == src_format){
            blit->blit_func = _blit_8i3_from_r5g6b5;
            return blit;
        }else if(NILE_FORMAT_r6g6b6 == src_format){
            blit->src_o[0] = 0;
            blit->src_o[1] = 1;
            blit->src_o[2] = 2;
            blit->blit_func = _blit_8i3_from_6i3;
            return blit;
        }
        blit->src_o[0] = NILE_OFFSET_CHANNEL(src_format,0);
        blit->src_o[1] = NILE_OFFSET_CHANNEL(src_format,1);
        blit->src_o[2] = NILE_OFFSET_CHANNEL(src_format,2);
        blit->src_o[4] = NILE_OFFSET_ALPHA(src_format);
        switch(NILE_GET_BYTES_PER_CHANNEL(dst_format)){
        case 1:
            switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
            case 1:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_8i3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_8i3a_multiply;
                    }else{
                        blit->blit_func = _blit_8i3a_unmultiply;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    blit->blit_func = _blit_8i3a_fill;
                }else{
                    blit->blit_func = _blit_8i3;
                }
                break;
            case 2:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_8i3a_from_16i3a;
                    }else{
                        return NULL;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    /* FIXME!! */
                    /* blit->blit_func = _blit_8i3a_from_16i3_fill; */
                    return NULL;
                }else{
                    blit->blit_func = _blit_8i3_from_16i3;
                }
                break;
            case 4:
                DEBUG_p("4 bytes per channel to 1: channeltype: %x", NILE_GET_CHANNELTYPE(src_format));
                switch(NILE_GET_CHANNELTYPE(src_format)){
                case NILE_CHANNELTYPE_UNKNOWN:
                    return NULL;
                case NILE_CHANNELTYPE_INTEGER:
                    if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                        /* Both have alpha channel */
                        if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                            /* blit->blit_func = _blit_8i3a_from_32i3a; */
                            return NULL;
                        }else if(dst_format & NILE_PREMULTIPLIED){
                            /* blit->blit_func = _blit_8i3a_from_32i3a_multiply; */
                            return NULL;
                        }else{
                            /* blit->blit_func = _blit_8i3a_from_32i3a_unmultiply; */
                            return NULL;
                        }
                    }else if(NILE_HAS_ALPHA(dst_format)){
                        /* blit->blit_func = _blit_8i3a_from_32i3_fill; */
                        return NULL;
                    }else{
                        /* blit->blit_func = _blit_8i3_from_32i3; */
                        return NULL;
                    }
                    break;
                case NILE_CHANNELTYPE_IEEE754:
                    if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                        /* Both have alpha channel */
                        if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                            blit->blit_func = _blit_8i3a_from_32f3a;
                        }else if(dst_format & NILE_PREMULTIPLIED){
                            blit->blit_func = _blit_8i3a_from_32f3a_multiply;
                        }else{
                            blit->blit_func = _blit_8i3a_from_32f3a_unmultiply;
                        }
                    }else if(NILE_HAS_ALPHA(dst_format)){
                        /* blit->blit_func = _blit_8i3a_from_32f3_fill; */
                        return NULL;
                    }else{
                        blit->blit_func = _blit_8i3_from_32f3;
                    }
                    break;
                }
                break;
            default:
                return NULL;
            }
            break;
        case 4: /* assume float */
            switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
            case 2:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_32f3a_from_16i3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_32f3a_from_16i3a_multiply;
                    }else{
                        return NULL;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    return NULL;
                }else{
                    blit->blit_func = _blit_32f3_from_16i3;
                }
                break;
            case 4:
                switch(NILE_GET_CHANNELTYPE(src_format)){
                case NILE_CHANNELTYPE_UNKNOWN:
                    return NULL;
                case NILE_CHANNELTYPE_INTEGER:
                    if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                        /* Bath have alpha channel */
                        if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                            /* blit->blit_func = _blit_32f3a_from_32i3a; */
                            return NULL;
                        }else if(dst_format & NILE_PREMULTIPLIED){
                            /* blit->blit_func = _blit_32f3a_32i3a_multiply; */
                            return NULL;
                        }else{
                            return NULL;
                        }
                    }else if(NILE_HAS_ALPHA(dst_format)){
                        return NULL;
                    }else{
                        if((src_format & NILE_ENDIANESS) == NILE_ENDIAN_NATIVE){
                            blit->blit_func = _blit_32f3_from_32i3;
                        }else{
                            blit->blit_func = _blit_32f3_from_32i3_swap;
                        }
                    }
                    break;
                case NILE_CHANNELTYPE_IEEE754:
                    if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                        /* Bath have alpha channel */
                        if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                            blit->blit_func = _blit_32f3a;
                        }else if(dst_format & NILE_PREMULTIPLIED){
                            blit->blit_func = _blit_32f3a_multiply;
                        }else{
                            return NULL;
                        }
                    }else if(NILE_HAS_ALPHA(dst_format)){
                        return NULL;
                    }else{
                        blit->blit_func = _blit_32f3;
                    }
                    break;
                }
                break;
            default:
                return NULL;
            }
            break;
        default:
            return NULL;
        }
    }
    return blit;
}

static void
_blit_unindex(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        memcpy(dst, blit->palette + *src++ * blit->dst_bpc, blit->dst_bpc);
        dst += blit->dst_bpc;
    }
}

nile_blit_t *
nile_blit_init_unindex_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(nile_blit_setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format) < 0)
        return NULL;
    blit->dst_bpc = dst_bpc;
    blit->src_bpc = 1;
    blit->blit_func = _blit_unindex;
    return blit;
}


void
nile_blit(nile_blit_t *blit, uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int width, int height, int flags)
{
    if(dst_stride < 0)
        dst_stride = blit->dst_bpc * width;
    if(src_stride < 0)
        src_stride = blit->src_bpc * width;
    if(flags & NILE_FLIPPED){
        src += src_stride * (height-1);
        src_stride *= -1;
    }
    while(height-->0){
        nile_blit_line(blit, dst, src, width);
        dst += dst_stride;
        src += src_stride;
    }
}



void
nile_blit_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count)
{
    nile_blit_t blit;
    nile_blit_init_rgb(&blit, dst_bpc, dst_format, src_bpc, src_format);
    nile_blit_line(&blit, dst, src, count);
}



static void
_blit_line_unindex(uint8_t *dst, const uint8_t *src, int bpc, const uint8_t *palette, int count)
{
    while(count-->0){
        memcpy(dst, &palette[*src++ * bpc], bpc);
        dst += bpc;
    }
}

void nile_blit_pixels_unindex(uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int bpc, const uint8_t *palette, int width, int height)
{
    while(height-->0){
        _blit_line_unindex(dst, src, bpc, palette, width);
        dst += dst_stride;
        src += src_stride;
    }
}
