/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <byteswap.h>
#include <float.h>                      /* DBL_MAX */
#include <stddef.h>                     /* size_t */
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcmp(), memcpy() */

#include "debug.h"
#include "nile.h"
//#include "nile-blit.h"


#define CHECK_MISMATCH(a,b) (((a)>0 && (b)>0 && (a) != (b)) || ((a)<=0 && (b)<=0))


static int
_calc_stride(int bytes, int stride)
{
    if(stride>0)
        return stride;
    switch(stride){
    case 0:
    case NILE_STRIDE_ALIGN8:
        break;
    case NILE_STRIDE_ALIGN16:
        if(bytes&1)
            bytes += 1;
        break;
    case NILE_STRIDE_ALIGN32:
        if(bytes&3)
            bytes += 4 - (bytes&3);
        break;
    case NILE_STRIDE_ALIGN64:
        if(bytes&7)
            bytes += 8 - (bytes&7);
        break;
    }
    return bytes;
}


static int
_calc_stride_bits(int bits, int stride)
{
    return _calc_stride((bits-1)/8+1, stride);
}


static void
_link_bitplanes(nile_t *ni)
{
    int offset = nile_get_stride(ni) * ni->height;
    int i;
    ni->bitplanes[0] = ni->pixels + ni->depth * sizeof(uint8_t*);
    for(i=1;i<ni->depth;++i)
        ni->bitplanes[i] = ni->bitplanes[i-1] + offset;
}

/* ---------------------------------------------------------------------- */

nile_t *
nile_init(nile_t *ni)
{
    memset(ni,0,sizeof(nile_t));
    return ni;
}


void
nile_done(nile_t *ni)
{
    if(ni->pixels_allocation)
        free(ni->pixels);
    if(ni->palette_allocation)
        free(ni->palette);
}


nile_t *
nile_new()
{
    return (nile_t*)calloc(1,sizeof(nile_t));
}


void
nile_free(nile_t* ni)
{
    nile_done(ni);
    free(ni);
}

/* ---------------------------------------------------------------------- */

const uint8_t *
nile_get_color(const nile_t *ni, int x, int y)
{
    int stride;
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_INDEXED:
        stride = nile_get_stride(ni);
        return ni->palette + ni->pixels[x + y*stride] * ni->bpc;
    case NILE_TYPE_DIRECT:
        stride = nile_get_stride(ni);
        return ni->pixels + x*ni->bpc + y*stride;
    case NILE_TYPE_BITPLANES:
    case NILE_TYPE_BITPLANES_DIRECT:
        /* FIXME!! */
        break;
    }
    return NULL;
}


int
nile_get_stride(const nile_t *ni)
{
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_INDEXED:
        return _calc_stride(ni->width, ni->stride);
    case NILE_TYPE_DIRECT:
        return _calc_stride(ni->width * ni->bpc, ni->stride);
    case NILE_TYPE_BITPLANES:
    case NILE_TYPE_BITPLANES_DIRECT:
        return _calc_stride_bits(ni->width, ni->stride);
    }
    return 0;
}


int
nile_get_skip(const nile_t *ni)
{
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_INDEXED:
        return _calc_stride(ni->width, ni->stride) - ni->width;
    case NILE_TYPE_DIRECT:
        return _calc_stride(ni->width * ni->bpc, ni->stride) - ni->width * ni->bpc;
    case NILE_TYPE_BITPLANES:
    case NILE_TYPE_BITPLANES_DIRECT:
        {
            int bytes = (ni->width-1)/8+1;
            return _calc_stride(bytes, ni->stride) - bytes;
        }
    }
    return 0;
}


double
nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2)
{
    int r1_o = NILE_OFFSET_CHANNEL(format1,0);
    int g1_o = NILE_OFFSET_CHANNEL(format1,1);
    int b1_o = NILE_OFFSET_CHANNEL(format1,2);
    int r2_o = NILE_OFFSET_CHANNEL(format2,0);
    int g2_o = NILE_OFFSET_CHANNEL(format2,1);
    int b2_o = NILE_OFFSET_CHANNEL(format2,2);
    if(NILE_GET_COLORSPACE(format1) != NILE_GET_COLORSPACE(format2))
        return DBL_MAX;
    switch(NILE_GET_COLORSPACE(format1)){
    case NILE_COLORSPACE_RGB:
        {
            double r = color1[r1_o]-color2[r2_o];
            double g = color1[g1_o]-color2[g2_o];
            double b = color1[b1_o]-color2[b2_o];
#if 1
            return 2.5*r*r + 4*g*g + 2.5*b*b;
#else
            return r*r + g*g + b*b;
#endif
        }
    default:
        break;
    }
    return DBL_MAX;
}


int
nile_best_color_index(const uint8_t *p, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format)
{
    double best_value = DBL_MAX;
    int best_index = -1;
    int i;
    for(i=0;i<colors;++i){
        double value = nile_color_variance(p, palette_format, color, color_format);
        if(value < best_value){
            best_value = value;
            best_index = i;
        }
        p += bpc;
    }
    return best_index;
}

/* ---------------------------------------------------------------------- */

int
nile_ensure_indexed(nile_t *ni,int width, int height)
{
#if 1
    int stride;
    if(NILE_TYPE_INDEXED == ni->type && ni->pixels && ni->width >= width && ni->height >= height)
        return 0;
    nile_drop_pixels(ni);
    stride = _calc_stride(width, ni->stride);
    ni->type = NILE_TYPE_INDEXED;
    ni->pixels_allocation = (size_t)stride * height;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
#else
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(colors > 256)
        return -1;
    if(ni->colors){
        if(ni->colors < colors)
            return -1;
    }else{
        ni->colors = colors;
    }
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    ni->indexed = malloc(ni->width*ni->height);
#endif
    return 0;
}


int
nile_ensure_direct(nile_t *ni,int width, int height, int bpc)
{
#if 1
    if(NILE_TYPE_DIRECT == ni->type && ni->pixels && ni->width >= width && ni->height && ni->bpc >= bpc)
        return 0;
    nile_drop_pixels(ni);
    ni->type = NILE_TYPE_DIRECT;
    ni->stride = _calc_stride(width * bpc, 0);
    ni->pixels_allocation = ni->stride * height;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
    ni->bpc = bpc;
#else
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    if(bpc>0)
        ni->bpc = bpc;
    ni->planar = malloc(ni->width * ni->height * ni->bpc);
#endif
    return 0;
}


int
nile_ensure_bitplanes(nile_t *ni,int width, int height, int depth, int deep)
{
#if 1
    int stride;
    if(NILE_TYPE_BITPLANES == ni->type && ni->bitplanes && ni->width >= width && ni->height >= height && ni->depth >= depth)
        return 0;
    nile_drop_pixels(ni);
    ni->type = NILE_TYPE_BITPLANES;
    stride = deep ? _calc_stride_bits(width, ni->stride) : 0;
    ni->pixels_allocation = depth * sizeof(uint8_t*) + (size_t)stride * height * depth;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
    ni->depth = depth;
    if(deep)
        _link_bitplanes(ni);
#else
    int stride = (width-1)/8+1;
    int size = stride * height * planes;
    int i;
    ni->width = width;
    ni->height = height;
    ni->planes = planes;
    ni->bitplanes = malloc(ni->planes * sizeof(uint8_t*));
    ni->bitplanes[0] = malloc(size);
    for(i=1;i<ni->planes;++i)
        ni->bitplanes[i] = ni->bitplanes[i-1] + stride*ni->height;
#endif
    return 0;
}


int
nile_ensure_palette(nile_t *ni,int colors, int bpc)
{
#if 1
    if(ni->palette && ni->colors >= colors && ni->bpc >= bpc)
        return 0;
    if(ni->palette_allocation){
        if(ni->palette_allocation < colors * bpc){
            ni->palette_allocation = colors * bpc;
            ni->palette = realloc(ni->palette, ni->palette_allocation);
        }
    }else{
        ni->palette_allocation = colors * bpc;
        ni->palette = malloc(ni->palette_allocation);
    }
    ni->colors = colors;
    ni->bpc = bpc;
#else
    if(CHECK_MISMATCH(ni->bpc, bpc))
        return -1;
    if(colors>0)
        ni->colors = colors;
    if(ni->colors <= 0)
        return -1;
    if(bpc>0)
        ni->bpc = bpc;
    if(ni->bpc <= 0)
        return -1;
    if(ni->palette){
        ni->palette = realloc(ni->palette, ni->colors * ni->bpc);
    }else{
        ni->palette = (uint8_t*)malloc(ni->colors * ni->bpc);
    }
#endif
    return 0;
}


void nile_drop_pixels(nile_t *ni){
    if(ni->pixels_allocation){
        free(ni->pixels);
        ni->pixels_allocation = 0;
    }
    ni->pixels = NULL;
    if(ni->stride > 0)
        ni->stride = 0;
}


void
nile_drop_palette(nile_t *ni)
{
    if(ni->palette_allocation){
        free(ni->palette);
        ni->palette_allocation = 0;
    }
}

/* ---------------------------------------------------------------------- */

void nile_copy_identity(nile_t *dst, const nile_t *src)
{
    dst->type = src->type;
    dst->width = src->width;
    dst->height = src->height;
    dst->stride = src->stride;
    dst->depth = src->depth;
    dst->colors = src->colors;
    dst->bpc = src->bpc;
    dst->format = src->format;
}


void
nile_copy_pixels(nile_t *dst, const nile_t *src, int deep)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    dst->type = src->type;
    if(deep && src->pixels_allocation){
        dst->pixels_allocation = src->pixels_allocation;
        dst->pixels = malloc(dst->pixels_allocation);
        memcpy(dst->pixels, src->pixels, dst->pixels_allocation);
        if(NILE_TYPE_BITPLANES == dst->type)
            _link_bitplanes(dst);
    }else{
        dst->pixels_allocation = 0;
        dst->pixels = src->pixels;
    }
    dst->width = src->width;
    dst->height = src->height;
    dst->stride = src->stride;
    if(NILE_TYPE_DIRECT == dst->type){
        dst->bpc = src->bpc;
        dst->format = src->format;
    }
}


void
nile_copy_palette(nile_t *dst, const nile_t *src, int deep)
{
    if(dst->palette_allocation)
        free(dst->palette);
    if(deep && src->palette_allocation){
        dst->palette_allocation = src->palette_allocation;
        dst->palette = malloc(dst->palette_allocation);
        memcpy(dst->palette, src->palette, dst->palette_allocation);
    }else{
        dst->palette_allocation = 0;
        dst->palette = src->palette;
    }
    dst->colors = src->colors;
    dst->bpc = src->bpc;
    dst->format = src->format;
}


void
nile_copy(nile_t *dst, const nile_t *src, int deep)
{
    memcpy(dst, src, sizeof(nile_t));
    if(deep){
        if(dst->pixels_allocation){
            dst->pixels = malloc(dst->pixels_allocation);
            memcpy(dst->pixels, src->pixels, dst->pixels_allocation);
            if(NILE_TYPE_BITPLANES == dst->type)
                _link_bitplanes(dst);
        }
        if(dst->palette_allocation){
            dst->palette = malloc(dst->palette_allocation);
            memcpy(dst->palette, src->palette, dst->palette_allocation);
        }
    }else{
        memcpy(dst, src, sizeof(nile_t));
        dst->pixels_allocation = dst->palette_allocation = 0;
    }
}


void
nile_move_pixels(nile_t *dst, nile_t *src)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    dst->type = src->type;
    dst->pixels_allocation = src->pixels_allocation;
    dst->pixels = src->pixels;
    dst->width = src->width;
    dst->height = src->height;
    dst->stride = src->stride;
    if(NILE_TYPE_BITPLANES == src->type || NILE_TYPE_BITPLANES_DIRECT){
        dst->depth = src->depth;
    }
    if(NILE_TYPE_DIRECT == src->type || NILE_TYPE_BITPLANES_DIRECT){
        dst->bpc = src->bpc;
        dst->format = src->format;
    }
    src->pixels_allocation = 0;
}


void
nile_move_palette(nile_t *dst, nile_t *src)
{
    if(dst->palette_allocation)
        free(dst->palette);
    dst->palette_allocation = src->palette_allocation;
    dst->palette = src->palette;
    dst->colors = src->colors;
    src->palette_allocation = 0;
    dst->bpc = src->bpc;
    dst->format = src->format;
}


void
nile_move(nile_t *dst, nile_t *src)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    if(dst->palette_allocation)
        free(dst->palette);
    memcpy(dst, src, sizeof(nile_t));
    src->pixels_allocation = src->palette_allocation = 0;
}
