/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef nile_h
#define nile_h

#include <math.h>                       /* sqrt() */
#include <endian.h>
#include <stddef.h>                     /* size_t */
#include <stdint.h>                     /* uint8_t */
#include <sys/types.h>                  /* off_t */


#define NILE_PREMULTIPLIED (1<<29)
typedef enum{
    NILE_COLORSPACE_UNKNOWN = (0<<26),
    NILE_COLORSPACE_RGB     = (1<<26),
    NILE_COLORSPACE_CMYK    = (2<<26),
    NILE_COLORSPACE_HSV     = (3<<26)
}nile_colorspace_t;
typedef enum{
    NILE_CHANNELTYPE_UNKNOWN = (0<<24),
    NILE_CHANNELTYPE_INTEGER = (1<<24),
    NILE_CHANNELTYPE_IEEE754 = (2<<24)
}nile_channeltype_t;
#define NILE_ENDIANESS     (1<<23)
#define NILE_ENDIAN_BIG 0
#define NILE_ENDIAN_LITTLE NILE_ENDIANESS
#if __BYTE_ORDER == __BIG_ENDIAN
# define NILE_ENDIAN_NATIVE NILE_ENDIAN_BIG
#else
# define NILE_ENDIAN_NATIVE NILE_ENDIAN_LITTLE
#endif

/*

31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
         \      / \   /    \      / \         / \         / \         / \         / \         /
R     P     CS      C   E     CB       Alpha     Channel 3   Channel 2   Channel 1   Channel 0

R  = Reserved for custom formats
P  = Premultiplied
CS = Color space
C  = Channel type (integer or IEEE-754 floting point)
E  = Endianness (0 = big endian, 1 = little endian)
CB = Bytes per color channel (offset shifts): up to 128 bytes (1024 bits) per channel

*/

#define NILE_FORMAT_rgba8(r,g,b,a) (NILE_COLORSPACE_RGB|NILE_CHANNELTYPE_INTEGER|0<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb8(r,g,b) NILE_FORMAT_rgba8(r,g,b,15)
#define NILE_FORMAT_rgba16(r,g,b,a) (NILE_COLORSPACE_RGB|NILE_CHANNELTYPE_INTEGER|1<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb16(r,g,b) NILE_FORMAT_rgba16(r,g,b,15)
#define NILE_FORMAT_rgba32(r,g,b,a) (NILE_COLORSPACE_RGB|NILE_CHANNELTYPE_INTEGER|2<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgba32f(r,g,b,a) (NILE_COLORSPACE_RGB|NILE_CHANNELTYPE_IEEE754|2<<20|(r)<<0|(g)<<4|(b)<<8|15<<12|(a)<<16)
#define NILE_FORMAT_rgb32(r,g,b) NILE_FORMAT_rgba32(r,g,b,15)
#define NILE_FORMAT_rgb32f(r,g,b) NILE_FORMAT_rgba32f(r,g,b,15)

#define NILE_GET_COLORSPACE(f) ((nile_colorspace_t)((f)&(7<<26)))
#define NILE_GET_CHANNELTYPE(f) ((nile_channeltype_t)((f)&(3<<24)))
#define NILE_GET_BYTES_PER_CHANNEL(f) (1<<(((f)>>20)&7))
#define NILE_GET_CHANNEL(f,c) (((f)>>((c)*4))&15)
#define NILE_OFFSET_CHANNEL(f,c) (NILE_GET_CHANNEL(f,c)*NILE_GET_BYTES_PER_CHANNEL(f))
#define NILE_OFFSET_RED(f)   NILE_OFFSET_CHANNEL(f,0)
#define NILE_OFFSET_GREEN(f) NILE_OFFSET_CHANNEL(f,1)
#define NILE_OFFSET_BLUE(f)  NILE_OFFSET_CHANNEL(f,2)
#define NILE_OFFSET_ALPHA(f) NILE_OFFSET_CHANNEL(f,4)

#define NILE_HAS_CHANNEL(f,c) (NILE_GET_CHANNEL(f,c)<15)
#define NILE_HAS_ALPHA(f) NILE_HAS_CHANNEL(f,4)

#define NILE_MAX_BYTES_PER_COLOR 64
#define NILE_STRIDE_ALIGN8  -8
#define NILE_STRIDE_ALIGN16 -16
#define NILE_STRIDE_ALIGN32 -32
#define NILE_STRIDE_ALIGN64 -64
#define NILE_FLIPPED       (1<<0)


typedef enum{
    NILE_TYPE_NONE = 0,
    NILE_TYPE_INDEXED,
    NILE_TYPE_DIRECT,
    NILE_TYPE_BITPLANES,
    NILE_TYPE_BITPLANES_DIRECT
}nile_type_t;

typedef enum{
    NILE_FORMAT_UNKNOWN = 0,
    /* RGB */
    NILE_FORMAT_r8g8b8          = NILE_FORMAT_rgb8(0,1,2),
    NILE_FORMAT_x8r8g8b8        = NILE_FORMAT_rgb8(1,2,3),
    NILE_FORMAT_r8g8b8a8        = NILE_FORMAT_rgba8(0,1,2,3),
    NILE_FORMAT_a8r8g8b8        = NILE_FORMAT_rgba8(1,2,3,0),
    NILE_FORMAT_be_r16g16b16    = NILE_FORMAT_rgb16(0,1,2)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_r16g16b16    = NILE_FORMAT_rgb16(0,1,2)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_r16g16b16    = NILE_FORMAT_rgb16(0,1,2)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_be_x16r16g16b16 = NILE_FORMAT_rgb16(1,2,3)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_x16r16g16b16 = NILE_FORMAT_rgb16(1,2,3)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_x16r16g16b16 = NILE_FORMAT_rgb16(1,2,3)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_be_r16g16b16a16 = NILE_FORMAT_rgba16(0,1,2,3)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_r16g16b16a16 = NILE_FORMAT_rgba16(0,1,2,3)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_r16g16b16a16 = NILE_FORMAT_rgba16(0,1,2,3)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_be_r32g32b32    = NILE_FORMAT_rgb32(0,1,2)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_r32g32b32    = NILE_FORMAT_rgb32(0,1,2)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_r32g32b32    = NILE_FORMAT_rgb32(0,1,2)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_f_r32g32b32     = NILE_FORMAT_rgb32f(0,1,2),
    NILE_FORMAT_be_r32g32b32a32 = NILE_FORMAT_rgba32(0,1,2,3)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_r32g32b32a32 = NILE_FORMAT_rgba32(0,1,2,3)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_r32g32b32a32 = NILE_FORMAT_rgba32(0,1,2,3)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_f_r32g32b32a32  = NILE_FORMAT_rgba32f(0,1,2,3),
    NILE_FORMAT_ne_r5g6b5       = -10565,
    /* MCGA style encoding */
    NILE_FORMAT_r6g6b6          = -10666,
    /* BGR */
    NILE_FORMAT_b8g8r8          = NILE_FORMAT_rgb8(2,1,0),
    NILE_FORMAT_x8b8g8r8        = NILE_FORMAT_rgb8(3,2,1),
    NILE_FORMAT_b8g8r8a8        = NILE_FORMAT_rgba8(2,1,0,3),
    NILE_FORMAT_a8b8g8r8        = NILE_FORMAT_rgba8(3,2,1,0),
    NILE_FORMAT_be_b16g16r16a16 = NILE_FORMAT_rgba16(2,1,0,3)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_b16g16r16a16 = NILE_FORMAT_rgba16(2,1,0,3)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_b16g16r16a16 = NILE_FORMAT_rgba16(2,1,0,3)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_ne_b32g32r32    = NILE_FORMAT_rgb32(2,1,0)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_ne_b32g32r32a32 = NILE_FORMAT_rgba32(2,1,0,3)|NILE_ENDIAN_NATIVE,
    /* gray */
    NILE_FORMAT_g8              = NILE_FORMAT_rgb8(0,0,0),
    NILE_FORMAT_g8a8            = NILE_FORMAT_rgba8(0,0,0,1),
    NILE_FORMAT_a8g8            = NILE_FORMAT_rgba8(1,1,1,0),
    NILE_FORMAT_be_g16          = NILE_FORMAT_rgb16(0,0,0)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_g16          = NILE_FORMAT_rgb16(0,0,0)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_g16          = NILE_FORMAT_rgb16(0,0,0)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_be_g16a16       = NILE_FORMAT_rgba16(0,0,0,1)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_g16a16       = NILE_FORMAT_rgba16(0,0,0,1)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_g16a16       = NILE_FORMAT_rgba16(0,0,0,1)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_ne_a16g16       = NILE_FORMAT_rgba16(1,1,1,0)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_be_g32          = NILE_FORMAT_rgb32(0,0,0)|NILE_ENDIAN_BIG,
    NILE_FORMAT_le_g32          = NILE_FORMAT_rgb32(0,0,0)|NILE_ENDIAN_LITTLE,
    NILE_FORMAT_ne_g32          = NILE_FORMAT_rgb32(0,0,0)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_ne_g32a32       = NILE_FORMAT_rgba32(0,0,0,1)|NILE_ENDIAN_NATIVE,
    NILE_FORMAT_f_g32           = NILE_FORMAT_rgb32f(0,0,0)|NILE_ENDIAN_NATIVE
}nile_format_t;

typedef struct{
#if 1
    nile_type_t type;
    size_t pixels_allocation;           /** Amount of memory allocated for the pixels. */
    union{
        uint8_t *pixels;
        uint8_t **bitplanes;
    };
    int width;
    int height;
    int stride;
    int depth;                          /** The number of bitplanes. */
    size_t palette_allocation;          /** Amount of memory allocated for the palette. */
    uint8_t *palette;
    int colors;
    int bpc;
    nile_format_t format;
#else
    int width;
    int height;
    nile_format_t format;
    int bpc;                            /** Bytes per color. */
    int colors;
    int planes;
    uint8_t *palette;
    uint8_t **bitplanes;
    uint8_t *indexed;
    uint8_t *planar;
#endif
}nile_t;

nile_t *nile_init(nile_t*);
void nile_done(nile_t*);
nile_t *nile_new(void);
void nile_free(nile_t*);

/* Utility functions. */

const uint8_t *nile_get_color(const nile_t*, int x, int y);
int nile_get_stride(const nile_t*);
int nile_get_skip(const nile_t*);
double nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2);
#define nile_color_distance(c1,f1,c2,f2) sqrt(nile_color_variance(c1,f1,c2,f2))
int nile_best_color_index(const uint8_t *palette, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format);

int nile_ensure_indexed(nile_t *,int width, int height);
int nile_ensure_direct(nile_t *,int width, int height, int bytes_per_color);
int nile_ensure_bitplanes(nile_t *,int width, int height, int depth, int deep);
int nile_ensure_palette(nile_t *,int colors, int bytes_per_color);
void nile_drop_pixels(nile_t*);
void nile_drop_palette(nile_t*);

/* Copy everything but the pixels and the palette. */
void nile_copy_identity(nile_t*, const nile_t*);
void nile_copy_pixels(nile_t*, const nile_t*, int deep);
void nile_copy_palette(nile_t*,const nile_t*,int deep);
void nile_copy(nile_t*,const nile_t*,int deep);
/* Copy the pixels (and maybe the palette) based on the type and format of
   the destination image. */
void nile_move_pixels(nile_t*,nile_t*);
void nile_move_palette(nile_t*,nile_t*);
void nile_move(nile_t*,nile_t*);

void nile_convert(nile_t*, const nile_t*, int flags);

#endif /* nile_h */
