/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nile-blit.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_fbm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-fuzzy-bitmap",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_fbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_fbm_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_get_int(int *dst, const char *s, const char *end){
    if(end[-1] != '\0')
        return -1;
    while(' ' == *s)
        ++s;
    *dst = atoi(s);
    return 0;
}


static int
_get_real(double *dst, const char *s, const char *end){
    if(end[-1] != '\0')
        return -1;
    while(' ' == *s)
        ++s;
    *dst = strtod(s, NULL);
    return 0;
}


static void
_blit_rgb(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[0][offs];
        dst[blit->dst_o[1]] = src[1][offs];
        dst[blit->dst_o[2]] = src[2][offs];
        dst += blit->dst_bpc;
        ++offs;
    }
}


static int
_fbm_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    uint8_t *channel[4];
    nile_t ni;
    nile_t cairo_ni;
    int channels;
    int plane_size;
    int x;
    nile_blit_t blit;
    int i;

    if(len < 256)
        return -1;
    if(memcmp(data, "%bitmap", 8))
        return -1;

    nile_init(&ni);

    if(_get_int(&ni.width, data + 8, data + 16) < 0)
        return -1;
    if(_get_int(&ni.height, data + 16, data + 24) < 0)
        return -1;
    DEBUG_p("dimensions: %dx%d", ni.width, ni.height);
    /* planes / channels */
    if(_get_int(&channels, data + 24, data + 32) < 0)
        return -1;
    DEBUG_p("planes / channels: %d", channels);
    /* bits */
    if(_get_int(&x, data + 32, data + 40) < 0)
        return -1;
    DEBUG_p("bits: %d", x);
    /* physbits */
    if(_get_int(&x, data + 40, data + 48) < 0)
        return -1;
    DEBUG_p("physbits: %d", x);
    /* rowlen / stride */
    if(_get_int(&ni.stride, data + 48, data + 60) < 0)
        return -1;
    DEBUG_p("rowlen / stride: %d", ni.stride);
    /* plane size */
    if(_get_int(&plane_size, data + 60, data + 72) < 0)
        return -1;
    DEBUG_p("plane size: %d", plane_size);
    /* colormap length */
    if(_get_int(&x, data + 72, data + 84) < 0)
        return -1;
    DEBUG_p("colormap length: %d", x);
    /* aspect ratio */
    if(_get_real(&h->info->pixel_ratio, data + 84, data + 96) < 0)
        return -1;
    DEBUG_p("aspect ratio: %f", h->info->pixel_ratio);
    /* title */
    /* credits */

    data += 256;
    len -= 256;

    if(plane_size * channels > len)
        return -1;

    switch(channels){
    case 1:
        ni.type = NILE_TYPE_DIRECT;
        ni.pixels = (uint8_t*)data;
        ni.bpc = 1;
        ni.format = NILE_FORMAT_g8;
        break;
    case 3:
        ni.type = NILE_TYPE_DIRECT;
        for(i=0; i<channels; ++i)
            channel[i] = (uint8_t*)data + plane_size * i;
        ni.bitplanes = channel;
        ni.bpc = 3;
        ni.format = NILE_FORMAT_r8g8b8;
        break;
    default:
        return -1;
    }

    h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, ni.width, ni.height);
    nile_init(&cairo_ni);
    nil_cairo_from_surface(&cairo_ni, h->surface);
    switch(channels){
    case 1:
        nile_blit_init_rgb(&blit, cairo_ni.bpc, cairo_ni.format, ni.bpc, ni.format);
        nile_blit(&blit, cairo_ni.pixels, nile_get_stride(&cairo_ni),
                  ni.pixels, nile_get_stride(&ni), ni.width, ni.height, 0);
        break;
    default:
        nile_blit_init_from_bitplanes_rgb(&blit, cairo_ni.bpc, cairo_ni.format, ni.bpc, ni.format);
        blit.blit_from_bitplanes_func = _blit_rgb;
        nile_blit_from_bitplanes(&blit, cairo_ni.pixels, nile_get_stride(&cairo_ni),
                                 ni.bitplanes, nile_get_stride(&ni), ni.width, ni.height, 0);
        break;
    }
    cairo_surface_mark_dirty(h->surface);

    h->info->width = ni.width;
    h->info->height = ni.height;

    /* stride = ((width + 15) / 16) * 16; */
    /* size = stride * height */

    return 0;
}


static cairo_surface_t *
_fbm_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _fbm_get_mime_types,
        .new = _fbm_new,
        .free = _fbm_free,
        .create_from_data = _fbm_create_from_data,
        .get_image_surface = _fbm_get_image_surface
    };
    return &p;
}
